
#include "ShapePointCloud.h"

#include "FileManagement.h"
#include "VKString.h"
#include "TimeProfiler.h"
#include "gaps.h"
#include "Sortable.h"

bool windows_output = false;
bool simpleobj = false;

void ReadAsciiOff(const VKString & filename, std::vector<R3Point> & vertices, std::vector<int> & faces);
void ReadAsciiPly(const VKString & filename, std::vector<R3Point> & vertices, std::vector<int> & faces);
void WriteAsciiObj(const VKString & filename, std::vector<R3Point> & vertices, std::vector<int> & faces);
void FuzzyInitGraph2Maps(const VKString & fromFile, const VKString & toFile);
void ReadModelVertices(const VKString & fuzzyDir, std::vector<ShapePointCloud*> & vertices);
void Remesh(std::vector<R3Point> & vertices, std::vector<int> & faces, int minvtx);

void ConvertCorrespondences(const VKString & corrsDirectory, const VKString& outFile);

int main(int argc, char** argv )
{
	bool remesh = false;
	int minvtx = -1;

	if (argc<2)
	{
		std::cout<<"Usage: ./"<<argv[0]<<" fromFile toFile [optional]\n";
		std::cout<<"Optional:\n";
		std::cout<<" -verbose : be verbose\n";
		std::cout<<" -windows_output : use \\r\\n for newline and stuff\n"<<std::endl;
		std::cout<<" -simpleobj : no geometries (for huang et al. 2012 input)\n"<<std::endl;
		std::cout<<" -minvtx nvtx: minimla number of vertices\n"<<std::endl;
		exit(0);
	}
	bool verbose = false;
	VKString fromFile(argv[1]);
	VKString toFile(argv[2]);
	
	// parse params
	for (int i=3; i<argc; i++)
	{
		if (strcmp(argv[i], "-verbose")==0)
			verbose = true;
		else if (strcmp(argv[i], "-windows_output")==0)
			windows_output = true;
		else if (strcmp(argv[i], "-simpleobj")==0)
			simpleobj = true;
		else if (strcmp(argv[i], "-minvtx")==0)
		{
			minvtx = atoi(argv[++i]);
			remesh = true;
		}
		else
			std::cout<<"[WARNING] Unknown arg: "<<argv[i]<<std::endl;
	}
	
//	std::cout<<"Converting "<<fromFile.c_str()<<" "<<toFile.c_str()<<std::endl;
	
	if (fromFile.endsWith(".off") && toFile.endsWith(".obj"))
	{
		std::vector<R3Point> vertices;
		std::vector<int> faces;
		ReadAsciiOff(fromFile, vertices, faces);
		if (remesh)
			Remesh(vertices, faces, minvtx);
		WriteAsciiObj(toFile, vertices, faces);
	}
	else if (fromFile.endsWith(".ply") && toFile.endsWith(".obj"))
	{
		std::vector<R3Point> vertices;
		std::vector<int> faces;
		ReadAsciiPly(fromFile, vertices, faces);
		if (remesh)
			Remesh(vertices, faces, minvtx);
		WriteAsciiObj(toFile, vertices, faces);
	}
	else if (fromFile.endsWith("init_rigid_xforms.bin") && toFile.endsWith("/"))
	{
		FuzzyInitGraph2Maps(fromFile, toFile);
	}
	else if ((fromFile.endsWith("MAPS/")||fromFile.endsWith("MAPS")) && toFile.endsWith(".bin"))
	{
		// MAPS corrs to fuzzy corrs
		ConvertCorrespondences(fromFile, toFile);
	}
	else
		assert(false);

	return 0;
}

void ReadAsciiOff(const VKString & filename, std::vector<R3Point> & vertices, std::vector<int> & faces)
{
	// Open file
	FILE *fp;
	if (!(fp = fopen(filename.c_str(), "r")))
	{
		std::cout<<"[ERROR] Unable to open (read) "<<filename.c_str()<<std::endl;
		assert(false);
	}
	
	// Read file
	int nverts = 0;
	int nfaces = 0;
	int nedges = 0;
	int line_count = 0;
	int vertex_count = 0;
	int face_count = 0;
	char buffer[1024];
	char header[64];
	while (fgets(buffer, 1023, fp)) {
		// Increment line counter
		line_count++;
		
		// Skip white space
		char *bufferp = buffer;
		while (isspace(*bufferp)) bufferp++;
		
		// Skip blank lines and comments
		if (*bufferp == '#') continue;
		if (*bufferp == '\0') continue;
		
		// Check section
		if (nverts == 0) {
			// Read header keyword
			if (strstr(bufferp, "OFF")) {
				// Check if counts are on first line
				int tmp;
				if (sscanf(bufferp, "%s%d%d%d", header, &tmp, &nfaces, &nedges) == 4) {
					nverts = tmp;
				}
			}
			else {
				// Read counts from second line
				if ((sscanf(bufferp, "%d%d%d", &nverts, &nfaces, &nedges) != 3) || (nverts == 0)) {
					std::cout<<"[ERROR] Unable to read header "<<line_count;
					std::cout<<" file="<<filename.c_str()<<std::endl;
					fclose(fp);
					assert(false);
				}
			}
		}
		else if (vertex_count < nverts) {
			// Read vertex coordinates
			double x, y, z;
			if (sscanf(bufferp, "%lf%lf%lf", &x, &y, &z) != 3) {
				std::cout<<"[ERROR] Unable to read vtx coords "<<line_count;
				std::cout<<" file="<<filename.c_str()<<std::endl;
				fclose(fp);
				assert(false);
			}
			
			// Create vertex
			vertices.push_back(R3Point(x, y, z));
			
			// Increment counter
			vertex_count++;
		}
		else if (face_count < nfaces) {
			// Read number of vertices in face
			int face_nverts = 0;
			bufferp = strtok(bufferp, " \t");
			if (bufferp) face_nverts = atoi(bufferp);
			else {
				std::cout<<"[ERROR] Unable to read face "<<line_count;
				std::cout<<" file="<<filename.c_str()<<std::endl;
				fclose(fp);
				assert(false);
			}
			
			// Read vertex indices for face
			bufferp = strtok(NULL, " \t");
			assert(bufferp);
			int v1 = atoi(bufferp);
			bufferp = strtok(NULL, " \t");
			assert(bufferp);
			int v2 = atoi(bufferp);
			bufferp = strtok(NULL, " \t");
			assert(bufferp);
			int v3 = atoi(bufferp);
			faces.push_back(v1);
			faces.push_back(v2);
			faces.push_back(v3);
			// Increment counter
			face_count++;
		}
		else {
			// Should never get here
			std::cout<<"[ERROR] extr text line "<<line_count;
			std::cout<<" file="<<filename.c_str()<<std::endl;
			assert(false);
		}
	}
	
	// Close file
	fclose(fp);
}

void ReadAsciiPly(const VKString & filename, std::vector<R3Point> & vertices, std::vector<int> & faces)
{
	// parse simple ply
	std::ifstream mPlyStream(filename.c_str());
	assert(mPlyStream.is_open());
	VKString tempStr;
	tempStr.readLine(mPlyStream);
	if (tempStr!="ply")
		std::cout<<"[ERROR] File: "<<filename.c_str()<<" has invalid header: "<<tempStr.c_str()<<std::endl;
	assert(tempStr=="ply");
	tempStr.readLine(mPlyStream);		assert(tempStr=="format ascii 1.0");
	tempStr.readLine(mPlyStream);		assert(tempStr.startsWith("element vertex"));
	tempStr = tempStr.replace("element vertex ", "");
	bool ok = false;
	int nvert = tempStr.toInt(&ok);
	assert(ok);
	
	tempStr.readLine(mPlyStream);		assert(tempStr=="property float x");
	tempStr.readLine(mPlyStream);		assert(tempStr=="property float y");
	tempStr.readLine(mPlyStream);		assert(tempStr=="property float z");
	tempStr.readLine(mPlyStream);		assert(tempStr.startsWith("element face"));
	tempStr = tempStr.replace("element face ", "");
	int nface = tempStr.toInt(&ok);
	assert(ok);

	tempStr.readLine(mPlyStream);		assert(tempStr=="property list uchar int vertex_indices");
	tempStr.readLine(mPlyStream);		assert(tempStr=="property int material_id");
	tempStr.readLine(mPlyStream);		assert(tempStr=="property int segment_id");
	tempStr.readLine(mPlyStream);		assert(tempStr=="end_header");
	
	for (int v=0; v<nvert; v++)
	{
		tempStr.readLine(mPlyStream);
		VKStringList currLine =tempStr.split(" ");
		assert(currLine.count()==3);
		double x =currLine[0].toDouble(&ok);	assert(ok);
		double y =currLine[1].toDouble(&ok);	assert(ok);
		double z =currLine[2].toDouble(&ok);	assert(ok);
		vertices.push_back(R3Point(x, y, z));
	}
	
	for (int f=0; f<nface; f++)
	{
		tempStr.readLine(mPlyStream);
		VKStringList currLine =tempStr.split(" ");
		assert(currLine.count()==6);
		assert(currLine[0] == "3");
		assert(currLine[4] == "0");
		assert(currLine[5] == "0");
		int v1 =currLine[1].toInt(&ok);	assert(ok && v1 >= 0 && v1 < nvert);
		int v2 =currLine[2].toInt(&ok);	assert(ok && v2 >= 0 && v2 < nvert);
		int v3 =currLine[3].toInt(&ok);	assert(ok && v2 >= 0 && v2 < nvert);
		faces.push_back(v1);
		faces.push_back(v2);
		faces.push_back(v3);
	}
	mPlyStream.close();
}

void WriteAsciiObj(const VKString & filename, std::vector<R3Point> & vertices, std::vector<int> & faces)
{
	R3Mesh * mesh = new R3Mesh();
	for (int i=0; i<(int)vertices.size(); i++)
		mesh->CreateVertex(vertices[i]);
	
	assert(faces.size()%3==0);
	for (int i=0; i<(int)faces.size(); i+=3)
		mesh->CreateFace(mesh->Vertex(faces[i+0]), mesh->Vertex(faces[i+1]), mesh->Vertex(faces[i+2]));
	
	mesh->WriteObjFile(filename.c_str());
	return;
	
	int nvert = (int)vertices.size();
	int nface = (int)faces.size() /3;
	
	VKString newlineChar = "\n";
	if (windows_output)
		newlineChar = "\r\n";
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	if (!simpleobj)
	{
		textStream<<"# Studio"<<newlineChar.c_str();
		textStream<<"g Polygonal_Model_1"<<newlineChar.c_str();
		textStream<<"# Number of geometric vertices: "<<nvert<<newlineChar.c_str();
	}
	else
		textStream<<"# "<<nvert<<" "<<nface<<newlineChar.c_str();
	for (int i=0; i<nvert; i++)
	{
		R3Point p = vertices[i];
		textStream<<"v "<<p.X()<<" "<<p.Y()<<" "<<p.Z()<<newlineChar.c_str();
	}
	if (!simpleobj)
	{
		textStream<<"# Number of texture vertices: 0"<<newlineChar.c_str();
		textStream<<"g Polygonal_Model_1 Triangles_0"<<newlineChar.c_str();
		textStream<<"# Number of triangles: "<<nface<<newlineChar.c_str();
	}
	for (int f = 0; f<nface; f++)
	{
		int v1 = faces[3 * f + 0];
		int v2 = faces[3 * f + 1];
		int v3 = faces[3 * f + 2];
		textStream<<"f "<<v1<<" "<<v2<<" "<<v3<<newlineChar.c_str();
	}
	textStream<<"# end of file"<<newlineChar.c_str()<<std::endl;
	textStream.close();
}

void FuzzyInitGraph2Maps(const VKString & fromFile, const VKString & toFile)
{	
	VKString newlineChar = "\n";
	if (windows_output)
		newlineChar = "\r\n";

	// read models + create kd trees
	VKString fuzzyDir = fromFile;
	fuzzyDir = fuzzyDir.replace("init_rigid_xforms.bin", "");
	std::vector<ShapePointCloud*> vertices;
	ReadModelVertices(fuzzyDir, vertices);
	
	// read binary alignments file -> write vertex-to-vertex maps
	FILE * binFile = fopen(fromFile.c_str(), "rb");
	assert(binFile!=NULL);
	
	int nmodels;
	fread(&nmodels, sizeof(int), 1, binFile);
	assert(nmodels==(int)vertices.size());
	for (int i=0; i<nmodels; i++)
	{
		int mfrom;
		fread(&mfrom, sizeof(int), 1, binFile);
		assert(mfrom==i);
		int naligns;
		fread(&naligns, sizeof(int), 1, binFile);
		for (int a=0; a<naligns; a++)
		{
			int mid2;
			double array[16];
			fread(&mid2, sizeof(int), 1, binFile);
			fread(array, sizeof(double), 16, binFile);
			R3Affine aff12 = R3Affine(R4Matrix(array), false);
			
			// compute vertex-> vertex map
			VKString oFile = toFile+"/"+VKString::number(i+1)+"_"+VKString::number(mid2+1)+".txt";
			std::ofstream outMapStream(oFile.c_str());
			assert(outMapStream.is_open());
			for (int v=0; v<vertices[i]->NumPoints(); v++)
			{
				R3Point p1 = vertices[i]->PointPosition(v);
				aff12.Apply(p1);
				const PointInSet * closest = vertices[mid2]->GetKdTree()->FindClosest(p1);
				assert(closest!=NULL);
				outMapStream<<v<<" "<<closest->id<<newlineChar.c_str();
			}
			outMapStream.close();
		}
	}
	
	fclose(binFile);
}

void ReadModelVertices(const VKString & fuzzyDir, std::vector<ShapePointCloud*> & vertices)
{
	VKStringList modelList;
	VKString database = fuzzyDir+"../../database.txt";
	std::ifstream mListStream(database.c_str());
	assert(mListStream.is_open());
	int counter = 0;
	while (!mListStream.eof())
	{
		VKString tempStr;
		tempStr.readToken(mListStream);
		if (tempStr=="")
			break;
		assert(tempStr==VKString("Model")+VKString::number(counter)+":");
		tempStr.readToken(mListStream);		assert(tempStr=="ModelName");
		tempStr.readToken(mListStream);
		modelList.push_back(tempStr);
		tempStr.readToken(mListStream);		assert(tempStr=="ModelID");
		bool valid;
		tempStr.readToken(mListStream);		assert(tempStr.toInt(&valid)==counter);
		assert(valid);
		counter++;
	}
	mListStream.close();
	
	VKString modelsDir = fuzzyDir+"../../Models/";
	std::cout<<"Reading: "<<modelsDir.c_str()<<std::endl;
	for (int i=0; i<modelList.count(); i++)
	{
		int lid = (int)vertices.size();
		vertices.push_back(new ShapePointCloud(NULL, "vtx"));
		VKString mFile = modelsDir+modelList[i]+"/Geometry/"+modelList[i]+".ply";
		std::vector<R3Point> vtxs;
		std::vector<int> faces;
		ReadAsciiPly(mFile, vtxs, faces);
		for (int v = 0; v < (int)vtxs.size(); v++)
			vertices[i]->AddPoint(ShapePoint(vtxs[v]));
	}
}

void Remesh(std::vector<R3Point> & vertices, std::vector<int> & faces, int minvtx)
{
	if ((int)vertices.size() > minvtx)
		return;
	std::cout<<"Remeshing: "<<vertices.size()<<" / "<<minvtx<<": "<<std::flush;
	R3Mesh * mesh = new R3Mesh();
	for (int i=0; i<(int)vertices.size(); i++)
		mesh->CreateVertex(vertices[i]);
	
	assert(faces.size()%3==0);
	for (int i=0; i<(int)faces.size(); i+=3)
		mesh->CreateFace(mesh->Vertex(faces[i+0]), mesh->Vertex(faces[i+1]), mesh->Vertex(faces[i+2]));

	std::vector<Sortable> srtedges;
	for (int i=0; i<mesh->NEdges(); i++)
		srtedges.push_back(Sortable(-mesh->EdgeLength(mesh->Edge(i)), mesh->Edge(i)));
	std::sort(srtedges.begin(), srtedges.end());
	int counter = 0;
	while (mesh->NVertices() <= minvtx)
	{
		counter++;
		if (counter%100 == 0)
			std::cout<<"."<<std::flush;
		assert(srtedges.size()>0);
		R3MeshVertex * v = mesh->SubdivideEdge((R3MeshEdge *)srtedges[0].ptr);
		srtedges.erase(srtedges.begin());
		for (int i = 0; i < mesh->VertexValence(v); i++)
			srtedges.push_back(Sortable(-mesh->EdgeLength(mesh->EdgeOnVertex(v, i)), mesh->EdgeOnVertex(v, i)));
		std::sort(srtedges.begin(), srtedges.end());
	}
	std::cout<<std::endl;
	vertices.clear();
	faces.clear();
	for (int i=0; i<mesh->NVertices(); i++)
		vertices.push_back(mesh->VertexPosition(mesh->Vertex(i)));
	
	for (int i=0; i<mesh->NFaces(); i++)
	for (int v=0; v<3; v++)
		faces.push_back(mesh->VertexID(mesh->VertexOnFace(mesh->Face(i), v)));
	
	std::cout<<"Output: "<<vertices.size()<<" / "<<minvtx<<std::endl;
	
	
//	Model3D * m = new Model3D("tmp");
//	ShapeManifoldMesh * manifoldMesh = new ShapeManifoldMesh();
//	manifoldMesh->SetMesh(mesh);
//	m->AddGeomShape(Model3D::m_sMainGeomName, manifoldMesh);
//	
//	ShapePointCloud * pnts = ShapePointCloud::CreateIterFar("iterfar", Model3D::m_sMainGeomName, m,
//															minvtx, ShapeDistance::m_sDefaultEuclidean);
//	std::vector<R3Point> newpnts;
//	for (int i=0; i<pnts->NumPoints(); i++)
//		newpnts.push_back(pnts->PointPOsition(i));
//	
//	for (int i=0; i<(int)newpnts.size(); i++)
		
}


void ConvertCorrespondences(const VKString & corrsDirectory, const VKString& outFile)
{
	VKString modelListFile = corrsDirectory+"/../ModelList.txt";
	VKString gtModellistFile = corrsDirectory+"/../gtModelList.txt";
	assert(FileManagement::FileExists(modelListFile.c_str()));
	assert(FileManagement::FileExists(gtModellistFile.c_str()));	// copy manually
	
	VKStringList origNames;
	VKStringList gtNames;
	std::vector<int> gtID2OrigID;
	std::map<VKString, int> origName2id;
	
	VKString mName;
	std::ifstream allModelsStream(modelListFile.c_str());
	assert(allModelsStream.is_open());
	while (!allModelsStream.eof())
	{
		mName.readLine(allModelsStream);
		if (mName=="")
			continue;
		origName2id[mName] = origNames.count();
		origNames.push_back(mName);
	}
	allModelsStream.close();
	
	std::ifstream gtModelsStream(gtModellistFile.c_str());
	assert(gtModelsStream.is_open());
	while (!gtModelsStream.eof())
	{
		mName.readLine(gtModelsStream);
		if (mName=="")
			continue;
		assert(origName2id.find(mName)!=origName2id.end());
		gtID2OrigID.push_back(origName2id[mName]);
		gtNames.push_back(mName);
	}
	gtModelsStream.close();

	
	// GIVEN: 		sample -> sample
	// FIND:  		point -> point
	// PROCEDURE (for all pairs):	sample -> normalized mesh point -> original mesh point
	
	std::cout<<"Reading: "<<gtNames.count()<<" shapes: "<<std::flush;
	// read original shapes
	VKString origObjDir = corrsDirectory+"/../OBJ_ASCII/";
	std::vector<R3Mesh*> origMeshes;
	for (int i=0; i<gtNames.count(); i++)
	{
		if (i%10==0)
			std::cout<<i<<" "<<std::flush;
		int origid = gtID2OrigID[i];
		VKString fName = origObjDir+VKString::number(origid+1)+".obj";
		assert (FileManagement::FileExists(fName.c_str()));
		origMeshes.push_back(new R3Mesh());
		origMeshes[i]->ReadFile(fName.c_str());
	}
	std::cout<<"\nGround Truth shapes: "<<gtNames.count()<<" / Total: "<<origNames.count()<<std::endl;
	 
	// read normalized shapes
	std::vector<R3Mesh*> normMeshes;
	VKString normObjDir = corrsDirectory+"/../NORMALIZED/";
	std::cout<<"Reading Normalized : "<<std::flush;
	for (int i=0; i<gtNames.count(); i++)
	{
		if (i%10==0)
			std::cout<<i<<" "<<std::flush;
		int origid = gtID2OrigID[i];
		VKString fName = normObjDir+VKString::number(origid+1)+".obj";
		assert (FileManagement::FileExists(fName.c_str()));
		normMeshes.push_back(new R3Mesh());
		normMeshes[i]->ReadFile(fName.c_str());
	}
	std::cout<<std::endl;
	std::cout<<"Reading samples"<<std::endl;
	// read samples (as vertices);
	std::vector<R3Mesh*> sampMeshes;
	VKString sampFilesDir = corrsDirectory+"/../SAMPLE/";
	for (int i=0; i<gtNames.count(); i++)
	{
		//VKString fName = sampFilesDir+VKString::number(i+1)+".txt";
		int origid = gtID2OrigID[i];
		VKString fName = sampFilesDir+VKString::number(origid+1)+".txt";

		assert (FileManagement::FileExists(fName.c_str()));
		sampMeshes.push_back(new R3Mesh());
		std::ifstream textStream(fName.c_str());
		assert(textStream.is_open());
		int id;
		double x, y, z;
		while (!textStream.eof())
		{
			textStream>>id>>x>>y>>z;
			if (textStream.eof())
				break;
			assert(id==sampMeshes[i]->NVertices());
			sampMeshes[i]->CreateVertex(R3Point(x, y, z));
		}
		assert(sampMeshes[i]->NVertices()==512);
		
		textStream.close();
	}

	std::cout<<"Creating search trees: "<<std::flush;
	// create mesh search trees
	std::vector<R3MeshSearchTree*> searchTrees;
	for (int i=0; i<gtNames.count(); i++)
	{
		if (i%10==0)
			std::cout<<i<<" "<<std::flush;
		searchTrees.push_back(new R3MeshSearchTree(normMeshes[i]));
	}
	std::cout<<std::endl;
	
	std::cout<<"Point2mesh projections:"<<std::flush;
	// map samples to points on original mesh
	std::vector<R3Mesh * > samples2origmesh;
	for (int i=0; i<(int)sampMeshes.size(); i++)
	{
		if (i%10==0)
			std::cout<<i<<" "<<std::flush;
		samples2origmesh.push_back(new R3Mesh());
		for (int p=0; p<sampMeshes[i]->NVertices(); p++)
		{
//			std::cout<<"p: "<<p<<" / "<<sampMeshes[i]->NVertices()<<std::endl;
			R3Point correspondence(0,0,0);
			R3Point sp = sampMeshes[i]->VertexPosition(sampMeshes[i]->Vertex(p));
			R3MeshIntersection closest;
			searchTrees[i]->FindClosest(sp, closest);
			assert(closest.type != R3_MESH_NULL_TYPE);
			if (closest.type == R3_MESH_VERTEX_TYPE)
			{
				int v = normMeshes[i]->VertexID(closest.vertex);
				correspondence = origMeshes[i]->VertexPosition(origMeshes[i]->Vertex(v));
			}
			else if (closest.type == R3_MESH_EDGE_TYPE)
			{
				R3MeshFace * face = normMeshes[i]->FaceOnEdge(closest.edge);
				int fid = normMeshes[i]->FaceID(face);
				R3Point b = normMeshes[i]->FaceBarycentric(face, closest.point);
				double bary[3];
				bary[0] = b.X();
				bary[1] = b.Y();
				bary[2] = b.Z();
				correspondence = origMeshes[i]->FacePoint(origMeshes[i]->Face(fid), bary);
			}
			else if (closest.type == R3_MESH_FACE_TYPE)
			{
				R3MeshFace * face = closest.face;
				int fid = normMeshes[i]->FaceID(face);
				R3Point b = normMeshes[i]->FaceBarycentric(face, closest.point);
				double bary[3];
				bary[0] = b.X();
				bary[1] = b.Y();
				bary[2] = b.Z();
				correspondence = origMeshes[i]->FacePoint(origMeshes[i]->Face(fid), bary);
			}
			else assert(false);
			samples2origmesh[i]->CreateVertex(correspondence);
		}
	}
	std::cout<<std::endl;
	std::cout<<"Reading pairwise maps: "<<std::flush;
	// read pairwise maps
	std::vector<std::vector<std::vector<int > > > pairwiseMaps;
	for (int i=0; i<gtNames.count(); i++)
		pairwiseMaps.push_back(std::vector<std::vector<int > >());
	
	for (int i=0; i<(int)sampMeshes.size(); i++)
	for (int j=0; j<(int)sampMeshes.size(); j++)
	{
		if (i%10==0 && j==0)
			std::cout<<i<<" "<<std::flush;
		pairwiseMaps[i].push_back(std::vector<int>());
		
		if (i==j)
			continue;
		
		int origid1 = gtID2OrigID[i];
		int origid2 = gtID2OrigID[j];

		VKString mapFile = corrsDirectory + "/"+VKString::number(origid1+1)+"_"+VKString::number(origid2+1)+".txt";
		assert (FileManagement::FileExists(mapFile.c_str()));
		std::ifstream textStream(mapFile.c_str());
		assert(textStream.is_open());
		
		for (int s=0; s<sampMeshes[i]->NVertices(); s++)
		{
			int s1, s2;
			textStream>>s1>>s2;
			assert(s1==s);
			pairwiseMaps[i][j].push_back(s2);
		}
		textStream.close();
	}
	std::cout<<std::endl;
	
	std::cout<<"Writing pairwiss maps"<<std::endl;
	// write pairwise maps
	FILE * binMapsFile = fopen(outFile.c_str(), "wb");
	assert(binMapsFile!=NULL);
	int nshapes = (int)sampMeshes.size();
	fwrite(&nshapes, sizeof(int), 1, binMapsFile);
	for (int i=0; i<(int)sampMeshes.size(); i++)
	{
		fwrite(&i, sizeof(int), 1, binMapsFile);
		int npnts = samples2origmesh[i]->NVertices();
		fwrite(&npnts, sizeof(int), 1, binMapsFile);
		assert(npnts==512);
		for (int p=0; p < npnts; p++)	// for each point print correspondence
		{
			double xyz1[3];
			R3MeshVertex * s1vtx = samples2origmesh[i]->Vertex(p);
			xyz1[0] = samples2origmesh[i]->VertexPosition(s1vtx).X();
			xyz1[1] = samples2origmesh[i]->VertexPosition(s1vtx).Y();
			xyz1[2] = samples2origmesh[i]->VertexPosition(s1vtx).Z();
			fwrite(xyz1, sizeof(double), 3, binMapsFile);
			fwrite(&nshapes, sizeof(int), 1, binMapsFile);
			for (int j=0; j<nshapes; j++)
			{
				if (i==j)
					fwrite(xyz1, sizeof(double), 3, binMapsFile);
				else
				{
					double xyz2[3];
					int s2 = pairwiseMaps[i][j][p];
					R3MeshVertex * s2vtx = samples2origmesh[j]->Vertex(s2);
					xyz2[0] = samples2origmesh[j]->VertexPosition(s2vtx).X();
					xyz2[1] = samples2origmesh[j]->VertexPosition(s2vtx).Y();
					xyz2[2] = samples2origmesh[j]->VertexPosition(s2vtx).Z();
					fwrite(xyz2, sizeof(double), 3, binMapsFile);
				}
			}
		}
	}
	int checksum = -29349234;
	fwrite(&checksum, sizeof(int), 1, binMapsFile);
	fclose(binMapsFile);
	std::cout<<"Write "<<outFile.c_str()<<" successful!"<<std::endl;
}







