#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"
#include "BasicNumerical.h"

#include "RendererOSMesa.h"
#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "AppScreenshots.h"
#include "VisShapeTemplate.h"

#include "STemplate.h"
#include "TMatch.h"
#include "TMatchEnergy.h"
#include "TMatchAlg.h"
#include "PartOfEdge.h"
#include "ConsistentPartsAlg.h"
#include "STemplateNode.h"

#include "ShapeTriangleSoup.h"

enum ParseDirection
{
	PARSE_My2Simple,
	PARSE_Simple2My
};

TMatchEnergy * ReadMatch(const VKString & mName, const VKString & pntsGeometryFile, const VKString & matchBinaryFile,
						 const VKString & meshGeometryFile, STemplate * st);

/**
 * Given per-point labels (should be accurate), transfer labels to a face. Utilizes CRF by Boykov et al. 2008
 */
int main(int argc, char** argv )
{
	if (argc<5)
	{
		std::cout<<"Usage:"<<std::endl;
		std::cout<<"   "<<argv[0]<<" [options]"<<std::endl;;
		std::cout<<"Optional:"<<std::endl;
		std::cout<<"    -readTmplt templateFile"<<std::endl;
		std::cout<<"    -readMatch matchFile"<<std::endl;
		std::cout<<"    -dbDir modelDir"<<std::endl;
		std::cout<<"    -model modelName"<<std::endl;
		std::cout<<"    -writeTmplt templateFile"<<std::endl;
		std::cout<<"    -writeMatch matchFile"<<std::endl;
		std::cout<<"    -writeCoAligned coAlignedPnts"<<std::endl;
		exit(0);
	}
	
	ParseDirection parse = PARSE_My2Simple;
	VKString readTmplt = "none";
	VKString readMatch = "none";
	VKString dbDir = "none";
	VKString model = "none";
	VKString writeMatch = "none";
	VKString writeTmplt = "none";
	VKString writeCoAligned  = "none";
	
	
	for (int i=1; i<argc; i++)
	{
		if (strcmp(argv[i], "-readTmplt")==0)
			readTmplt = argv[++i];
		else if (strcmp(argv[i], "-readMatch")==0)
			readMatch = argv[++i];
		else if (strcmp(argv[i], "-dbDir")==0)
			dbDir = argv[++i];
		else if (strcmp(argv[i], "-model")==0)
			model = argv[++i];
		else if (strcmp(argv[i], "-writeMatch")==0)
			writeMatch = argv[++i];
		else if (strcmp(argv[i], "-writeTmplt")==0)
			writeTmplt = argv[++i];
		else if (strcmp(argv[i], "-writeCoAligned")==0)
			writeCoAligned = argv[++i];
		else
			std::cout<<"[WARNING] Unknown command: "<<argv[i]<<std::endl;
	}
	
	// variables to fill
	STemplate * st = NULL;
	TMatch * match = NULL;
	
	// read template
	if (readTmplt!="none")
		st = STemplate::ReadTemplate(readTmplt, NULL);
	
	// read match
	if (readMatch!="none")
	{
		assert(st!=NULL);
		VKString pntsGeometryFile = dbDir + "/Models/"+model+"/Geometry/main_Even1000.pnts";
		VKString meshGeometryFile = dbDir + "/Models/"+model+"/Geometry/main.polysoup.off";
		match = ReadMatch(model, pntsGeometryFile, readMatch, meshGeometryFile, st)->GetMatch();
	}
	
	// write template
	if (writeTmplt != "none")
	{
		std::cout<<"Writing template: "<<writeTmplt.c_str()<<std::endl;
		assert(st!=NULL);
		// binary: .simpl.tmplt
		// ascii: .tmplt.txt
		VKString binFile = writeTmplt;
		VKString asciiFile = writeTmplt+".txt";
		
		// write ascii meta-file (part names / template names)
		std::ofstream asciiTmplt(asciiFile.c_str());
		assert(asciiTmplt.is_open());
		// write nparts (total)
		asciiTmplt<<"NParts "<<(st->NNodes()-1)<<"\n";	// exclude root
		
		// write part names
		for (int i=1; i<st->NNodes(); i++)	// exclude root
			asciiTmplt<<st->GetNode(i)->NodeName().c_str()<<" ";
		//		asciiTmplt<<"\n";
		//		// write nconfigs
		//		asciiTmplt<<"NTemplates "<<st->NSeedConfigs()<<"\n";
		//		for (int i=0; i<st->NSeedConfigs(); i++)
		//			st->ConfigName(int configID)
		asciiTmplt.close();
		
		int nconfigs = 1;
		int offset = 0;
		if (st->NSeedConfigs()>1)
		{
			nconfigs = st->NSeedConfigs()-1;
			offset = 1;
		}
		
		FILE * binTmplt = fopen(binFile.c_str(), "wb");
		assert(binTmplt!=NULL);
		
		// write number of configs
		fwrite(&nconfigs, sizeof(int), 1, binTmplt);	// WRITE NCONFIGS
		
		// write each config
		std::vector<int> emptyset;
		double * triplet = new double[9];
		for (int i=0; i<nconfigs; i++)
		{
			int cid = offset + i;
			TMatch * seed = st->CreateSeedConfig(cid, NULL, emptyset, R3identity_affine);
			const LocalGeometryClassifier * classifier = st->GetLocalClassifier(cid);
			//std::cout<<"Classifier: "<<cid<<std::endl;
			
			int nparts = seed->NTemplateNodes()-1;	// exclude root
			fwrite(&nparts, sizeof(int), 1, binTmplt);	// WRITE NPARTS
			
			// exclude root
			for (int p=1; p<nparts+1; p++)
			{
				int nodeid = seed->Deformation(p).NodeID();
				int partid = seed->Deformation(p).PartID();
				int nparts = seed->Deformation(p).NParts();
				assert(partid==0 && nparts == 1);
				fwrite(&nodeid, sizeof(int), 1, binTmplt);	// WRITE PART_ID
				const PartBox & box = seed->NodeEllipsoid(p);
				int type = box.m_BoxType;
				if (!box.IsIncluded())
					type = 0;
				fwrite(&type, sizeof(int), 1, binTmplt);	// WRITE PART TYPE: 0 - not included, 1 - box, 3 - null
				
				//std::cout<<"Part: "<<p<<" Type: "<<type<<std::endl;
				if (box.m_BoxType==PartBox::BOXTYPE_BoxSurface && box.IsIncluded())
				{
					// mean values
					R3Point pnt = box.UndeformedCentroid();
					pnt = box.Deformation()->Orig2Deformed(pnt);
					R3Vector scale = box.UndeformedScale();
					R3Triad orient = box.UndeformedOrientation();
					box.Deformation()->Orig2Deformed(scale);
					box.Deformation()->Orig2Deformed(orient);
					
					// variances
					const STemplateNode * node = st->GetNode(nodeid);
					assert(node!=NULL);
					const PartOfEdge * edge = node->PartOfParent();
					assert(edge!=NULL);
					int anc1, anc2;
					R3Vector svar = node->ScaleVariance(cid, anc1);
					double prob = edge->GetProbNParts(cid)->Prob(nparts);
					R3Vector pvar = edge->PosVariance(cid, nparts, partid, anc2);
					
					// write
					triplet[0]=pnt[0];    triplet[1]=pnt[1];    triplet[2]=pnt[2];
					fwrite(triplet, sizeof(double), 3, binTmplt);	// WRITE MEAN POSITION
					triplet[0]=pvar[0];   triplet[1]=pvar[1];   triplet[2]=pvar[2];
					fwrite(triplet, sizeof(double), 3, binTmplt);	// WRITE VARIANCE IN POS
					
					triplet[0]=scale[0];  triplet[1]=scale[1];  triplet[2]=scale[2];
					fwrite(triplet, sizeof(double), 3, binTmplt);	// WRITE MEAN SCALE
					triplet[0]=svar[0];	  triplet[1]=svar[1];   triplet[2]=svar[2];
					fwrite(triplet, sizeof(double), 3, binTmplt);	// WRITE VARIANCE IN SCALE
					
					triplet[0]=orient[0][0];  triplet[1]=orient[0][1];  triplet[2]=orient[0][2];
					triplet[3]=orient[1][0];  triplet[4]=orient[1][1];  triplet[5]=orient[1][2];
					triplet[6]=orient[2][0];  triplet[7]=orient[2][1];  triplet[8]=orient[2][2];
					fwrite(triplet, sizeof(double), 9, binTmplt);	// WRITE ORIENTATION
					
					//std::cout<<"Gaussian for: "<<p<<" - "<<node->NodeName().c_str()<<std::endl;
					const ProbGaussian * gauss = classifier->GetGaussian(p);
					int ndims = gauss->NDims();
					int nsamp = gauss->NSamples();
					const double * mean = gauss->GetMostProbable();
					const double * vars = gauss->GetVariance();
					
					// WRITE LOCAL SHAPE FEATURES
					fwrite(&ndims, sizeof(int), 1, binTmplt);		// WRITE N DIMENSIONS
					fwrite(&nsamp, sizeof(int), 1, binTmplt);		// WRITE N SAMPLES
					fwrite(mean, sizeof(double), ndims, binTmplt);	// WRITE MEAN FEATURE
					fwrite(vars, sizeof(double), ndims, binTmplt);	// WRITE FEATURE VARIANCE
				}
			}
		}
		
		delete [] triplet;
		fclose(binTmplt);
	}
	
	// write match
	if (writeMatch != "none")
	{
		FILE * binMatch = fopen(writeMatch.c_str(), "wb");
		assert(binMatch!=NULL);
		
		double * triplet = new double[16];
		assert(match!=NULL);
		
		// write tempalte id
		int cid = match->SeedID();
		fwrite(&cid, sizeof(int), 1, binMatch);	// WRITE SEED ID
		
		// write affine transformation
		const R4Matrix & mtx = match->GlobalXForm().Matrix();
		for (int r=0; r<4; r++)
			for (int c=0; c<4; c++)
				triplet[c * 4 + r] = mtx[r][c];
		fwrite(triplet, sizeof(double), 16, binMatch);	// WRITE AFFINE XFORM
		
		// write deformation
		int nparts = match->NTemplateNodes()-1;
		fwrite(&nparts, sizeof(int), 1, binMatch);	// WRITE NPARTS
		
		// exclude root
		for (int p=1; p<nparts+1; p++)
		{
			int nodeid = match->Deformation(p).NodeID();
			int partid = match->Deformation(p).PartID();
			int nparts = match->Deformation(p).NParts();
			assert(partid==0 && nparts == 1);
			fwrite(&nodeid, sizeof(int), 1, binMatch);	// WRITE PART_ID
			const PartBox & box = match->NodeEllipsoid(p);
			
			int type = box.m_BoxType;
			if (!box.IsIncluded())
				type = 0;
			fwrite(&type, sizeof(int), 1, binMatch);	// WRITE PART TYPE: 0 - not included, 1 - box, 3 - null
			
			if (box.m_BoxType==PartBox::BOXTYPE_BoxSurface && box.IsIncluded())
			{
				// mean values
				R3Point pnt = box.UndeformedCentroid();
				pnt = box.Deformation()->Orig2Deformed(pnt);
				R3Vector scale = box.UndeformedScale();
				R3Triad orient = box.UndeformedOrientation();
				box.Deformation()->Orig2Deformed(scale);
				box.Deformation()->Orig2Deformed(orient);
				
				triplet[0]=pnt[0];    triplet[1]=pnt[1];    triplet[2]=pnt[2];
				fwrite(triplet, sizeof(double), 3, binMatch);	// WRITE MEAN POSITION
				triplet[0]=scale[0];  triplet[1]=scale[1];  triplet[2]=scale[2];
				fwrite(triplet, sizeof(double), 3, binMatch);	// WRITE MEAN SCALE
				
				triplet[0]=orient[0][0];  triplet[1]=orient[0][1];  triplet[2]=orient[0][2];
				triplet[3]=orient[1][0];  triplet[4]=orient[1][1];  triplet[5]=orient[1][2];
				triplet[6]=orient[2][0];  triplet[7]=orient[2][1];  triplet[8]=orient[2][2];
				fwrite(triplet, sizeof(double), 9, binMatch);	// WRITE ORIENTATION
			}
		}
		
		// write points
		const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
		assert(pnts!=NULL);
		int npnts = pnts->NumPoints();
		fwrite(&npnts, sizeof(int), 1, binMatch);	// WRITE NPNTS
		for (int pnt=0; pnt < pnts->NumPoints(); pnt++)
		{
			R3Point pos = pnts->PointPosition(pnt);
			triplet[0] = pos.X();   triplet[1] = pos.Y();   triplet[2] = pos.Z();
			fwrite(triplet, sizeof(double), 3, binMatch);	// WRITE POS
			int partID = match->LocalNodeID(pnt);
			fwrite(&partID, sizeof(int), 1, binMatch);	// WRITE LOCAL PART ID (WRT ORDERING OF THIS TEMPLATE)
		}
		
		delete [] triplet;
		fclose(binMatch);
	}
	
	// write co-aligned points
	if (writeCoAligned!="none")
	{
		assert(match!=NULL && st!=NULL);
		
		const ShapePointCloud * matchPnts = match->GetMatchedGeomAsPnts();
		assert(matchPnts!=NULL);
		
		const ShapeMesh * matchMesh = matchPnts->GetMesh();
		assert(matchMesh!=NULL);
		
		// create template samples
		ShapePointCloud * deformedTmplt = new ShapePointCloud(NULL, "deformedTmplt");
		ShapePointCloud * canonicalTmplt = new ShapePointCloud(NULL, "canonicalTmplt");
		
		for (int i=0; i<match->NTemplatePoints(); i++)
		{
			int deformID, locPntID;
			match->LocalPntID(i, deformID, locPntID);
			bool included = match->Deformation(deformID).IsIncluded();
			if (included)
			{
				deformedTmplt->AddPoint(ShapePoint(match->DeformedTemplateSample(i)));
				canonicalTmplt->AddPoint(ShapePoint(st->TemplatePointInCanonicalSpace(match, i)));
			}
		}
		
		const R3Kdtree<const PointInSet *> * deformedKdTree = deformedTmplt->GetKdTree();
		
		FILE * pntsFile = fopen(writeCoAligned.c_str(), "wb");
		assert(pntsFile!=NULL);
		int npnts = matchPnts->NumPoints();
		fwrite(&npnts, sizeof(int), 1, pntsFile);
		R3Point meshCenter = matchMesh->Centroid();
		double meshRadius = matchMesh->AverageRadius(meshCenter);
		for (int i=0; i<matchPnts->NumPoints(); i++)
		{
			// map every point in the match
			R3Point fp = (matchPnts->PointPosition(i)+(-meshCenter)) / meshRadius;
			match->GlobalXForm().Apply(fp);
			const PointInSet * closest = deformedKdTree->FindClosest(fp);
			// deformed template point
			assert(closest!=NULL);
			// vector from template to aligned model in deformed space
			R3Vector offset = fp - closest->GetPosition();
			// same point undeformed
			R3Point pos = canonicalTmplt->PointPosition(closest->id) + offset;
			
			double xyz[3];	xyz[0] = pos.X();	xyz[1] = pos.Y();	xyz[2] = pos.Z();
			fwrite(xyz, sizeof(double), 3, pntsFile);
		}
		fclose(pntsFile);
	}
}



TMatchEnergy * ReadMatch(const VKString & mName, const VKString & pntsFile, const VKString & matchBinaryFile,
						 const VKString & meshFile, STemplate * st)
{
	
	// read geometriee
	//STemplate * st = STemplate::ReadTemplate(readTemplate, NULL);
	Model3D * m = new Model3D(mName, 0);
	ModelDatabase * db = new ModelDatabase();
	db->AddModel(m);
	m->ReadGeometry(GeomShape::SHAPE_TYPE_TRIANGLE_SOUP, Model3D::m_sMainGeomName, meshFile);
	m->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, "main_Even1000.pnts", pntsFile, Model3D::m_sMainGeomName);
	
	// read match
	FILE * mFile = fopen(matchBinaryFile.c_str(), "rb");
	assert(mFile!=NULL);
	
	TMatchEnergy * energy = TMatchEnergy::ReadEnergyAndMatch(mFile, st, db);
	assert(energy!=NULL);
	assert(energy->GetMatch()->GetMatchedGeomAsPnts()!=NULL);
	fclose(mFile);
	return energy;
}





