#ifndef __MRF_OPTIMIZATION_H
#define __MRF_OPTIMIZATION_H

#include <iostream>
#include <assert.h>
#include <vector>
#include <map>
#include <string.h>
#include <math.h>

#include "VKGraph.h"

#include "GCoptimization.h"
class MRFProblemArray;

/**
 * Formulate MRF problem
 */
class MRFProblem
{
	public:
		MRFProblem(int nNodes, int nLabels);
		virtual ~MRFProblem();
		virtual MRFProblemArray * SafeCastToArray();
		
		virtual int NodeCost(int nodeID, int labelID) const = 0;
		virtual int EdgeCost(int nodeID1, int nodeID2, int label1ID, int label2ID) const = 0;
		virtual int NNeighbors(int nodeID) const = 0;
		virtual int Neighbor(int nodeID, int nID, int * weight) const = 0;
		virtual int LabelCost(int labelID) const;	 // returns 0 if not implemented
	
		virtual int NNodes();
		virtual int NLabels();
		
	protected:	
		int m_iNNodes;
		int m_iNLabels;
};

/**
 * A simple MRF problem: matrix of label -> node costs
 */
class MRFProblemArray : public MRFProblem
{
	public:	
		MRFProblemArray(int nNodes, int nLabels, VKGraph * neighborhood, 
						int * dataCosts=NULL, int * labelsCosts=NULL);
		MRFProblemArray(int nNodes, int nLabels, const VKGraph * neighborhood, 
						const double * dataCosts, const double * labelsCosts, double d2int);
		virtual ~MRFProblemArray();
		virtual MRFProblemArray * SafeCastToArray();
		virtual int EdgeCost(int nodeID1, int nodeID2, int label1ID, int label2ID) const;
	
		virtual void SetNewDataCosts(const double * newCosts);
	
		virtual const int * DataCosts() const;
		virtual const int * LabelsCosts() const;

		virtual int NodeCost(int nodeID, int labelID) const;
		virtual int NNeighbors(int nodeID) const;
		virtual int Neighbor(int nodeID, int nID, int * weight) const;
		virtual int LabelCost(int label1ID, int label2ID) const;
		
	protected:
		int * m_aDataCost;
		int * m_aLabelsCost;
		const VKGraph * m_pGraph;
		double m_RescaleDouble2Int;
};

/**
 * Optimizer: solve MRF problem using graph cuts (just a wrapper).
 * See GCO_README.TXT for citations, code license, etc...
 */
class MRFOptimizer
{
	public:
		MRFOptimizer(MRFProblem * problem);
		virtual ~MRFOptimizer();
	
		virtual void Optimize(int nIterations = -1);

		virtual void UpdateDataCosts();
		virtual void SetLabel(int nodeID, int labelID);
		virtual int Label(int nodeID) const;
		virtual int Energy(int * data = NULL, int * smooth = NULL, int * label = NULL) const;
	
	protected:
		MRFProblem * m_pProblem;
		int * m_aNodeLabels;
		GCoptimizationGeneralGraph * m_pGeneralGraphOptim;
};

#endif



