#include "DiffusionMap.h"
#include "Sortable.h"
#include "LinAlg.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <fstream>
#include <iostream>


DiffusionMap::DiffusionMap(int N, int nEigs, double diffTime,
						   const std::vector<int> & numPointsInGroup,
						   DiffusionMap::DiffusedSearchType srch,
						   SparseMtxDataTypes decomp)
{
	m_SearchType = srch;
	m_MtxLib = decomp;
	
	m_iN = N;	
	std::vector<int> groupCopy = numPointsInGroup;
	if (groupCopy.size()==0)
		groupCopy.push_back(m_iN);
	
	m_iNGroups = (int)groupCopy.size();
	m_aPointToGroup = new int[m_iN];
	m_aGroupToOffset = new int[m_iNGroups];
	m_aGroupToNumPoints = new int[m_iNGroups];
	int pntCounter=0;
	for (int i=0; i<(int)groupCopy.size(); i++)
	{
		m_aGroupToNumPoints[i] = groupCopy[i];
		m_aGroupToOffset[i] = pntCounter;
		for (int p=pntCounter; p<pntCounter+m_aGroupToNumPoints[i]; p++)
			m_aPointToGroup[p] = i;
			 
		pntCounter+=m_aGroupToNumPoints[i];
	}
	if (pntCounter!=m_iN)
		std::cout<<"[ERROR] pntCounter="<<pntCounter<<" != "<<m_iN<<std::endl;
	assert(pntCounter==m_iN);

	m_fDiffTime = diffTime;	
	m_iNEigs = nEigs;
	
	m_pCorrVals = NULL;
	m_pEigenVectors = NULL;
	m_pEigenValues = NULL;
		
#ifdef INCLUDE_FLANN
	m_pEmbeddedSpaceANN = NULL;
	m_pANNDataset = NULL;
	m_pANNOutIndices = NULL;
	m_pANNOutDistances = NULL;
	m_pANNQuery = NULL;
	
	m_pEmbeddedSpaceANNPerGroup = new flann::Index<flann::L2<double> > *[m_iNGroups];
	m_pANNDatasetPerGroup = new flann::Matrix<double> *[m_iNGroups];
	for (int i=0; i<m_iNGroups; i++)
	{
		m_pEmbeddedSpaceANNPerGroup[i] = NULL;
		m_pANNDatasetPerGroup[i] = NULL;
	}
#endif
	
	m_aQueryFeatures = NULL;
	m_aClosestPntIDs = NULL;
	m_aClosestDistances = NULL;

	FillInputOutputPtrs();
}
DiffusionMap::~DiffusionMap()
{
	FreeData();
}

int DiffusionMap::NEntries() const
{
	return m_iN;
}

int DiffusionMap::NEigs() const
{
	return m_iNEigs;
}

void DiffusionMap::FreeData()
{
	if (m_pCorrVals!=NULL)
		delete m_pCorrVals;
	
	if (m_pEigenVectors!=NULL)
		delete m_pEigenVectors;

	if (m_pEigenValues!=NULL)
		delete m_pEigenValues;
	
	if (m_aQueryFeatures!=NULL)
		delete m_aQueryFeatures;
		
	if (m_aPointToGroup!=NULL)
		delete [] m_aPointToGroup;

	if (m_aGroupToOffset!=NULL)
		delete [] m_aGroupToOffset;
	
	if (m_aGroupToNumPoints!=NULL)
		delete [] m_aGroupToNumPoints;
		
	ClearNNCache();
#ifdef INCLUDE_FLANN	
	if (m_pEmbeddedSpaceANNPerGroup!=NULL)
		delete [] m_pEmbeddedSpaceANNPerGroup;
	if (m_pANNDatasetPerGroup!=NULL)
		delete [] m_pANNDatasetPerGroup;
#endif
}

void DiffusionMap::AddCorrValue(int i, int j, double val)
{
	assert(m_MtxLib==SPARSE_MTX_GENERAL);
	if (m_pCorrVals==NULL)
		m_pCorrVals = new LinAlgMatrixSparseRealGeneral(m_iN, m_iN);
	
	m_pCorrVals->AddVal(i, j, val);
}

void DiffusionMap::ClearAllCorrValues()
{
	if (m_pCorrVals!=NULL)
	{
		delete m_pCorrVals;
		m_pCorrVals = NULL;
	}
}

void DiffusionMap::Embed(bool addIdentities, bool symmetrize, 
						 bool rowNormalize, bool normalizeEigenvectors,
						 double * timeInSec, const char * prefix)
{
	assert(m_pEigenVectors==NULL && m_pEigenValues==NULL);
	assert(m_pCorrVals!=NULL);
	SparseMtxDataTypes tp = m_pCorrVals->GetMatrixDataType();
	assert(tp==SPARSE_MTX_GENERAL && m_MtxLib==SPARSE_MTX_GENERAL);
			// correspondence matrix
	LinAlgMatrixSparseRealGeneral * C = (LinAlgMatrixSparseRealGeneral*)m_pCorrVals;
			// processed correspondence matrix (row-normalized, etc...)
	LinAlgMatrixSparseRealGeneral * S = new LinAlgMatrixSparseRealGeneral(*C);
	
	// prepare corrs matrix
		// add identities
	if (addIdentities)
	{
		//double maxVal = S->GetMax();
		for (int i=0; i<m_iN; i++)
		{
			if (S->GetVal(i, i)==0)
				S->AddVal(i, i, 1.);
		}	
	}
		// symmetrize: S = .5 * (S + S')
	if (symmetrize)
		S->Symmetrize();

		// row-normalize (temporary store for normalization of eigenvectors)
	LinAlgVectorReal * rn = new LinAlgVectorReal(m_iN, rowNormalize ? 0. : 1.);
	if (rowNormalize)
		S->RowNormalize(rn);
	
	// Find eigenvectors / eigenvalues
	m_pEigenVectors = new LinAlgMatrixReal(m_iN, m_iNEigs);
	m_pEigenValues = new LinAlgVectorReal(m_iNEigs);

	S->Eigs(*m_pEigenValues, *m_pEigenVectors, m_iNEigs, prefix, timeInSec);

	// Normalization of eigenvectors (as Nader et al.)
	if (normalizeEigenvectors)
	{
		for (int i=0; i<m_pEigenVectors->Cols(); i++)
		{
			double norm = pow((*rn)(i), -.5);
			for (int j=0; j<m_pEigenVectors->Rows(); j++)
				(*m_pEigenVectors)(j,i) *= norm;
		}	
	}
	
	// clear temp data
	delete S;
	delete rn;
}

void DiffusionMap::SetTime(double t)
{
	if (m_fDiffTime!=t)
	{
		m_fDiffTime = t;
		ClearNNCache();
	}
}

void DiffusionMap::SetSearchType(DiffusedSearchType srchTp)
{
	if (m_SearchType!=srchTp)
	{
		m_SearchType = srchTp;
		ClearNNCache();
	}
}

double DiffusionMap::DiffTime() const
{
	return m_fDiffTime;
}

void DiffusionMap::ClearNNCache()
{
#ifdef INCLUDE_FLANN
	if (m_pEmbeddedSpaceANN!=NULL)
	{
		delete m_pEmbeddedSpaceANN;
		m_pEmbeddedSpaceANN = NULL;
	}
	
	for (int i=0; i<m_iNGroups; i++)
	{
		if (m_pEmbeddedSpaceANNPerGroup[i]!=NULL)
		{
			delete m_pEmbeddedSpaceANNPerGroup[i];
			m_pEmbeddedSpaceANNPerGroup[i] = NULL;
		}
		if (m_pANNDatasetPerGroup[i]!=NULL)
		{
			delete m_pANNDatasetPerGroup[i];
			m_pANNDatasetPerGroup[i] = NULL;
		}
	}
#endif
}


void DiffusionMap::Eigs(LinAlgMatrixReal ** eigVecs, LinAlgVectorReal ** eigVals)
{
	*eigVals = m_pEigenValues;
	*eigVecs = m_pEigenVectors;
}
void DiffusionMap::Corrs(LinAlgMatrixSparseReal ** corrs)
{
	*corrs = m_pCorrVals;
}

double DiffusionMap::ComputeExactDistance(int i, int j) const
{
	double evalNorm = (*m_pEigenValues)(1);
	double dist = 0;
	for (int e=1; e<m_pEigenValues->Rows(); e++)	// top N eigenvectors
	{
		double eigVal = pow((*m_pEigenValues)(e)/evalNorm, m_fDiffTime);
		double e1 = (*m_pEigenVectors)(i, e);
		double e2 = (*m_pEigenVectors)(j, e);
		dist += pow(eigVal * (e1-e2), 2);
	}
	return sqrt(dist);
}

void DiffusionMap::ClosestPntsKNN(int pntID, int numNN,
								  int ** closestIDs, double ** closestDists, int * nPoints)
{
	assert(closestIDs!=NULL && closestDists!=NULL && nPoints!=NULL);
	
	// slow approach: examine every pair - calculate distances - sort
	if (m_SearchType==SEARCH_EXHAUSTIVE)
	{
		assert(false);
	}
#ifdef INCLUDE_FLANN
	else if (m_SearchType==SEARCH_USE_FLANN)	// fast approach: use FLANN
	{
		BuildFLANNIndex();
		
		assert(m_pEmbeddedSpaceANN!=NULL && m_pANNQuery!=NULL && m_pANNDataset!=NULL 
			   && m_pANNOutIndices!=NULL && m_pANNOutDistances!=NULL);
		
		// fill m_pANNQuery
		for (int e=0; e<m_iNEigs; e++)
			(*m_pANNQuery)[0][e] = (*m_pANNDataset)[pntID][e];

		assert(numNN <= m_iN);
		// find nearest neighbors (can set FLANN_CHECKS_UNLIMITED)
		// if approximate: set flann::KDTreeIndexParams(4..16), flann::SearchParams(numNN*10(?))
		flann::Index<flann::L2<double> > * nnData = m_pEmbeddedSpaceANN;
		nnData->knnSearch(*m_pANNQuery, *m_pANNOutIndices, *m_pANNOutDistances, 
						  numNN, flann::SearchParams(128));
		*nPoints = numNN;
		// note: m_aClosestPntIDs and m_aClosestDistances correspond to 
		//			m_pANNOutIndices and m_pANNOutDistances
		(*closestIDs) = m_aClosestPntIDs;
		(*closestDists) = m_aClosestDistances;
	}
#endif
	else
		assert(false);	

}

void DiffusionMap::ClosestPntsInGroupKNN(int pntID, int numNN, int groupID, 
										 int ** closestIDs, double ** closestDists, int * nPoints)
{
	assert(closestIDs!=NULL && closestDists!=NULL && nPoints!=NULL);
	if (m_SearchType==SEARCH_EXHAUSTIVE)
	{
		assert(false);	// todo...
	}
#ifdef INCLUDE_FLANN
	else if (m_SearchType==SEARCH_USE_FLANN)	// fast approach: use FLANN
	{
		int srcGroupID = m_aPointToGroup[pntID];
		BuildFLANNForGroup(groupID);
		BuildFLANNForGroup(srcGroupID);
		
		assert(m_pEmbeddedSpaceANNPerGroup[srcGroupID]!=NULL
			   && m_pEmbeddedSpaceANNPerGroup[groupID]!=NULL
			   && m_pANNDatasetPerGroup[srcGroupID]!=NULL
			   && m_pANNDatasetPerGroup[groupID]!=NULL 
			   && m_pANNOutIndices!=NULL && m_pANNOutDistances!=NULL && m_pANNQuery!=NULL);
		
		assert(numNN <= m_iN);
		int srcElID = pntID - m_aGroupToOffset[srcGroupID];
		assert(srcElID>=0);
		
		// fill m_pANNQuery
		for (int e=0; e<m_iNEigs; e++)
		{
			double fv = (*m_pANNDatasetPerGroup[srcGroupID])[srcElID][e];
			(*m_pANNQuery)[0][e] = fv;
		}
		int nInGroup = m_aGroupToNumPoints[groupID];
		assert(numNN <= nInGroup);
		
		// find nearest neighbors (can set FLANN_CHECKS_UNLIMITED)
		// if approximate: set flann::KDTreeIndexParams(4..16), flann::SearchParams(numNN*10(?))
		flann::Index<flann::L2<double> > * nnData = m_pEmbeddedSpaceANNPerGroup[groupID];
		nnData->knnSearch(*m_pANNQuery, *m_pANNOutIndices, *m_pANNOutDistances, numNN,	
						  flann::SearchParams(128));
		*nPoints = numNN;
		// note: m_aClosestPntIDs and m_aClosestDistances correspond to 
		//			m_pANNOutIndices and m_pANNOutDistances
		(*closestIDs) = m_aClosestPntIDs;
		(*closestDists) = m_aClosestDistances;

	}
#endif
	else
		assert(false);	
}

void DiffusionMap::FillInputOutputPtrs()
{
	if (m_aQueryFeatures==NULL)
		m_aQueryFeatures = new double[m_iNEigs];
	
	if (m_aClosestPntIDs==NULL)
		m_aClosestPntIDs = new int[m_iN];

	if (m_aClosestDistances==NULL)
		m_aClosestDistances = new double[m_iN];
}

#ifdef INCLUDE_FLANN
void DiffusionMap::FillInputOutputPtrsFLANN()
{	
	assert(m_aClosestPntIDs!=NULL && m_aClosestDistances!=NULL && m_aQueryFeatures!=NULL);
	if (m_pANNOutIndices == NULL)
		m_pANNOutIndices = new flann::Matrix<int>(m_aClosestPntIDs, 1, m_iN);
	
	if (m_pANNOutDistances == NULL)
		m_pANNOutDistances = new flann::Matrix<double>(m_aClosestDistances, 1, m_iN);
	
	if (m_pANNQuery==NULL)
		m_pANNQuery = new flann::Matrix<double>(m_aQueryFeatures, 1, m_iNEigs);
}

void DiffusionMap::BuildFLANNIndex()
{
	if (m_pEmbeddedSpaceANN==NULL)	// construct ANN index
	{
		FillInputOutputPtrsFLANN();
		
		if (m_pANNDataset == NULL)
			m_pANNDataset = new flann::Matrix<double>(new double[m_iN*m_iNEigs], 
													  m_pEigenVectors->Rows(), 
													  m_pEigenVectors->Cols());

		for (int i=0; i<m_pEigenVectors->Rows(); i++)
			for (int j=0; j<m_pEigenVectors->Cols(); j++)
			{
				double evalT = pow((*m_pEigenValues)(j), m_fDiffTime);
				double evec = (*m_pEigenVectors)(i,j);
				(*m_pANNDataset)[i][j] = evalT * evec;
			}
		
		m_pEmbeddedSpaceANN = new flann::Index<flann::L2<double> >(*m_pANNDataset, 
																   flann::KDTreeIndexParams(1));
		m_pEmbeddedSpaceANN->buildIndex();
	}
}

void DiffusionMap::BuildFLANNForGroup(int groupID)
{
	if (m_pEmbeddedSpaceANNPerGroup[groupID]==NULL)	// construct ANN index
	{
		FillInputOutputPtrsFLANN();
		
		if (m_pANNDatasetPerGroup[groupID] == NULL)
		{
			int nEls = m_aGroupToNumPoints[groupID];
			int nFeats = m_pEigenVectors->Cols();
			assert(m_iNEigs==nFeats);
			double * vals = new double[nEls*nFeats];
			m_pANNDatasetPerGroup[groupID] = new flann::Matrix<double>(vals, nEls, nFeats);
		}
		
		for (int i=0; i<m_aGroupToNumPoints[groupID]; i++)
		for (int j=0; j<m_pEigenVectors->Cols(); j++)
		{
			int nID = m_aGroupToOffset[groupID] + i;
			double evalT = pow((*m_pEigenValues)(j), m_fDiffTime);
			double evec = (*m_pEigenVectors)(nID, j);
			(*(m_pANNDatasetPerGroup[groupID]))[i][j] = evalT * evec;
		}
		
		flann::Matrix<double> & data = *(m_pANNDatasetPerGroup[groupID]);
		m_pEmbeddedSpaceANNPerGroup[groupID] = new flann::Index<flann::L2<double> >(data, 
																flann::KDTreeIndexParams(1));
		m_pEmbeddedSpaceANNPerGroup[groupID]->buildIndex();
	}

}
#endif

void DiffusionMap::WriteBinaryData(FILE * file) const
{
//	// to save: input params
//	int m_iN;
//	int m_iNEigs;
//	double m_fDiffTime;
//	SparseMtxDataTypes m_MtxLib;
//	DiffusedSearchType m_SearchType;	
//	
//	// to save: correspondence values
//	LinAlgMatrixSparseReal * m_pCorrVals;
//	
//	// to save: eigen-decomposition (embedded values)
//	LinAlgMatrixReal * m_pEigenVectors;
//	LinAlgVectorReal * m_pEigenValues;
//	
//	// to save: grouping (e.g. points per model)	
//	int m_iNGroups;
//	int * m_aPointToGroup;
//	int * m_aGroupToOffset;
//	int * m_aGroupToNumPoints;
	assert(false);
}

void DiffusionMap::ReadBinaryData(FILE * file) const
{
	assert(false);
}

