/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"
#include "VKStringList.h"

#include "CollectionPairwise.h"
#include "Model3D.h"
#include "DBCorrs.h"
#include "CorrsCreator.h"
#include "Sortable.h"


CollectionPairwise::CollectionPairwise(const VKString & pairwiseCollName)
{
	m_sPairwiseCollName = pairwiseCollName;
	m_pDB = NULL;
}

CollectionPairwise::~CollectionPairwise()
{
	
}

const VKString & CollectionPairwise::CollectioName() const
{
	return m_sPairwiseCollName;
}

int CollectionPairwise::AddPairCorrs(CorrsPair * pairCorr)
{
	int id = (int)m_PairwiseCorrs.size();
	
	GeomShape * shapeFrom = pairCorr->ShapeFrom();
	assert(shapeFrom!=NULL);
	const Model3D * modelFrom = shapeFrom->GetModel();
	assert(modelFrom!=NULL);
	const VKString & fromStr = modelFrom->ModelName();

	const GeomShape * shapeTo = pairCorr->ShapeTo();
	assert(shapeTo!=NULL);
	const Model3D * modelTo = shapeTo->GetModel();
	assert(modelTo!=NULL);
	const VKString & toStr = modelTo->ModelName();

	m_FromModelToCorrs[fromStr].push_back(id);
	m_ToModelToCorrs[toStr].push_back(id);
	m_PairToCorrs[fromStr][toStr].push_back(id);
	
	m_PairwiseCorrs.push_back(pairCorr);
	
	return id;
}

int CollectionPairwise::NumCorrs() const
{
	return (int)m_PairwiseCorrs.size();
}

int CollectionPairwise::NumCorrsFrom(const VKString & modelFrom) const
{
	return (int)GetCorrsFrom(modelFrom).size();

}

int CollectionPairwise::NumCorrsTo(const VKString & modelTo) const
{
	return (int)GetCorrsTo(modelTo).size();
}

int CollectionPairwise::NumCorrsForPair(const VKString & modelFrom, 
										const VKString & modelTo) const
{
	return (int)GetCorrPairs(modelFrom, modelTo).size();
}

CorrsPair * CollectionPairwise::GetCorr(int pairID)
{
	assert(pairID >=0 && pairID < (int)m_PairwiseCorrs.size());
	return m_PairwiseCorrs[pairID];
}

const CorrsPair * CollectionPairwise::GetCorr(int pairID) const
{
	assert(pairID >=0 && pairID < (int)m_PairwiseCorrs.size());
	return m_PairwiseCorrs[pairID];	
}
 
const std::vector<int> & CollectionPairwise::GetCorrsFrom(const VKString & modelFrom) const
{
	std::map<VKString, std::vector<int> >::iterator iter;
	iter = ((CollectionPairwise*)this)->m_FromModelToCorrs.find(modelFrom);
	if (iter!=m_FromModelToCorrs.end())
		return iter->second;
	return m_EmptyCorrs;
}

const std::vector<int> & CollectionPairwise::GetCorrsTo(const VKString & modelTo) const
{
	std::map<VKString, std::vector<int> >::iterator iter;
	iter = ((CollectionPairwise*)this)->m_ToModelToCorrs.find(modelTo);
	if (iter!=m_ToModelToCorrs.end())
		return iter->second;
	return m_EmptyCorrs;
}

const std::vector<int> & CollectionPairwise::GetCorrPairs(const VKString & modelFrom, 
														  const VKString & modelTo) const
{
	std::map<VKString, std::map<VKString, std::vector<int> > >::iterator iter1;
	iter1 = ((CollectionPairwise*)this)->m_PairToCorrs.find(modelFrom);
	if(iter1==m_PairToCorrs.end())
		return m_EmptyCorrs;	
	std::map<VKString, std::vector<int> >::iterator iter2;
	iter2 = iter1->second.find(modelTo);
	if (iter2 == iter1->second.end())
		return m_EmptyCorrs;
	
	return iter2->second;
}

void CollectionPairwise::SortMapsForEachPair(const VKString & evalPnts, bool beVerbouse)
{
	int counter = 0;
	int nCorrs = NumCorrs();
	TimeProfiler timer;
	std::map<VKString, std::map<VKString, std::vector<int> > >::iterator iter1;
	for (iter1 = m_PairToCorrs.begin(); iter1 != m_PairToCorrs.end(); iter1++)
	{
		std::map<VKString, std::vector<int> >::iterator iter2;
		for (iter2 = iter1->second.begin(); iter2 != iter1->second.end(); iter2++)
		{
			std::vector<Sortable> sortableMaps;
			for (int i=0; i<(int)iter2->second.size(); i++)
			{
				int mID = iter2->second[i];
				const CorrsPair * corrs = GetCorr(mID);
				double mVal = corrs->DefaultIntegratedValue(evalPnts);
				sortableMaps.push_back(Sortable(-mVal, NULL, mID));
				counter++;
				if (beVerbouse)
					timer.WriteProgress("sortmaps", counter, nCorrs);
			}
			std::sort(sortableMaps.begin(), sortableMaps.end());
			iter2->second.clear();
			for (int i=0; i<(int)sortableMaps.size(); i++)
				iter2->second.push_back(sortableMaps[i].id);
		}
	}

}

void CollectionPairwise::GetBestMapForEachPair(std::vector<int> & mapIDs) const
{
	std::map<VKString, std::map<VKString, std::vector<int> > >::iterator iter1;
	for (iter1 = ((CollectionPairwise*)this)->m_PairToCorrs.begin();
		 iter1 != ((CollectionPairwise*)this)->m_PairToCorrs.end(); iter1++)
	{
		std::map<VKString, std::vector<int> >::iterator iter2;
		for (iter2 = iter1->second.begin(); iter2 != iter1->second.end(); iter2++)
		{
			int bestMapID = -1;
			double bestMapValue = -1;
			
			for (int i=0; i<(int)iter2->second.size(); i++)
			{
				int mID = iter2->second[i];
				const CorrsPair * corrs = GetCorr(mID);
				double mVal = corrs->DefaultIntegratedValue();
				if (mVal > bestMapValue)
				{
					bestMapID = mID;
					bestMapValue = mVal;
				}
			}
			assert(bestMapID>=0);
			mapIDs.push_back(bestMapID);
		}
	}
}

void CollectionPairwise::WriteConcatenated(const VKString & filename,
										   ShapeCorrs::CoreCorrsType writeFormat)
{
	FILE * file = fopen(filename.c_str(), "wb");
	assert(file!=NULL);
	WriteConcatenatedHeader(file, writeFormat);
	WriteMaps(file, writeFormat);	
	fclose(file);
}

void CollectionPairwise::WriteConcatenatedHeader(FILE * file,
												 ShapeCorrs::CoreCorrsType writeFormat)
{
	int nGeoms;
	VKStringList modelNames; 
	VKStringList geomNames;
	int totalNMaps = NumCorrs();
	
	std::vector<int> nMapsPerType;
	VKStringList corrNames;
	VKStringList classNames;
	
	std::vector<int> fromToPairs;
	
	std::map<VKString, int> onlyModelNameToID;
	std::map<VKString, int> modelNameToID;
	VKString prevClassName="asfsdf";
	VKString prevCorrName="84nhvkiasdj";
	// go over all maps to fill data above (to be written)
	for (int i=0; i<NumCorrs(); i++)
	{
		CorrsPair * corr = GetCorr(i);
		
		// get geometry names (model + geometry)
		VKString g1Name = corr->ShapeFrom()->GeometryName();
		VKString g2Name = corr->ShapeTo()->GeometryName();
		VKString m1Name = corr->ShapeFrom()->GetModel()->ModelName();
		VKString m2Name = corr->ShapeTo()->GetModel()->ModelName();
		
		VKString str1 = m1Name + "&" + g1Name;
		VKString str2 = m2Name + "&" + g2Name;

		// see if geometry has IDs
		std::map<VKString, int>::iterator iter1 = modelNameToID.find(str1);
		std::map<VKString, int>::iterator iter2 = modelNameToID.find(str2);
		
		int id1, id2;
		
			// add new if id is not assigned
		if (iter1 == modelNameToID.end())
		{
			onlyModelNameToID[m1Name] = modelNames.count();
			modelNameToID[str1] = modelNames.count();
			id1 = modelNames.count();
			modelNames.push_back(m1Name);
			geomNames.push_back(g1Name);
		}
		else
			id1 = iter1->second;
		
			// add new if id is not assigned
		if (iter2 == modelNameToID.end())
		{
			onlyModelNameToID[m2Name] = modelNames.count();
			modelNameToID[str2] = modelNames.count();
			id2 = modelNames.count();
			modelNames.push_back(m2Name);
			geomNames.push_back(g2Name);
		}
		else
			id2 = iter2->second;

		// save pair IDs
		fromToPairs.push_back(id1);
		fromToPairs.push_back(id2);

		VKString corrName = corr->GetCorrsName();
		VKString className = "unknown";
		if (writeFormat==ShapeCorrs::CORECORRS_CorrsPair)
			className = "CorrsPair";
		else if (writeFormat==ShapeCorrs::CORECORRS_CorrsPairMap)
			className = "CorrsPairMap";
		else if (writeFormat==ShapeCorrs::CORECORRS_CorrsOther)
			className = corr->GetCorrsClassName();
		assert(className!="unknown");
		
		if (corrName != prevCorrName || className != prevClassName)
		{
			prevCorrName = corrName;
			prevClassName = className;
			nMapsPerType.push_back(1);
			classNames.push_back(className);
			corrNames.push_back(corrName);
		}
		else
		{
			assert(nMapsPerType.size() > 0);
			nMapsPerType[(int)nMapsPerType.size()-1]++;
		}
	}
	
	// write lists: models, maps, etc...
	nGeoms = modelNames.count();
	
	assert(modelNames.count()==geomNames.count() && nGeoms==geomNames.count());
	assert(corrNames.count()==classNames.count() 
		   && corrNames.count()==(int)nMapsPerType.size());
	assert((2*totalNMaps)==(int)fromToPairs.size());
	
	fwrite(&(DBCorrs::s_iCorrsCollHeader), sizeof(unsigned int), 1, file);
	fwrite(&nGeoms, sizeof(int), 1, file);
	VKStringList::WriteList(file, modelNames);
	VKStringList::WriteList(file, geomNames);
	
	fwrite(&totalNMaps, sizeof(int), 1, file);
	VKStringList::WriteList(file, nMapsPerType);
	VKStringList::WriteList(file, corrNames);
	VKStringList::WriteList(file, classNames);
	VKStringList::WriteList(file, fromToPairs);
	
	// write sorting
	std::map<VKString, std::map<VKString, std::vector<int> > >::iterator iter1; 
	int size1 = (int)m_PairToCorrs.size();
	fwrite(&size1, sizeof(int), 1, file);
	for (iter1 = m_PairToCorrs.begin(); iter1 != m_PairToCorrs.end(); iter1++)
	{		
		std::map<VKString, std::vector<int> >::iterator iter2;
		int size2 = (int)iter1->second.size();
		fwrite(&size2, sizeof(int), 1, file);
		for (iter2 = iter1->second.begin(); iter2 != iter1->second.end(); iter2++)
		{
			std::map<VKString, int>::iterator iterID1 = onlyModelNameToID.find(iter1->first);
			std::map<VKString, int>::iterator iterID2 = onlyModelNameToID.find(iter2->first);
			
			assert(iterID1!=onlyModelNameToID.end() && iterID2 != onlyModelNameToID.end());
			
			int id1 = iterID1->second;
			int id2 = iterID2->second;
			int nCorrs = iter2->second.size();
			int * sortedOrder = new int[nCorrs];
			for (int i=0; i<nCorrs; i++)
				sortedOrder[i] = iter2->second[i];
			fwrite(&id1, sizeof(int), 1, file);
			fwrite(&id2, sizeof(int), 1, file);
			fwrite(&nCorrs, sizeof(int), 1, file);
			fwrite(sortedOrder, sizeof(int), nCorrs, file);
			fflush(file);
			delete sortedOrder;
		}
	}
	
	fflush(file);
}

void CollectionPairwise::WriteMaps(FILE * file, ShapeCorrs::CoreCorrsType writeFormat)
{
	if (writeFormat==ShapeCorrs::CORECORRS_CorrsPair)
	{
		for (int i=0; i<NumCorrs(); i++)
		{
			GetCorr(i)->WriteShapeCorrsAsCorrsPair(file);
			fflush(file);
		}
	}
	else if (writeFormat==ShapeCorrs::CORECORRS_CorrsPairMap)
	{
		std::cout<<"[ERROR] Cannot write CorrsPairMap (todo): "<<writeFormat<<std::endl;
		assert(false);
	}
	else if (writeFormat==ShapeCorrs::CORECORRS_CorrsOther)
	{
		for (int i=0; i<NumCorrs(); i++)
		{
			GetCorr(i)->WriteShapeCorrs(file);		
			fflush(file);
		}
	}
	else
	{
		std::cout<<"[ERROR] Unknown corrs format: "<<writeFormat<<std::endl;
		assert(false);
	}
}

CollectionPairwise * CollectionPairwise::ReadConcatenated(const VKString & corrName,
														  ModelDatabase * db, 
														  const VKString & filename,
														  DBCorrs * dbCorrs)
{
	CollectionPairwise * collection = new CollectionPairwise(corrName);

	int nGeoms, totalNMaps;
	VKStringList modelNames;
	VKStringList geomNames;
	VKStringList corrNames;
	VKStringList classNames;
	
	std::vector<int> nMapsPerType;
	std::vector<int> fromToPairs;
	std::map<VKString, std::map<VKString, std::vector<int> > > pairsToCorrs;
	
	FILE * file = ReadConcatenatedHeader(filename, nGeoms, modelNames, geomNames,
										 totalNMaps, nMapsPerType, corrNames,
										 classNames, fromToPairs, pairsToCorrs);
	int currTypeID = 0;
	assert((2*totalNMaps)==(int)fromToPairs.size());
	assert(corrNames.count()==classNames.count()
		   && (int)nMapsPerType.size() == classNames.count());
	assert(modelNames.count()==geomNames.count());
	// read every map
	for (int i=0; i<totalNMaps; i++)
	{
		int id1 = fromToPairs[2*i+0];
		int id2 = fromToPairs[2*i+1];
		
		assert(id1>=0 && id1<modelNames.count());
		assert(id2>=0 && id2<modelNames.count());
		VKString m1Name = modelNames[id1];
		VKString m2Name = modelNames[id2];
		VKString g1Name = geomNames[id1];
		VKString g2Name = geomNames[id2];
		
		assert(currTypeID >=0 && currTypeID < classNames.count());
		VKString className = classNames[currTypeID];
		VKString corrName = corrNames[currTypeID];
		
		nMapsPerType[currTypeID]--;
		if (nMapsPerType[currTypeID]==0)
			currTypeID++;
		
		Model3D * m1 = db->GetModel(m1Name);
		Model3D * m2 = db->GetModel(m2Name);
		assert(m1!=NULL && m2!=NULL);
		GeomShape * g1 = m1->GetGeomShape(g1Name);
		GeomShape * g2 = m2->GetGeomShape(g2Name);
		assert(g1!=NULL && g2!=NULL);
		
		CorrsPair * pairCorr;
		CorrsCreator * reader = CorrsCreator::GetCreator(className);
		assert(reader!=NULL);
		if (reader->ClassType()==ShapeCorrs::CORECORRS_CorrsPair)
			collection->AddPairCorrs(reader->ReadCorrsPair(corrName, g1, g2, file, dbCorrs));
		else if (reader->ClassType()==ShapeCorrs::CORECORRS_CorrsPairMap)
			collection->AddPairCorrs(reader->ReadCorrsPairMap(corrName, g1, g2, file, dbCorrs));
		else
		{
			std::cout<<"[ERROR] Wrong class type: "<<className.c_str()<<std::endl;
			assert(false);
		}
	}
	
	// re-sort
	collection->m_PairToCorrs = pairsToCorrs;
	
	return collection;
}


FILE * CollectionPairwise::ReadConcatenatedHeader(const VKString & filename,  
												  int & nGeoms,
												  VKStringList & modelNames, 
												  VKStringList & geomNames,
												  int & totalNMaps,
												  std::vector<int> & nMapsPerType,
												  VKStringList & corrNames,
												  VKStringList & classNames,
												  std::vector<int> & fromToPairs,
						std::map<VKString, std::map<VKString, std::vector<int> > > & pairsToCorrsSorted)
{
	FILE * file = fopen(filename.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot read: "<<filename.c_str()<<std::endl;
	assert(file!=NULL);
	unsigned int header;
	fread(&header, sizeof(unsigned int), 1, file);
	assert(header == DBCorrs::s_iCorrsCollHeader);
	
	fread(&nGeoms, sizeof(int), 1, file);
	VKStringList::ReadList(file, modelNames);
	VKStringList::ReadList(file, geomNames);
	assert(nGeoms==modelNames.count() 
		   && nGeoms==geomNames.count());
	fread(&totalNMaps, sizeof(int), 1, file);
	VKStringList::ReadList(file, nMapsPerType);
	VKStringList::ReadList(file, corrNames);
	VKStringList::ReadList(file, classNames);
	VKStringList::ReadList(file, fromToPairs);
	
	assert((2*totalNMaps)==(int)fromToPairs.size());
	assert((int)nMapsPerType.size()==corrNames.count());
	assert((int)nMapsPerType.size()==classNames.count());
	
	// read sorting
	int size1 = 0;
	int size2 = 0;
	int fromTo[2];
	fread(&size1, sizeof(int), 1, file);
	for (int i=0; i<size1; i++)
	{
		fread(&size2, sizeof(int), 1, file);
		for (int j=0; j<size2; j++)
		{
			fread(fromTo, sizeof(int), 2, file);
			assert(fromTo[0] < modelNames.count());
			assert(fromTo[1] < modelNames.count());
			const VKString & fromStr = modelNames[fromTo[0]];
			const VKString & toStr = modelNames[fromTo[1]];
			int nCorrs;
			fread(&nCorrs, sizeof(int), 1, file);
			int * corrs = new int[nCorrs];
			fread(corrs, sizeof(int), nCorrs, file);
			for (int k=0; k<nCorrs; k++)
				pairsToCorrsSorted[fromStr][toStr].push_back(corrs[k]);
			delete corrs;
		}
	}
	
	return file;
}


