/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VKStringList.h"

#include "DBCorrs.h"
#include "CorrsCreator.h"
#include "CorrsCollection.h"

void CorrsCollection::FillShapes(ModelDatabase * db, const VKString & geomName,
								 std::vector<GeomShape*> & shapes)
{
	for (int i=0; i<db->NumModels(); i++)
		shapes.push_back(db->GetModel(i)->GetGeomShape(geomName));
}

CorrsCollection::CorrsCollection(const VKString & corrsName,
								 const std::vector<GeomShape*> & shapes, DBCorrs * corrs)
: ShapeCorrs(corrsName, ShapeCorrs::CORECORRS_CorrsCollection, corrs)
{
	m_Shapes = shapes;
	
	// enumeration params
	m_iMaxModelID = -1;
	m_ModelIDToOffset = NULL;
	m_aModelIDToShapeID = NULL;
	m_iMaxGlobPntID = -1;
	m_GlobPntIDToShapeID = NULL;
	m_GlobPntIDToLocPntID = NULL;

	// cache params
	m_aCachedCorrVals = NULL;
	m_aCachedCorrs = NULL;
	m_aNumCachedCorrs = NULL;
	EnumeratePoints();	
}

CorrsCollection::~CorrsCollection()
{
	if (m_ModelIDToOffset!=NULL)
		delete [] m_ModelIDToOffset;
	
	if (m_GlobPntIDToShapeID!=NULL)
		delete [] m_GlobPntIDToShapeID;
	
	if (m_GlobPntIDToLocPntID!=NULL)
		delete [] m_GlobPntIDToLocPntID;
	
	if (m_aCachedCorrVals!=NULL)
	{
		for (int i=0; i<NumGlobPnts(); i++)
			if (m_aCachedCorrVals[i]!=NULL) delete [] m_aCachedCorrVals[i];
		delete [] m_aCachedCorrVals;
	}
	if (m_aCachedCorrs!=NULL)
	{
		for (int i=0; i<NumGlobPnts(); i++)
			if (m_aCachedCorrs[i]!=NULL) delete [] m_aCachedCorrs[i];
		delete [] m_aCachedCorrs;
	}
	
	if (m_aNumCachedCorrs!=NULL)
		delete [] m_aNumCachedCorrs;
	
	if (m_aModelIDToShapeID!=NULL)
		delete [] m_aModelIDToShapeID;
}

VKString CorrsCollection::GetCorrsClassName() const
{
	return "CorrsCollection";
}

int CorrsCollection::NumShapes()
{
	return (int)m_Shapes.size();
}

GeomShape * CorrsCollection::GetShape(int localID)
{
	assert(localID < (int)m_Shapes.size());
	return m_Shapes[localID];
}

int CorrsCollection::LocalShapeID(int modelID) const
{
	if (m_aModelIDToShapeID==NULL)
		return -1;
	assert(m_aModelIDToShapeID!=NULL);
	if (modelID>=m_iMaxModelID)
		return -1;
	return m_aModelIDToShapeID[modelID];
}

GeomShape * CorrsCollection::GetShapeByModelID(int modelID)
{
	assert(m_aModelIDToShapeID!=NULL);
	if (modelID>=m_iMaxModelID)
		return NULL;
	int localID = m_aModelIDToShapeID[modelID];
	if (localID<0)
		return NULL;
	return GetShape(localID);
}

void CorrsCollection::Domain(std::vector<const GeomShape *> & domainShapes ) const 
{
	for (int i=0; i<(int)m_Shapes.size(); i++)
		domainShapes.push_back(m_Shapes[i]);
}

void CorrsCollection::Range(std::vector<const GeomShape *> & rangeShapes ) const 
{
	for (int i=0; i<(int)m_Shapes.size(); i++)
		rangeShapes.push_back(m_Shapes[i]);
}

int CorrsCollection::GetCorrs(const ShapePoint & pnt,
							  std::vector<ShapePoint> & corrs, 
							  const VKString & corrValName,
							  std::vector<double> *corrVals)
{
	return GetCachedCorrs(pnt, corrs, corrValName, corrVals);
}

// enumeration
bool CorrsCollection::Enumerated() const
{
	return (m_ModelIDToOffset!=NULL);
}

int CorrsCollection::NumGlobPnts() const
{
	return m_iMaxGlobPntID;
}

void CorrsCollection::EnumeratePoints()
{
	assert(m_ModelIDToOffset==NULL);
	
	// find maximal model ID is small (for quick enumeration)
	
	if (m_Shapes.size() == 0)
		return;
	
	m_iMaxGlobPntID = 0;
	m_iMaxModelID = -1;
	for (int i=0; i<(int)m_Shapes.size(); i++)
	{
		m_iMaxModelID = vkMax(m_iMaxModelID, m_Shapes[i]->GetModel()->ModelID());
		const ShapePointCloud * pnts = m_Shapes[i]->SafeCastToPointCloud();
		assert(pnts!=NULL);
		m_iMaxGlobPntID += pnts->NumPoints();
	}
	m_iMaxModelID++;
	
	// check if there are models / points and not more than 1,000,000 models
	assert(m_iMaxGlobPntID>0);
	assert(m_iMaxModelID>=0);
	if (m_iMaxModelID > 1000000)
	{
		std::cout<<"[ERROR] max model ID is too high ("<<m_iMaxModelID<<")";
		std::cout<<", enumeration will not be efficient"<<std::endl;
		assert(false);
	}
	
	// fill model id to shape id
	m_aModelIDToShapeID = new int[m_iMaxModelID];
	for (int i=0; i<m_iMaxModelID; i++)
		m_aModelIDToShapeID[i] = -1;
	for (int i=0; i<(int)m_Shapes.size(); i++)
		m_aModelIDToShapeID[m_Shapes[i]->GetModel()->ModelID()] = i;
	
	// create offset arrays (map ShapePoint to id)
	m_ModelIDToOffset = new int[m_iMaxModelID];
	for (int i=0; i<m_iMaxModelID; i++)
		m_ModelIDToOffset[i] = -1;
	
	// create global ids arrays (map id to ShapePoint)
	m_GlobPntIDToShapeID = new int[m_iMaxGlobPntID];
	m_GlobPntIDToLocPntID = new int[m_iMaxGlobPntID];
	for (int i=0; i<m_iMaxGlobPntID; i++)
	{
		m_GlobPntIDToShapeID[i] = -1;
		m_GlobPntIDToLocPntID[i] = -1;
	}
	
	// fill offsets and id to shape maps
	int offset = 0;
	for (int i=0; i<(int)m_Shapes.size(); i++)
	{
		int mID = m_Shapes[i]->GetModel()->ModelID();
		if (m_ModelIDToOffset[mID]!=-1)
			std::cout<<"[ERROR] Shape "<<i<<" Model "<<mID<<" already inited"<<std::endl;
		assert(m_ModelIDToOffset[mID]==-1);		// check all shapes correspond to a unique model
		m_ModelIDToOffset[mID] = offset;
		ShapePointCloud * pnts = m_Shapes[i]->SafeCastToPointCloud();
		assert(pnts!=NULL);		// surfaces need to be discretized
		
		assert(offset + pnts->NumPoints() <= m_iMaxGlobPntID);
		for (int pID=0; pID < pnts->NumPoints(); pID++)
		{
			m_GlobPntIDToShapeID[offset + pID] = i;
			m_GlobPntIDToLocPntID[offset + pID] = pID;
		}
		
		offset += pnts->NumPoints();
	}	
	
}

int CorrsCollection::PointToId(const ShapePoint & pnt) const
{
	assert(Enumerated());
	
	assert(m_ModelIDToOffset!=NULL);
	
	// find offset
	const ShapePointCloud * pnts = pnt.GetPntsSet();
	assert(pnts!=NULL);
	int modelID = pnts->GetModel()->ModelID();
	assert(modelID>=0);
	
	assert(modelID>=0 && modelID < m_iMaxModelID);
	int offset = m_ModelIDToOffset[modelID];
	assert(offset>=0);
	
	// find point id
	int pntID = pnt.PointID();
	assert(pntID>=0);
	
	// return global id 
	return offset + pntID;
}

ShapePoint CorrsCollection::IdToPoint(int globalPntID) const
{
	assert(Enumerated());
	
	assert(m_GlobPntIDToShapeID!=NULL && m_GlobPntIDToLocPntID!=NULL);
	assert(globalPntID>=0 && globalPntID < NumGlobPnts());
	
	// get local ids
	int sID = m_GlobPntIDToShapeID[globalPntID];
	int pID = m_GlobPntIDToLocPntID[globalPntID];
	
	// get geometry
	assert(sID>=0 && sID<(int)m_Shapes.size());
	const GeomShape * g = m_Shapes[sID];
	
	// cast to point cloud
	const ShapePointCloud *pnts = ((GeomShape*)g)->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	// get point in the cloud
	assert(pID>=0 && pID<pnts->NumPoints());
	return pnts->GetPoint(pID);
	
}

/// caching
void CorrsCollection::AutoFillCache(const VKString & corrValueName)
{
	assert(Enumerated());
	assert(m_aNumCachedCorrs!=NULL);
	assert(corrValueName=="default");
	for (int i=0; i<NumGlobPnts(); i++)
	{
		std::vector<ShapePoint> corrs;
		std::vector<double> corrVals;
		GetCorrs(IdToPoint(i), corrs, corrValueName, &corrVals);
		if (!IsCacheFilled(i))
			SetCachedCorrs(i, corrs, corrVals);
	}	
}

// also use InitCache to clear cache
void CorrsCollection::InitCache(const VKString & corrValueName)
{
	assert(Enumerated());	
	assert(m_aNumCachedCorrs==NULL);
	m_aCachedCorrVals = new double*[NumGlobPnts()];
	m_aCachedCorrs = new int*[NumGlobPnts()];
	m_aNumCachedCorrs = new int[NumGlobPnts()];

	for (int i=0; i<NumGlobPnts(); i++)
	{
		m_aNumCachedCorrs[i] = -1;
		m_aCachedCorrVals[i] = NULL;
		m_aCachedCorrs[i] = NULL;
	}
}

void CorrsCollection::SetCachedCorrs(int fromPntID, 
									 const std::vector<ShapePoint> & corrs,
									 const std::vector<double> & vals)
{
	assert(m_aNumCachedCorrs!=NULL);
	assert(m_aNumCachedCorrs[fromPntID] == -1);
	assert(m_aCachedCorrVals[fromPntID] == NULL);
	assert(m_aCachedCorrs[fromPntID] == NULL);
	assert(vals.size()==corrs.size());
	
	int nEntries = (int)corrs.size();
	m_aNumCachedCorrs[fromPntID] = nEntries;
	if (nEntries>0)
	{
		m_aCachedCorrs[fromPntID] = new int[nEntries];
		m_aCachedCorrVals[fromPntID] = new double[nEntries];

		for (int i=0; i<nEntries; i++)
		{
			m_aCachedCorrs[fromPntID][i] = PointToId(corrs[i]);
			m_aCachedCorrVals[fromPntID][i] = vals[i];
		}
	}
}

void CorrsCollection::GetCachedCorrs(int fromPntID, 
									 std::vector<ShapePoint> & corrs,
									 std::vector<double> * vals)
{
	assert(m_aNumCachedCorrs!=NULL);
	assert(IsCacheFilled(fromPntID));
	int nCorrs = m_aNumCachedCorrs[fromPntID];
	for (int i=0; i<nCorrs; i++)
	{
		corrs.push_back(IdToPoint(m_aCachedCorrs[fromPntID][i]));
		if (vals!=NULL)
			vals->push_back(m_aCachedCorrVals[fromPntID][i]);
	}
}

bool CorrsCollection::IsCacheFilled(int fromPntID) const
{
	bool valid = m_aNumCachedCorrs!=NULL && fromPntID>=0 && fromPntID < NumGlobPnts();
	return valid && (m_aNumCachedCorrs[fromPntID] > -1);
}

int CorrsCollection::GetCachedCorrs(const ShapePoint & pnt,
									std::vector<ShapePoint> & corrs, 
									const VKString & corrValName,
									std::vector<double> *corrVals)
{
	assert(corrValName=="default");
	GetCachedCorrs(PointToId(pnt), corrs, corrVals);
	return (int)corrs.size();
}

///// Read / write
void CorrsCollection::WriteShapeCorrs(FILE * file)
{
	WriteShapeCorrsAsCollection(file);
}

void CorrsCollection::ReadShapeCorrs(FILE * file)
{
	ReadShapeCorrsAsCollection(file);
}

void CorrsCollection::WriteShapeCorrsAsCollection(FILE * file)
{
	// how to handle a header? need to know the class / shapes
	// everything needs to be cached
	WriteCollectionHeader(file, "CorrsCollection");
	
	assert(m_aCachedCorrVals!=NULL);
	assert(m_aCachedCorrs!=NULL);
	assert(m_aNumCachedCorrs!=NULL);
	
	int nFrom = NumGlobPnts();
	fwrite(&nFrom, sizeof(int), 1, file);
	fwrite(m_aNumCachedCorrs, sizeof(int), nFrom, file);
	
	for (int i=0; i<nFrom; i++)
	{
		assert(m_aNumCachedCorrs[i]>=0);
		if (m_aNumCachedCorrs[i]>0)
		{
			assert(m_aCachedCorrVals[i]!=NULL && m_aCachedCorrs[i]!=NULL);
			fwrite(m_aCachedCorrVals[i], sizeof(double), m_aNumCachedCorrs[i], file);
			fwrite(m_aCachedCorrs[i], sizeof(int), m_aNumCachedCorrs[i], file);
		}
	}
}

void CorrsCollection::ReadShapeCorrsAsCollection(FILE * file)
{
	// (header was already loaded)
	// load from cached
	InitCache();
	
	assert(m_aCachedCorrVals!=NULL);
	assert(m_aCachedCorrs!=NULL);
	assert(m_aNumCachedCorrs!=NULL);
	
	int nFrom;
	fread(&nFrom, sizeof(int), 1, file);
	assert(nFrom==NumGlobPnts());
	fread(m_aNumCachedCorrs, sizeof(int), nFrom, file);
	
	for (int i=0; i<nFrom; i++)
	{
		assert(m_aNumCachedCorrs[i]>=0);
		if (m_aNumCachedCorrs[i]>0)
		{
			m_aCachedCorrVals[i] = new double[m_aNumCachedCorrs[i]];
			m_aCachedCorrs[i] = new int[m_aNumCachedCorrs[i]];
			fread(m_aCachedCorrVals[i], sizeof(double), m_aNumCachedCorrs[i], file);
			fread(m_aCachedCorrs[i], sizeof(int), m_aNumCachedCorrs[i], file);
		}
	}
}

CorrsCollection * CorrsCollection::ReadCollection(ModelDatabase * db, 
												  const VKString & filename, DBCorrs * corrs)
{
	int nGeoms;
	VKStringList modelNames;
	VKStringList geomNames;
	VKString className;
	VKString corrsName;
	FILE * file = ReadCollectionHeader(filename, nGeoms, modelNames, 
									   geomNames, className, corrsName);
	assert(modelNames.count()==nGeoms && geomNames.count()==nGeoms);
	
	std::vector<GeomShape*> shapes;
	
	for (int i=0; i<nGeoms; i++)
	{
		Model3D * model = db->GetModel(modelNames[i]);
		assert(model!=NULL);
		GeomShape * shape = model->GetGeomShape(geomNames[i]);
		assert(shape!=NULL);
		shapes.push_back(shape);
	}
	
	CorrsCreator * reader = CorrsCreator::GetCreator(className);
	assert(reader!=NULL);
	CorrsCollection * collection = NULL;
	
	if (reader->ClassType()==ShapeCorrs::CORECORRS_CorrsCollection)
		collection = reader->ReadCorrsCollection(corrsName, shapes, file, corrs);
	else if (reader->ClassType()==ShapeCorrs::CORECORRS_CorrsCollectionMap)
		collection = reader->ReadCorrsCollectionMap(corrsName, shapes, file, corrs);
	else
	{
		std::cout<<"[ERROR] Wrong class type: "<<className.c_str()<<std::endl;
		assert(false);
	}
	assert(collection!=NULL);
	return collection;
}

void CorrsCollection::WriteCollectionHeader(FILE * file,
											const VKString & className)
{
	int nGeoms = (int)m_Shapes.size();
	VKStringList modelNames;
	VKStringList geomNames;
	
	VKString corrsName = GetCorrsName();
	VKString readClassName = className;
	if (readClassName=="Native")
		readClassName = GetCorrsClassName();
	
	for (int i=0; i<(int)m_Shapes.size(); i++)
	{
		modelNames.push_back(m_Shapes[i]->GetModel()->ModelName());
		geomNames.push_back(m_Shapes[i]->GeometryName());
	}

	fwrite(&(DBCorrs::s_iCorrsHeader), sizeof(unsigned int), 1, file);
	fwrite(&nGeoms, sizeof(int), 1, file);
	VKStringList::WriteList(file, modelNames);
	VKStringList::WriteList(file, geomNames);
	VKString::WriteString(corrsName, file);
	VKString::WriteString(readClassName, file);
}

FILE * CorrsCollection::ReadCollectionHeader(const VKString & filename, 
											 int & nGeoms,
											 VKStringList & modelNames, 
											 VKStringList & geomNames,
											 VKString & className,
											 VKString & corrsName)
{
	FILE * file = fopen(filename.c_str(), "rb");
	assert(file!=NULL);
	
	int header;	
	fread(&header, sizeof(unsigned int), 1, file);
	assert(header==DBCorrs::s_iCorrsHeader);
	fread(&nGeoms, sizeof(int), 1, file);
	VKStringList::ReadList(file, modelNames);
	VKStringList::ReadList(file, geomNames);
	corrsName = VKString::ReadString(file);
	className = VKString::ReadString(file);
	
	return file;
}


