/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "GeomShape.h"

#include "FileManagement.h"

GeomShape::GeomShapeTypes GeomShape::GetGeomShapeType(const VKString & str)
{
	if (str=="SHAPE_TYPE_POINT_CLOUD")
		return SHAPE_TYPE_POINT_CLOUD;
	else if (str=="SHAPE_TYPE_POINT_CLOUD_ALLVTX")
		return SHAPE_TYPE_POINT_CLOUD_ALLVTX;
	else if (str=="SHAPE_TYPE_POINT_CLOUD_ALLFACE")
		return SHAPE_TYPE_POINT_CLOUD_ALLFACE;
	else if (str=="SHAPE_TYPE_MANIFOLD"
			 || str=="UNIFORM_MANIFOLD")
		return SHAPE_TYPE_MANIFOLD;
	else if (str=="SHAPE_TYPE_IMAGE")
		return SHAPE_TYPE_IMAGE;
	else if (str=="SHAPE_TYPE_DESCRIPTOR")
		return SHAPE_TYPE_DESCRIPTOR;
//	else if (str=="SHAPE_TYPE_MID_EDGE_FLATTENED")
//		return SHAPE_TYPE_MID_EDGE_FLATTENED;
	else if (str=="SHAPE_TYPE_TRIANGLE_SOUP")
		return SHAPE_TYPE_TRIANGLE_SOUP;
	else if (str=="SHAPE_TYPE_OTHER")
		return SHAPE_TYPE_OTHER;
	else
		assert(false);
}

VKString GeomShape::GetGeomShapeTypeStr(GeomShapeTypes tp)
{
	if (tp==SHAPE_TYPE_POINT_CLOUD)
		return "SHAPE_TYPE_POINT_CLOUD";
	else if (tp==SHAPE_TYPE_POINT_CLOUD_ALLVTX)
		return "SHAPE_TYPE_POINT_CLOUD_ALLVTX";
	else if (tp==SHAPE_TYPE_POINT_CLOUD_ALLFACE)
		return "SHAPE_TYPE_POINT_CLOUD_ALLFACE";
	else if (tp==SHAPE_TYPE_MANIFOLD)
		return "SHAPE_TYPE_MANIFOLD";
	else if (tp==SHAPE_TYPE_IMAGE)
		return "SHAPE_TYPE_IMAGE";
	else if (tp==SHAPE_TYPE_DESCRIPTOR)
		return "SHAPE_TYPE_DESCRIPTOR";
//	else if (tp==SHAPE_TYPE_MID_EDGE_FLATTENED)
//		return "SHAPE_TYPE_MID_EDGE_FLATTENED";
	else if (tp==SHAPE_TYPE_TRIANGLE_SOUP)
		return "SHAPE_TYPE_TRIANGLE_SOUP";
	else if (tp==SHAPE_TYPE_OTHER)
		return "SHAPE_TYPE_OTHER";
	else
		assert(false);
}

VKString GeomShape::GetGeomExtension(GeomShapeTypes tp)
{
	if (tp==SHAPE_TYPE_POINT_CLOUD
		|| tp==SHAPE_TYPE_POINT_CLOUD_ALLVTX
		|| tp==SHAPE_TYPE_POINT_CLOUD_ALLFACE)
		return ".pnts";
	else if (tp==SHAPE_TYPE_MANIFOLD)
		return ".manifold.ifs";
	else if (tp==SHAPE_TYPE_TRIANGLE_SOUP)
		return ".polysoup.off";
	else if (tp==SHAPE_TYPE_DESCRIPTOR)
		return ".polysoup.off";
	else if (tp==SHAPE_TYPE_IMAGE)
		return ".jpg";
	else if (tp==SHAPE_TYPE_DESCRIPTOR)
		return ".sd";
	else
		assert(false);
}

GeomShape::GeomShape(GeomShape::GeomShapeTypes shapeTp, 
					 const VKString & geomName,
					 const Model3D * model)
{
	m_ShapeTp = shapeTp;
	m_sGeometryName = geomName;
	m_pModel = model;
	m_sSrcFile = "none";
	m_MemoryStatus = SHAPEMEM_IgnoreMemoryCommands;
}

GeomShape::~GeomShape()
{
	
}

VKString GeomShape::GeometryName() const
{
	return m_sGeometryName;
}

VKString GeomShape::GeometryFile() const
{
	return GeometryName() + GetGeomExtension(GetShapeType());
}

void GeomShape::SetGeometryModel(Model3D * model)
{
	m_pModel = model;
}

void GeomShape::SetGeometryName(const VKString & newName)
{
	m_sGeometryName = newName;
}

ShapeMesh * GeomShape::SafeCastToMesh()
{
	return NULL;
}

ShapeManifoldMesh * GeomShape::SafeCastToManifold()
{
	return NULL;
}

ShapeTriangleSoup * GeomShape::SafeCastToTriangleSoup()
{
	return NULL;
}

ShapePointCloud * GeomShape::SafeCastToPointCloud()
{
	return NULL;
}

ShapeImg * GeomShape::SafeCastToImg()
{
	return NULL;
}

const ShapeMesh * GeomShape::SafeCastToMesh() const
{
	return ((GeomShape*)this)->SafeCastToMesh();
}

const ShapeManifoldMesh * GeomShape::SafeCastToManifold() const
{
	return ((GeomShape*)this)->SafeCastToManifold();	
}

const ShapeTriangleSoup * GeomShape::SafeCastToTriangleSoup() const
{
	return ((GeomShape*)this)->SafeCastToTriangleSoup();
}


const ShapePointCloud * GeomShape::SafeCastToPointCloud() const
{
	return ((GeomShape*)this)->SafeCastToPointCloud();
}

const ShapeImg * GeomShape::SafeCastToImg() const
{
	return ((GeomShape*)this)->SafeCastToImg();
}

GeomShape::GeomShapeTypes GeomShape::GetShapeType() const
{
	return m_ShapeTp;
}

const Model3D * GeomShape::GetModel() const
{
//	assert(m_pModel!=NULL);
	return m_pModel;
}

GeomShape::MemoryStatus GeomShape::MemStatus() const
{
	return m_MemoryStatus;
}

void GeomShape::FreeMemory()
{
	assert(false);
//	if (m_MemoryStatus==SHAPEMEM_IgnoreMemoryCommands)
//		return;
//	assert(m_MemoryStatus == SHAPEMEM_Loaded);
//	m_MemoryStatus = SHAPEMEM_NotLoaded;
//	// NOTE: not useful if you do not cleanup memory (reimplement in children)!
}

void GeomShape::ReadIntoMemory()
{
	if (m_MemoryStatus==SHAPEMEM_IgnoreMemoryCommands
		|| m_MemoryStatus==SHAPEMEM_Loaded)
		return;

	assert(FileManagement::FileExists(m_sSrcFile.c_str()));
	assert(m_MemoryStatus == SHAPEMEM_NotLoaded);
	m_MemoryStatus = SHAPEMEM_Loaded;
	ReadGeometryFile(m_sSrcFile);
}

void GeomShape::SetSrcFile(const VKString & filename)
{
	m_sSrcFile = filename;
}

int GeomShape::NDistMetrics()
{
	return (int)m_Distances.size();
}

void GeomShape::AddDistanceMetric(const VKString & metricName,
								  ShapeDistance * dist)
{
	m_DistNameToID[metricName] = NDistMetrics();
	m_Distances.push_back(dist);
}

ShapeDistance * GeomShape::GetDistanceMetric(int metricID)
{
	assert(metricID>=0 && metricID < (int)m_Distances.size());
	return m_Distances[metricID];
}

ShapeDistance * GeomShape::GetDistanceMetric(const VKString & metricName)
{
	std::map<VKString, int>::iterator iter = m_DistNameToID.find(metricName);
	//assert(iter!=m_DistNameToID.end());
	if (iter!=m_DistNameToID.end())
		return GetDistanceMetric(iter->second);
	else
		return NULL;
}

int GeomShape::NShapeVals()
{
	return (int)m_Values.size();
}

void GeomShape::AddShapeVal(const VKString & valsName,
							ShapeValue * vals)
{
	m_ValsNameToID[valsName] = NShapeVals();
	m_Values.push_back(vals);
}

ShapeValue * GeomShape::GetShapeVal(int valsID)
{
	assert(valsID < (int)m_Values.size() && valsID>=0);
	return m_Values[valsID];
}

ShapeValue * GeomShape::GetShapeVal(const VKString & valsName)
{
	std::map<VKString, int>::iterator iter = m_ValsNameToID.find(valsName);
	if (iter!=m_ValsNameToID.end())
		return GetShapeVal(iter->second);
	else
		return NULL;
}

const ShapeValue * GeomShape::GetShapeVal(int valsID) const
{
	return ((GeomShape*)this)->GetShapeVal(valsID);
}

const ShapeValue * GeomShape::GetShapeVal(const VKString & valsName) const
{
	return ((GeomShape*)this)->GetShapeVal(valsName);
}

void GeomShape::RearrangeAxes(char newOrder[3], int newSigns[3])
{
	std::cout<<"[ERROR] Geometry: "<<GeometryName().c_str()<<" axes reorder not implemented"<<std::endl;
	assert(false);
}
