/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __GeomShape_H
#define __GeomShape_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "VKString.h"
#include "ShapePoint.h"

class Model3D;
class ShapeMesh;
class ShapeDistance;
class ShapeValue;
class ShapeManifoldMesh;
class ShapeImg;
class ShapeTriangleSoup;

class GeomShape
{
	public:
		enum GeomShapeTypes
		{
			SHAPE_TYPE_POINT_CLOUD,
			SHAPE_TYPE_POINT_CLOUD_ALLVTX,
			SHAPE_TYPE_POINT_CLOUD_ALLFACE,
			SHAPE_TYPE_MANIFOLD,
			SHAPE_TYPE_TRIANGLE_SOUP,
			SHAPE_TYPE_IMAGE,
			SHAPE_TYPE_DESCRIPTOR,
			SHAPE_TYPE_OTHER
		};
	
		enum MemoryStatus
		{
			SHAPEMEM_IgnoreMemoryCommands,
			SHAPEMEM_Loaded,
			SHAPEMEM_NotLoaded
		};
	
		static VKString GetGeomExtension(GeomShapeTypes tp);
		static GeomShapeTypes GetGeomShapeType(const VKString & str);
		static VKString GetGeomShapeTypeStr(GeomShapeTypes tp);
	
		GeomShape(GeomShapeTypes shapeTp, 
				  const VKString & geomName = "none",
				  const Model3D * model = NULL);
		virtual ~GeomShape();
	
		virtual VKString GeometryName() const;
		virtual VKString GeometryFile() const;
		virtual void SetGeometryName(const VKString & newName);
		virtual void SetGeometryModel(Model3D * model);
	
		virtual MemoryStatus MemStatus() const;
		virtual void FreeMemory();	// should be re-implemented in a child
		virtual void ReadIntoMemory();
		virtual void SetSrcFile(const VKString & filename);

		virtual void WriteGeometryFile(const VKString & geomFile) = 0;
		virtual void ReadGeometryFile(const VKString & geomFile) = 0;
		
		virtual GeomShapeTypes GetShapeType() const;
		virtual const Model3D * GetModel() const;
		virtual ShapeMesh * SafeCastToMesh();
		virtual ShapeManifoldMesh * SafeCastToManifold();
		virtual ShapeTriangleSoup * SafeCastToTriangleSoup();
		virtual ShapePointCloud * SafeCastToPointCloud();
		virtual ShapeImg * SafeCastToImg();

		virtual const ShapeMesh * SafeCastToMesh() const;
		virtual const ShapeManifoldMesh * SafeCastToManifold() const;
		virtual const ShapeTriangleSoup * SafeCastToTriangleSoup() const;
		virtual const ShapePointCloud * SafeCastToPointCloud() const;
		virtual const ShapeImg * SafeCastToImg() const;

		virtual int NDistMetrics();
		virtual void AddDistanceMetric(const VKString & metricName,
									   ShapeDistance * dist);
		// ok if returns null
		virtual ShapeDistance * GetDistanceMetric(int metricID);
		virtual ShapeDistance * GetDistanceMetric(const VKString & metricName);

		virtual int NShapeVals();
		virtual void AddShapeVal(const VKString & valsName, ShapeValue * vals);
		// ok if returns null
		virtual ShapeValue * GetShapeVal(int valsID);
		virtual ShapeValue * GetShapeVal(const VKString & valsName);
		virtual const ShapeValue * GetShapeVal(int valsID) const;
		virtual const ShapeValue * GetShapeVal(const VKString & valsName) const;
	
		// not very general - but seems necessary...
		virtual void RearrangeAxes(char newOrder[3], int newSigns[3]);
	
	protected:
		const Model3D * m_pModel;
		GeomShapeTypes m_ShapeTp;
		VKString m_sGeometryName;
	
		MemoryStatus m_MemoryStatus;
		VKString m_sSrcFile;
		
		std::vector<ShapeDistance *> m_Distances;
		std::vector<ShapeValue *> m_Values;
		std::map<VKString, int> m_DistNameToID;
		std::map<VKString, int> m_ValsNameToID;
	
	// todo: symemtric maps + inter-geometry maps
};


#endif

