/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapeInterpolator.h"
#include "ShapePointCloud.h"
#include "ShapeMesh.h"
#include "ShapeDistance.h"
#include "Model3D.h"

ShapeInterpolator::ShapeInterpolator(GeomShape * shapeFrom, GeomShape * shapeTo, 
									 InterpolatorType tp)
{
	m_Type = tp;
	m_pShapeFrom = shapeFrom;
	m_pShapeTo = shapeTo;
}

ShapeInterpolator::~ShapeInterpolator()
{
	
}

GeomShape * ShapeInterpolator::ShapeFrom()
{
	return m_pShapeFrom;	
}

GeomShape * ShapeInterpolator::ShapeTo()
{
	return m_pShapeTo;
}


//////// P2P w/ gaussian smoothing
ShapeInterpP2PGaussian::ShapeInterpP2PGaussian(ShapePointCloud * pntsFrom, ShapePointCloud * pntsTo, 
											   const VKString & distMetric, double sigma)
: ShapeInterpolator(pntsFrom, pntsTo, ShapeInterpolator::INTERP_P2PGaussian)
{
	m_bDistancesPrecomputed = false;
	m_fSigma = sigma;
	m_pPntsFrom = pntsFrom;
	m_pPntsTo = pntsTo;
	assert(m_pPntsTo!=NULL && m_pPntsFrom!=NULL);
	
	if (m_fSigma==-1)
		m_fSigma = .2;
		
	// geodesic distance (if same mesh)
	const ShapeMesh * fromMesh = m_pPntsFrom->GetMesh();
	const ShapeMesh * toMesh = m_pPntsTo->GetMesh();
	assert(toMesh!=NULL && fromMesh!=NULL);	// non-null meshes
	if (distMetric=="auto")
	{
		if (fromMesh==toMesh && ((ShapeMesh*)fromMesh)->SafeCastToManifold()!=NULL)
			m_pDistanceMetric = ((ShapeMesh*)toMesh)->GetDistanceMetric(ShapeDistance::m_sDefaultGeodesic);
		else
			m_pDistanceMetric = ((ShapeMesh*)toMesh)->GetDistanceMetric(ShapeDistance::m_sDefaultEuclidean);
	}
	else
		m_pDistanceMetric = ((ShapeMesh*)toMesh)->GetDistanceMetric(distMetric);
	assert(m_pDistanceMetric!=NULL);

	m_aWeights = new std::vector<double>[m_pPntsTo->NumPoints()];
	m_aNeighbors = new std::vector<ShapePoint>[m_pPntsTo->NumPoints()];
}

ShapeInterpP2PGaussian::~ShapeInterpP2PGaussian()
{
	delete [] m_aWeights;
	delete [] m_aNeighbors;
}

void ShapeInterpP2PGaussian::GetInterpWeights(const ShapePoint & pntOnToShape,
											  const std::vector<ShapePoint> ** savePntsOnFromShape,
											  const std::vector<double> ** saveWeights) const
{
	int pntID = -1;
	bool toAllVtx = (m_pPntsTo->GetShapeType()==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLVTX);
	
	if (toAllVtx)
		pntID = pntOnToShape.VertexID();
	else
		pntID = pntOnToShape.PointID();
	
	if (!toAllVtx && !pntOnToShape.SamePntsSet(m_pPntsTo))
	{
		std::cout<<"[ERROR] Wrong 'to' point. \n\tExpecting: ";
		std::cout<<m_pPntsTo->GetModel()->ModelName().c_str();
		std::cout<<"::"<<m_pPntsTo->GeometryName().c_str();
		const ShapePointCloud * pnts = pntOnToShape.GetPntsSet();
		std::cout<<"\n\tFound: "<<pnts->GetModel()->ModelName().c_str();
		std::cout<<"::"<<pnts->GeometryName().c_str()<<std::endl;		
		assert(false);
	}
	assert(pntID>=0 && pntID<m_pPntsTo->NumPoints());

//	std::cout<<"w["<<pntID<<"] = "<<m_aWeights[pntID].size()<<std::endl;
//	std::cout<<"n["<<pntID<<"] = "<<m_aNeighbors[pntID].size()<<std::endl;
	// fill row
	if (m_aWeights[pntID].size()==0)
	{
		double sumW = 0;
		assert(m_pDistanceMetric!=NULL);
		if (!m_bDistancesPrecomputed)
			((ShapeInterpP2PGaussian*)this)->PrecomputeDistances();
		
		double maxDist = m_fSigma*2.;
		double unnormMaxDist = maxDist / m_pDistanceMetric->NormalizedDistance(1.);
		
		RNArray<const PointInSet *> neighbors;
		const R3Kdtree<const PointInSet *> * fromKdTree = m_pPntsFrom->GetKdTree();
		fromKdTree->FindAll(pntOnToShape.GetPosition(), 0, unnormMaxDist, neighbors);
		
//		std::cout<<"maxDist = "<<maxDist<<" nentries="<<neighbors.NEntries()<<std::endl;
		
		for (int nID=0; nID < neighbors.NEntries(); nID++)
		{
			ShapePoint pnt = neighbors[nID]->GetModelPoint();
			double d = m_pDistanceMetric->NormalizedDistance(pntOnToShape, pnt);
			if (d < unnormMaxDist)
			{
				double w = exp(-pow(d/m_fSigma, 2.));
				m_aWeights[pntID].push_back(w);
				m_aNeighbors[pntID].push_back(pnt);
				sumW += w;
			}
		}
		
		for (int i=0; i<(int)m_aWeights[pntID].size(); i++)
			m_aWeights[pntID][i] /= sumW;
	}
	
	// return precomputed values
	assert(m_aWeights[pntID].size()==m_aNeighbors[pntID].size());
	assert(m_aNeighbors[pntID].size()>0);
	*savePntsOnFromShape = &(m_aNeighbors[pntID]);
	*saveWeights = &(m_aWeights[pntID]);
}

void ShapeInterpP2PGaussian::PrecomputeDistances()
{
	assert(!m_bDistancesPrecomputed);
	m_bDistancesPrecomputed = true;
	ShapePointCloud * precomp = m_pPntsFrom;
	ShapePoint nullPnt = m_pPntsTo->GetPoint(0);
	if (m_pPntsFrom->NumPoints() > m_pPntsTo->NumPoints())
	{
		precomp = m_pPntsTo;
		nullPnt = m_pPntsFrom->GetPoint(0);
	}
	
	for (int i=0; i<precomp->NumPoints(); i++)
		m_pDistanceMetric->NormalizedDistance(precomp->GetPoint(i), nullPnt);
}




