/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifdef OPEN_COLLADA

#include "COLLADAFW.h"
#include "COLLADASaxFWLLoader.h"

#include "VkFunctions.h"
#include "FileManagement.h"

#include "Model3DColladaParser.h"

#include "ShapeMesh.h"
#include "Model3D.h"
#include "ShapeTriangleSoup.h"
#include "ShapeManifoldMesh.h"
#include "ModelDatabase.h"
#include "ShapeMaterials.h"

void Model3DColladaParserStatic::LoadModelFile(Model3D * model, const VKString & colladaFile,
											   const VKString & geometryName, int colladaReadFlags)
{
	Model3DColladaParser * modelWriter = new Model3DColladaParser(colladaFile.FilePath(),
																  colladaReadFlags);
	
	assert(modelWriter!=NULL);
	COLLADASaxFWL::Loader * colladaLoader = new COLLADASaxFWL::Loader();
	assert(colladaLoader!=NULL);
	COLLADAFW::Root root(colladaLoader, modelWriter);
	bool loaded = root.loadDocument(colladaFile.c_str());
	if (!loaded)
		std::cout<<"[WARNING] Collada Parsing Failed"<<std::endl;
	
	if ((colladaReadFlags &  ModelDatabase::READ_MESH_AS_SOUP)!=0
		|| (colladaReadFlags &  ModelDatabase::READ_MESH_AS_MANIFOLD)!=0)
	{
		ShapeMesh * mesh = modelWriter->ColladaGeometry();
		assert(mesh!=NULL);
		mesh->SetGeometryName(geometryName);
		mesh->SetGeometryModel(model);
		model->AddGeomShape(geometryName, mesh);
		
		assert(modelWriter->ColladaMaterial()!=NULL);
		mesh->AddShapeVal(modelWriter->ColladaMaterial()->GetShapeValueName(), 
						  modelWriter->ColladaMaterial());
		
		assert(modelWriter->ColladaSceneGraphLabels()!=NULL);
		mesh->AddShapeVal(modelWriter->ColladaSceneGraphLabels()->GetShapeValueName(), 
						  modelWriter->ColladaSceneGraphLabels());
	}
	delete modelWriter;
	delete colladaLoader;
	//assert(loaded);
}

Model3DColladaParser::GeometryIDs::Face::Face(int v1, int v2, int v3)
{
	pos[0] = v1;
	pos[1] = v2;
	pos[2] = v3;
	
	uv[0] = -1;
	uv[1] = -1;
	uv[2] = -1;

	norm[0] = -1;
	norm[1] = -1;
	norm[2] = -1;

	rgb[0] = -1;
	rgb[1] = -1;
	rgb[2] = -1;
}

int & Model3DColladaParser::GeometryIDs::Face::operator [](int i)
{
	return pos[i];
}



Model3DColladaParser::Model3DColladaParser(const VKString & colladaDir, 
										   int colladaReadFlags)
{
	m_ColladaDir = colladaDir;
	m_iColladaReadFlags = colladaReadFlags;
	m_pGeometry = NULL;
	m_pMaterial = NULL;
	m_pSceneGraphLabels = NULL;
}

Model3DColladaParser::~Model3DColladaParser()
{
	// note: member pointers will be re-used by the caller
	// see LoadColladaFile
}

ShapeMesh * Model3DColladaParser::ColladaGeometry()
{
	return m_pGeometry;
}
ShapeMaterials * Model3DColladaParser::ColladaMaterial()
{
	return m_pMaterial;
}

ShapeLabels * Model3DColladaParser::ColladaSceneGraphLabels()
{
	return m_pSceneGraphLabels;
}

void Model3DColladaParser::CreateMaterials()
{
	assert(m_pGeometry!=NULL);
	m_pMaterial = new ShapeMaterials(m_pGeometry, ShapeValue::m_sDefaultMaterials);
	
	std::map<COLLADAFW::UniqueId, int> img2ID;
	std::map<COLLADAFW::UniqueId, int> mat2ID;
	
	std::map<COLLADAFW::UniqueId, ColladaEffect>::iterator iter;
	for (iter = m_mEffectIDToMaterial.begin(); iter != m_mEffectIDToMaterial.end(); iter++)
	{
		MaterialBRDF * brdf = new MaterialBRDF();
		brdf->m_iMaterialID = m_pMaterial->AddMaterial(brdf);
		mat2ID[iter->first] = brdf->m_iMaterialID;

		const ColladaEffect & effect = iter->second;
		AddColorOrTexture(effect.m_EmitClr, effect, brdf->m_EmitClr, *m_pMaterial, img2ID);
		
		if (effect.m_Shader == COLLADAFW::EffectCommon::SHADER_BLINN
			|| effect.m_Shader == COLLADAFW::EffectCommon::SHADER_PHONG
			|| effect.m_Shader == COLLADAFW::EffectCommon::SHADER_LAMBERT)
		{
			AddColorOrTexture(effect.m_DiffuseClr, effect, brdf->m_DiffuseClr, *m_pMaterial, img2ID);
			AddColorOrTexture(effect.m_AmbientClr, effect, brdf->m_AmbientClr, *m_pMaterial, img2ID);
		}
		
		if (effect.m_Shader == COLLADAFW::EffectCommon::SHADER_BLINN
			|| effect.m_Shader == COLLADAFW::EffectCommon::SHADER_PHONG)
		{
			AddColorOrTexture(effect.m_SpecularClr, effect, brdf->m_SpecularClr, *m_pMaterial, img2ID);
			brdf->m_fShininess = effect.m_fShininess;
		}
		AddColorOrTexture(effect.m_ReflectiveClr, effect, brdf->m_ReflectiveClr, *m_pMaterial, img2ID);
		brdf->m_Reflectivity = effect.m_Reflectivity;
		AddColorOrTexture(effect.m_OpacityClr, effect, brdf->m_OpacityClr, *m_pMaterial, img2ID);
		brdf->m_IndexOfRefraction = effect.m_IndexOfRefraction;
	}
	
	// add uv coordinates (per face)
	std::vector<int> perFaceMaterialID;
	for (int i=0; i<(int)m_FaceToMaterial.size(); i++)
	{
		std::map<COLLADAFW::UniqueId, COLLADAFW::UniqueId>::iterator iterM2E;
		iterM2E = m_mMaterialUniqueIDToEffectID.find(m_FaceToMaterial[i]);
		assert(iterM2E!=m_mMaterialUniqueIDToEffectID.end());
		
		std::map<COLLADAFW::UniqueId, int>::iterator iter = mat2ID.find(iterM2E->second);
		assert(iter!=mat2ID.end());
		perFaceMaterialID.push_back(iter->second);
	}
		
	m_pMaterial->SetFaceMaterials(m_GlobalUVs, perFaceMaterialID, m_FaceToUVs);
}

void Model3DColladaParser::AddColorOrTexture(const COLLADAFW::ColorOrTexture & cOrImg,
											 const ColladaEffect & effect,
											 ColorOrTexture & myCOrImg,
											 ShapeMaterials & materialList, 
											 std::map<COLLADAFW::UniqueId, int> & img2myID)
{
	if (cOrImg.isColor())
	{
		myCOrImg.m_Type = ColorOrTexture::BRDF_Color;
		myCOrImg.m_Color.SetRed(cOrImg.getColor().getRed());
		myCOrImg.m_Color.SetGreen(cOrImg.getColor().getGreen());
		myCOrImg.m_Color.SetBlue(cOrImg.getColor().getBlue());
	}
	else if (cOrImg.isTexture())
	{
		myCOrImg.m_Type = ColorOrTexture::BRDF_Texture;
		COLLADAFW::TextureMapId tMapID = cOrImg.getTexture().getTextureMapId();
		COLLADAFW::SamplerID sampID = cOrImg.getTexture().getSamplerId();
		COLLADAFW::UniqueId tImgID = effect.m_Images[sampID];
		
		std::map<COLLADAFW::UniqueId, int>::iterator iter = img2myID.find(tImgID);
		
		int textureID;
		if (iter==img2myID.end())
		{
			std::map<COLLADAFW::UniqueId, ColladaImage>::iterator iterImg;
			iterImg = m_mIDToImage.find(tImgID);
			assert(iterImg != m_mIDToImage.end());
			textureID = m_pMaterial->AddTextureImg(iterImg->second.m_pImg);
			img2myID[tImgID] = textureID;
		}
		else
			textureID = iter->second;
		myCOrImg.m_iTextureID = textureID;
	}
}


void Model3DColladaParser::ParseNodeHierarchy()
{
	R4Matrix matrixStack = R4identity_matrix;
	for (int i=0; i<(int)m_RootVisualNodes.size(); i++)
		TransformNode(m_RootVisualNodes[i], matrixStack);//, nameStack, depthToSegStack);
}




SceneGraphLabel * Model3DColladaParser::TransformNode(const COLLADAFW::UniqueId nodeID,
													  const R4Matrix & matrixStack)
{	
	std::map<COLLADAFW::UniqueId, NodeHierarchy>::iterator iter = m_mIDToNodeHierarchy.find(nodeID);
	assert(iter!=m_mIDToNodeHierarchy.end());
	NodeHierarchy & node = iter->second;

	SceneGraphLabel * sgNode = new SceneGraphLabel();
	sgNode->m_sLabel = node.m_sNodeName;
	int myNodeID = (int)m_AllSceneGraphNodes.size();
	sgNode->m_iID = myNodeID;
	m_AllSceneGraphNodes.push_back(sgNode);
		
	R4Matrix newMtx = matrixStack * node.m_XForm;
	
	for (int i=0; i<(int)node.m_NodeIDs.size(); i++)
	{
		SceneGraphLabel * cNode = TransformNode(node.m_NodeIDs[i], newMtx);
		sgNode->m_apChildren.push_back(cNode);
		cNode->m_pParent = sgNode;
	}
	
	assert(node.m_GeometryIDs.size()==node.m_GeomBindinginInfo.size());
	for (int i=0; i<(int)node.m_GeometryIDs.size(); i++)
	{
		// add geometries to the node
		assert(ColladaGeometry()!=NULL);
		int firstFace = ColladaGeometry()->NFaces();
		// create faces (and all associated data)
		int gID = TransformGeometry(node.m_GeometryIDs[i], 
									node.m_GeomBindinginInfo[i], newMtx, myNodeID);
		int nFaces = ColladaGeometry()->NFaces()-firstFace;
		sgNode->m_aiGeometryID.push_back(gID);
		sgNode->m_iFirstFaceInGeometry.push_back(firstFace);
		sgNode->m_iNFacesInGeometry.push_back(nFaces);
	}
	
	for (int i=0; i<(int)node.m_ChildrenIDs.size(); i++)
	{
		SceneGraphLabel * cNode = TransformNode(node.m_ChildrenIDs[i], newMtx);
		sgNode->m_apChildren.push_back(cNode);
		cNode->m_pParent = sgNode;
	}
	return sgNode;
}

int Model3DColladaParser::TransformGeometry(const COLLADAFW::UniqueId geometryID,
							std::map<COLLADAFW::MaterialId, COLLADAFW::UniqueId> & matMap,
											const R4Matrix & matrixStack, int segmentID)
{
	R3Affine gapsXForm(matrixStack);
	std::map<COLLADAFW::UniqueId, int>::iterator iter;
	iter = m_mIDToGeometry.find(geometryID);
	assert(iter!=m_mIDToGeometry.end());
	
	assert(iter->second < (int)m_AllGeometries.size());
	GeometryIDs & gDescr = m_AllGeometries[iter->second];
	
	std::vector<int> vts;
	ShapeMesh * geom = ColladaGeometry();
	assert(geom!=NULL);
	
	for (int v=0; v<(int)gDescr.m_Positions.size(); v++)
	{
		R3Point p = gDescr.m_Positions[v];
		gapsXForm.Apply(p);
		int v = geom->AddVertex(p.X(), p.Y(), p.Z());
		vts.push_back(v);
	}
	
	std::vector<int> nIDs;
	for (int v=0; v<(int)gDescr.m_Normals.size(); v++)
	{
		nIDs.push_back((int)m_GlobalNormals.size());
		m_GlobalNormals.push_back(gDescr.m_Normals[v]);
	}

	std::vector<int> uvIDs;
	for (int v=0; v<(int)gDescr.m_UVCoords.size(); v++)
	{
		uvIDs.push_back((int)m_GlobalUVs.size());
		m_GlobalUVs.push_back(gDescr.m_UVCoords[v]);
	}

	std::vector<int> clrIDs;
	for (int v=0; v<(int)gDescr.m_Colors.size(); v++)
	{
		clrIDs.push_back((int)m_GlobalColors.size());
		m_GlobalColors.push_back(gDescr.m_Colors[v]);
	}
	
	for (int fID=0; fID<(int)gDescr.m_Faces.size(); fID++)
	{
		GeometryIDs::Face & f = gDescr.m_Faces[fID];
		if (f[0]==f[1] || f[0]==f[2] || f[1]==f[2])
			continue;	// don't create 
		
		assert((int)m_FaceToMaterial.size()==geom->NFaces());
		geom->AddFace(vts[f[0]], vts[f[1]], vts[f[2]]);
		std::map<COLLADAFW::MaterialId, COLLADAFW::UniqueId>::iterator iterMat;
		iterMat = matMap.find(f.materialID);
		if (iterMat==matMap.end())
			std::cout<<"[ERROR] Cannot find material id: "<<f.materialID<<std::endl;
		assert(iterMat!=matMap.end());

		m_FaceToSceneGraphLeaf.push_back(segmentID);		
		m_FaceToMaterial.push_back(iterMat->second);

		if (f.uv[0] < 0)
		{
			m_FaceToUVs.push_back(-1);
			m_FaceToUVs.push_back(-1);
			m_FaceToUVs.push_back(-1);
		}
		else
		{
			assert(f.uv[0]>=0 && f.uv[0] < uvIDs.size());
			assert(f.uv[1]>=0 && f.uv[1] < uvIDs.size());
			assert(f.uv[2]>=0 && f.uv[2] < uvIDs.size());
			m_FaceToUVs.push_back(uvIDs[f.uv[0]]);
			m_FaceToUVs.push_back(uvIDs[f.uv[1]]);
			m_FaceToUVs.push_back(uvIDs[f.uv[2]]);
		}
		
		if (f.norm[0]<0)
		{
			m_FaceToNormals.push_back(-1);
			m_FaceToNormals.push_back(-1);
			m_FaceToNormals.push_back(-1);
		}
		else
		{
			assert(f.norm[0]>=0 && f.norm[0] < nIDs.size());
			assert(f.norm[1]>=0 && f.norm[1] < nIDs.size());
			assert(f.norm[2]>=0 && f.norm[2] < nIDs.size());
			m_FaceToNormals.push_back(nIDs[f.norm[0]]);
			m_FaceToNormals.push_back(nIDs[f.norm[1]]);
			m_FaceToNormals.push_back(nIDs[f.norm[2]]);
		}
		
		if (f.rgb[0] < 0)
		{
			m_FaceToColors.push_back(-1);
			m_FaceToColors.push_back(-1);
			m_FaceToColors.push_back(-1);
		}
		else
		{
			assert(f.rgb[0]>=0 && f.rgb[0] < clrIDs.size());
			assert(f.rgb[1]>=0 && f.rgb[1] < clrIDs.size());
			assert(f.rgb[2]>=0 && f.rgb[2] < clrIDs.size());
			m_FaceToColors.push_back(clrIDs[f.rgb[0]]);
			m_FaceToColors.push_back(clrIDs[f.rgb[1]]);
			m_FaceToColors.push_back(clrIDs[f.rgb[2]]);
		}
	}
	return iter->second;
}


void Model3DColladaParser::UnimplementedCall(const VKString & name)
{
	if (!m_UnimplementedStuffCalled.contains(name))
		m_UnimplementedStuffCalled.push_back(name);
}

/** This method will be called if an error in the loading process occurred and the loader cannot
 continue to to load. The writer should undo all operations that have been performed.
 @param errorMessage A message containing informations about the error that occurred.
 */
void Model3DColladaParser::cancel(const COLLADAFW::String& errorMessage)
{
	std::cout<<"[WARNING] Collada parsing failed"<<std::endl;
	std::cout<<errorMessage<<std::endl;
}

/** This is the method called. The writer hast to prepare to receive data.*/
void Model3DColladaParser::start()
{
	assert(m_pGeometry==NULL);
	if (m_iColladaReadFlags & ModelDatabase::READ_MESH_AS_SOUP)
		m_pGeometry = new ShapeTriangleSoup();
	else if (m_iColladaReadFlags & ModelDatabase::READ_MESH_AS_MANIFOLD)
		m_pGeometry = new ShapeManifoldMesh();
	assert(m_pGeometry!=NULL);
}

/** This method is called after the last write* method. No other methods will be called after this.*/
void Model3DColladaParser::finish()
{
	VKString defaultPrefix = "LabelsFromScene";

	ParseNodeHierarchy();
	CreateMaterials();
	std::vector<SceneGraphGeometry*> allGeoms;
	for (int i=0; i<(int)m_AllGeometries.size(); i++)
	{
		SceneGraphGeometry * currGeom = new SceneGraphGeometry();
		allGeoms.push_back(currGeom);
		currGeom->m_sGeometryName = m_AllGeometries[i].m_sName;
		currGeom->m_iID = i;
	}
	
	m_pSceneGraphLabels = new ShapeLabels(ColladaGeometry(), 
										  ShapeValue::m_sDefaultSceneGraphLabels);
	m_pSceneGraphLabels->SetFaceSegments(m_FaceToSceneGraphLeaf);
	m_pSceneGraphLabels->SetAllLabels(m_AllSceneGraphNodes);
	m_pSceneGraphLabels->SetAllGeometries(allGeoms);
	
	// ENABLE CODE BELOW FOR SHOWING UNIMPLEMENTED CALLS
//	if (m_UnimplementedStuffCalled.count()>0)
//	{
//		std::cout<<"[WARNING] Parsing COLLADA, unimplemented features: ";
//		for (int i=0; i<m_UnimplementedStuffCalled.count(); i++)
//			std::cout<<m_UnimplementedStuffCalled[i].c_str()<<" ";
//		std::cout<<std::endl;
//    }
}

/** When this method is called, the writer must write the global document asset.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeGlobalAsset ( const COLLADAFW::FileInfo* asset )
{
	UnimplementedCall("FileInfo");
	return true;
}

/** When this method is called, the writer must write the scene.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeScene ( const COLLADAFW::Scene* scene )
{
	// pretty-much pointless - just gives a unique id of VisualScene and KinematicScene
	return true;
}

/** When this method is called, the writer must write the entire visual scene.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeVisualScene ( const COLLADAFW::VisualScene* visualScene )
{
	const COLLADAFW::NodePointerArray & nodes = visualScene->getRootNodes();
	for (int i=0; i<nodes.getCount(); i++)
		m_RootVisualNodes.push_back(AddNode(nodes[i]));
	return true;
}

/** When this method is called, the writer must handle all nodes contained in the 
 library nodes.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeLibraryNodes ( const COLLADAFW::LibraryNodes* libraryNodes )
{
	const COLLADAFW::NodePointerArray & nodes = libraryNodes->getNodes();
	for (int i=0; i<nodes.getCount(); i++)
		m_RootLibraryNodes.push_back(AddNode(nodes[i]));
	return true;
}

COLLADAFW::UniqueId Model3DColladaParser::AddNode(const COLLADAFW::Node * node)
{
	// TODO: transformations of the node?
	COLLADAFW::UniqueId nodeID = node->getUniqueId();
	
	assert(m_mIDToNodeHierarchy.find(nodeID)==m_mIDToNodeHierarchy.end());
	m_mIDToNodeHierarchy[nodeID] = NodeHierarchy();
	
	NodeHierarchy & saveNode = m_mIDToNodeHierarchy[nodeID];
	saveNode.m_sNodeName = node->getName().c_str();
	
	COLLADABU::Math::Matrix4 nodeMatrix = node->getTransformationMatrix();
	R4Matrix gapsMatrix;
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			gapsMatrix[i][j] = nodeMatrix.getElement(i,j);
	saveNode.m_XForm = gapsMatrix;
	
	for (int i=0; i<node->getInstanceNodes().getCount(); i++)
		saveNode.m_NodeIDs.push_back(node->getInstanceNodes()[i]->getInstanciatedObjectId());
	
	const COLLADAFW::InstanceGeometryPointerArray & geoms = node->getInstanceGeometries();
	for (int i=0; i<geoms.getCount(); i++)
	{
		int id = (int)saveNode.m_GeometryIDs.size();
		assert(saveNode.m_GeometryIDs.size()==saveNode.m_GeomBindinginInfo.size());
		saveNode.m_GeometryIDs.push_back(geoms[i]->getInstanciatedObjectId());
		saveNode.m_GeomBindinginInfo.push_back(std::map<COLLADAFW::MaterialId, COLLADAFW::UniqueId>());
		
		const COLLADAFW::MaterialBindingArray& bind = geoms[i]->getMaterialBindings();
		std::map<COLLADAFW::MaterialId, COLLADAFW::UniqueId> & matMap = saveNode.m_GeomBindinginInfo[id];
		for (int bID=0; bID < bind.getCount(); bID++)
		{
			matMap[bind[bID].getMaterialId()] = bind[bID].getReferencedMaterial();
		}
	}
	
	for (int i=0; i<node->getChildNodes().getCount(); i++)
		saveNode.m_ChildrenIDs.push_back(AddNode(node->getChildNodes()[i]));
	
	return nodeID;
}

/** When this method is called, the writer must write the geometry.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeGeometry ( const COLLADAFW::Geometry* geometry )
{
	int locGeomID = (int)m_AllGeometries.size();
	m_AllGeometries.push_back(GeometryIDs());
	assert(m_mIDToGeometry.find(geometry->getUniqueId())==m_mIDToGeometry.end());
	m_mIDToGeometry[geometry->getUniqueId()] = locGeomID;
	
	GeometryIDs & geomIDs = m_AllGeometries[locGeomID];
	geomIDs.m_iGeomListID = locGeomID;
	geomIDs.m_sName = geometry->getName().c_str();

	COLLADAFW::Geometry::GeometryType gType = geometry->getType();
	if (gType==COLLADAFW::Geometry::GEO_TYPE_MESH)
	{
		// TODO: deduce label
		COLLADAFW::Mesh * mesh = (COLLADAFW::Mesh *) geometry;
		COLLADAFW::FloatArray * vtxPositions = mesh->getPositions().getFloatValues();
		COLLADAFW::FloatArray * vtxNormals = mesh->getNormals().getFloatValues();
		COLLADAFW::FloatArray * vtxColors = mesh->getColors().getFloatValues();
		COLLADAFW::FloatArray * uvCoords = mesh->getUVCoords().getFloatValues();
		assert(mesh->getPositions().getDoubleValues()==NULL
			   && mesh->getNormals().getDoubleValues()==NULL
			   && mesh->getColors().getDoubleValues()==NULL 
			   && mesh->getUVCoords().getDoubleValues()==NULL);
		
		int NVert = mesh->getPositions().getValuesCount();
		assert(NVert%3==0);				NVert /= 3;
		
		int NNorm = mesh->getNormals().getValuesCount();
		assert(NNorm%3==0);				NNorm /= 3;
		
		int NColors = mesh->getColors().getValuesCount();
		assert(NColors%3==0);			NColors /= 3;
		
		int NTextureCoords = mesh->getUVCoords().getValuesCount();
		assert(NTextureCoords%2==0);	NTextureCoords /= 2;
		
		assert(vtxPositions!=NULL);
		for (int i=0; i<NVert; i++)
		{
			R3Point p((*vtxPositions)[i*3+0], (*vtxPositions)[i*3+1], (*vtxPositions)[i*3+2]);			
			geomIDs.m_Positions.push_back(p);
		}
		
		if (vtxNormals!=NULL)
		{
			for (int i=0; i<NNorm; i++)
			{
				R3Vector n((*vtxNormals)[i*3+0], (*vtxNormals)[i*3+1], (*vtxNormals)[i*3+2]);
				geomIDs.m_Normals.push_back(n);	
			}
		}
		
		if (uvCoords!=NULL)
		{
			for (int i=0; i<NTextureCoords; i++)
			{
				double u = (*uvCoords)[i*2+0];
				double v = (*uvCoords)[i*2+1];
				geomIDs.m_UVCoords.push_back(R2Point(u, v));
			}
		}
		
		if (vtxColors!=NULL)
		{
			for (int i=0; i<NColors; i++)
			{
				RNRgb c((*vtxColors)[i*3+0], (*vtxColors)[i*3+1], (*vtxColors)[i*3+2]);
				geomIDs.m_Colors.push_back(c);
			}
		}
		
		COLLADAFW::MeshPrimitiveArray & primitives = mesh->getMeshPrimitives();

		int NFaces = mesh->getFacesCount();		

		for (int i=0; i<primitives.getCount(); i++)
		{
			COLLADAFW::MeshPrimitive * primitive = primitives[i];
			COLLADAFW::MeshPrimitive::PrimitiveType tp = primitive->getPrimitiveType();
			COLLADAFW::MaterialId materialID = primitive->getMaterialId();

			if (tp==COLLADAFW::MeshPrimitive::TRIANGLES)
			{
				int NFaces = primitive->getFaceCount();
				COLLADAFW::UIntValuesArray & pids = primitive->getPositionIndices();
				assert(pids.getCount()==NFaces*3);
				for (int f=0; f<NFaces; f++)
				{				
					int fID = (int)geomIDs.m_Faces.size();
					geomIDs.m_Faces.push_back(GeometryIDs::Face(pids[3*f+0], pids[3*f+1], pids[3*f+2]));
					geomIDs.m_Faces[fID].materialID = materialID;
				}
				
				if (primitive->hasNormalIndices())
				{
					assert(vtxNormals!=NULL);
					COLLADAFW::UIntValuesArray & ids = primitive->getNormalIndices();
					assert(ids.getCount()==NFaces*3);
					for (int f=0; f<NFaces; f++)
					for (int p=0; p<3; p++)
						geomIDs.m_Faces[f].norm[p] = ids[3*f + p];
				}
				
				if (primitive->hasColorIndices())
				{
					assert(vtxColors!=NULL);
					COLLADAFW::IndexListArray& colorArrays = primitive->getColorIndicesArray();
					std::cout<<"[ERROR] Don't know how to parse colors"<<std::endl;
					std::cout<<"NFaces="<<NFaces<<std::endl;
					std::cout<<"Colors arrays: "<<colorArrays.getCount()<<std::endl;
					
					for (int aID=0; aID<colorArrays.getCount(); aID++)
						std::cout<<"\tA["<<aID<<"] cnt="<<colorArrays[aID]->getIndicesCount()<<std::endl;
					
					assert(false);	// todo: don't understand the datastructure
					
//					COLLADAFW::UIntValuesArray & ids = primitive->getColorIndicesArray();
//					assert(ids.getCount()==NFaces*3);
//					for (int f=0; f<NFaces; f++)
//						for (int p=0; p<3; p++)
//							geomIDs.m_Faces[f].norm[p] = ids[3*f + p];
				}

				if (primitive->hasUVCoordIndices())
				{
					assert(uvCoords!=NULL);
					COLLADAFW::IndexListArray& uvArrays = primitive->getUVCoordIndicesArray();
					bool canParse = (uvArrays.getCount()==1)&&(uvArrays[0]->getIndicesCount()==3*NFaces);
					
					if (canParse)
					{
						const COLLADAFW::UIntValuesArray & ids = uvArrays[0]->getIndices();
						assert(ids.getCount()==3*NFaces);
						for (int f=0; f<NFaces; f++)
						for (int p=0; p<3; p++)
							geomIDs.m_Faces[f].uv[p] = ids[3*f + p];
					}
					else
					{
						std::cout<<"[ERROR] Don't know how to parse uv coords"<<std::endl;
						std::cout<<"NFaces="<<NFaces<<std::endl;
						std::cout<<"UV arrays: "<<uvArrays.getCount()<<std::endl;
					
						for (int aID=0; aID<uvArrays.getCount(); aID++)
							std::cout<<"\tA["<<aID<<"] cnt="<<uvArrays[aID]->getIndicesCount()<<std::endl;
						assert(canParse);
					}
				}
			}
			else 
				UnimplementedCall(VKString("NonTriMesh_")+VKString::number((int)tp));
		}
	}
	else 
	{
		std::cout<<"[WARNING] Unknown geometry: "<<gType<<std::endl;
		assert(false);
	}
	return true;
}

/** When this method is called, the writer must write the material.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeMaterial( const COLLADAFW::Material* material )
{
	COLLADAFW::UniqueId effectID = material->getInstantiatedEffect();
	COLLADAFW::UniqueId materialUID = material->getUniqueId();
	m_mMaterialUniqueIDToEffectID[materialUID] = effectID;
	
//	std::cout<<"\tMaterial Effect ID: "<<effectID<<std::endl;
//	UnimplementedCall("Material");	
	return true;
}

/** When this method is called, the writer must write the effect.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeEffect( const COLLADAFW::Effect* effect )
{
	const COLLADAFW::CommonEffectPointerArray& effects = effect->getCommonEffects();
	
	COLLADAFW::UniqueId effectID = effect->getUniqueId();
	
	assert(effects.getCount()==1);
	
	int i=0;	
	m_mEffectIDToMaterial[effectID] = ColladaEffect();
	ColladaEffect & m = m_mEffectIDToMaterial[effectID];
	
	m.m_Shader = effects[i]->getShaderType();
	m.m_EffectID = effectID;
	m.m_EmitClr			= effects[i]->getEmission();			// Blinn, Constant, Phong, Lambert
	if (m.m_Shader==COLLADAFW::EffectCommon::SHADER_BLINN
		|| m.m_Shader==COLLADAFW::EffectCommon::SHADER_PHONG
		|| m.m_Shader==COLLADAFW::EffectCommon::SHADER_LAMBERT)
	{
		m.m_AmbientClr		= effects[i]->getAmbient();				// Blinn, Phong, Lambert
		m.m_DiffuseClr		= effects[i]->getDiffuse();				// Blinn, Phong, Lambert
	}
	if (m.m_Shader==COLLADAFW::EffectCommon::SHADER_BLINN
		|| m.m_Shader==COLLADAFW::EffectCommon::SHADER_PHONG)
	{
		m.m_SpecularClr		= effects[i]->getSpecular();			// Blinn, Phong
		assert(effects[i]->getShininess().getType()==COLLADAFW::FloatOrParam::FLOAT);
		m.m_fShininess		= effects[i]->getShininess().getFloatValue();	// Blinn, Phong
	}
	m.m_ReflectiveClr	= effects[i]->getReflective();			// Blinn, Constant, Phong, Lambert
	assert(effects[i]->getReflectivity().getType()==COLLADAFW::FloatOrParam::FLOAT);
	m.m_Reflectivity	= effects[i]->getReflectivity().getFloatValue(); // All: B, C, P, L
	m.m_OpacityClr		= effects[i]->getOpacity();				// Blinn, Constant, Phong, Lambert
	assert(effects[i]->getIndexOfRefraction().getType()==COLLADAFW::FloatOrParam::FLOAT);
	m.m_IndexOfRefraction = effects[i]->getIndexOfRefraction().getFloatValue();	// All: B, C, P, L
	
	const COLLADAFW::SamplerPointerArray& samplers = effects[i]->getSamplerPointerArray();
	for (int s=0; s<samplers.getCount(); s++)
	{
//		std::cout<<"FilterMin["<<s<<"] : "<<samplers[s]->getMinFilter()<<std::endl;
//		std::cout<<"FilterMag["<<s<<"] : "<<samplers[s]->getMagFilter()<<std::endl;
//		std::cout<<"FilterMip["<<s<<"] : "<<samplers[s]->getMipFilter()<<std::endl;
//		std::cout<<"WrapS["<<s<<"] : "<<samplers[s]->getWrapS()<<std::endl;
//		std::cout<<"WrapT["<<s<<"] : "<<samplers[s]->getWrapT()<<std::endl;
//		std::cout<<"WrapP["<<s<<"] : "<<samplers[s]->getWrapP()<<std::endl;
//		std::cout<<"getMipmapMaxlevel["<<s<<"] : "<<samplers[s]->getMipmapMaxlevel()<<std::endl;
//		std::cout<<"getMipmapBias["<<s<<"] : "<<samplers[s]->getMipmapBias()<<std::endl;
		m.m_Images.push_back(samplers[s]->getSourceImage());
	}
	//UnimplementedCall("Effect");
	return true;
}

/** When this method is called, the writer must write the camera.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeCamera( const COLLADAFW::Camera* camera )
{
	UnimplementedCall("Camera");
	return true;
}

/** When this method is called, the writer must write the image.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeImage( const COLLADAFW::Image* image )
{
	m_mIDToImage[image->getUniqueId()] = ColladaImage();
	ColladaImage & img = m_mIDToImage[image->getUniqueId()];
	
	COLLADABU::URI uri = image->getImageURI();
	VKString imgFile = m_ColladaDir + "/" + uri.getURIString().c_str();
	img.m_ImgID = image->getUniqueId();
	
	if (!FileManagement::FileExists(imgFile.c_str()))
	{
		UnimplementedCall(VKString("[ERROR] Could not read texture: ")+imgFile);
		img.m_pImg = new R2Image(1, 1, 3);
		img.m_pImg->SetPixelRGB(0, 0, RNRgb(.5, .5, .5));
		return true;
	}
	img.m_pImg = new R2Image();

	if (!imgFile.endsWith("jpg") && !imgFile.endsWith("JPG"))
	{
		VKString newFile = imgFile+".jpg";
		if (!FileManagement::FileExists(newFile.c_str()))
		{
			system((VKString("convert ")+imgFile+" "+newFile).c_str());
			if (!FileManagement::FileExists(newFile.c_str()))
			{
				UnimplementedCall(VKString("[ERROR] Cannot convert to jpeg: ")+imgFile);
				img.m_pImg = new R2Image(1, 1, 3);
				img.m_pImg->SetPixelRGB(0, 0, RNRgb(.5, .5, .5));
				return true;
			}
		}
		imgFile = newFile;	
	}
	img.m_pImg->ReadJPEG(imgFile.c_str());
	
	if (img.m_pImg->NComponents()==1)	// convert grayscale
	{		
		int nEls = img.m_pImg->Width() * img.m_pImg->Height();
		R2Image * convImg = new R2Image(img.m_pImg->Width(), img.m_pImg->Height(), 3);
		unsigned char * source = (unsigned char*)img.m_pImg->Pixels();
		unsigned char * target = (unsigned char*)convImg->Pixels();
		
		for (int p=0; p<nEls; p++)
		{
			unsigned char g = source[p];
			for (int c=0; c<3; c++)
				target[3*p+c] = g;
		}
		delete img.m_pImg;
		img.m_pImg = convImg;
	}
	assert(img.m_pImg->NComponents()==3);
	return true;
}

/** When this method is called, the writer must write the light.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeLight( const COLLADAFW::Light* light )
{
	UnimplementedCall("Light");
	return true;
}

/** When this method is called, the writer must write the Animation.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeAnimation( const COLLADAFW::Animation* animation )
{
	UnimplementedCall("Animation");
	return true;
}

/** When this method is called, the writer must write the AnimationList.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeAnimationList( const COLLADAFW::AnimationList* animationList )
{
	UnimplementedCall("AnimationList");
	return true;
}

/** When this method is called, the writer must write the skin controller data.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeSkinControllerData(const COLLADAFW::SkinControllerData* 
														skinControllerData )
{
	UnimplementedCall("SkinControllerData");
	return true;
}

/** When this method is called, the writer must write the controller.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeController( const COLLADAFW::Controller* controller )
{
	UnimplementedCall("Controller");
	return true;
}

/** When this method is called, the writer must write the formulas. All the formulas of the entire
 COLLADA file are contained in @a formulas.
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeFormulas( const COLLADAFW::Formulas* formulas )
{
	UnimplementedCall("Formulas");
	return true;
}

/** When this method is called, the writer must write the kinematics scene. 
 @return The writer should return true, if writing succeeded, false otherwise.*/
bool Model3DColladaParser::writeKinematicsScene( const COLLADAFW::KinematicsScene* kinematicsScene )
{
	UnimplementedCall("KinematicsScene");
	return true;
}

#endif

