/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CorrsAffineAlign.h"
#include "ShapePointCloud.h"
#include "LinAlgMatrixReal.h"
#include "ShapeMesh.h"

AffineCorrsCreator::AffineCorrsCreator()
: CorrsCreator(ShapeCorrs::CORECORRS_CorrsPair, "CorrsAffineAlign")
{
	
}

CorrsPair * AffineCorrsCreator::ReadCorrsPair(const VKString & corrName, 
											   GeomShape * shapeFrom, 
											   GeomShape * shapeTo,
											   FILE * file, DBCorrs * )
{
	CorrsPair * corrPair = new CorrsAffineAlign(corrName, shapeFrom, shapeTo);
	corrPair->ReadShapeCorrs(file);
	return corrPair;
}

void AffineCorrsCreator::CreateCorrsPair(const VKString & corrName, 
										 GeomShape * shapeFrom, GeomShape * shapeTo, CorrParams * params,
										 std::vector<CorrsPair *> & saveCorrs, DBCorrs * )
{
	std::vector<CorrsAffineAlign* > corrs;
	ShapePointCloud * pnts1 = shapeFrom->SafeCastToPointCloud();
	ShapePointCloud * pnts2 = shapeTo->SafeCastToPointCloud();
	assert(pnts1!=NULL && pnts2!=NULL);
	assert(params->m_sParamClassName=="ParamsAffineAlign");
	CorrsAffineAlign::CreateCorrs(corrName, pnts1, pnts2, (ParamsAffineAlign *) params, corrs);
	for (int i=0; i<(int)corrs.size(); i++)
		saveCorrs.push_back(corrs[i]);
}



ParamsAffineAlign::ParamsAffineAlign(const VKString & paramSettingName)
: CorrParams(paramSettingName, "CorrsAffineAlign", "ParamsAffineAlign")
{
	m_UpVector = NULL;
	m_fMaxDist = .4;
	m_bICP_Translation = true;
	m_bICP_Rotation = true;
	m_bICP_Scale = true;
	m_bPCA_Translation = true;
	m_bPCA_Rotation = true;
	m_bPCA_Scale = true;
}

R3Point * CorrsAffineAlign::m_CorrBuffer1 = NULL;
R3Point * CorrsAffineAlign::m_CorrBuffer2 = NULL;
int CorrsAffineAlign::m_iCorrBufferSize = 0;

void CorrsAffineAlign::CreateCorrs(const VKString & corrName,
								   ShapePointCloud * pnts1, ShapePointCloud * pnts2,
								   ParamsAffineAlign * params, 
								   std::vector<CorrsAffineAlign* > & corrs)
{
	CreateCorrsFromPCA(corrName, pnts1, pnts2, params->m_fMaxDist, params->m_UpVector,
					   params->m_bICP_Translation, params->m_bICP_Rotation, params->m_bICP_Scale,
					   params->m_bPCA_Translation, params->m_bPCA_Rotation, params->m_bPCA_Scale,
					   corrs);
}




void CorrsAffineAlign::CreateCorrsFromPCA(const VKString & corrName,
										  ShapePointCloud * pnts1, ShapePointCloud * pnts2,
										  double distNorm, R3Vector * upVector,
										  bool pca_translation, bool pca_rotation, bool pca_scale,
										  bool icp_translation, bool icp_rotation, bool icp_scale,
										  std::vector<CorrsAffineAlign* > & corrs)
{
	R3Affine affine12 = R4identity_matrix;
	std::vector<R3Affine> affine12s;
	
	GetPCAAlignments(pnts1, pnts2, pca_translation, 
					 pca_rotation, pca_scale, upVector, affine12s);
	
	if (icp_translation || icp_rotation || icp_scale)
	{
		for (int i=0; i<(int)affine12s.size(); i++)
			RefineWithICP(pnts1, pnts2, affine12s[i], 
						  icp_translation, icp_rotation, icp_scale, upVector);
	}
	
	CorrsAffineAlign * curr = NULL;
	for (int i=0; i<(int)affine12s.size(); i++)
	{
		curr = new CorrsAffineAlign(corrName, pnts1, pnts2, distNorm, affine12s[i]);
		// todo: sort by quality, check for similarity (prune if too similar)
		corrs.push_back(curr);
	}
}

void CorrsAffineAlign::GetPCAAlignments(const R3Point & centroid1, double radius1, const R3Triad & axes1, 
										const R3Point & centroid2, double radius2, const R3Triad & axes2, 
										bool pca_translation, bool pca_rotation, bool pca_scale,
										int upAxis, std::vector<R3Affine> & affine12s)
{
	RNScalar scale1 = (pca_scale && (radius1 > 0)) ? 1.0 / radius1 : 1;
	if (!pca_scale)
	{
		scale1 = 1; 
		radius1 = 1;
		radius2 = 1;
	}
	assert(scale1!=0);
	
	// create transformation given axes 2
	R3Affine affine02 = R3identity_affine;
	if (pca_translation) affine02.Translate(centroid2.Vector());
	if (pca_rotation==1) affine02.Transform(axes2.Matrix());
	if (pca_scale) affine02.Scale(radius2);
	
	if (pca_rotation) 
	{
		int mapID=0;
		for (int dim1 = 0; dim1 < 3; dim1++) 
		for (int dim2 = 0; dim2 < 3; dim2++) 
		{
			if (dim1 == dim2) 
				continue;
			
			if ((upAxis==0 && dim1!=0) || (upAxis==1 && dim2!=1)
				|| (upAxis==2 && (dim1==2 || dim2==2)))	// up vectors must align (if provided)
				continue;
			
			for (int dir1 = -1; dir1 <= 1; dir1+=2) 
			for (int dir2 = -1; dir2 <= 1; dir2+=2) 
			{
				if ((upAxis==0 && dir1!=1) || (upAxis==1 && dir2!=1)
					|| (upAxis==2 && ((dim1==0 && dir1*dir2<0)||(dim1!=0 && dir1*dir2>0))))	// up vectors have the same direction
					continue;
								
				// Create triad of axes for flip
				R3Vector axis1a = (dir1 == 1) ? axes1[dim1] : -axes1[dim1];
				R3Vector axis1b = (dir2 == 1) ? axes1[dim2] : -axes1[dim2];
				R3Vector axis1c = axis1a % axis1b;
				R3Triad triad1(axis1a, axis1b, axis1c);
				
				// Compute transformation for mesh1 (with flip)
				R3Affine affine10 = R3identity_affine;
				if (pca_scale) affine10.Scale(scale1);
				
				affine10.Transform(InverseMatrix(triad1.Matrix()));
				if (pca_translation) affine10.Translate(-centroid1.Vector());
				
				// Compute composite transformation and its inverse
				R3Affine flipped_affine12 = affine02;
				flipped_affine12.Transform(affine10);
				//R3Affine flipped_affine21(R4null_matrix);
				affine12s.push_back(flipped_affine12);
			}
		}
	}
	else // scale or just alignment
	{
		// Compute transformation for mesh1
		R3Affine affine10 = R3identity_affine;
		if (pca_scale) affine10.Scale(scale1);
		if (pca_translation) affine10.Translate(-centroid1.Vector());
		
		// Compute composite transformation
		R3Affine affine12 = affine02;
		affine12.Transform(affine10);
		affine12s.push_back(affine12);
	}
}


void CorrsAffineAlign::GetPCAAlignments(ShapePointCloud * pnts1, ShapePointCloud * pnts2,
										bool pca_translation, bool pca_rotation, bool pca_scale,
										R3Vector * upVector,
										std::vector<R3Affine> & affine12s)
{
	// Compute info for mesh1
	R3Point centroid1 = pnts1->Centroid();
	RNScalar radius1 = (pca_scale) ? pnts1->AverageRadius(centroid1) : 1;
	
	R3Triad axes1 = (pca_rotation==1) ? pnts1->PCAAxes(upVector) : R3xyz_triad;	
	
	// Compute info for mesh2
	R3Point centroid2 = pnts2->Centroid();
	RNScalar radius2 = (pca_scale) ? pnts2->AverageRadius(centroid2) : 1;
	R3Triad axes2 = (pca_rotation == 1) ? pnts2->PCAAxes(upVector) : R3xyz_triad;
	assert(radius2!=0);	
	
	GetPCAAlignments(centroid1, radius1, axes1, centroid2, radius2, axes2, 
					 pca_translation, pca_rotation, pca_scale, upVector==NULL ? -1 : 0, affine12s);
}

bool CorrsAffineAlign::RefineWithICP(ShapePointCloud * pnts1, ShapePointCloud * pnts2,
									 R3Affine & affine12,
									 bool icp_translation, bool icp_rotation, bool icp_scale,
									 R3Vector * upVector)
{
	int nPnts1 = pnts1->NumPoints();
	int nPnts2 = pnts2->NumPoints();
	
	// Allocate arrays of correspondences
	int ncorrespondences = nPnts1 + nPnts2;
	
	if (m_iCorrBufferSize < 2 * ncorrespondences)
	{
		if (m_CorrBuffer1!=NULL)
			delete [] m_CorrBuffer1;
		if (m_CorrBuffer2!=NULL)
			delete [] m_CorrBuffer2;
		
		m_iCorrBufferSize = 2*ncorrespondences;
		m_CorrBuffer1 = new R3Point[m_iCorrBufferSize];
		m_CorrBuffer2 = new R3Point[m_iCorrBufferSize];
	}
	
	assert(m_CorrBuffer1 && m_CorrBuffer2);
	
	
	// Iterate until max_iterations or converged
	RNBoolean converged = FALSE;
	const int max_iterations = 100;
	for (int iteration = 0; iteration < max_iterations; iteration++) 
	{
		// Get arrays of correspondences
		R3Point *correspondences1 = &m_CorrBuffer1[iteration%2 * ncorrespondences];
		R3Point *correspondences2 = &m_CorrBuffer2[iteration%2 * ncorrespondences];
		
		// Update correspondences for aligning transformation
		CreateP2PCorrs(pnts1, pnts2, affine12, 
					   correspondences1, correspondences2, ncorrespondences);
		
		
		// Update aligning transformation for correspondences
		R4Matrix matrix;
		if (upVector==NULL)
			matrix = R3AlignPoints(ncorrespondences, correspondences2, correspondences1, 
								   NULL, icp_translation, icp_rotation, icp_scale);
		else
		{
			// todo: just re-align so that upVec = Z
			assert(upVector->X()==0 && upVector->Y()==0);
			matrix = R3AlignInXY(ncorrespondences, correspondences2, correspondences1, 
								 NULL, icp_translation, icp_rotation, icp_scale);
		}
		
		affine12.Reset(matrix);
		
		// Check for convergence
		converged = FALSE;
		if (iteration > 0) 
		{
			converged = TRUE;
			R3Point *prev_correspondences1 = &m_CorrBuffer1[(1-(iteration%2)) * ncorrespondences];
			R3Point *prev_correspondences2 = &m_CorrBuffer2[(1-(iteration%2)) * ncorrespondences];
			for (int i = 0; i < ncorrespondences; i++) 
			{
				if (!R3Contains(correspondences1[i], prev_correspondences1[i])) 
				{ converged = FALSE; break; }
				if (!R3Contains(correspondences2[i], prev_correspondences2[i])) 
				{ converged = FALSE; break; }
			}
		}
		if (converged) break;
	}
		
	// Return whether converged
	return converged;	
}

R4Matrix CorrsAffineAlign::R3AlignInXY(int npoints, R3Point *points1, R3Point *points2, RNScalar *weights,
									   RNBoolean align_center, RNBoolean align_rotation, int align_scale)
{
//	// Get count of points
//	int count = points1.NEntries();
//	if (points2.NEntries() < count) count = points2.NEntries();
//	
//	// Check number of points
//	if (count < 1) align_center = 0;
//	if (count < 2) align_scale = 0;
//	if (count < 3) align_rotation = 0;
	
	// Compute centers
	R3Point center1(0.0, 0.0, 0.0);
	R3Point center2(0.0, 0.0, 0.0);
	if (align_center)
	{
		double sumW = 0;
		for (int i=0; i<npoints; i++)
		{
			double w = (weights==NULL) ? 1. : weights[i];
			center1 += points1[i] * w;
			center2 += points2[i] * w;
			sumW += w;
		}
		center1 /= sumW;
		center2 /= sumW;
	}	
	
	// Compute scales
	RNScalar s1 = 1;
	RNScalar s2 = 1;
	if (align_scale)
	{
		double sumW = 0;
		for (int i=0; i<npoints; i++)
		{
			double w = (weights==NULL) ? 1. : weights[i];
			s1 += R3Distance(center1, points1[i]) * w;
			s2 += R3Distance(center2, points2[i]) * w;
			sumW += w;
		}
		s1 /= sumW;
		s2 /= sumW;
	}
	
	// Compute cross-covariance of two point sets
	R4Matrix rotation = R4identity_matrix;
	if (align_rotation) 
	{
		// Compute covariance matrix
		RNScalar w;
		RNScalar m[4] = { 0 };
			// vk: new tweak... why count?
//		for (int i = 0; i < count; i++)
		for (int i = 0; i < npoints; i++)
		{
			R3Vector p1 = (points1[i] - center1) / s1;
			R3Vector p2 = (points2[i] - center2) / s2;
			if (weights == NULL) { w = 1; }
			else { w = weights[i]; }
			w *= w;
			for(int j = 0; j < 2; j++) 
			{
				for(int k = 0; k < 2; k++) 
				{
					m[k*2 + j]+=p1[j]*p2[k]*w;
				}
			}
		}
		
		// Normalize covariance matrix
			// vk: new tweak..
//		for (int j = 0; j < 4; j++) m[j] /= count;
		for (int j = 0; j < 4; j++) m[j] /= npoints;
		
		// Calculate SVD of covariance matrix
		RNScalar Um[4];
		RNScalar Wm[2];
		RNScalar Vmt[4];
		RNSvdDecompose(2, 2, m, Um, Wm, Vmt);
		
		// https://sakai.rutgers.edu/access/content/group/7bee3f05-9013-4fc2-8743-3c5078742791/material/svd_ls_rotation.pdf
		R3Matrix Ut(Um[0], Um[2], 0, Um[1], Um[3], 0, 0, 0, 1); 
		R3Matrix V(Vmt[0], Vmt[2], 0, Vmt[1], Vmt[3], 0, 0, 0, 1); 
		R3Matrix VUt = V * Ut;
		R3Matrix D = R3identity_matrix;
		D[1][1] = R3MatrixDet2(VUt[0][0], VUt[0][1], VUt[1][0], VUt[1][1]);
		R3Matrix R = V * D * Ut;
		rotation[0][0] = R[0][0];
		rotation[0][1] = R[0][1];
		rotation[1][0] = R[1][0];
		rotation[1][1] = R[1][1];
	}
	
	// Compute result
	R4Matrix result = R4identity_matrix;
	if (align_center) result.Translate(center1.Vector());
	if (align_scale) result.Scale(s1/s2);
	if (align_rotation) result.Transform(rotation);
	if (align_center) result.Translate(-(center2.Vector()));
	
	// Return resulting matrix that takes points2 to points1
	return result;	
}

R4Matrix CorrsAffineAlign::InverseMatrix(const R4Matrix & mtxGaps)
{
	LinAlgMatrixReal mtxMy(4,4);
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			mtxMy(i,j) = mtxGaps[i][j];
	
	mtxMy.Inverse();
	
	R4Matrix mtxInvGaps;
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			mtxInvGaps[i][j] = mtxMy(i,j);
	
	return mtxInvGaps;
}

R3Affine CorrsAffineAlign::InverseAffine(const R3Affine & affine)
{
	const R4Matrix & mtxGaps = affine.Matrix();
	
	return R3Affine(InverseMatrix(mtxGaps), affine.IsMirrored());
}

int CorrsAffineAlign::CreateP2PCorrs(ShapePointCloud * pnts1, ShapePointCloud * pnts2,
									 const R3Affine& affine12, 
									 R3Point *correspondences1, R3Point *correspondences2, 
									 int max_correspondences)
{	
	int np1 = pnts1->NumPoints();
	int np2 = pnts2->NumPoints();
	R3Affine affine21 = InverseAffine(affine12);
	
	// Build kdtree for points1, points2
	const R3Kdtree<const PointInSet *> * tree1 = pnts1->GetKdTree();
	const R3Kdtree<const PointInSet *> * tree2 = pnts2->GetKdTree();
	assert(tree1!=NULL && tree2!=NULL);
	
	// Initialize number of correspondences
	assert(max_correspondences >= np1 + np2);
	int ncorrespondences = 0;
	
	// Compute correspondences for points1 -> mesh2
	for (int i = 0; i < np1; i++) 
	{
		R3Point position1 = pnts1->PointPosition(i);
		position1.Transform(affine12);
		const PointInSet *closest = tree2->FindClosest(position1, 0, FLT_MAX);
		assert(closest!=NULL);
		assert(ncorrespondences < max_correspondences);
		correspondences1[ncorrespondences] = pnts1->PointPosition(i);
		correspondences2[ncorrespondences] = closest->GetPosition();
		ncorrespondences++;
	}
	
	// Compute correspondences for points2 -> mesh1
	for (int i = 0; i < np2; i++)
	{
		R3Point position2 =pnts2->PointPosition(i);
		position2.Transform(affine21);
		const PointInSet *closest = tree1->FindClosest(position2, 0, FLT_MAX);
		assert(closest!=NULL);
		assert(ncorrespondences < max_correspondences);
		correspondences1[ncorrespondences] = closest->GetPosition();
		correspondences2[ncorrespondences] = pnts2->PointPosition(i);
		ncorrespondences++;
	}
	
	// Return number of correspondences
	assert(ncorrespondences == np1 + np2);
	assert(ncorrespondences == max_correspondences);
	return ncorrespondences;
}

CorrsAffineAlign::CorrsAffineAlign(const VKString & corrName,
								   GeomShape * shapeFrom, GeomShape * shapeTo)
: CorrsPair(corrName, shapeFrom, shapeTo)
{
	m_Affine12 = R3Affine(R4identity_matrix);
	m_fDistNorm = 1.;
	m_fIntegratedValue = -1;
}

CorrsAffineAlign::CorrsAffineAlign(const VKString & corrName,
								   GeomShape * shapeFrom, GeomShape * shapeTo, 
								   double distNorm, R3Affine affine12)
: CorrsPair(corrName, shapeFrom, shapeTo)
{
	m_Affine12 = affine12;
	
	ShapePointCloud * pnts2 = SafeCastPntsTo();
	assert(pnts2!=NULL);
	R3Point c = pnts2->Centroid();
	double r = pnts2->AverageRadius(c);
	m_fDistNorm = distNorm * r;	
	m_fIntegratedValue = -1;
}

CorrsAffineAlign::~CorrsAffineAlign()
{
	
}

int CorrsAffineAlign::GetCorrs(const ShapePoint & pnt,
							   std::vector<ShapePoint> & corrs, 
							   const VKString & corrValName,
							   std::vector<double> *corrVals)
{
	// transform from point
	R3Point p1 = pnt.GetPosition();
	m_Affine12.Apply(p1);
	
	// get kd tree on 'to' points
	ShapePointCloud * pnts2 = SafeCastPntsTo();
	assert(pnts2!=NULL);
	const R3Kdtree<const PointInSet *> * tree2 = pnts2->GetKdTree();
	assert(tree2!=NULL);

	// map from point
	RNArray<const PointInSet *> corrs2;
	tree2->FindAll(p1, 0, m_fDistNorm, corrs2);
	bool defaultCorr = (corrValName=="default") && (corrVals!=NULL);
	for (int i=0; i<corrs2.NEntries(); i++)
	{
		corrs.push_back(corrs2[i]->GetModelPoint());
		
		if (defaultCorr)
		{
			R3Point p2 = corrs2[i]->GetPosition();
			double dist = R3Distance(p1, p2) / m_fDistNorm;
			corrVals->push_back(exp(-dist*dist));
		}
	}
	return (int)corrs.size();
}

ShapePoint CorrsAffineAlign::BestCorr(const ShapePoint & pnt,
									  GeomShape * toShape,
									  const VKString & corrValName,
									  double * corrValue)
{
	// todo: can also map point cloud to mesh
	
	// transform from point
	R3Point p1 = pnt.GetPosition();
	m_Affine12.Apply(p1);
	
	// get kd tree on 'to' points
	ShapePointCloud * pnts2 = SafeCastPntsTo();
	assert(pnts2!=NULL);
	const R3Kdtree<const PointInSet *> * tree2 = pnts2->GetKdTree();
	assert(tree2!=NULL);
	
	// map from point
	const PointInSet * bestCorr = tree2->FindClosest(p1);
	bool defaultCorr = (corrValName=="default") && (corrValue!=NULL);
	
	if (bestCorr!=NULL)
	{	
		if (defaultCorr && corrValue!=NULL)
		{
			R3Point p2 = bestCorr->GetPosition();
			double dist = R3Distance(p1, p2) / m_fDistNorm;
			*corrValue = exp(-dist*dist);
		}
		return bestCorr->GetModelPoint();
	}
		
	return ShapePoint();
}

double CorrsAffineAlign::DefaultIntegratedValue(const VKString & ) const
{
	if (m_fIntegratedValue==-1)
	{
		CorrsAffineAlign * ncThis = (CorrsAffineAlign*)this;
		ShapePointCloud * pnts1 = ncThis->SafeCastPntsFrom();
		ShapePointCloud * pnts2 = ncThis->SafeCastPntsTo();
		assert(pnts1!=NULL && pnts2!=NULL);
		
		const GeomShape * mainShape1 = ncThis->ModelFrom()->GetGeomShape(Model3D::m_sMainGeomName);
		const GeomShape * mainShape2 = ncThis->ModelTo()->GetGeomShape(Model3D::m_sMainGeomName);
		assert(mainShape1!=NULL && mainShape2!=NULL);

		const ShapeMesh * mainMesh1 = mainShape1->SafeCastToMesh();
		const ShapeMesh * mainMesh2 = mainShape2->SafeCastToMesh();
		assert(mainMesh1!=NULL && mainMesh2!=NULL);
		
		R3Affine aff12 = GetTransformation();
		R3Affine aff21 = InverseAffine(aff12);
		
		double tighterDistNorm = .5 * m_fDistNorm;
		
		double score1 = 0;
		for (int i=0; i<pnts1->NumPoints(); i++)
		{
			R3Point p1 = pnts1->PointPosition(i);
			aff12.Apply(p1);
			R3Point p2 = mainMesh2->GetPointPositionOnMesh(p1);
			
			double dist = R3Distance(p1, p2) / tighterDistNorm;
			score1 += exp(-dist*dist);
		}
		score1 /= pnts1->NumPoints();
		
		double score2 = 0;
		for (int i=0; i<pnts2->NumPoints(); i++)
		{
			R3Point p2 = pnts2->PointPosition(i);
			R3Point p2new = p2;
			aff21.Apply(p2new);
			R3Point p1 = mainMesh1->GetPointPositionOnMesh(p2new);
			aff12.Apply(p1);
			
			double dist = R3Distance(p1, p2) / tighterDistNorm;
			score2 += exp(-dist*dist);
		}
		score2 /= pnts2->NumPoints();
		
		((double&)m_fIntegratedValue) = (score1+score2) / 2.;
	}
	
	return m_fIntegratedValue;
}

void CorrsAffineAlign::WriteShapeCorrs(FILE * file)
{
//	PairMapParams params("CorrsAffineAlign", GetCorrsName(), ShapeFrom(), ShapeTo());
//	CorrsPair::WriteHeader(file, params);
	
	double m[16];
	const R4Matrix& mtx = m_Affine12.Matrix();
	
	for (int i=0; i<4; i++)
	for (int j=0; j<4; j++)	
		m[i*4 + j] = mtx[i][j];
	fwrite(&m_fDistNorm, sizeof(double), 1, file);
	fwrite(m, sizeof(double), 16, file);
	fwrite(&m_fIntegratedValue, sizeof(double), 1, file);
}

void CorrsAffineAlign::ReadShapeCorrs(FILE * file)
{
	double m[16];
	fread(&m_fDistNorm, sizeof(double), 1, file);
	fread(m, sizeof(double), 16, file);
	
	m_Affine12 = R3Affine(R4Matrix(m[0],  m[1],  m[2],  m[3],
								   m[4],  m[5],  m[6],  m[7],
								   m[8],  m[9],  m[10], m[11],
								   m[12], m[13], m[14], m[15]));
	fread(&m_fIntegratedValue, sizeof(double), 1, file);
}

const R3Affine & CorrsAffineAlign::GetTransformation() const
{
	return m_Affine12;
}

VKString CorrsAffineAlign::GetCorrsClassName() const
{
	return "CorrsAffineAlign";
}


