/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TMatch.h"

#include "VkFunctions.h"
#include "LinAlg.h"

#include "CoAlignedParts.h"

#include "CorrsAffineAlign.h"

#include "PartTransforms.h"
#include "STemplate.h"
#include "PartOfEdge.h"
#include "PartConnectorEdge.h"

#include "TMatchEnergy.h"
#include "ConsistentPartsAlg.h"


void CoAlignedParts::PrintLog(const VKString & str, std::ofstream * fileStream)
{
	std::cout<<str.c_str()<<std::flush;
	if (fileStream!=NULL)
		(*fileStream)<<str.c_str();
}


MatchedPart::MatchedPart(int matchID, int deformID)
{
	m_iMatchID = matchID;
	m_iDeformID = deformID;

}

PartCluster::PartCluster(int clusterID, int seedID, int nodeID, int partID, int nParts)
{
	m_iClusterID = clusterID;
	m_iSeedID = seedID;
	m_iNodeID = nodeID;
	m_iPartID = partID;
	m_iNParts = nParts;
}

PartClustering::PartClustering(int clusteringID, int seedID, int nodeID, int partID, int nParts)
{
	m_iClusteringID = clusteringID;
	m_iSeedID = seedID;
	m_iNodeID = nodeID;
	m_iPartID = partID;
	m_iNParts = nParts;
	m_bPrecomputedMatch2Cluster = false;
}

void PartClustering::GetMatch(int matchID, int nodeID, int * clusterID, int * partInClusterID, bool & valid) const
{
	valid = true;
	PartClustering * nc = (PartClustering*)this;
	if (!m_bPrecomputedMatch2Cluster)
	{
		for (int i=0; i<(int)m_AllClusters.size(); i++)
		for (int j=0; j<(int)m_AllClusters[i].m_AllParts.size(); j++)
		{
			int cmid = m_AllClusters[i].m_AllParts[j].m_iMatchID;
			int cpid = m_AllClusters[i].m_AllParts[j].m_iDeformID;
			while (cmid >= m_Match2ClusterAndPartID.size())
				nc->m_Match2ClusterAndPartID.push_back(std::map<int, ClusterPartPair>());
			ClusterPartPair & cpp = nc->m_Match2ClusterAndPartID[cmid][cpid];
			cpp.m_iClusterID = i;
			cpp.m_iPartID = j;
		}

		nc->m_bPrecomputedMatch2Cluster = true;
	}
	
	assert (matchID < (int)m_Match2ClusterAndPartID.size());
	std::map<int, ClusterPartPair>::iterator iter = nc->m_Match2ClusterAndPartID[matchID].find(nodeID);
	assert(iter != nc->m_Match2ClusterAndPartID[matchID].end());
	*clusterID = iter->second.m_iClusterID;
	*partInClusterID = iter->second.m_iPartID;
}

CoAlignedParts::CoAlignedParts(STemplate * st, std::vector<TMatchEnergy * > & matches)
: m_pST(st), m_Matches(matches)
{

}

CoAlignedParts::~CoAlignedParts()
{
	
}

int CoAlignedParts::NClusterings() const
{
	return (int)m_AllClusterings.size();
}

const PartClustering & CoAlignedParts::GetClustering(int clusteringID) const
{
	assert(clusteringID>=0 && clusteringID < (int)m_AllClusterings.size());
	return m_AllClusterings[clusteringID];
}


int CoAlignedParts::GetClusteringID(int seedID, int nodeID, int partID, int nParts) const
{
	for (int i=0; i<(int)m_AllClusterings.size(); i++)
	{
		if (m_AllClusterings[i].m_iSeedID == seedID && m_AllClusterings[i].m_iNodeID == nodeID
			&& m_AllClusterings[i].m_iPartID == partID && m_AllClusterings[i].m_iNParts == nParts)
			return i;
	}
	return -1;
}

int CoAlignedParts::NClusters(int clusteringID) const
{
	if (clusteringID<0 || clusteringID>=NClusterings())
		return 0;
	return (int)GetClustering(clusteringID).m_AllClusters.size();
}

const PartCluster & CoAlignedParts::GetCluster(int clusteringID, int clusterID) const
{
	const PartClustering & c = GetClustering(clusteringID);
	assert(clusterID>=0 && clusterID < (int)c.m_AllClusters.size());
	return c.m_AllClusters[clusterID];
}

int CoAlignedParts::NPartsInCluster(int clusteringID, int clusterID) const
{
	assert(clusteringID>=0 && clusteringID<NClusterings());
	const PartClustering & c = GetClustering(clusteringID);
	assert(clusterID>=0 && clusterID<(int)c.m_AllClusters.size());
	return (int)c.m_AllClusters[clusterID].m_AllParts.size();
}

const MatchedPart & CoAlignedParts::GetPart(int clusteringID, int clusterID, int partID) const
{
	const PartCluster & c = GetCluster(clusteringID, clusterID);
	assert(partID>=0 && partID < (int)c.m_AllParts.size());
	return c.m_AllParts[partID];
}


void CoAlignedParts::ComputeConfidences(int clusteringID, int clusterID, int CCFflags)
{
	std::vector<double> xVals;
	std::vector<double> yVals;
	std::vector<double> zVals;
	
	// pre-compute aligned points
	std::vector<ShapePointCloud * > partPnts;
	//std::vector<ShapePointCloud * > partPntsUVW;
	int nparts = NPartsInCluster(clusteringID, clusterID);
	for (int i=0; i<nparts; i++)
	{
		ShapePointCloud * alignedPnts = new ShapePointCloud(NULL, "alignedpnts");
		//ShapePointCloud * uvwPnts = new ShapePointCloud(NULL, "alignedpnts");
		partPnts.push_back(alignedPnts);
		//partPntsUVW.push_back(uvwPnts);
		MatchedPart & p1 = (MatchedPart &)GetPart(clusteringID, clusterID, i);
		assert(p1.m_iMatchID>=0 && p1.m_iDeformID>=0);
		TMatch * m1 = m_Matches[p1.m_iMatchID]->GetMatch();
		const TemplateNodeDeform & d1 = m1->Deformation(p1.m_iDeformID);
		if (p1.m_PntsSubset.size()==0)
		{
			for (int p=0; p<m1->NShapePrimitives(); p++)
			{
				if (m1->LocalNodeID(p) == p1.m_iDeformID)
					p1.m_PntsSubset.push_back(p);
			}
		}
		for (int ps=0; ps<(int)p1.m_PntsSubset.size(); ps++)
		{
			double uvw[3];
			R3Point pos = m1->AlignedToNodePointPosition(p1.m_iDeformID, p1.m_PntsSubset[ps], 0);
			xVals.push_back(pos.X());
			yVals.push_back(pos.Y());
			zVals.push_back(pos.Z());
			alignedPnts->AddPoint(pos);
			//partPntsUVW->AddPoint(R3Point(uvw[0], uvw[1], uvw[2]));
		}
	}
	
	// create grid
	std::sort(xVals.begin(), xVals.end());
	std::sort(yVals.begin(), yVals.end());
	std::sort(zVals.begin(), zVals.end());
	assert(xVals.size()>0);
	int minID = (int)xVals.size() * .1;
	int maxID = vkMin((int)xVals.size() * .9, (int)xVals.size()-1);
	double minX = xVals[minID];
	double minY = yVals[minID];
	double minZ = zVals[minID];
	double maxX = xVals[maxID];
	double maxY = yVals[maxID];
	double maxZ = zVals[maxID];
	
	double dx = maxX - minX;
	double dy = maxY - minY;
	double dz = maxZ - minZ;
	
	if (dx == 0) dx = .1;
	if (dy == 0) dz = .1;
	if (dz == 0) dy = .1;
	double expand = .5;
	minX -= dx * expand;
	maxX += dx * expand;
	minY -= dy * expand;
	maxY += dy * expand;
	minZ -= dz * expand;
	maxZ += dz * expand;
	
	int gridRes = 32;
	double dmax = vkMax(dx, vkMax(dy, dz));
	int xRes = gridRes * dx / dmax;
	int yRes = gridRes * dy / dmax;
	int zRes = gridRes * dz / dmax;
	
	double thresh = .1;
	
	R3Grid * aveDistGrid = new R3Grid(xRes, yRes, zRes, R3Box(minX, minY, minZ, maxX, maxY, maxZ));
	R3Grid * tempGrid = new R3Grid(xRes, yRes, zRes, R3Box(minX, minY, minZ, maxX, maxY, maxZ));
	aveDistGrid->Clear();
	
	// draw to grid
	for (int i=0; i<nparts; i++)
	{
		tempGrid->Clear();
		//timer.WriteProgress("conf", i, nparts);
		MatchedPart & p1 = (MatchedPart &)GetPart(clusteringID, clusterID, i);
		assert(p1.m_iMatchID>=0 && p1.m_iDeformID>=0);
		TMatch * m1 = m_Matches[p1.m_iMatchID]->GetMatch();
		const TemplateNodeDeform & d1 = m1->Deformation(p1.m_iDeformID);

		for (int ps=0; ps<(int)p1.m_PntsSubset.size(); ps++)
			tempGrid->RasterizeWorldPoint(partPnts[i]->PointPosition(ps), 1);
		//tempGrid->Blur(gridRes / 10);
		//tempGrid->Blur(2. * thresh / dmax);
		
		tempGrid->SquaredDistanceTransform();
		tempGrid->Gauss(thresh, false);
		
		aveDistGrid->Add(*tempGrid);
	}
	
	// for all pairs - compute distance
	double minValue = 10E10;
	double maxValue = -10e10;
	double totalValue = 0;
	int totalCount = 0;
	
	//std::cout<<"Computing confidence:"<<std::endl;
	TimeProfiler timer;
	
	for (int i=0; i<nparts; i++)
	{
		//timer.WriteProgress("conf", i, nparts);
		MatchedPart & p1 = (MatchedPart &)GetPart(clusteringID, clusterID, i);
		assert(p1.m_iMatchID>=0 && p1.m_iDeformID>=0);
		TMatch * m1 = m_Matches[p1.m_iMatchID]->GetMatch();
		const TemplateNodeDeform & d1 = m1->Deformation(p1.m_iDeformID);
		
		p1.m_Confidences.clear();
		for (int ps=0; ps<(int)p1.m_PntsSubset.size(); ps++)
		{	
			double aveDist = aveDistGrid->WorldValue(partPnts[i]->PointPosition(ps));
//			double aveDist = 0;
//			for (int j=0; j<nparts; j++)	// TODO: use volume to have linear complexity
//			{
//				const PointInSet * closest = partPnts[j]->GetKdTree()->FindClosest(p);
//				if (closest != NULL)
//					aveDist += exp(-pow(R3Distance(p, closest->GetPosition())/thresh, 2.));
//				else
//					aveDist += 0;
//			}
			aveDist /= nparts;
			p1.m_Confidences.push_back(aveDist);
			p1.m_ToSubsetMap[p1.m_PntsSubset[ps]]  = ps;
			
			minValue = vkMin(minValue, aveDist);
			maxValue = vkMax(maxValue, aveDist);
			totalValue = aveDist;
			totalCount++;
		}
	}
	//std::cout<<std::endl;
	
	// record min / max value
	PartCluster & c = (PartCluster &)GetCluster(clusteringID, clusterID);
	c.m_AveConfidence = totalValue / totalCount;
	c.m_MinConfidence = minValue;
	c.m_MaxConfidence = maxValue;
	for (int i=0; i<(int)partPnts.size(); i++)
	{
		delete partPnts[i];
		//delete partPntsUVW[i];
	}
}

double CoAlignedParts::GetPartConfidenceValue(int clusteringID, int clusterID, int partID,
											  int matchID, bool & valid) const
{
	MatchedPart & part = (MatchedPart &)GetPart(clusteringID, clusterID, partID);
	valid = part.m_iMatchID==matchID;
	if (!valid) return -1;
	double conf = 0;
	for (int i=0; i<(int)part.m_Confidences.size(); i++)
		conf += part.m_Confidences[i];
	return conf;
}

double CoAlignedParts::GetConfidenceValue(int clusteringID, int clusterID, int partID,
										  int matchID, int pointID, bool & valid) const
{
	MatchedPart & part = (MatchedPart &)GetPart(clusteringID, clusterID, partID);
	valid = part.m_iMatchID==matchID;
	if (!valid) return -1;
	std::map<int, int>::iterator iter = part.m_ToSubsetMap.find(pointID);
	valid = iter != part.m_ToSubsetMap.end();
	if (!valid) return -1;
	valid = iter->second>=0 && iter->second < (int)part.m_Confidences.size();
	if (!valid) return -1;
	return part.m_Confidences[iter->second];
}

void CoAlignedParts::GetConfidenceValBounds(int clusteringID, int clusterID, double * minValue, double * maxValue) const
{
	const PartCluster & c = GetCluster(clusteringID, clusterID);
	*minValue = c.m_MinConfidence;
	*maxValue = c.m_MaxConfidence;
}

void CoAlignedParts::GetClusterForPoint(int clusteringID, int matchID, int pointID,
										int * clusterID, int * partInClusterID, bool & valid) const
{
	//const PartClustering & c = GetClustering(clusteringID);
	valid = matchID>=0 && matchID < (int)m_Matches.size();
	if (!valid) return;
	int lnid = m_Matches[matchID]->GetMatch()->LocalNodeID(pointID);

	GetClusterWithPart(clusteringID, matchID, lnid, clusterID, partInClusterID, valid);
	
}

void CoAlignedParts::GetClusterWithPart(int clusteringID, int matchID, int nodeID,
										int * clusterID, int * partInClusterID, bool & valid) const
{
	valid = (matchID>=0 && matchID < (int)m_Matches.size());
	if (!valid) return;
	
	GetClustering(clusteringID).GetMatch(matchID, nodeID, clusterID, partInClusterID, valid);
}


int CoAlignedParts::GetClusteringFromMatchesID(int configID)
{
	std::map<int, int>::iterator iter = m_Seed2FromMatchesClustering.find(configID);
	if (iter != m_Seed2FromMatchesClustering.end())
		return iter->second;
	else
	{
		int clusteringID = CreateClustersFromMatches(configID);
		m_Seed2FromMatchesClustering[configID] = clusteringID;
		return clusteringID;
	}
}



int CoAlignedParts::CreateClustersFromMatches(int configID)
{
	int clusteringID = (int)m_AllClusterings.size();
	m_AllClusterings.push_back(PartClustering(clusteringID, configID, -1, -1, -1));

	m_Seed2FromMatchesClustering[configID] = clusteringID;
	
	std::map<int, std::map<int, std::map<int, int> > > deform2cluster;
	
	// figure out number of clusters
	for (int i=0; i<(int)m_Matches.size(); i++)
	for (int j=0; j<(int)m_Matches[i]->GetMatch()->NTemplateNodes(); j++)
	{
		const TemplateNodeDeform & deform = m_Matches[i]->GetMatch()->Deformation(j);
		std::map<int, std::map<int, std::map<int, int> > >::iterator iter1 = deform2cluster.find(deform.NodeID());
		
		if (iter1 == deform2cluster.end())
		{
			deform2cluster[deform.NodeID()] = std::map<int, std::map<int, int> >();
			iter1 = deform2cluster.find(deform.NodeID());
		}
		assert (iter1 != deform2cluster.end());
		
		std::map<int, std::map<int, int> >::iterator iter2 = iter1->second.find(deform.NParts());
		if (iter2 == iter1->second.end())
		{
			iter1->second[deform.NParts()] = std::map<int, int>();
			iter2 = iter1->second.find(deform.NParts());
		}
		assert(iter2 != iter1->second.end());
		
		std::map<int, int>::iterator iter3 = iter2->second.find(deform.PartID());
		if (iter3 == iter2->second.end())
		{
			std::vector<PartCluster> & c = m_AllClusterings[clusteringID].m_AllClusters;
			int clusterID = (int)c.size();
			c.push_back(PartCluster(clusterID, configID, deform.NodeID(), deform.PartID(), deform.NodeID()));
			iter2->second[deform.PartID()] = clusterID;
		}
	}
	
	// save parts to clusters
	for (int i=0; i<(int)m_Matches.size(); i++)
	for (int j=0; j<(int)m_Matches[i]->GetMatch()->NTemplateNodes(); j++)
	{
		const TemplateNodeDeform & deform = m_Matches[i]->GetMatch()->Deformation(j);
		std::map<int, std::map<int, std::map<int, int> > >::iterator iter1 = deform2cluster.find(deform.NodeID());
		assert(iter1!=deform2cluster.end());
		std::map<int, std::map<int, int> >::iterator iter2 = iter1->second.find(deform.NParts());
		assert(iter2 != iter1->second.end());
		std::map<int, int>::iterator iter3 = iter2->second.find(deform.PartID());
		assert (iter3 != iter2->second.end());
		
		std::vector<MatchedPart> & ap = m_AllClusterings[clusteringID].m_AllClusters[iter3->second].m_AllParts;
		ap.push_back(MatchedPart(i, j));
	}

	return clusteringID;
}

void CoAlignedParts::ClusterStyles(int configID, int nodeID, int partID, int nParts)
{
	int methodFlags = ConsistentPartsAlg::CPM_InitWhole | ConsistentPartsAlg::CPM_ClusterSpectralAgglomerative;
	methodFlags = methodFlags | ConsistentPartsAlg::CPM_AveClusterTakeAll;
		
	int clusteringID = GetClusteringID(configID, nodeID, partID, nParts);
	if (clusteringID==-1)
	{
		clusteringID = (int)m_AllClusterings.size();
		m_AllClusterings.push_back(PartClustering(clusteringID, configID, nodeID, partID, nParts));
	}
	
	m_AllClusterings[clusteringID].m_AllClusters.clear();

	// find final clustering
	ConsistentPartsAlg consistentParts(m_pST, m_Matches, methodFlags, true);
	std::vector<PointsCluster> finalClusters;
	consistentParts.ExtractConsistentParts(nodeID, 0, 1, finalClusters);

	int nClusters = 0;
	for (int i=0; i<(int)finalClusters.size(); i++)
		nClusters = vkMax(nClusters, finalClusters[i].m_iClusterID+1);
		
	for (int i=0; i<nClusters; i++)
		m_AllClusterings[clusteringID].m_AllClusters.push_back(PartCluster(i, configID, nodeID, partID, nParts));
	
	std::vector<PartCluster> & ac = m_AllClusterings[clusteringID].m_AllClusters;
	for (int i=0; i<(int)finalClusters.size(); i++)
	{
		int cID = finalClusters[i].m_iClusterID;
		int mID = finalClusters[i].m_iModelID;
		int lID = finalClusters[i].m_iLocalNodeID;
		ac[cID].m_AllParts.push_back(MatchedPart(mID, lID));
	}
	
//	std::cout<<"Pruning clusters: "<<m_PartStyleClusters.size()<<std::endl;
//	for (int i=0; i<(int)m_PartStyleClusters.size(); i++)
//		std::cout<<"Cluster["<<i<<"] = "<<m_PartStyleClusters[i].size()<<std::endl;
//	int removed = 0;
//	for (int i=0; i<(int)m_PartStyleClusters.size(); i++)
//		if (m_PartStyleClusters[i].size()<=1)
//		{
//			removed++;
//			m_PartStyleClusters.erase(m_PartStyleClusters.begin()+i);
//			i--;
//		}
//	std::cout<<"Removed: "<<removed<<" clusters. Final count: "<<m_PartStyleClusters.size()<<std::endl;
}

void CoAlignedParts::ClusterNullNodes(int configID)
{
	assert(false);
}


VKString CoAlignedParts::UpdateTemplateFromAllMatches(const VKString & updateLog, int nShapes)
{
	// based on clustering
	bool updates = false;
	if (m_Matches.size()==0)
		return "";
	
	std::ofstream * textStream = NULL;
	if (updateLog!="none")
	{
		FileManagement::CreateDirsOnPathToFile(updateLog.c_str());
		textStream = new std::ofstream(updateLog.c_str());
		assert(textStream->is_open());
	}
	
	// learn from every match
	std::vector<int> oldTerms;
	PrintLog(VKString("Updating from: ")+VKString::number((int)m_Matches.size())+" matches.\n", textStream);
	TimeProfiler timer;

	// Agglomerative clustering
	std::vector<int> es;
	std::vector<ClusterOfModels> seeds;
	// Identify seeds
	if (m_pST->NSeedConfigs() > 1)
	{
		for (int i=1; i<m_pST->NSeedConfigs(); i++)
		{
			int lid = (int)seeds.size();
			seeds.push_back(ClusterOfModels());
			seeds[lid].m_pSeedDeform = m_pST->CreateSeedConfig(i, NULL, es, R3identity_affine);
			seeds[lid].m_iCurrentSeedID = i;
			seeds[lid].m_sClusterName = m_pST->ConfigName(seeds[lid].m_iCurrentSeedID);
		}
	}
	else
	{
		assert(seeds.size()==0);
		seeds.push_back(ClusterOfModels());
		seeds[0].m_pSeedDeform = m_pST->CreateSeedConfig(0, NULL, es, R3identity_affine);
		seeds[0].m_iCurrentSeedID = -1;
		seeds[0].m_sClusterName = "Default";
	}

	double diffThresh = TMatchEnergy::DeformationDistanceThreshold();
	int minNInCluster = 3;	// vk: const
//	if (nShapes < 50)
//		minNInCluster = 5;	// vk: const
	
	// assign to seeds: also create seeds without parts
	std::vector<int> initMatches;
	for (int i=0; i<(int)m_Matches.size(); i++)
		initMatches.push_back(i);
	std::vector<int> similar2seedsButDifferentParts;
	std::vector<int> available;

	AddMatchesToBestSeeds(initMatches, diffThresh, seeds, available, similar2seedsButDifferentParts, textStream);
	VKString nRemainingStr = VKString::number((int)available.size());
	VKString nToAddStr = VKString::number((int)similar2seedsButDifferentParts.size());
	PrintLog(VKString("ToCluster 0: ")+nRemainingStr+" DiffConfig: "+nToAddStr+"\n", textStream);
	
	// add new configs: due to parts included / not included
		// remove redundant configs
	for (int i=0; i<(int)similar2seedsButDifferentParts.size(); i++)
	{
		for (int j=i+1; j<(int)similar2seedsButDifferentParts.size(); j++)
		{
			int m1 = similar2seedsButDifferentParts[i];
			int m2 = similar2seedsButDifferentParts[j];
			bool sameParts, sameIncluded;
			TMatchEnergy::Compatible(m_Matches[m1]->GetMatch(), m_Matches[m2]->GetMatch(), sameParts, sameIncluded);
			
			if (sameParts && sameIncluded)
			{
				available.push_back(m2);
				similar2seedsButDifferentParts.erase(similar2seedsButDifferentParts.begin() + j);
				j--;
			}
		}
	}
		// add new configs
	VKString nStr = VKString::number((int)similar2seedsButDifferentParts.size());
	if (similar2seedsButDifferentParts.size()>0)
		PrintLog(VKString("New Configs due to DiffParts: ")+nStr+"\n", textStream);
	for (int i=0; i<(int)similar2seedsButDifferentParts.size(); i++)
	{
		available.push_back(similar2seedsButDifferentParts[i]);
		int lid = (int)seeds.size();
		seeds.push_back(ClusterOfModels());
		seeds[lid].m_pSeedDeform = m_Matches[similar2seedsButDifferentParts[i]]->GetMatch()->CreateCopy();
		seeds[lid].m_iCurrentSeedID = -1;
		VKString mName = seeds[lid].m_pSeedDeform->ModelName();
		seeds[lid].m_sClusterName = VKString("OtherConfig_")+mName;
		PrintLog(VKString("DiffPartsCluster: ")+VKString::number(i)+": "+mName+"\n", textStream);
	}
	
	if (similar2seedsButDifferentParts.size()>0)
	{
		std::vector<int> initAvailable = available;
		available.clear();
		similar2seedsButDifferentParts.clear();
		AddMatchesToBestSeeds(initAvailable, diffThresh, seeds, available, similar2seedsButDifferentParts, textStream);
		for (int i=0; i<(int)similar2seedsButDifferentParts.size(); i++)
			available.push_back(similar2seedsButDifferentParts[i]);
	}
	
	PrintLog(VKString("Clustering Remaining: ")+VKString::number((int)available.size())+"\n", textStream);	
	
	// cluster remaining: agglomerative
	// agglomerative clustering
	std::vector<std::vector<int> > outClusters;
	AgglomerativeClustering(available, diffThresh, outClusters, textStream);
	PrintLog(VKString("Resulting Clusters (")+VKString::number((int)outClusters.size())+"):\n", textStream);
	for (int i=0; i<(int)outClusters.size(); i++)
	{
		int mid = available[outClusters[i][0]];
		int lid = (int)seeds.size();

		// print ids to stream
		VKString printStr = VKString("Cluster ")+VKString::number(i)+": ";
		printStr += (outClusters[i].size()<minNInCluster ? VKString("Too Small\n") : VKString("Included\n"));
		printStr += VKString(" (")+VKString::number((int)outClusters[i].size())+")";
		PrintLog(printStr, textStream);
		
		for (int j=0; j<(int)outClusters[i].size(); j++)
		{
			int mid = available[outClusters[i][j]];
			printStr = VKString("    M[")+VKString::number(j)+"]: "+m_Matches[mid]->GetMatch()->ModelName()+"\n";
			PrintLog(printStr, textStream);
		}
			
		if (outClusters[i].size()<minNInCluster)
			continue;
		
		seeds.push_back(ClusterOfModels());
		seeds[lid].m_pSeedDeform = m_Matches[mid]->GetMatch();
		seeds[lid].m_iCurrentSeedID = -1;
		seeds[lid].m_sClusterName = VKString("Learned_")+m_Matches[mid]->GetMatch()->ModelName();
		
		for (int j=0; j<(int)outClusters[i].size(); j++)
		{
			mid = available[outClusters[i][j]];
			seeds[lid].m_ClusterOfModels.push_back(ModelInCluster());
			seeds[lid].m_ClusterOfModels[j].m_pModelMatch = m_Matches[mid];
			seeds[lid].m_ClusterOfModels[j].m_iModelID = mid;
		}
	}
	
	VKString updString="";
	int nextID = m_pST->NSeedConfigs();
	for (int c=0; c<(int)seeds.size(); c++)
	{
		if (seeds[c].m_iCurrentSeedID==-1)	// new seed - clearly an update
			updString += VKString::number(nextID++)+" ";
		else
		{
			double updConstant = .2;	// vk: const
			int ndescribed = m_pST->MaxDescribedInstances(seeds[c].m_iCurrentSeedID);
			if ((int)seeds[c].m_ClusterOfModels.size()>(int)(updConstant * ndescribed))
				updString += VKString::number(seeds[c].m_iCurrentSeedID)+" ";
		}
	}
	if (updString!="")
		updString = VKString("InitConfigs: ")+VKString::number(m_pST->NSeedConfigs())+"\nNUpdated: " + updString;

	
	// update template
	m_Matches.clear();
	for (int c=0; c<(int)seeds.size(); c++)
		UpdateConfig(seeds[c], -1, textStream);	// vk: const - energy thresh
	if (textStream!=NULL)
		textStream->close();
	return updString;
}

void CoAlignedParts::AddMatchesToBestSeeds(const std::vector<int> & matches, double diffThresh,
										   std::vector<ClusterOfModels> & seeds, std::vector<int> & notIncluded,
										   std::vector<int> & notIncludedNewConfigs, std::ofstream  * logStream)
{
	std::cout<<"Add To Best Seeds: "<<matches.size()<<std::endl;
	for (int mlid=0; mlid<(int)matches.size(); mlid++)
	{
		int m = matches[mlid];
		double minDist = -1;
		int bestSeed = -1;
		
		double minDistDifferentParts = -1;
		int bestSeedDifferentParts = -1;
		for (int i=0; i<(int)seeds.size(); i++)
		{
			bool sameParts, sameIncluded;
			double dist = TMatchEnergy::DeformationDistance(m_Matches[m]->GetMatch(),
															seeds[i].m_pSeedDeform, sameParts, sameIncluded);
			//sameIncluded = true; // vk: do not ignore this - see comment in the TMatchEnergy
			if (sameParts)
			{
				if (sameIncluded && (bestSeed==-1 || minDist>dist))	// same parts included
				{
					minDist = dist;
					bestSeed = i;
				}
				else if (!sameIncluded && (bestSeedDifferentParts==-1 || minDistDifferentParts>dist))
				{
					minDistDifferentParts = dist;
					bestSeedDifferentParts = i;
				}
			}
		}
		if (minDist < diffThresh && bestSeed>=0)	// adding to cluster
		{
			ClusterOfModels & seed = seeds[bestSeed];
			int lid = (int)seed.m_ClusterOfModels.size();
			seed.m_ClusterOfModels.push_back(ModelInCluster());
			seed.m_ClusterOfModels[lid].m_pModelMatch = m_Matches[m];
			seed.m_ClusterOfModels[lid].m_iModelID = m;
			VKString mName = m_Matches[m]->GetMatch()->ModelName();
			PrintLog(VKString("   Add ")+mName+" to seed "+VKString::number(bestSeed), logStream);
			PrintLog(VKString(" D=")+VKString::number(minDist)+" / "+VKString::number(diffThresh)+"\n", logStream);
		}
		else if (minDistDifferentParts>=0 && minDistDifferentParts < diffThresh && bestSeedDifferentParts >= 0)	// need to include config
		{
			VKString mName = m_Matches[m]->GetMatch()->ModelName();
			PrintLog(VKString("   NewPartConfig ")+mName+" (best: "+VKString::number(bestSeedDifferentParts)+"\n", logStream);
			notIncludedNewConfigs.push_back(m);
		}
		else
			notIncluded.push_back(m);
	}
}

void CoAlignedParts::AgglomerativeClustering(const std::vector<int> & elements, double terminationThreshold,
											 std::vector<std::vector<int> > & outClusters, std::ofstream * logStream)
{
	int N = (int)elements.size();
	std::vector<Sortable> pairwiseDists;
	
	std::vector<int> element2cluster;
	std::vector<double> pairwiseDistsForElements;
	// compute pairwise distances
	//std::cout<<"PairDist"<<std::endl;
	if (N <= 10)
		PrintLog("ClusteringDists:\n", logStream);

	TimeProfiler timer;
	for (int i=0; i < (int)elements.size(); i++)
	{
		timer.WriteProgress("clustering", i, (int)elements.size());
		element2cluster.push_back(i);
		outClusters.push_back(std::vector<int>());
		outClusters[i].push_back(i);
		for (int j=i+1; j < (int)elements.size(); j++)
		{
			int id = vkMin(i,j) * N + vkMax(i,j);
			int m1id = elements[i];
			int m2id = elements[j];
			bool sameParts, sameIncluded;
			TMatch * match1 = m_Matches[m1id]->GetMatch();
			TMatch * match2 = m_Matches[m2id]->GetMatch();
			double dist = TMatchEnergy::DeformationDistance(match1, match2, sameParts, sameIncluded);
			//sameIncluded = true;	// vk: do not ignore this - see comment in the TMatchEnergy
			if (!sameParts || !sameIncluded)
				dist = terminationThreshold * 1000.;
			pairwiseDists.push_back(Sortable(dist, NULL, id));
			while (id >= (int)pairwiseDistsForElements.size())
				pairwiseDistsForElements.push_back(-1);
			pairwiseDistsForElements[id] = dist;
			
//			if (N <= 10)
//			{
//				VKString str = VKString("D(")+match1->ModelName();
//				str += VKString(", ")+match2->ModelName()+"): "+VKString::number(dist)+"\n";
//				PrintLog(str, logStream);
//			}
		}
	}
	std::cout<<std::endl;
	
	// agglomerative clustering
	while (pairwiseDists.size()>0)
	{
		// find two nearest clusters
		std::sort(pairwiseDists.begin(), pairwiseDists.end());
		if (pairwiseDists[0].value > terminationThreshold)	// can't merge closest - end
			break;
		
		int e1id = pairwiseDists[0].id / N;
		int e2id = pairwiseDists[0].id % N;
		
		assert(e1id>=0 && e1id < (int)element2cluster.size() && e2id>=0 && e2id < (int)element2cluster.size());
		int c1id = vkMin(element2cluster[e1id], element2cluster[e2id]);
		int c2id = vkMax(element2cluster[e1id], element2cluster[e2id]);
		assert(c1id>=0 && c1id < (int)outClusters.size() && c2id>=0 && c2id < (int)outClusters.size());
		
		if (N < 10)
		{
			// DEBUG: print clusters
			PrintLog("Clusters:\n", logStream);
			for (int i=0; i<(int)outClusters.size(); i++)
			for (int j=0; j<(int)outClusters[i].size(); j++)
			{
				if (j==0)
					PrintLog(VKString("     C")+VKString::number(i)+":", logStream);
				PrintLog(m_Matches[elements[outClusters[i][j]]]->GetMatch()->ModelName(), logStream);
				if (j==(int)outClusters[i].size()-1)
					PrintLog(VKString("\n"), logStream);
				else
					PrintLog(VKString(" "), logStream);
			}
//			std::cout<<"Elements -> Clusters: [ ";
//			for (int i=0; i<(int)element2cluster.size(); i++)
//				std::cout<<element2cluster[i]<<" ";
//			std::cout<<"]"<<std::endl;
			PrintLog(VKString("Err = ")+VKString::number(pairwiseDists[0].value)+" / ", logStream);
			PrintLog(VKString::number(terminationThreshold)+"\n", logStream);
			PrintLog(VKString("Merging clusters: ")+VKString::number(c1id)+" "+VKString::number(c2id)+"\n", logStream);
			PrintLog("--------", logStream);
			// END DEBUGING
		}
		
		// merge clusters: outClusters, element2cluster, pairwiseDists
		//	  remove pairwise distances. remember which ones to update
		std::vector<int> idWithC1;
		std::map<int, double> dist2cluster;
		for (int i=0; i<(int)pairwiseDists.size(); i++)
		{
			int itere1 = pairwiseDists[i].id / N;
			int itere2 = pairwiseDists[i].id % N;
			assert(itere1>=0 && itere1 < (int)element2cluster.size() && itere2>=0 && itere2 < (int)element2cluster.size());
			int iterc1 = element2cluster[itere1];
			int iterc2 = element2cluster[itere2];
			if (iterc2 == c2id || iterc1 == c2id)	// need to delete
			{
				if (iterc2 == c2id)
					dist2cluster[iterc1] = pairwiseDists[i].value;
				else
					dist2cluster[iterc2] = pairwiseDists[i].value;
				pairwiseDists.erase(pairwiseDists.begin()+i);
				i--;
			}
			else if (iterc1 == c1id || iterc1 == c1id)
				idWithC1.push_back(i);
		}
		
		//    update pairwise distances (take max)
		for (int i=0; i<(int)idWithC1.size(); i++)
		{
			int itere1 = pairwiseDists[idWithC1[i]].id / N;
			int itere2 = pairwiseDists[idWithC1[i]].id % N;
			assert(itere1>=0 && itere1 < (int)element2cluster.size() && itere2>=0 && itere2 < (int)element2cluster.size());
			int iterc1 = element2cluster[itere1];
			int iterc2 = element2cluster[itere2];
			
			std::map<int, double>::iterator iter = dist2cluster.end();
			if (iterc1==c1id)
				iter = dist2cluster.find(iterc2);
			else
			{
				assert(iterc2==c1id);
				iter = dist2cluster.find(iterc1);
			}
			assert(iter != dist2cluster.end());
			// take max value for distances between clusters
			pairwiseDists[idWithC1[i]].value = vkMax(pairwiseDists[idWithC1[i]].value, iter->second);
		}
			
		//    update elements to cluster mapping
		for (int i=0; i<(int)outClusters[c2id].size(); i++)
			element2cluster[outClusters[c2id][i]] = c1id;
		
		//    copy elements in outClusters
		for (int i=0; i<(int)outClusters[c2id].size(); i++)
			outClusters[c1id].push_back(outClusters[c2id][i]);
		
		// erase cluster
		outClusters.erase(outClusters.begin() + c2id);
		
		// update elements 2 cluster
		for (int i=0; i<(int)element2cluster.size(); i++)
			if (element2cluster[i]>c2id)
				element2cluster[i]--;
			else
				assert(element2cluster[i] < c2id);
	}
}

void CoAlignedParts::UpdateConfig(ClusterOfModels & cluster, double energyUpdThresh, std::ofstream * textStream)
{
	// update from final clusters
	// initialize seed
	std::vector<TemplateNodeDeform*> emptySet;
	TMatch * bm = cluster.m_pSeedDeform->CreateCopy(emptySet);
	
	int seedID = bm->SeedID();
	VKString nModelsStr = VKString(" ")+VKString::number((int)cluster.m_ClusterOfModels.size())+" models";
	// create new seed (if not from existing seeds)
	if (cluster.m_iCurrentSeedID==-1)
	{
		seedID = m_pST->NSeedConfigs();
		bm->SetSeedID(seedID);
		VKString newSeedName = VKString("Learned_")+bm->ModelName();
		if (bm->ModelName()=="none")
			newSeedName = VKString("Default_")+VKString::number(seedID);
		VKString str = VKString("Adding Seed: ")+newSeedName+" id=";
		str += VKString::number(m_pST->NSeedConfigs())+" with "+nModelsStr+"\n";
		PrintLog(str, textStream);
		m_pST->AddOrReplaceSeed(m_pST->NSeedConfigs(), newSeedName, bm, cluster.m_pSeedDeform->SeedID() );
	}
	else if (seedID==0)
		PrintLog(VKString("Updating Seed: Default")+nModelsStr+"\n", textStream);
	else if (seedID > 0 && seedID < m_pST->NSeedConfigs())
		PrintLog(VKString("Updating Seed: ")+m_pST->ConfigName(seedID)+nModelsStr+"\n", textStream);
	else
		assert(false);
	
	for (int mid=0; mid<(int)cluster.m_ClusterOfModels.size(); mid++)
	{
		ModelInCluster & m = cluster.m_ClusterOfModels[mid];
		m.m_pModelMatch->GetMatch()->SetSeedID(seedID);
		assert(m.m_pModelMatch->GetMatch()->GetMatchedGeomAsPnts()!=NULL);
		m_pST->UpdateAllStats(m.m_pModelMatch->GetMatch(), true);
		m_Matches.push_back(m.m_pModelMatch);
	}
}







