/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VkFunctions.h"

#include "PartBox.h"
#include "ShapePointCloud.h"
#include "ProbDistribution.h"
#include "TMatch.h"
#include "PartTransforms.h"
#include "TMatchEnergy.h"

PartBox * PartBox::m_pNullPart = NULL;
PartBox * PartBox::m_pEmptyPart = NULL;

PartBox * PartBox::GetNullPart()
{
	if (m_pNullPart==NULL)
	{
		m_pNullPart = new PartBox(NULL);
		m_pNullPart->InitNull();
	}
	return m_pNullPart;
}

PartBox * PartBox::GetEmptyPart()
{
	if (m_pEmptyPart!=NULL)
	{
		m_pEmptyPart = new PartBox(NULL);
		m_pEmptyPart->InitEmpty();
	}
	return m_pEmptyPart;
}

/////////// PART BOX ////////
PartBox::PartBox(PartsCollection * partStorage)
{
	m_pLocalGeometry = NULL;
	
	m_pSamplingOfABox = NULL;
	m_iPartGeometry = -1;
	m_pCollection = partStorage;

	m_BoxType = BOXTYPE_Invalid;
	m_Centroid = R3Point(0,0,0);
	m_Orientation = R3Triad(R3Vector(1,0,0), R3Vector(0,1,0), R3Vector(0,0,1));
	m_Scale = R3Vector(0, 0, 0);
	m_fFittingError = 0;
	
	m_pBoxDeformation = &(IdentityTemplateDeform::s_IdentityDeform);
}

PartBox::~PartBox()
{
	if (m_pSamplingOfABox!=NULL)
		delete m_pSamplingOfABox;
	if (m_pLocalGeometry!=NULL)
		delete m_pLocalGeometry;
}

PartBox * PartBox::CreateCopy(PartsCollection * newStorage, int partoffset) const
{
	PartBox * bc = new PartBox(newStorage);
	bc->m_BoxType = m_BoxType;
	bc->m_Centroid = m_Centroid;
	bc->m_Orientation = m_Orientation;
	bc->m_Scale = m_Scale;
	bc->m_iPartGeometry = m_iPartGeometry + partoffset;
	
	bc->m_pLocalGeometry = NULL;
	bc->m_pSamplingOfABox = NULL;
	bc->m_fFittingError = 0;
	bc->m_pBoxDeformation = &(IdentityTemplateDeform::s_IdentityDeform);
	return bc;
}

PartBox * PartBox::CreateCopy() const
{
	PartBox * newBox = new PartBox(m_pCollection);
	newBox->InitCopy(*this);
	return newBox;
}

void PartBox::InitPart(int partID, double fitErr)
{
	assert(m_pCollection!=NULL);
	m_BoxType = BOXTYPE_MeshPoints;
	m_iPartGeometry = partID;
	m_fFittingError = fitErr;
	
	m_Centroid = m_pCollection->GetPart(partID)->Centroid();
	m_Orientation = m_pCollection->GetPart(partID)->Orientation(); 
	m_Scale = m_pCollection->GetPart(partID)->Scale();
}

const PartGeometry * PartBox::GetPart() const
{
	if (m_BoxType!=BOXTYPE_MeshPoints)
		return NULL;
	assert(m_pCollection!=NULL);
	return m_pCollection->GetPart(m_iPartGeometry);
}

void PartBox::FindOptimalBoxFit(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
								R3Point & center, R3Triad & orient, R3Vector & scaleOut, double & fitErr)
{
	assert(ptsIDs.size()>0);
	// find principal axes
	double variances[3];
	double centroid[3];
	R3Point c =  pnts->Centroid();
	double aveRad = pnts->AverageRadius(c);
	R3Triad pcaAxes = pnts->PCAFromPoints(ptsIDs, variances, centroid);
	pcaAxes.Normalize();
	
	R3Kdtree<const PointInSet *> * kdTree = pnts->CreateSubsetKdTree(ptsIDs);
	R3Point sc = R3Point(centroid[0]-c.X(), centroid[1]-c.Y(), centroid[2]-c.Z()) / aveRad;
	
	for (int i=0; i<3; i++)
		variances[i] = 1.*sqrt(variances[i]) / aveRad;
	R3Vector scale(variances[0], variances[1], variances[2]);
	
//	if (variances[0]!=0 || variances[1]!=0 || variances[2]!=0)	// optimize
//	{
//		bool planarBox = true;
//		
//		// try to get a tighter fit
//		int nIter = 1000;
//		int iter = 0;
//		int nChanges = 1;
//		// steps: centroid (x,y,z), triad (axis=x,y,z, angle), scale (x, y, z)
//		//while (iter < nIter && nChanges > 0)
//		{
//			nChanges = 0;
//			
//			// not optimization for orientations...
//			for (int optimStep = 0; optimStep<1; optimStep++)
//				for (int axis=0; axis<3; axis++)
//				{				
//					double step;
//					double minstep;	
//					if (optimStep==0)	// scale
//					{
//						step = .1;
//						minstep = .02;	// vk: const	
//						if (scale[axis]==0)
//							continue;	// do not rescale if already 0
//					}
//					else if (optimStep==1)	// rotation
//					{
//						step = 3.14 / 2.;
//						minstep = 3.14 / 16;	// vk: const
//					}
//					
//					double current = AveDistance(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale);
//					double prev = EvalStep(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale, optimStep, axis, -step);
//					double next = EvalStep(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale, optimStep, axis, step);
//					
//					while (iter < nIter)
//					{
//						if (current <= next && current <= prev)
//						{
//							if (step <= minstep)
//								break;
//							else
//							{
//								step = vkMax(step/2., minstep);
//								prev = EvalStep(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale, optimStep, axis, -step);
//								next = EvalStep(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale, optimStep, axis, step);
//							}
//						}
//						else if (next < current && next <= prev)
//						{
//							prev = current;
//							current = next;
//							ChooseStep(sc, pcaAxes, scale, optimStep, axis, step);
//							next = EvalStep(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale, optimStep, axis, step);
//							iter++;
//							nChanges++;
//						}
//						else if (prev < current && prev < next)
//						{
//							next = current;
//							current = prev;
//							ChooseStep(sc, pcaAxes, scale, optimStep, axis, -step);
//							prev = EvalStep(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale, optimStep, axis, -step);
//							iter++;
//							nChanges++;
//						}
//						else
//							assert(false);
//					}
//				}
//		}
//	}
	
	
	double err = AveDistance(pnts, ptsIDs, kdTree, c, aveRad, sc, pcaAxes, scale);
	
	delete kdTree;
	
	fitErr = err;
	center = sc;
	orient = pcaAxes;
	scaleOut = scale;
}

int PartBox::FindBestPartFit(int partID, R3Point & center, R3Triad & orient, R3Vector & scale, double & fitErr)
{ 	
	assert(false);
}

double PartBox::GetFittingErr() const
{
	return m_fFittingError;
}

double PartBox::EvalStep(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
						 R3Kdtree<const PointInSet *> * kdTree, const R3Point & center, double aveRad,
						 const R3Point & boxCenter, const R3Triad & boxAxes, const R3Vector & boxScale,
						 int optimStep, int dim, double step)
{		
	if (optimStep == 0) // scale
	{
		R3Vector test = boxScale;
		test[dim] = boxScale[dim] + step;
		return AveDistance(pnts, ptsIDs, kdTree, center, aveRad, boxCenter, boxAxes, test);
	}
	else if (optimStep == 1)	// orientation
	{
		R3Triad test = boxAxes;
		test.Rotate(dim, step);
		return AveDistance(pnts, ptsIDs, kdTree, center, aveRad, boxCenter, test, boxScale);
	}
	else
		assert(false);	// unknown step
}

double PartBox::ChooseStep(R3Point & boxCenter, R3Triad & boxAxes, R3Vector & boxScale, 
						   int optimStep, int dim, double step)
{
	if (optimStep == 0)	// scale
		boxScale[dim] = boxScale[dim] + step;
	else if (optimStep == 1)	// orientation
		boxAxes.Rotate(dim, step);
	else
		assert(false);	// unknown step		
}


void PartBox::InitNull()
{
	m_BoxType=BOXTYPE_NullBox;
}

void PartBox::InitEmpty()
{
	R3Triad xyz(R3Vector(1, 0, 0), R3Vector(0, 1, 0), R3Vector(0, 0, 1));
	InitBox(R3Point(0, 0, 0), xyz, R3Vector(-1, -1, -1), 0);
	m_BoxType = BOXTYPE_EmptyBox;
}

void PartBox::InitBox(const R3Point & centroid, const R3Triad & orient, const R3Vector & pcaScale, double fitErr)
{
	m_Centroid = centroid;
	m_Orientation = orient;
	m_Scale = pcaScale;

	m_BoxType = BOXTYPE_BoxSurface;
	m_fFittingError = fitErr;	
}

PartBox & PartBox::operator = (const PartBox & bbox)
{
	this->InitCopy(bbox);
	return *this;
}

void PartBox::InitCopy(const PartBox & bbox)
{
	if (bbox.IsType(PartBox::BOXTYPE_NullBox))
		InitNull();
	else if (bbox.IsType(PartBox::BOXTYPE_EmptyBox))
		InitEmpty();
	else if (bbox.IsType(PartBox::BOXTYPE_BoxSurface))
		InitBox(bbox.m_Centroid, bbox.m_Orientation, bbox.m_Scale, bbox.m_fFittingError);
	else if (bbox.IsType(PartBox::BOXTYPE_MeshPoints))
	{
		InitPart(bbox.m_iPartGeometry, bbox.m_fFittingError);
		//m_Centroid = R3Point(0,0,0);
	}
	else
		assert(false);
//	std::cout<<"Copying: "<<this<<" from "<<(&bbox)<<" ptr - "<<bbox.m_pLocalGeometry<<std::endl;
	m_pLocalGeometry = NULL;
	//m_pLocalGeometry = bbox.m_pLocalGeometry;
}

void PartBox::SetOrigin(const R3Point & newOrigin)
{
	m_Centroid = newOrigin;
}

void PartBox::SetScale(const R3Vector & newScale)
{
	m_Scale = newScale;
}

void PartBox::ApplyDeform(const TemplateNodeDeform * deform)
{
	if (IsType(PartBox::BOXTYPE_BoxSurface) || IsType(PartBox::BOXTYPE_MeshPoints))
	{
		// get transformation: assume that parent is at root
        m_pBoxDeformation = deform;
	}
	else
		assert(IsType(PartBox::BOXTYPE_NullBox) || IsType(PartBox::BOXTYPE_EmptyBox));
}

void PartBox::ApplyTransform(const R3Affine & transform)
{
	if (IsType(PartBox::BOXTYPE_BoxSurface))
	{
		transform.Apply(m_Centroid);
		m_Orientation.Transform(transform);
		transform.Apply(m_Scale);
	}
	else if (IsType(PartBox::BOXTYPE_MeshPoints))
		assert(false);//todo
}

R3Vector PartBox::GetScale() const
{
	if (m_pBoxDeformation!=NULL)
		return m_pBoxDeformation->GetScale();
	else
		return m_Scale;
}

double PartBox::Area() const
{
	if (m_BoxType == BOXTYPE_NullBox || m_BoxType == BOXTYPE_EmptyBox)
		return 0;
	if (m_BoxType==BOXTYPE_BoxSurface)
	{
		R3Vector s = GetScale();
		return 2*(s[0]*s[1] + s[1]*s[2] + s[0]*s[2]);
	}
	else if (m_BoxType==BOXTYPE_MeshPoints)
	{
		assert(m_iPartGeometry>=0);
		return m_pCollection->GetPart(m_iPartGeometry)->Area();
	}
	else
		assert(false);
}

void PartBox::CreateTemplateSamples(int nSamples)
{
	if (m_BoxType==BOXTYPE_BoxSurface)
		CreateBoxSurfaceSamples(nSamples);
	else
		assert(m_BoxType==BOXTYPE_NullBox || m_BoxType==BOXTYPE_EmptyBox || m_BoxType==BOXTYPE_MeshPoints);
}

void PartBox::ComputeLocalShapeDescriptors(const ShapePointCloud * pnts, int offset, int npoints)
{
	if (m_pLocalGeometry!=NULL)
		delete m_pLocalGeometry;
	
	if (npoints>0)
		m_pLocalGeometry = new PartLocalGeometry(pnts, offset, npoints);
}

int PartBox::NTemplatePointsOnNode() const
{
	if (m_BoxType==BOXTYPE_MeshPoints)
	{
		assert(m_pCollection!=NULL);
		
		return m_pCollection->GetPart(m_iPartGeometry)->GetNormalizedPart()->NumPoints();
	}
	else if (m_BoxType==BOXTYPE_BoxSurface)
	{
		assert(m_pSamplingOfABox!=NULL);
		return (int)m_pSamplingOfABox->samples.size();
	}
	else
		assert(m_BoxType==BOXTYPE_NullBox || m_BoxType==BOXTYPE_EmptyBox);
	return 0;	
}

int PartBox::NearestTemplatePointLocalID(const R3Point & pos) const
{
	R3Point undefPnt = m_pBoxDeformation->Deformed2Orig(pos);
	if (m_BoxType == BOXTYPE_BoxSurface)
	{
		assert(Area()>0);
		assert(m_pSamplingOfABox!=NULL);
		int planeDimID1, sign, udim, vdim;
		double u, v;
		PartBox::NearestPointOnBox(undefPnt, m_Centroid, m_Orientation, m_Scale,
								   planeDimID1, sign, udim, vdim, u, v);
		return GetNearestPointID(planeDimID1, sign, u/2.+.5, v/2.+.5);
	}
	else if (m_BoxType == BOXTYPE_MeshPoints)
	{
		assert(m_iPartGeometry>=0 && m_pCollection!=NULL);
		const PartGeometry * partGeometry = m_pCollection->GetPart(m_iPartGeometry);
		R3Point newPos = undefPnt;
		return partGeometry->NearestPointID(newPos);
	}
	std::cout<<"[ERROR] Unknown box type ("<<m_BoxType<<"): "<<std::flush;
	std::cout<<GetBoxTypeStr().c_str()<<std::endl;
	assert(false);
	return -1;
}


R3Point PartBox::UndeformedTemplatePointPosition(int pntID) const
{
	if (m_BoxType==BOXTYPE_BoxSurface)
	{
		assert(m_pSamplingOfABox!=NULL);
		assert(pntID >= 0 && pntID < m_pSamplingOfABox->samples.size());
		const PointOnPartBox & p = m_pSamplingOfABox->samples[pntID];
		return TemplatePointPos(p);
	}
	else if (m_BoxType==BOXTYPE_MeshPoints)
	{
		assert(m_pCollection!=NULL);
		const PartGeometry * part = m_pCollection->GetPart(m_iPartGeometry);
		return part->PointPosition(pntID);
	}
}

R3Point PartBox::DeformedTemplatePointPosition(int pntID) const
{
	return m_pBoxDeformation->Orig2Deformed(UndeformedTemplatePointPosition(pntID));
}

void PartBox::Normalize(const R3Point & centroid, double aveRad)
{
	//R3Point ;
	//R3Triad m_Orientation;
	//m_Centroid[0] -= centroid[0];
	//m_Centroid[1] -= centroid[1];
	//m_Centroid[2] -= centroid[2];
	//m_Centroid /= aveRad;
	//m_Scale /= aveRad;
}

void PartBox::Reset()
{
	//m_iNInstances = 0;
	m_BoxType==BOXTYPE_Invalid;
    m_pBoxDeformation = &(IdentityTemplateDeform::s_IdentityDeform);;
	//if (m_pPCAScale!=NULL)
	//	m_pPCAScale->Reset();
}

const TemplateNodeDeform * PartBox::Deformation() const
{
	assert(m_pBoxDeformation!=NULL);
    return m_pBoxDeformation;
}

const R3Triad & PartBox::UndeformedOrientation() const
{
    return m_Orientation;
}

const R3Point & PartBox::UndeformedCentroid() const
{
    return m_Centroid;
}

const R3Vector & PartBox::UndeformedScale() const
{
    return m_Scale;
}

bool PartBox::IsType(BoxType type) const
{
	return m_BoxType==type;
}

bool PartBox::IsIncluded() const
{
//	std::cout<<"IsIncluded: Ptr="<<m_pBoxDeformation<<" - "<<m_pBoxDeformation->IsIncluded()<<std::endl;
	return m_pBoxDeformation->IsIncluded();
}

VKString PartBox::GetBoxTypeStr() const
{
	switch(m_BoxType)
	{
		case BOXTYPE_Invalid:		// not detected yet
			return "Invalid";
		case BOXTYPE_BoxSurface:	// box
			return "BoxSurface";
		case BOXTYPE_MeshPoints:
			return "MeshPoints";
		case BOXTYPE_NullBox:
			return "NullBox";
		case BOXTYPE_EmptyBox:
			return "EmptyBox";			
	}
	return "ERROR BOX TYPE";
}

R3Vector PartBox::Average(const R3Vector & v1, double w1, const R3Vector & v2, double w2) 
{
	R3Vector v;
	if (v1.Dot(v2)<0)
		v = (v1*w1 - v2*w2)/(w1+w2);
	else
		v = (v1*w1 + v2*w2)/(w1+w2);
	v.Normalize();
	return v;
}

void PartBox::WriteEllipsoid(FILE * file) const
{
	// Write box stats
	fwrite(&m_BoxType, sizeof(BoxType), 1, file);
	fwrite(&m_iPartGeometry, sizeof(int), 1, file);
	
	if (m_BoxType!=BOXTYPE_NullBox && m_BoxType!=BOXTYPE_Invalid)
	{
		double center[3], orientation[9], scale[3];
		for (int i=0; i<3; i++)
		{
			scale[i] = m_Scale[i];
			center[i] = m_Centroid[i];
			for (int j=0; j<3; j++)
				orientation[i*3+j] = m_Orientation[i][j];	
		}
		fwrite(center, sizeof(double), 3, file);
		fwrite(orientation, sizeof(double), 9, file);
		fwrite(scale, sizeof(double), 3, file);
	}
}

PartBox * PartBox::ReadEllipsoid(FILE * file, PartsCollection * partStorage)
{
	// Write box stats
	PartBox * box = new PartBox(partStorage);
	fread(&(box->m_BoxType), sizeof(BoxType), 1, file);
	fread(&(box->m_iPartGeometry), sizeof(int), 1, file);
	
	if (box->m_BoxType!=BOXTYPE_NullBox && box->m_BoxType!=BOXTYPE_Invalid)
	{
		double c[3], o[9], s[3];
		fread(c, sizeof(double), 3, file);
		fread(o, sizeof(double), 9, file);
		fread(s, sizeof(double), 3, file);
		box->m_Centroid = R3Point(c[0], c[1], c[2]);
		box->m_Orientation = R3Triad(R3Vector(o[0], o[1], o[2]), R3Vector(o[3], o[4], o[5]), R3Vector(o[6], o[7], o[8]));
		box->m_Scale = R3Vector(s[0], s[1], s[2]);
//		std::cout<<"Ellipsoid scale: "<<s[0]<<" "<<s[1]<<" "<<s[2]<<std::endl;
	}
	return box;
}

int PartBox::GetNearestPointID(int planeDimID, int sign, double u, double v) const
{
	assert(m_pSamplingOfABox!=NULL);
	assert(u>=0 && u<=1. && v>=0 && v<=1.);
	
	int plane = planeDimID*2 + (sign>0 ? 0 : 1);
	bool validPlane = (plane>=0 && plane < m_pSamplingOfABox->nx.size());
	if (!validPlane)
		return -1;
	int nx = m_pSamplingOfABox->nx[plane];
	int ny = m_pSamplingOfABox->ny[plane];
	int offset = m_pSamplingOfABox->offset[plane];
	int ix = u * nx;
	int iy = v * ny;
	if (u==1.)
		ix -= 1;
	if (v==1.)
		iy -= 1;
	
	if (nx==0 || ny==0)
		return -1;
	
	return offset + ix * ny + iy;	
}

R3Point PartBox::TemplatePointPos(const PointOnPartBox & p) const
{
	int plane = p.planeID/2;
	int sign = (p.planeID%2 == 0 ? 1 : -1);
	
	int a1, a2;
	PartBox::GetUVDims(plane, a1, a2);
	
	double x = (p.x - .5) * 2.;
	double y = (p.y - .5) * 2.;
	
	R3Point pos = m_Centroid + sign * m_Orientation[plane] * m_Scale[plane];
	pos = pos + m_Orientation[a1] * m_Scale[a1] * x + m_Orientation[a2] * m_Scale[a2] * y;
	return pos;
}

//R3Triad PartBox::TemplatePointDir(const PointOnPartBox & p, double pntEigs[3]) const
//{
//	for (int i=0; i<3; i++)
//		pntEigs[i] = p.eigs[i];
//	
//	return p.orient;
//}

void PartBox::GetUVDims(int planeDim, int & uid, int & vid)
{
	uid = 1, vid = 2;
	if (planeDim>=1)
		uid = 0;
	if (planeDim==2)
		vid = 1;
}


R3Point PartBox::NearestPointOnBox(const R3Point & pos, const R3Point & c, 
								   const R3Triad & dir, const R3Vector & scale,
								   int & planeID, int & sign, int & uid, int & vid, double & u, double & v)
{
	R3Vector locPnt(pos.X() - c.X(), pos.Y() - c.Y(), pos.Z() - c.Z());
	double xyz[3];
	xyz[0] = locPnt.Dot(dir[0]) / scale[0];
	xyz[1] = locPnt.Dot(dir[1]) / scale[1];
	xyz[2] = locPnt.Dot(dir[2]) / scale[2];
	
	bool in[3];
	in[0] = (xyz[0]>=-1 && xyz[0]<=1);
	in[1] = (xyz[1]>=-1 && xyz[1]<=1);
	in[2] = (xyz[2]>=-1 && xyz[2]<=1);
	planeID = -1;
	
	// closest to a unique plane
	double x=-1, y=-1, z=-1;
	if (in[0] && in[1] && in[2])	// inside - nearest plane
	{
		x = vkAbs(xyz[0]*scale[0] - scale[0]);
		y = vkAbs(xyz[1]*scale[1] - scale[1]);
		z = vkAbs(xyz[2]*scale[2] - scale[2]);
		if (x <= y && x <= z)						// x==0 plane
			planeID = 0;		
		else if (y <= z)							// y==0 plane
			planeID = 1;
		else									// z==0 plane
			planeID = 2;
	}
	else if (in[1] && in[2])					// x==0 plane
		planeID = 0;
	else if (in[0] && in[2])					// y==0 plane
		planeID = 1;
	else if (in[0] && in[1])					// z==0 plane
		planeID = 2;
	else if (in[2])								// intersection of some x==0 and y==0 planes
	{	planeID = 0; /*plane2 = 1;*/	}	
	else if (in[1])								// intersection of some x==0 and z==0 planes
	{	planeID = 0; /*plane2 = 2;*/	}
	else if (in[0])								// intersection of some z==0 and y==0 planes
	{	planeID = 1; /*plane2 = 2;*/	}
	
	// Find point ID based on projection to the plane
	// planeID>=0 && plane2<0	-> single plane, 
	// planeID>=0 && plane2>=0  -> corner of 2 planes
	// planeID<0 && plane2<0	-> corner of 3 planes
	
	if (planeID<0)
		planeID = 0;
	
	sign = (xyz[planeID] >= 0 ? 1 : -1);
	
	// get uv coordinates (on plane)
	GetUVDims(planeID, uid, vid);
	u = vkMin(vkMax(xyz[uid], -1.), 1.);
	v = vkMin(vkMax(xyz[vid], -1.), 1.);
	
//	return c;
//	if (R3Distance(pos, R3Point(0.0164727, 0.634509, 1.56024)) < .001)
//	{
//		std::cout<<"Sign="<<sign<<" dimids=["<<planeID<<" "<<uid<<" "<<vid<<"]"<<std::endl;
//		std::cout<<"   Scale=["<<scale[planeID]<<" "<<scale[uid]<<" "<<scale[vid]<<"] "<<std::endl;
//		std::cout<<"   dir[uid]="<<dir[uid].X()<<" "<<dir[uid].Y()<<" "<<dir[uid].Z()<<" "<<std::endl;
//		std::cout<<"   dir[vid]="<<dir[vid].X()<<" "<<dir[vid].Y()<<" "<<dir[vid].Z()<<" "<<std::endl;
//		std::cout<<"   dir[vid]="<<dir[vid].X()<<" "<<dir[vid].Y()<<" "<<dir[vid].Z()<<" "<<std::endl;
//	}
		
	R3Point tPos = c + sign * dir[planeID] * scale[planeID];
	tPos = tPos + dir[uid] * (scale[uid] * u) + dir[vid] * (scale[vid] * v);
	return tPos;
}


		
double PartBox::AveDistance(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
							R3Kdtree<const PointInSet *> * kdTree, const R3Point & center, double aveRad,
							const R3Point & boxCenter, const R3Triad & boxAxes, const R3Vector & boxScale)
{
	if (boxScale[0]<0 || boxScale[1]<0 || boxScale[2]<0)
		return 1000000.;
	if (boxScale[0]==0 && boxScale[1]==0 && boxScale[2]==0)
		return 0;
//	std::cout<<"npnts="<<ptsIDs.size()<<std::endl;
//	std::cout<<"scale=["<<boxScale[0]<<" "<<boxScale[1]<<" "<<boxScale[2]<<"] "<<std::endl;
	
	int planeID, sign, uid, vid;
	double u, v;

	// points-to-template
	double pnts2template = 0;
	for (int i=0; i<(int)ptsIDs.size(); i++)
	{
		R3Vector pos = (pnts->PointPosition(ptsIDs[i]) - center) / aveRad;
		
		R3Point ppos(pos.X(), pos.Y(), pos.Z());
		R3Point boxPos = NearestPointOnBox(ppos, boxCenter, boxAxes, boxScale, 
										   planeID, sign, uid, vid, u, v);
		pnts2template += pow(R3Distance(boxPos, ppos), 2.);
	}
	
	// template-to-points
	double template2pnts = 0;
	int nTPnts = ptsIDs.size() * 5;
	double totalArea = 2.*(boxScale[0]*boxScale[1] + boxScale[0]*boxScale[2] + boxScale[1]*boxScale[2]);
	
	int finalTemplatePoints = 0;
	for (int plane = 0; plane<3; plane++)
	for (int sign = -1; sign<=1; sign+=2)
	{
		int a1, a2;
		PartBox::GetUVDims(plane, a1, a2);
		
		double currArea = boxScale[a1] * boxScale[a2];
		int nNewPnts = (totalArea==0 ? 1 : ((int)(nTPnts * currArea / totalArea) + 1));
		
		int nx = 1; 
		if (boxScale[a2] > 0)
			nx = vkMax((int)(sqrt(nNewPnts * boxScale[a1] / boxScale[a2]) + .5), 1);;
		int ny = 1; 
		if (boxScale[a1] > 0)
			ny = vkMax((int)(sqrt(nNewPnts * boxScale[a2] / boxScale[a1]) + .5), 1);

//		std::cout<<"1. nx/ny=["<<nx<<" "<<ny<<"] nnew="<<nNewPnts;
//		std::cout<<" scale=["<<boxScale[a1]<<"/"<<boxScale[a2];
//		std::cout<<" sqrt="<<(sqrt(nNewPnts * boxScale[a1] / boxScale[a2]) + .5)<<std::endl;		
		if (nx<ny)
			ny = nNewPnts / nx;
		else 
			nx = nNewPnts / ny;
//		std::cout<<"2. nx/ny=["<<nx<<" "<<ny<<"]"<<std::endl;
		assert(nx>=1 && ny>=1);
				
		// add nx * ny grid
		double dx = 1. / nx;
		double dy = 1. / ny;
		double xOffset = (sign==1 ? (dx/4.) : (3.*dx/4.));
		double yOffset = (sign==1 ? (dy/4.) : (3.*dy/4.));
		
		for (int x=0; x<nx; x++)
		for (int y=0; y<ny; y++)
		{
			double u = ((xOffset + dx * x)-.5) * 2.;
			double v = ((yOffset + dy * y)-.5) * 2.;
			
			R3Point boxPos = boxCenter + boxAxes[a1] * boxScale[a1] * u + boxAxes[a2] * boxScale[a2] * v;
			R3Point boxInPntsSpace = boxPos * aveRad + center;
			
			const PointInSet * closest = kdTree->FindClosest(boxInPntsSpace, 0, 1000000000);
			assert(closest!=NULL);
			R3Vector pos = (closest->GetPosition() - center) / aveRad;
			R3Point ppos(pos.X(), pos.Y(), pos.Z());
			pnts2template += pow(R3Distance(boxPos, ppos), 2.);
			finalTemplatePoints++;
		}
	}

	return pnts2template / (int)ptsIDs.size() + template2pnts/finalTemplatePoints;
}

void PartBox::CreateBoxSurfaceSamples(int nSamples)
{
	// distribute points
	assert(IsType(BOXTYPE_BoxSurface));
	
	m_pSamplingOfABox = new SampledBox();
	std::vector<PointOnPartBox> & ellSmpls = m_pSamplingOfABox->samples;
	
//	ShapePointCloud * pnts = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD, "allt", "none", NULL);
	
	R3Vector s = GetScale();
	double totalArea = Area();
	if (totalArea==0)
		return;
	//assert(totalArea>0);
		
	// for each plane of a box: create point samples
	for (int planeID = 0; planeID<6; planeID++)
	{
		int plane = planeID / 2;
		int sign = planeID % 2;
		
		int a1, a2;
		GetUVDims(plane, a1, a2);
		
		// how many points to add
		double currArea = s[a1] * s[a2];
		int nNewPnts =  vkMax((int)((currArea * nSamples) / totalArea+.5), 1);
		//std::cout<<"     Plane: "<<planeID<<". Area="<<currArea<<" nSamples="<<nSamples<<std::endl;
		
		// grid dimensions
		int nx = vkMax((int)(sqrt(nNewPnts * s[a1] / s[a2]) + .5), 1);
		int ny = vkMax((int)(sqrt(nNewPnts * s[a2] / s[a1]) + .5), 1);
		if (nx<ny)
			ny = nNewPnts / nx;
		else 
			nx = nNewPnts / ny;
		
		// store samples
		m_pSamplingOfABox->nx.push_back(nx);
		m_pSamplingOfABox->ny.push_back(ny);
		m_pSamplingOfABox->offset.push_back((int)ellSmpls.size());
		
		// add nx * ny grid
		double dx = 1. / nx;
		double dy = 1. / ny;
		double xOffset = (sign==0 ? (dx/4.) : (3.*dx/4.));
		double yOffset = (sign==0 ? (dy/4.) : (3.*dy/4.));
		
		for (int x=0; x<nx; x++)
		for (int y=0; y<ny; y++)
		{
			// id = ix * ny + iy
			int lID = (int)ellSmpls.size();
			ellSmpls.push_back(PointOnPartBox(planeID, xOffset + dx * x, yOffset + dy * y));
//			pnts->AddPoint(ShapePoint(TemplatePointPos(ellSmpls[lID]), pnts, pnts->NumPoints()));
			nNewPnts--;
		}
		
		// add extra points
		assert(nNewPnts>=0);
		while (nNewPnts>0)
		{
			double xRnd = vkAbs((double)(rand() % 1000) / (double)1000.);
			double yRnd = vkAbs((double)(rand() % 1000) / (double)1000.);;
			int lID = (int)ellSmpls.size();
			ellSmpls.push_back(PointOnPartBox(planeID, xRnd, yRnd));
//			pnts->AddPoint(ShapePoint(TemplatePointPos(ellSmpls[lID]), pnts, pnts->NumPoints()));
			nNewPnts--;
		}
		assert(nNewPnts == 0);
	}

//	if (pnts!=NULL && pnts->NumPoints()>0)
//		~ = new PartLocalGeometry(pnts, true);
//	delete pnts;
}

const PartLocalGeometry * PartBox::GetLocalGeometry() const
{
	//std::cout<<"Returning: "<<this<<" - "<<m_pLocalGeometry<<std::endl;
	return m_pLocalGeometry;
}

PointOnPartBox::PointOnPartBox()
{
	this->planeID = -1;
	this->x = 0;
	this->y = 0;
//	eigs[0] = -1;
//	eigs[1] = -1;
//	eigs[2] = -1;	
}

PointOnPartBox::PointOnPartBox(int planeID, double x, double y)
{
	this->planeID = planeID;
	this->x = x;
	this->y = y;
//	eigs[0] = -1;
//	eigs[1] = -1;
//	eigs[2] = -1;
}
		
