/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PartBox_H
#define __PartBox_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapePointCloud.h"

#include "PartsCollection.h"

#include "TemplateNodeDeform.h"

#include "PartLocalGeometry.h"

class PartConnectorEdge;
class PartOfEdge;
class ContProbDistribution;
class TemplateNodeDeform;

/**
 * Sampling of a box-like template part: a single sample point
 */
struct PointOnPartBox
{
	PointOnPartBox();
	PointOnPartBox(int planeID, double x, double y);
	int planeID;
	double x, y;
	
//	// cached extrinsic data
//	R3Triad orient;	
//	double eigs[3];			
};

/**
 * Sampling of a box-like template part: whole box 
 */
struct SampledBox
{
	std::vector<int> nx;
	std::vector<int> ny;
	std::vector<int> offset;
	std::vector<PointOnPartBox> samples;
};

/**
 * Bounding box descriptor. Describes shape of a single node...
 */ 
class PartBox
{
	public:
		static PartBox * GetNullPart();
		static PartBox * GetEmptyPart();
	
		enum BoxType
		{
			BOXTYPE_Invalid,		// not detected yet
			BOXTYPE_BoxSurface,
			BOXTYPE_MeshPoints,
			BOXTYPE_NullBox,
			BOXTYPE_EmptyBox,
			BOXTYPE_Excluded
//			BOXTYPE_Rectangle,		// all e-values are different
//			BOXTYPE_Cylinder,		// smallest e-values are the same
//			BOXTYPE_Plane,			// largest e-values are the same
//			BOXTYPE_Cube,			// all e-values are the same
		};
	
		PartBox(PartsCollection * partStorage);
		virtual ~PartBox();
	
		virtual PartBox * CreateCopy(PartsCollection * newStorage, int partoffset) const;
		virtual PartBox * CreateCopy() const;
	// init box to something
		//virtual double InitStats(const ShapeMesh * mesh, const std::vector<int> & faces);
		//virtual double InitStats(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, double thresh);
		
		static void FindOptimalBoxFit(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
									  R3Point & center, R3Triad & orient, R3Vector & scale, double & fitErr);
		static int FindBestPartFit(int partID, R3Point & center, R3Triad & orient, R3Vector & scale, double & fitErr);
	
		virtual void InitNull();
		virtual void InitEmpty();
		virtual void InitBox(const R3Point & centroid, const R3Triad & orient, const R3Vector & pcaScale, double fitErr);
		virtual void InitPart(int partID, double fitErr);
	
		virtual const PartGeometry * GetPart() const;
	
		virtual PartBox & operator = (const PartBox & box);
		virtual void InitCopy(const PartBox & box);
		virtual void SetOrigin(const R3Point & newOrigin);
		virtual void SetScale(const R3Vector & newScale);
		virtual R3Vector GetScale() const;
		virtual void ApplyDeform(const TemplateNodeDeform * pDeform);
		virtual void ApplyTransform(const R3Affine & transform);
		virtual double Area() const;
	
		virtual void CreateTemplateSamples(int nSamples);
		virtual void ComputeLocalShapeDescriptors(const ShapePointCloud * pnts, int offset, int npoints);
	
		virtual int NTemplatePointsOnNode() const;
		virtual R3Point UndeformedTemplatePointPosition(int pntID) const;
		virtual R3Point DeformedTemplatePointPosition(int pntID) const;
		virtual int NearestTemplatePointLocalID(const R3Point & pos) const;
		virtual void Normalize(const R3Point & centroid, double aveRad);
	
	
	// update existing box
		//virtual void UpdateStats(const PartBox & bbox);
		//virtual void UpdateStats(const TemplateNodeDeform & deform);
		virtual void Reset();
	
	// get box properties. NOTE: scale can have 0 variance, and thus just define a single box
		//virtual R3Triad Orientation() const;
		//virtual R3Point Centroid() const;
		//virtual R3Vector Scale() const;
    
		virtual const TemplateNodeDeform * Deformation() const;
    
        virtual const R3Triad & UndeformedOrientation() const;
        virtual const R3Point & UndeformedCentroid() const;
        virtual const R3Vector & UndeformedScale() const;
    
	// box type
		//virtual bool NullBox() const;
		virtual bool IsType(BoxType type) const;
		virtual bool IsIncluded() const;
		virtual VKString GetBoxTypeStr() const;
		virtual double GetFittingErr() const;
	
	
	// read / write
		virtual void WriteEllipsoid(FILE * file) const;
		static PartBox * ReadEllipsoid(FILE * file, PartsCollection * partStorage);
	
		static void GetUVDims(int planeDim, int & uid, int & vid);
		static R3Point NearestPointOnBox(const R3Point & pos, const R3Point & boxCenter, 
										 const R3Triad & boxAxes, const R3Vector & boxScale,
										 int & planeID, int & sign, int & uDir, int & vDir, double & u, double & v);
		
		virtual const PartLocalGeometry * GetLocalGeometry() const;
	
	
//	protected:
		virtual int GetNearestPointID(int planeDimID, int sign, double u, double v) const;
		
		virtual R3Point TemplatePointPos(const PointOnPartBox & point) const;
		//virtual R3Triad TemplatePointDir(const PointOnPartBox & point, double pntEigs[3]) const;
		virtual void CreateBoxSurfaceSamples(int nSamples);

		BoxType m_BoxType;
		//int m_iNInstances;
		
		// position / orientation...
        
		R3Point m_Centroid;
		R3Triad m_Orientation;
		R3Vector m_Scale;
		
		// anisotropic scaling
		//ContProbDistribution* m_pPCAScale;
		
		// If pnts: 
		int m_iPartGeometry;
		PartsCollection * m_pCollection;
		PartLocalGeometry * m_pLocalGeometry;
				
		SampledBox * m_pSamplingOfABox;
	
		double m_fFittingError;
		
		const TemplateNodeDeform * m_pBoxDeformation;

	
		// SOME STATIC HELPER FUNCTIONS
		static double AveDistance(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
								  R3Kdtree<const PointInSet *> * kdTree, const R3Point & center, double aveRad,
								  const R3Point & boxCenter, const R3Triad & boxAxes, const R3Vector & boxScale);
	
		static double EvalStep(const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
							   R3Kdtree<const PointInSet *> * kdTree, const R3Point & center, double aveRad,
							   const R3Point & boxCenter, const R3Triad & boxAxes, const R3Vector & boxScale,
							   int optimStep, int dim, double step);
		static double ChooseStep(R3Point & boxCenter, R3Triad & boxAxes, R3Vector & boxScale, 
								 int optimStep, int dim, double step);
	

		//virtual void ClassifyBox();
		static R3Vector Average(const R3Vector & v1, double w1, const R3Vector & v2, double w2);
		static PartBox * m_pNullPart;
		static PartBox * m_pEmptyPart;
};


#endif
