/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PartLocalGeometry_H
#define __PartLocalGeometry_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"
#include "ShapeLocalPCA.h"
#include "ProbDistribution.h"

/**
 * Describes local shape features for a set of points
 */
class PartLocalGeometry
{
	public:
		friend class LocalGeometryClassifier;
		friend class LocalGeometryTrainingSet;
	
		static int GetNhdSizeConstant();
		static double GetNhdDist();
	
		PartLocalGeometry(const ShapePointCloud * pnts, int offset, int npoints, bool freePnts = false);
		~PartLocalGeometry();
		
		virtual int NPnts() const;
		static int NLocalFeatures();
		virtual R3Triad LocalOrient(int pntid, double * eigs, const R3Affine & xform) const;
		virtual const double * LocalOrientAsFeatures(int pntID, const R3Affine & xform) const;
		
		
	
	protected:
		virtual const ShapeLocalPCA * GetLocalPCAVals() const;
	
		double * m_aTempFeatureStorage;
	
		bool m_bFreePnts;
		const ShapePointCloud * m_pPnts;
		int m_iOffset;
		int m_iNPoints;
	
		ShapeLocalPCA * m_pLocalPCA;
		VKString m_PCAName;
	
		static VKString s_sDefaultLocPCAName;

};


/**
 * Mapping of nodes
 */
class TrainingNodeIDMap
{
	public:
		TrainingNodeIDMap(const std::vector<int> & nodeIDs, int nMaxNodes);
		~TrainingNodeIDMap();
		virtual TrainingNodeIDMap * CreateCopy() const;
		virtual int Classifier2OrigNodeID(int claID) const;
		virtual int OrigNodeID2Classifier(int nodeID) const;
		virtual int NumUniqueClasses() const;
		virtual int MaxOrigNodeIDs() const;
		virtual int Update(const std::vector<int> & nodeIDs);
	
		virtual void WriteNodeIDMap(FILE * file) const;
		static TrainingNodeIDMap * ReadNodeIDMap(FILE * file);
	
	protected:
		TrainingNodeIDMap();
		int * m_aiClassifier2OrigNode;
		int * m_aiOrigNode2Classifier;
		int m_iNClasses;
		int m_iMaxOrigNodes;
};

/**
 * Training dataset
 */
class LocalGeometryTrainingSet// : public Thea::JointBoost::TrainingData
{
	public:
		friend class LocalGeometryClassifier;
	
		LocalGeometryTrainingSet(const std::vector<const PartLocalGeometry *> & shapes,
								 const std::vector<int> & nodeIDs, int maxOrigNodes);
		LocalGeometryTrainingSet(const LocalGeometryTrainingSet & training);
		~LocalGeometryTrainingSet();

		virtual TrainingNodeIDMap * GetNodeIDMapping() const;
		virtual long numExamples() const;
		virtual long numFeatures() const;
	
		virtual void getFeature(long feature_index, std::vector<double> & values) const;
		
		virtual void getClasses(std::vector<long> & classes) const;
	
	protected:
		TrainingNodeIDMap * m_pTrainingIDMapping;
	
		double * m_afExampleFeatures;
		int * m_aiExampleClasses;
	
		const std::vector<const PartLocalGeometry *> m_Shapes;
		const std::vector<int> m_NodeIDs;
		int m_iNFeatures;
		int m_iNExamples;
	
};

/**
 * classifies geometry based on local features
 */
class LocalGeometryClassifier
{
	public:
		enum ClassifierType
		{
			CLASSIFY_NULL 	= 0,	// 
			CLASSIFY_JB 	= 1,	// joint boost
			CLASSIFY_GMM 	= 2		// mixture of gaussians
		};
	
		LocalGeometryClassifier(const LocalGeometryTrainingSet & trainingData);
		virtual ~LocalGeometryClassifier();
		virtual LocalGeometryClassifier * CreateCopy();
	
		virtual const double * LogProbs(int NFeat, const double * features) const;
	
		// Read / Write classifier
		virtual void WriteClassifier(FILE * file) const;
		static LocalGeometryClassifier * ReadClassifier(FILE * file);
		virtual void UpdateClassifier(const LocalGeometryTrainingSet & trainingData);
	
		virtual const ProbGaussian * GetGaussian(int localNodeID) const;
	
	protected:
		LocalGeometryClassifier();
		ClassifierType m_ClassifierType;
	
		double * m_afTempProbStorageClassifier;
		double * m_afTempProbStorageReturn;
	
		TrainingNodeIDMap * m_pTrainingIDMapping;
		int m_iNFeatures;
		int m_iNNodes;

	// MIXTURE OF GAUSSIANS CLASSIFIER
		void TrainGMM(const LocalGeometryTrainingSet & trainingData);
		void LogProbsGMM(int NFeat, const double * features) const; // NOTE: fills m_afTempProbStorageClassifier
		std::vector<ProbGaussian *> m_GMM;
	
	// JOINT BOOST CLASSIFIER
		void TrainJointBoost(const LocalGeometryTrainingSet & trainingData);
		void LogProbsJointBoost(int NFeat, const double * features) const; // NOTE: fills m_afTempProbStorageClassifier
//		Thea::JointBoost * m_pJB;
};

// TODO: PairwiseGeometryClassifier


#endif


