/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VkFunctions.h"
#include "PartOfEdge.h"
#include "STemplateNode.h"
#include "STemplate.h"
#include "TMatch.h"
#include "TMatchEnergy.h"


PartOfEdge::PartOfEdge(STemplateNode * parent, STemplateNode * child, STemplate * st)
{
	m_pAlwaysOnePart = NULL;
	m_pParent = parent;
	m_pChildTemplate = child;
	m_pST = st;
}

PartOfEdge::~PartOfEdge()
{
	if (m_pAlwaysOnePart!=NULL)
		delete m_pAlwaysOnePart;
	
	for (int i=0; i<(int)m_apPartProb.size(); i++)
		delete m_apPartProb[i];
	
	for (int i=0; i<m_apPartTransforms.size(); i++)
	for (std::map<int, PartTransforms*>::iterator iter = m_apPartTransforms[i].begin();
		 iter != m_apPartTransforms[i].end(); iter++)
		delete iter->second;
}

PartOfEdge * PartOfEdge::CreateCopy(STemplate * newST, std::map<int, int> & old2newnode,
									std::map<int, int> & new2oldnode, int configoffset) const
{
	assert(old2newnode.find(m_pParent->NodeID()) != old2newnode.end());
	assert(old2newnode.find(m_pChildTemplate->NodeID()) != old2newnode.end());
	int parentID = old2newnode[m_pParent->NodeID()];
	int childID = old2newnode[m_pChildTemplate->NodeID()];
	
	PartOfEdge * edge = new PartOfEdge((STemplateNode*)newST->GetNode(parentID), (STemplateNode*)newST->GetNode(childID), newST);
	
	edge->m_apPartProb.push_back(m_apPartProb[0]->CreateCopy());
	
	for (int i=0; i<configoffset; i++)	// initial configs: empty
	{
		edge->m_apPartProb.push_back(new ProbDiscreteStats());
		edge->m_apPartProb[i]->UpdateDistribution(0, true);
	}
	
	assert(m_pAlwaysOnePart==NULL);	// no null nodes should be copied
	
//	std::cout<<"NTransforms: "<<m_apPartTransforms.size()<<std::endl;
	for (int i=1; i<(int)m_apPartProb.size(); i++)	// copy part probabilities
		edge->m_apPartProb.push_back(m_apPartProb[i]->CreateCopy());

	int lastAdded = (int)edge->m_apPartTransforms.size();
	edge->m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
	std::map<int, PartTransforms*> & ncXForms = (std::map<int, PartTransforms*> &)m_apPartTransforms[0];
	for (std::map<int, PartTransforms*>::iterator iter = ncXForms.begin(); iter != ncXForms.end(); iter++)
	{
		PartTransforms * pt = iter->second->CreateCopy();
		edge->m_apPartTransforms[lastAdded][iter->first] = pt;
		const double *  mean = pt->GetPartTransform(0)->SafeCastToGaussian()->GetMostProbable();
//		std::cout<<"Copying transform: ["<<mean[0]<<" "<<mean[1]<<" "<<mean[2]<<"] "<<std::endl;
//		virtual const double * GetMostProbable() const;

	}
	
	for (int i=0; i<configoffset; i++)  // initial configs: empty
		edge->m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
	
	for (int i=1; i<(int)m_apPartTransforms.size(); i++)	// copy part transformations
	{
		int lastAdded = (int)edge->m_apPartTransforms.size();
		edge->m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
		
		std::map<int, PartTransforms*> & ncXForms = (std::map<int, PartTransforms*> &)m_apPartTransforms[i];
		for (std::map<int, PartTransforms*>::iterator iter = ncXForms.begin(); iter != ncXForms.end(); iter++)
			edge->m_apPartTransforms[lastAdded][iter->first] = iter->second->CreateCopy();
	}
	std::cout<<"Edge Transforms ("<<edge->m_pChildTemplate->NodeName().c_str()<<"): "<<edge->m_apPartTransforms.size()<<std::endl;
	return edge;
}

STemplateNode * PartOfEdge::ParentTemplate()
{
	return m_pParent;
}

const STemplateNode * PartOfEdge::ParentTemplate() const
{
	return m_pParent;
}

const STemplateNode * PartOfEdge::ChildTemplate() const
{
	return m_pChildTemplate;
}

STemplateNode * PartOfEdge::ChildTemplate()
{
	return m_pChildTemplate;
}

int PartOfEdge::NConfigs() const
{
	return (int)m_apPartProb.size();
}

const DiscreteProbDistribution * PartOfEdge::GetProbNParts(int configID) const
{
	if (m_pAlwaysOnePart!=NULL)
		return m_pAlwaysOnePart;
	if (!(configID>=0 && configID < (int)m_apPartProb.size()))
	{
		std::cout<<"[ERROR] Invalid configID = "<<configID<<" node="<<ChildTemplate()->NodeName().c_str()<<std::endl;
		std::cout<<"        NParts: "<<m_apPartProb.size()<<std::endl;
	}
	assert(configID>=0 && configID < (int)m_apPartProb.size() && m_apPartProb[configID]!=NULL);
	return m_apPartProb[configID];
}

const double * PartOfEdge::GetMeanTransform(int configID, int nParts, int partID, int & flags, int & ndims) const
{
	const PartTransforms * pt = GetTransform(configID, nParts, GlobalConstants::s_iMinNSamplesForMean, NULL);
	flags = pt->XFormFlags();
	ndims = pt->GetPartTransform(partID)->NDims();
	return pt->GetPartTransform(partID)->GetMostProbable();
} 

double PartOfEdge::LogProb(int configID, int nParts, int partID, const double * sample, int xformflags) const
{
//	const PartTransforms * pt = GetTransform(configID, nParts, GlobalConstants::s_iMinNSamples, NULL);
//	assert(pt->XFormFlags() == xformflags);
//	return pt->GetPartTransform(partID)->LogProb(sample);

	const PartTransforms * ptVar = GetTransform(configID, nParts, GlobalConstants::s_iMinNSamplesForVariance, NULL);
	const PartTransforms * ptMean = GetTransform(configID, nParts, GlobalConstants::s_iMinNSamplesForMean, NULL);
	assert(ptVar!=NULL && ptMean!=NULL && ptVar->XFormFlags()==xformflags && ptMean->XFormFlags()==xformflags);
	assert(ptVar->GetPartTransform(partID)!=NULL && ptMean->GetPartTransform(partID)!=NULL);
	assert(ptVar->GetPartTransform(partID)->SafeCastToGaussian()!=NULL);
	assert(ptMean->GetPartTransform(partID)->SafeCastToGaussian()!=NULL);
	 
	const ProbGaussian * ptVarG = ptVar->GetPartTransform(partID)->SafeCastToGaussian();
	const ProbGaussian * ptMeanG = ptMean->GetPartTransform(partID)->SafeCastToGaussian();
	
//	if (configID==3)
//		std::cout<<"mean value: ["<<ptMeanG->GetMostProbable()[0]<<" "<<ptMeanG->GetMostProbable()[1]<<" "<<ptMeanG->GetMostProbable()[2]<<"] "<<std::endl;;
	return ProbGaussian::LogProb(ptVarG->GetVariance(), ptMeanG->GetMostProbable(),
								 ptVarG->NDims(), ptVarG->NSamples(), sample);
}

R3Vector PartOfEdge::PosVariance(int configID, int nParts, int partID, int & ancestor) const
{
	const PartTransforms * pt = NULL;
	pt = GetTransform(configID, nParts, GlobalConstants::s_iMinNSamplesForVariance, &ancestor);
	
	assert(pt->XFormFlags() == PartTransforms::PARTXFORM_Translation);
	const ProbGaussian * gauss = pt->GetPartTransform(partID)->SafeCastToGaussian();
	assert(gauss!=NULL);
	const double * var = gauss->GetVariance();
	return R3Vector(var[0], var[1], var[2]);
}


const PartTransforms * PartOfEdge::GetTransform(int configID, int nParts, int nMinSamples, int * ancestor) const
{
	assert(m_pST!=NULL);
	int cid = configID;
	const PartTransforms * retVal = NULL;
	if (m_pChildTemplate->IsNull())
		cid = 0;
	while (true)
	{
		if (!(cid >= 0 && cid < (int)m_apPartTransforms.size()))
		{
			std::cout<<"[ERROR] Invalid config. Node="<<m_pChildTemplate->NodeName().c_str();
			std::cout<<" cid = "<<cid<<" / "<<m_apPartTransforms.size()<<std::endl;
		}
		assert(cid >= 0 && cid < (int)m_apPartTransforms.size());
		
		// check if enough samples
		PartOfEdge* ncThis = ((PartOfEdge*)this);
		std::map<int, PartTransforms*>::iterator iter = ncThis->m_apPartTransforms[cid].find(nParts);
		if (iter==m_apPartTransforms[cid].end())
		{
			std::cout<<"[ERROR] Could not get the following transform: "<<cid<<" "<<nParts<<std::endl;
			std::cout<<"    ConfigName: "<<m_pST->ConfigName(cid).c_str();
			std::cout<<" Node: "<<m_pChildTemplate->NodeName().c_str()<<std::endl;
		}
		assert(iter!=m_apPartTransforms[cid].end());
		retVal = iter->second;
		if (iter->second->NSamples() >= nMinSamples)
			break;
		
		// if not - update cid
		if (cid>0)
			//cid = m_pST->ConfigParent(cid);
			cid = 0;
		else
			break;
	}
	
	if (ancestor!=NULL)
		*ancestor = cid;
	assert(retVal!=NULL);
	return retVal;
}

void PartOfEdge::SetNPartsAlwaysOne()
{
	assert(m_pAlwaysOnePart==NULL);
	m_pAlwaysOnePart = new ProbDiscreteStats();
	m_pAlwaysOnePart->UpdateDistribution(1, true);
}
void PartOfEdge::UpdateNParts(int configID, int nparts, bool include)
{
	if (m_pAlwaysOnePart!=NULL)
	{
		while ((int)m_apPartProb.size() <= configID)
			m_apPartProb.push_back(m_pAlwaysOnePart);
	}
	else
	{
		assert(configID>=0);
		while ((int)m_apPartProb.size() <= configID)
			m_apPartProb.push_back(new ProbDiscreteStats());
		m_apPartProb[configID]->UpdateDistribution(nparts, include);
	}
	if (configID>0 && m_pST->ConfigParent(configID)>=0)
	{
		assert(configID != m_pST->ConfigParent(configID));
		UpdateNParts(m_pST->ConfigParent(configID), nparts, include);
	}
}

void PartOfEdge::UpdateXForm(int configID, const STemplateNode * parent,
							 const STemplateNode * child, int nParts, int partID)
{
	// update transformation
	PartTransforms * xform = NULL;
	
	assert(configID>=0);
	while (configID>=(int)m_apPartTransforms.size())
		m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
	assert(configID>=0 && configID < (int)m_apPartTransforms.size());
		
	std::map<int, PartTransforms*>::iterator iter = m_apPartTransforms[configID].find(nParts);
	if (iter!=m_apPartTransforms[configID].end())
		xform = iter->second;
	else
	{
		xform = new PartTransforms();
		m_apPartTransforms[configID][nParts] = xform;
	}

    PartBox rootEllipsoid(NULL);
    R3Triad xyz(R3Vector(1, 0, 0), R3Vector(0, 1, 0), R3Vector(0, 0, 1));
    rootEllipsoid.InitBox(R3Point(0,0,0), xyz, R3Vector(1, 1, 1), 0);

    xform->UpdateTransformStats(partID, (parent == NULL) ? rootEllipsoid : *(parent->GetPart(configID)),
                                (child == NULL) ? rootEllipsoid : *(child->GetPart(configID)));
	// recurse: disabled
//	if (configID>0 && m_pST->ConfigParent(configID)>=0)
//		UpdateXForm(m_pST->ConfigParent(configID), parent, child, nParts, partID);
}

void PartOfEdge::UpdateXForm(int configID, const TemplateNodeDeform & deform, bool include)
{
	assert(configID>=0);
	while (configID >=  (int)m_apPartTransforms.size())
		m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
		
	std::map<int, PartTransforms*>::iterator iter = m_apPartTransforms[configID].find(deform.NParts());
	if (iter == m_apPartTransforms[configID].end())
	{
		m_apPartTransforms[configID][deform.NParts()] = new PartTransforms();
		iter = m_apPartTransforms[configID].find(deform.NParts());
	}
	assert(iter != m_apPartTransforms[configID].end());
	
	iter->second->UpdateTransformStats(deform, include);

	// recurse: disabled
//	if (configID>0 && m_pST->ConfigParent(configID)>=0)
//		UpdateXForm(m_pST->ConfigParent(configID), deform, include);

}

void PartOfEdge::Reset()
{

	for (int i=0; i<(int)m_apPartTransforms.size(); i++)
	for (std::map<int, PartTransforms*>::iterator iter = m_apPartTransforms[i].begin();
		 iter != m_apPartTransforms[i].end(); iter++)
		iter->second->Reset();
}

void PartOfEdge::Normalize(const R3Point & centroid, double aveRadius)
{
	for (int i=0; i<(int)m_apPartTransforms.size(); i++)
	for (std::map<int, PartTransforms*>::iterator iter = m_apPartTransforms[i].begin();
		 iter != m_apPartTransforms[i].end(); iter++)
		iter->second->Normalize(centroid, aveRadius);
}

void PartOfEdge::WritePartOfEdge(FILE * file) const
{
	int parentID = m_pParent->NodeID();
	int childID = m_pChildTemplate->NodeID(); 
	fwrite(&parentID, sizeof(int), 1, file);
	fwrite(&childID, sizeof(int), 1, file);
	
	// save part multiplicity probabilities
	bool alwaysOne = m_pAlwaysOnePart!=NULL;
	fwrite(&alwaysOne, sizeof(bool), 1, file);

	int nconfigs = (int)m_apPartProb.size();
	fwrite(&nconfigs, sizeof(int), 1, file);
	if (!alwaysOne)
	{
		//VKString name = ChildTemplate()->NodeName();
		for (int i=0; i<nconfigs; i++)
			m_apPartProb[i]->WriteToFile(file);
	}
	
	if (nconfigs != (int)m_apPartTransforms.size())
		std::cout<<"[ERROR] Inconsistent # configs: "<<nconfigs<<" vs "<<m_apPartTransforms.size()<<std::endl;
	assert(nconfigs==(int)m_apPartTransforms.size());
	for (int i=0; i<nconfigs; i++)
	{
		int nNumPartsConfigs = (int)m_apPartTransforms[i].size();
		fwrite(&nNumPartsConfigs, sizeof(int), 1, file);

		for (std::map<int, PartTransforms*>::iterator iter = ((PartOfEdge*)this)->m_apPartTransforms[i].begin();
			 iter != m_apPartTransforms[i].end(); iter++)
		{
			// save number of parts, and part all transformations
			fwrite(&(iter->first), sizeof(int), 1, file);
			iter->second->WriteTransforms(file);
		}
	}
}

PartOfEdge * PartOfEdge::ReadPartOfEdge(FILE * file, STemplate * st, STemplate::STemplateVersions version)
{
	
	int parentID;
	int childID;
	fread(&parentID, sizeof(int), 1, file);
	fread(&childID, sizeof(int), 1, file);
	
	STemplateNode * child = (STemplateNode*) st->GetNode(childID);
	PartOfEdge * pe = new PartOfEdge((STemplateNode*) st->GetNode(parentID), child, st);
	
	// read part multiplicity probabilities
	bool alwaysOne;
	fread(&alwaysOne, sizeof(bool), 1, file);
	if (alwaysOne)
		pe->SetNPartsAlwaysOne();

	int nconfigs = -1;
	fread(&nconfigs, sizeof(int), 1, file);
	if (!alwaysOne)
	{
		for (int i=0; i<nconfigs; i++)
			pe->m_apPartProb.push_back(DiscreteProbDistribution::CreateFromFile(file));
	}
	else
	{
		for (int i=0; i<nconfigs; i++)
			pe->m_apPartProb.push_back(pe->m_pAlwaysOnePart);
	}
	
	if (version < STemplate::STMPLT_V0_3)	// not a multi-config version!
	{
		int nparts, nNumPartsConfigs;

		fread(&nNumPartsConfigs, sizeof(int), 1, file);
		std::map<int, PartTransforms*> readXForms;
		for (int j=0; j<nNumPartsConfigs; j++)
		{
			fread(&nparts, sizeof(int), 1, file);
			readXForms[nparts] = PartTransforms::ReadTransforms(file, version);
		}

		for (int i=0; i<nconfigs; i++)
		{
			pe->m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
			for (std::map<int, PartTransforms*>::iterator iter = readXForms.begin();
				 iter != readXForms.end(); iter++)
				pe->m_apPartTransforms[i][iter->first] = iter->second->CreateCopy();
		}
	}
	else
	{
		assert(version==STemplate::STMPLT_V0_3 || version==STemplate::STMPLT_V0_4);
		int nNumPartsConfigs, nparts;
		for (int i=0; i<nconfigs; i++)
		{
			pe->m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
			fread(&nNumPartsConfigs, sizeof(int), 1, file);
			for (int j=0; j<nNumPartsConfigs; j++)
			{
				// save number of parts, and part all transformations
				fread(&nparts, sizeof(int), 1, file);
				pe->m_apPartTransforms[i][nparts] = PartTransforms::ReadTransforms(file, version);
//				if (i==3 && st->GetNode(childID)->NodeName()=="seat")
//				{
//					ProbGaussian * gauss = (ProbGaussian*) pe->m_apPartTransforms[i][nparts]->m_apXForms[0];
//					gauss->m_aMeans[2] += 100;
//				}
			}
		}
	}
	return pe;
}

void PartOfEdge::DuplicateConfigs(int from)
{
	assert(from >= 0 && from < (int)m_apPartTransforms.size() );
	
	for (int i=(int)m_apPartTransforms.size(); i<m_pST->NSeedConfigs(); i++)
	{
		m_apPartTransforms.push_back(std::map<int, PartTransforms*>());
		for (std::map<int, PartTransforms*>::iterator iter = m_apPartTransforms[i].begin();
			 iter != m_apPartTransforms[i].end(); iter++)
			m_apPartTransforms[i][iter->first] = iter->second->CreateCopy();
	}
}



