/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "STemplateNode.h"
#include "PartOfEdge.h"
#include "STemplate.h"
#include "TMatch.h"
#include "TMatchEnergy.h"

// initialize empty template (maybe a part)
STemplateNode::STemplateNode(const VKString & mainName, int nodeID, STemplate * st)
{
	m_pST = st;
	m_NodeName = mainName;
	m_iNodeID = nodeID;
	m_pParentPart = NULL;
	m_bIsNull = false;
}

STemplateNode::~STemplateNode()
{
	
}

STemplateNode * STemplateNode::CreateCopyNoEdges(STemplate * stNew, int partoffset, int configoffset, int newNodeID)
{
	STemplateNode * stn = new STemplateNode(m_NodeName, newNodeID, stNew);
	stn->m_iNDescribedInstances = m_iNDescribedInstances;
	stn->m_bIsNull = m_bIsNull;

	assert(m_apPartScales.size()==m_apEllipsoids.size());
	
	PartsCollection * newStorage = stNew->GetAllParts();
	for (int i=0; i<configoffset+1; i++)
	{
		stn->m_apEllipsoids.push_back(m_apEllipsoids[0]->CreateCopy(newStorage, partoffset));
		stn->m_apPartScales.push_back(m_apPartScales[0]->CreateCopy());
	}

	for (int i=1; i<(int)m_apEllipsoids.size(); i++)
		stn->m_apEllipsoids.push_back(m_apEllipsoids[i]->CreateCopy(newStorage, partoffset));

	for (int i=1; i < (int)m_apPartScales.size(); i++)
		stn->m_apPartScales.push_back(m_apPartScales[i]->CreateCopy());
	
	return stn;
}

void STemplateNode::CopyEdges(int partoffset, int configoffset, std::map<int, int> & old2newnode,
							  std::map<int, int> & new2oldnode, STemplate * origTmplt)
{
	assert(new2oldnode.find(m_iNodeID)!=new2oldnode.end());
	STemplateNode * origNode = (STemplateNode*)origTmplt->GetNode(new2oldnode[m_iNodeID]);

	if (origNode->m_pParentPart!=NULL)
		m_pParentPart = origNode->m_pParentPart->CreateCopy(m_pST, old2newnode, new2oldnode, configoffset);

	if (m_pParentPart!=NULL)
	{
		assert(m_pParentPart->ParentTemplate()!=NULL);
		m_pParentPart->ParentTemplate()->m_SubpartTypes.push_back(m_pParentPart);
	}
}

// node info
const VKString & STemplateNode::NodeName() const
{
	return m_NodeName;
}

int STemplateNode::NodeID() const
{
	return m_iNodeID;
}

void STemplateNode::SetNodeID(int nodeID)
{
	m_iNodeID = nodeID;
}

void STemplateNode::SetName(const VKString & newName)
{
	m_NodeName = newName;
}

const PartBox * STemplateNode::GetPart(int configID) const
{
	if (m_bIsNull)
		return PartBox::GetNullPart();
	if (!(configID>=0 && configID < (int)m_apEllipsoids.size()))
	{
		std::cout<<"[ERROR] Invalid config: "<<configID<<" / "<<m_apEllipsoids.size();
		std::cout<<". Node="<<NodeName().c_str()<<std::endl;
		assert(false);
	}
	return m_apEllipsoids[configID];
}

int STemplateNode::NDescribedInstances(int configID) const
{
	if (configID < 0 || configID>=(int)m_apPartScales.size() || m_apPartScales[configID]==NULL)
		return 0;
	return m_apPartScales[configID]->NSamples();
}

R3Vector STemplateNode::GetMeanScale(int configID) const
{
	if (m_bIsNull)
		return R3Vector(0,0,0);
	
	const double * pt = GetScaleDistr(configID, GlobalConstants::s_iMinNSamplesForMean, NULL)->GetMostProbable();
	return R3Vector(pt[0], pt[1], pt[2]);
}

double STemplateNode::LogProb(int configID, const R3Vector & sample) const
{
	assert(!m_bIsNull);
	if (sample[0] < 0 || sample[1] < 0 || sample[2] < 0)
		return RN_INFINITY;
		
	const ContProbDistribution * ptVar = GetScaleDistr(configID, GlobalConstants::s_iMinNSamplesForVariance, NULL);
	const ContProbDistribution * ptMean = GetScaleDistr(configID, GlobalConstants::s_iMinNSamplesForMean, NULL);
	assert(ptVar!=NULL && ptMean!=NULL && ptVar->SafeCastToGaussian()!=NULL && ptMean->SafeCastToGaussian()!=NULL);
	
	const ProbGaussian * ptVarG = ptVar->SafeCastToGaussian();
	const ProbGaussian * ptMeanG = ptMean->SafeCastToGaussian();
	
	static double dblSample[3];
	dblSample[0] = sample.X();
	dblSample[1] = sample.Y();
	dblSample[2] = sample.Z();

	return ProbGaussian::LogProb(ptVarG->GetVariance(), ptMeanG->GetMostProbable(),
								 ptVarG->NDims(), ptVarG->NSamples(), dblSample);
}

R3Vector STemplateNode::ScaleVariance(int configID, int & ancestor) const
{
	if (m_bIsNull)
		return R3Vector(0,0,0);

	const ContProbDistribution * distr = GetScaleDistr(configID, GlobalConstants::s_iMinNSamplesForVariance, &ancestor);
	assert(distr->SafeCastToGaussian()!=NULL);
	const double * svar = distr->SafeCastToGaussian()->GetVariance();
	//std::cout<<"RETURN Variance: ["<<svar[0]<<" "<<svar[0]<<" "<<svar[0]<<"] "<<std::endl;
	return R3Vector(svar[0], svar[1], svar[2]);
}

const ContProbDistribution * STemplateNode::GetScaleDistr(int configID, int nMinSamples, int * ancestor) const
{
	assert(!m_bIsNull);
	assert(m_pST!=NULL);
	int cid = configID;
	while (cid >= 0)
	{
		if (!(cid >= 0 && cid < (int)m_apPartScales.size()))
		{
			std::cout<<"[ERROR] Invalid config: "<<cid<<" (started at: ";
			std::cout<<configID<<") / "<<m_apPartScales.size()<<". IsNull="<<m_bIsNull<<std::endl;
			assert(false);
		}
		// check if enough samples
		if (m_apPartScales[cid]->NSamples() >= nMinSamples)
			break;
		
		// if not - update cid
		//cid = m_pST->ConfigParent(cid);
		if (cid == 0)
			break;
		cid = 0;
	}
	
	if (ancestor!=NULL)
		*ancestor = cid;
	return m_apPartScales[cid];
}

void STemplateNode::InitNullGeometry()
{
	m_bIsNull = true;
}

bool STemplateNode::IsNull() const
{
	return m_bIsNull;
}
	
// hierarchy structure
void STemplateNode::AddPartOfChild(STemplateNode * child, STemplate * st, int configID, bool nullBox)
{
	PartOfEdge * edge = new PartOfEdge(this, child, st);
	m_SubpartTypes.push_back(edge);
	child->m_pParentPart = edge;
	
	if (nullBox)	// if null geometry - do not expect any updates
	{
		edge->SetNPartsAlwaysOne();
		edge->UpdateNParts(configID, 1, true);
		edge->UpdateXForm(configID, NULL, NULL, 1, 0);
	}
}

PartOfEdge * STemplateNode::PartOfParent()
{
	return m_pParentPart;
}

const PartOfEdge * STemplateNode::PartOfParent() const
{
	return m_pParentPart;
}

int STemplateNode::NPartOfChildren() const
{
	return (int)m_SubpartTypes.size();
}

void STemplateNode::RemoveChild(STemplateNode * child)
{
	for (int i=0; i<(int)m_SubpartTypes.size(); i++)
		if (m_SubpartTypes[i]->ChildTemplate()==child)
		{
			m_SubpartTypes.erase(m_SubpartTypes.begin()+i);
			return;
		}
	std::cout<<"[WARNING] Could not remove: "<<child->NodeName().c_str()<<" is NOT a part of "<<NodeName().c_str()<<std::endl;
}

const PartOfEdge * STemplateNode::PartOfChild(int cid) const
{
	assert(cid<(int)m_SubpartTypes.size() && cid>=0);
	return m_SubpartTypes[cid];
}

PartOfEdge * STemplateNode::PartOfChild(int cid)
{
	assert(cid<(int)m_SubpartTypes.size() && cid>=0);
	return m_SubpartTypes[cid];
}

void STemplateNode::InitShapeAndXForm(int configID, const STemplateNode * parent,
									  int nParts, int partID, const PartBox & geomPart)
{
	m_iNDescribedInstances = 1;
	assert(configID >= 0);
	while (configID >= (int)m_apPartScales.size())
		m_apPartScales.push_back(NULL);
	assert(m_apPartScales[configID]==NULL);
	
	while (configID >= (int)m_apEllipsoids.size())
		m_apEllipsoids.push_back(NULL);
	assert(m_apEllipsoids[configID]==NULL);
	m_apEllipsoids[configID] = new PartBox(m_pST->GetAllParts());
	m_apEllipsoids[configID]->InitCopy(geomPart);
	
	double meanScale[3];
	double varianceScale[3];
	for (int i=0; i<3; i++)
	{
		meanScale[i] = m_apEllipsoids[configID]->UndeformedScale()[i];
		varianceScale[i] = 0;
	}

	m_apPartScales[configID] = new ProbGaussian(3, meanScale, varianceScale, 1);
	
	if (m_pParentPart!=NULL)
		m_pParentPart->UpdateXForm(configID, parent, this, nParts, partID);
}

void STemplateNode::InitShapeAndXForm(int configID,const STemplateNode * parent, int nParts, int partID,
									  const ShapePointCloud * pnts, const std::vector<int> & ptsIDs, 
									  double fUseMeshThresh, const std::vector<int> & subsetFaces)
{
	m_iNDescribedInstances = 1;
	
	while (configID >= (int)m_apEllipsoids.size())
		m_apEllipsoids.push_back(NULL);
	assert(m_apEllipsoids[configID]==NULL);
	m_apEllipsoids[configID] = new PartBox(m_pST->GetAllParts());
	
	assert(configID >= 0);
	while (configID >= (int)m_apPartScales.size())
		m_apPartScales.push_back(NULL);

	// update bbox scales (note: centroids should align by the time box is transformed).
	// Not sure if orientation is considered to be part of box or transformation (currently not used)
	if ((ptsIDs.size()>0))
	{
		R3Point centroid;
		R3Triad orient;
		R3Vector scale;
		double fitErr = 0;
		
		PartBox::FindOptimalBoxFit(pnts, ptsIDs, centroid, orient, scale, fitErr);
		
		if (fitErr<fUseMeshThresh || fUseMeshThresh<0)
			m_apEllipsoids[configID]->InitBox(centroid, orient, scale, fitErr);
		else
		{
			PartGeometry * part = new PartGeometry(pnts, ptsIDs, NodeName(), centroid, orient, scale, subsetFaces);
			int partGeomID = m_pST->GetAllParts()->AddPart(part);
			m_apEllipsoids[configID]->InitPart(partGeomID, 0);
		}		
		
		double meanScale[3];
		double varianceScale[3];
		for (int i=0; i<3; i++)
		{
			meanScale[i] = m_apEllipsoids[configID]->UndeformedScale()[i];
			varianceScale[i] = 0;
		}
				
		assert(m_apPartScales[configID]==NULL);
		m_apPartScales[configID] = new ProbGaussian(3, meanScale, varianceScale, 1);
	}
	else
	{
		std::cout<<"[WARNING] No points on: "<<NodeName().c_str()<<std::endl;
		m_apEllipsoids[configID]->InitEmpty();
	}

	// update transformation to parent
	if (m_pParentPart!=NULL)
		m_pParentPart->UpdateXForm(configID, parent, this, nParts, partID);
}

void STemplateNode::UpdateShapeAndXForm(int configID, const TemplateNodeDeform & deform, bool include)
{
	assert(configID>=0);
	if (!m_bIsNull)
	{
		while (configID>=(int)m_apPartScales.size())
		{
			assert(m_apPartScales.size()==m_apEllipsoids.size());
			m_apPartScales.push_back(NULL);
			m_apEllipsoids.push_back(NULL);
		}

		if (m_apPartScales[configID]==NULL)
		{
			assert(m_apEllipsoids[configID]==NULL);
			m_apPartScales[configID] = new ProbGaussian(3);
			m_apEllipsoids[configID] = new PartBox(m_pST->GetAllParts());
			int parent = m_pST->ConfigParent(configID);
			assert(parent>=0 && parent < (int)m_apEllipsoids.size());
			m_apEllipsoids[configID]->InitCopy(*(m_apEllipsoids[parent]));
		}
	}
	
	if (m_pParentPart!=NULL)
		m_pParentPart->UpdateXForm(configID, deform, include);
	
	if (m_bIsNull)
		return;

	assert(m_apPartScales[configID]!=NULL);
	
	m_iNDescribedInstances += (include ? 1 : -1);
	assert(m_iNDescribedInstances>=0);
	
	double scale[3];
	assert(m_apPartScales[configID]->NDims()==3);
	for (int i=0; i<3; i++)
		scale[i] = deform.GetScale()[i];
	
	m_apPartScales[configID]->UpdateDistribution(scale, include);
	m_apEllipsoids[configID]->SetScale(GetMeanScale(configID));
	
	// recurse.. - disabled
//	if (configID>0 && m_pST->ConfigParent(configID)>=0)
//		UpdateShapeAndXForm(m_pST->ConfigParent(configID), deform, include);
}

void STemplateNode::Reset()
{
	if (m_pParentPart!=NULL)
		m_pParentPart->Reset();
	
	for (int i=0; i<(int)m_apEllipsoids.size(); i++)
		if (m_apEllipsoids[i]!=NULL)
			m_apEllipsoids[i]->Reset();

	for (int i=0; i<(int)m_apPartScales.size(); i++)
		m_apPartScales[i]->Reset();
	
	m_iNDescribedInstances = 0;
}

void STemplateNode::Normalize(const R3Point & centroid, double aveRadius)
{
	if (m_pParentPart!=NULL)
		m_pParentPart->Normalize(centroid, aveRadius);

	for (int i=0; i<(int)m_apPartScales.size(); i++)
	{
		for (int j=0; j<3; j++)
			m_apPartScales[i]->Rescale(j, 0, 1./aveRadius);
	}
	for (int i=0; i<(int)m_apEllipsoids.size(); i++)
		if (m_apEllipsoids[i]!=NULL)
			m_apEllipsoids[i]->Normalize(centroid, aveRadius) ;
}

void STemplateNode::WriteNode(FILE * file) const
{
	VKString::WriteString(m_NodeName, file);
	fwrite(&m_iNodeID, sizeof(int), 1, file);

	fwrite(&m_iNDescribedInstances, sizeof(int), 1, file);

	if (m_apEllipsoids.size() != m_apPartScales.size())
		std::cout<<"[ERROR] NEll = "<<m_apEllipsoids.size()<<" NSca="<<m_apPartScales.size()<<std::endl;
	assert(m_apEllipsoids.size() == m_apPartScales.size());
	int nconfigs = (int)m_apEllipsoids.size();

	fwrite(&m_bIsNull, sizeof(bool), 1, file);
	
	fwrite(&nconfigs, sizeof(int), 1, file);
	assert(nconfigs==(int)m_apPartScales.size());
	for (int i=0; i<nconfigs; i++)
	{
		if (m_apPartScales[i]==NULL)
			((STemplateNode*)this)->m_apPartScales[i] = new ProbGaussian(3);
		assert(m_apPartScales[i]!=NULL);
		m_apPartScales[i]->WriteToFile(file);
	}
	assert(nconfigs==(int)m_apEllipsoids.size());
	for (int i=0; i<nconfigs; i++)
	{
		if (m_apEllipsoids[i]==NULL)
			((STemplateNode*)this)->m_apEllipsoids[i] = new PartBox(m_pST->GetAllParts());
		assert(m_apEllipsoids[i]!=NULL);
		m_apEllipsoids[i]->WriteEllipsoid(file);
	}
}

STemplateNode * STemplateNode::ReadNode(FILE * file, STemplate * st, STemplate::STemplateVersions v)
{
	VKString nodeName = VKString::ReadString(file);
	int nodeID;
	fread(&nodeID, sizeof(int), 1, file);
	STemplateNode * stn = new STemplateNode(nodeName, nodeID, st);
	
	if (v < STemplate::STMPLT_V0_3)	// not a multi-config version!
	{
		fread(&(stn->m_iNDescribedInstances), sizeof(int), 1, file);
		
		ContProbDistribution * prob = NULL;

		bool scaleExists;
		fread(&scaleExists, sizeof(bool), 1, file);
		if (scaleExists)
		{
			ContProbDistribution * prob = ContProbDistribution::CreateFromFile(file);
			for (int i=0; i<st->NSeedConfigs(); i++)
				stn->m_apPartScales.push_back(prob->CreateCopy());
		}
		
		bool nullNode = false;
		PartBox * box = PartBox::ReadEllipsoid(file, st->GetAllParts());
//		std::cout<<"Node: "<<nodeName.c_str()<<std::endl;
		for (int i=0; i<st->NSeedConfigs() && !nullNode; i++)
		{
			stn->m_apEllipsoids.push_back(box->CreateCopy());
			nullNode = stn->m_apEllipsoids[0]->IsType(PartBox::BOXTYPE_NullBox);
		}
		if (nullNode)
		{
			stn->m_apPartScales.clear();
			stn->m_apEllipsoids.clear();
			stn->m_bIsNull = true;
		}
		
//		if (scaleExists)
//		{
//			assert(stn->m_pPCAScale!=NULL);
//			const double * scale = stn->m_pPCAScale->GetMostProbable();
//			for (int i=0; i<(int)stn->m_apEllipsoids.size(); i++)
//				stn->m_apEllipsoids[i]->SetScale(R3Vector(scale[0], scale[1], scale[2]));
//		}
	}
	else
	{
		assert(v==STemplate::STMPLT_V0_3 || v==STemplate::STMPLT_V0_4);
		fread(&(stn->m_iNDescribedInstances), sizeof(int), 1, file);
		fread(&(stn->m_bIsNull), sizeof(bool), 1, file);
		int nconfigs;
		fread(&nconfigs, sizeof(int), 1, file);
				
		for (int i=0; i<nconfigs; i++)
			stn->m_apPartScales.push_back(ContProbDistribution::CreateFromFile(file));
//		std::cout<<"Node: "<<nodeName.c_str()<<std::endl;
		for (int i=0; i<nconfigs; i++)
			stn->m_apEllipsoids.push_back(PartBox::ReadEllipsoid(file, st->GetAllParts()));

	}

	assert(stn!=NULL);
	return stn;
}

void STemplateNode::DuplicateConfigs(int from)
{
	if (m_bIsNull)
		return;
	
	assert(from >=0 && from < (int)m_apPartScales.size() );
	for (int i=(int)m_apPartScales.size(); i<m_pST->NSeedConfigs(); i++)
		m_apPartScales.push_back(m_apPartScales[from]->CreateCopy());
	
	assert(from >=0 && from < (int)m_apEllipsoids.size() );
	for (int i=(int)m_apEllipsoids.size(); i<m_pST->NSeedConfigs(); i++)
		m_apEllipsoids.push_back(m_apEllipsoids[from]->CreateCopy());
}

void STemplateNode::WriteSubparts(FILE * file) const
{
	int nSubparts = (int)m_SubpartTypes.size();
	fwrite(&nSubparts, sizeof(int), 1, file);
	
	for (int s=0; s<nSubparts; s++)
		m_SubpartTypes[s]->WritePartOfEdge(file);
}

void STemplateNode::ReadSubparts(FILE * file, STemplate * st, STemplate::STemplateVersions v)
{
	int nSubparts;
	fread(&nSubparts, sizeof(int), 1, file);
	
	for (int s=0; s<nSubparts; s++)
	{
		PartOfEdge * edge = PartOfEdge::ReadPartOfEdge(file, st, v);
		edge->ChildTemplate()->m_pParentPart = edge;
		m_SubpartTypes.push_back(edge);
	}
		
}



