/***************************************************************************
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Sortable.h"
#include "TimeProfiler.h"
#include "MRFOptimization.h"

#include "CorrsAffineAlign.h"

#include "PartOfEdge.h"
#include "TMatchAlg.h"
#include "TMatch.h"
#include "TMatchEnergy.h"
#include "STemplate.h"
#include "STemplateNode.h"


TMatchAlgParams::TMatchAlgParams()
{
	m_EStepAlg = ESTEP_GraphCutsMRF;
	//m_MStepAlg = MSTEP_GradientDescend;
	m_MStepAlg = MSTEP_ICP;
	m_bVerb = true;
	
	m_bInitAlignAndConfigQuickly = false;
	
	m_iExploreConfigs = -1;	// all configs
	//m_iExploreConfigs = 9;		// nov 15 debug
	
	//m_InitAlignments = ALIGN_IdentityOnly;	// only try identity transformation
	m_InitAlignments = ALIGN_AllXY;
	
	// Default (most expensive)
	m_iOptimFlags = (OPT_Pnts | OPT_PntsEdges | OPT_GlobalScale | OPT_GlobalRotate | OPT_GlobalTrans
					 | OPT_NodeExistence | OPT_NodeXForms | OPT_NodeShapes /*| OPT_RescaleNull*/);
	//m_iOptimFlags = OPT_Pnts | OPT_PntsEdges;
	//m_iOptimFlags = OPT_Pnts;
	//m_InitAlignments = ALIGN_AllXYZ;
	
	m_iIterations = 100;
	
	m_iMRFOptimDbl2Int = 1.;
	m_iMRFIterations = 1000;
	
	m_iICPIterations = 1000;
	
	m_iNGreedyIterations = 1000;	// nov 15 debug
	//m_iNGreedyIterations = 100;
	m_bRevert = false;
	//m_bDecoupleNodeEnergy = true;
	
	m_pMRFProblem = NULL;
	m_pMRFOptimizer = NULL;
	m_bMRFOptimizerInitLabels = false;
	
}

TimeProfiler TMatchAlg::s_sProfiler; 

void TMatchAlg::GetPointArrays(R3Point ** pntsT, R3Point ** pntsM, int arraySize)
{
	if (arraySize > s_PointsArraySize)
	{
		if (s_PointsT==NULL || s_PointsM==NULL)
		{
			delete [] s_PointsT;
			delete [] s_PointsM;
		}
		s_PointsT = new R3Point[arraySize];
		s_PointsM = new R3Point[arraySize];
		s_PointsArraySize = arraySize;
	}
	*pntsT = s_PointsT;
	*pntsM = s_PointsM;
}
R3Point * TMatchAlg::s_PointsT = NULL;
R3Point * TMatchAlg::s_PointsM = NULL;
int TMatchAlg::s_PointsArraySize = 0;


void TMatchAlg::Optimize(const STemplate * sTemplate, GeomShape * geom, 
						 TMatchAlgParams params, std::vector<TMatchEnergy *> & outSortedMatches)
{
	s_sProfiler.startedProcess("1_Optimize");
	s_sProfiler.startedProcess("1_OptimizeMeasure");
	ShapePointCloud * pnts = geom->SafeCastToPointCloud();
	assert(pnts!=NULL);

	std::vector<R3Affine> alignments;
	std::cout<<"Seed alignments"<<std::endl;
	SeedAlignments(sTemplate, geom, params, alignments);
	
	std::vector<Sortable> sortedEnergies;
	bool bInitQuickly = params.m_bInitAlignAndConfigQuickly;
	int optimFlags = params.m_iOptimFlags;

	int bestID = -1;
	double bestEnergy = -1;
	
	if (bInitQuickly)
		params.m_iOptimFlags = TMatchAlgParams::OPT_Pnts | TMatchAlgParams::OPT_NodeExistence;
	
	for (int i=0; i<(int)alignments.size(); i++)
	{
		//std::cout<<"Trying alignment: "<<i<<" / "<<alignments.size()<<std::endl;
		std::vector<TMatchEnergy *> outMatches;
		OptimizeAligned(sTemplate, geom, params, alignments[i], outMatches);
		for (int j=0; j<(int)outMatches.size(); j++)
		{
			// only points dist penalty for energy
			int origFlags = outMatches[j]->GetTerms();
			outMatches[j]->SetTerms(TMatchEnergy::DefaultMatchQualityEnergy());
			sortedEnergies.push_back(Sortable(outMatches[j]->Energy(), outMatches[j]));
			outMatches[j]->SetTerms(origFlags);
		}
	}
	std::sort(sortedEnergies.begin(), sortedEnergies.end());
	
	if (bInitQuickly)
	{
		params.m_iOptimFlags = optimFlags;
		assert(sortedEnergies.size()>0);
		TMatch * match =((TMatchEnergy*)sortedEnergies[0].ptr)->GetMatch();
		sortedEnergies[0].ptr = OptimizeConfigured(sTemplate, match, params);
	}
	
	for (int i=0; i<(int)sortedEnergies.size(); i++)
		outSortedMatches.push_back((TMatchEnergy *)sortedEnergies[i].ptr);
	s_sProfiler.finishedProcess("1_Optimize");
	s_sProfiler.finishedProcess("1_OptimizeMeasure");
}

//// HELPER FUNCTIONS /////
void TMatchAlg::OptimizeAligned(const STemplate * sTemplate, GeomShape * geom, TMatchAlgParams params,
								const R3Affine & alignment, std::vector<TMatchEnergy *> & outMatches)
{
	if (params.m_iExploreConfigs<0)	// exhaustively try all configs
	{
		int offset = 0;
		if (sTemplate->NSeedConfigs()>1)
			offset = 1;
		for (int i=offset; i<sTemplate->NSeedConfigs(); i++)
		{
			// create tmatch energy
			TMatch * seedConfig = sTemplate->CreateSeedConfig(i, geom, std::vector<int>(), alignment);
			TMatchEnergy * energy = OptimizeConfigured(sTemplate, seedConfig, params);
			outMatches.push_back(energy);
		}
	}
	else
	{
		assert(params.m_iExploreConfigs < sTemplate->NSeedConfigs());
		TMatch * seedConfig = sTemplate->CreateSeedConfig(params.m_iExploreConfigs, geom, std::vector<int>(), alignment);
		TMatchEnergy * energy = OptimizeConfigured(sTemplate, seedConfig, params);
		outMatches.push_back(energy);
	}
}

void TMatchAlg::OptimizeLabeled(const STemplate * st, ShapeLabels * labels, TMatchAlgParams params,
								std::vector<TMatchEnergy *> & outSortedMatches, char relabel)
{
	assert(false);	// TODO: e.g. CreateTemplate - should there be mesh?
	if (params.m_bVerb)
		std::cout<<"Optimizing for template / deformation of a labeled mesh"<<std::endl;
	
	int eTerms = TMatchEnergy::ET_Pnts | TMatchEnergy::ET_NodeFit2Pnts;
	
	// create template from new geometry
	// TODO: maybe use meshes all the way?
	STemplate * newst = STemplate::CreateTemplate(labels, false, false, -1, NULL);
	TMatch * newConfig = newst->CreateSeedConfig(0, labels->GetShape(), std::vector<int>(), R3identity_affine);
	ShapePointCloud * pnts = labels->GetShape()->SafeCastToPointCloud();
	assert(pnts!=NULL);
	
	// find optimal template -> template matching (constraint: same labels, boxes from input data)
		// find all distinct parts in new node
//	VKStringList partNames;
	std::map<VKString, std::vector<int> > newParts;
	std::vector<std::vector<int> > gtNode2Pnts;
	for (int i=0; i<newConfig->NTemplateNodes(); i++)	
	{
		gtNode2Pnts.push_back(std::vector<int>());
		if (newConfig->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox))
			continue;
		VKString label = newst->GetNode(newConfig->Deformation(i).NodeID())->NodeName();
		newParts[label].push_back(i);
	}
	
	// ground truth labels
	for (int i=0; i<newConfig->NShapePrimitives(); i++)
	{
		// local point ID -> local node ID
		int origNodeID = (int)labels->Value(pnts->GetPoint(newConfig->Loc2GlobPid(i)), 0);
		int localNodeID = origNodeID - 1;	// todo: check
		assert(localNodeID < (int)gtNode2Pnts.size());
		if (localNodeID>=0)
			gtNode2Pnts[localNodeID].push_back(i);		
	}
	
	// for all possible template configs - find part-to-part matches...
	std::vector<R3Affine> alignments;
	SeedAlignments(st, pnts, params, alignments);
	//alignments.push_back(R3identity_affine);
	
	std::vector<Sortable> sortedMatches;
	std::vector<std::vector<int> > unmatched;
	std::vector<std::map<int, int> > allNew2OldMatches;
	// ids of nodes in newConfig
	for (int align=0; (int)align < alignments.size(); align++)
	{
		for (int seed=0; seed<st->NSeedConfigs(); seed++)
		{
			int id = (int)unmatched.size();
			unmatched.push_back(std::vector<int>());
			allNew2OldMatches.push_back(std::map<int, int>());
			
			// find distinct parts in new config
			TMatch * oldConfig = st->CreateSeedConfig(seed, labels->GetShape(), std::vector<int>(), alignments[align]);
			std::map<VKString, std::vector<int> > oldParts;
			
			int nullID = -1;
			// keep all labels
			for (int i=0; i<oldConfig->NTemplateNodes(); i++)
			{
				VKString label = st->GetNode(oldConfig->Deformation(i).NodeID())->NodeName();
				oldParts[label].push_back(i);
				// find null node
				nullID = (oldConfig->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox) ? i : nullID);
			}
			assert(nullID!=-1);

			// match parts 'greedily'
			TMatchEnergy * oldEnergy = new TMatchEnergy(oldConfig);	
			
			// assign everything to NULL
			for (int p=0; p<oldConfig->NShapePrimitives(); p++)
				oldConfig->SetLabel(p, nullID);
			
			// for every part in newConfig (created from new geometry)
			for (std::map<VKString, std::vector<int> >::iterator iterNew = newParts.begin();
				 iterNew != newParts.end(); iterNew++)
			{
				// match it to every similarly labeled part in oldConfig (created from template)
				std::map<VKString, std::vector<int> >::iterator iterOld = oldParts.find(iterNew->first);
				if (iterOld == oldParts.end())	// no match in the template
				{
					for (int noMatchID = 0; noMatchID<(int)iterNew->second.size(); noMatchID++)
						unmatched[id].push_back(iterNew->second[noMatchID]);
					continue;
				}
				
				// measure energy with just that part labeled
				int nNew = (int)iterNew->second.size();
				int nOld = (int)iterOld->second.size();
				if (nNew==1 && nOld==1)
				{
					allNew2OldMatches[id][iterNew->second[0]] = iterOld->second[0];
					continue;
				}
					
				double * pairwiseEnergy = new double[nNew * nOld];
				for (int newNode = 0; newNode<nNew; newNode++)
				for (int oldNode = 0; oldNode<nOld; oldNode++)
				{
					// update labeling
					int newPartID = iterNew->second[newNode];	// input geometry
					int oldPartID = iterOld->second[oldNode];	// shape template
					
					// for all points w/ newPartID - assign oldPartID
					assert(newPartID >= 0 && newPartID < (int)gtNode2Pnts.size());
					for (int pID = 0; pID < gtNode2Pnts[newPartID].size(); pID++)
						oldConfig->SetLabel(gtNode2Pnts[newPartID][pID], oldPartID);
					oldConfig->UpdateLabels(false);
					
					// compute energy
					pairwiseEnergy[newNode * nOld + oldNode] = -oldEnergy->Energy(eTerms);
					
					// revert to NULL
					for (int pID = 0; pID < gtNode2Pnts[newPartID].size(); pID++)
						oldConfig->SetLabel(gtNode2Pnts[newPartID][pID], nullID);
					oldConfig->UpdateLabels(false);
				}
				
				// find best part matching
				std::vector<int> new2oldMatches;
				PairwiseMatching(nNew, nOld, pairwiseEnergy, new2oldMatches);
				delete [] pairwiseEnergy;
				
				// assign best part matching
				assert(nNew==(int)new2oldMatches.size());
				for (int lID = 0; lID < (int)new2oldMatches.size(); lID++)
				{
					int newPartID = iterNew->second[lID];
					assert(newPartID >= 0 && newPartID < (int)gtNode2Pnts.size());
					if (new2oldMatches[lID] == -1)
						unmatched[id].push_back(newPartID);
					else
					{
						int oldPartID = iterOld->second[new2oldMatches[lID]];
						assert(allNew2OldMatches[id].find(newPartID)==allNew2OldMatches[id].end());
						allNew2OldMatches[id][newPartID] = oldPartID;
					}					
				}				
			}
			
			// assign final labels
			for (int n=0; n<(int)gtNode2Pnts.size(); n++)
			{
				std::map<int, int>::iterator iter = allNew2OldMatches[id].find(n);
				int oldNodeID = nullID;
				if (iter != allNew2OldMatches[id].end())
					oldNodeID = iter->second;
				
				for (int p=0; p<(int)gtNode2Pnts[n].size(); p++)
					oldConfig->SetLabel(gtNode2Pnts[n][p], oldNodeID);
			}
			oldConfig->UpdateLabels(false);
			
			double eval = oldEnergy->Energy(eTerms);
			sortedMatches.push_back(Sortable(eval, oldEnergy, id));
		}
	}
	std::sort(sortedMatches.begin(), sortedMatches.end());
	
	assert(sortedMatches.size()>0);
	int bestID = sortedMatches[0].id;
	for (int i=0; i<(int)sortedMatches.size(); i++)
		outSortedMatches.push_back((TMatchEnergy*)sortedMatches[i].ptr);
	
	assert(sortedMatches.size()>0);
	TMatchEnergy * bestMatchEnergy = outSortedMatches[0];
	
	// deform template to fit the new point cloud
	int mStepIterations;
	int pntsEnergyType = 0;
	pntsEnergyType |= (bestMatchEnergy->GetTerms() & TMatchEnergy::ET_PntsDistance);
	pntsEnergyType |= (bestMatchEnergy->GetTerms() & TMatchEnergy::ET_PntsOrientation);
	
	bestMatchEnergy->SetTerms(TMatchEnergy::ET_Pnts | TMatchEnergy::ET_NodeFit2Pnts | pntsEnergyType);
	if (params.m_bVerb)
		std::cout<<"Optimizing deformation energy: "<<bestMatchEnergy->Energy()<<std::flush;
	MStep_DeformTemplate(bestMatchEnergy, params, mStepIterations);
	if (params.m_bVerb)
		std::cout<<" -> "<<bestMatchEnergy->Energy()<<std::endl;

	// TODO BELOW: add unmatched nodes as 'new' node parts
//	// for best old -> new template match
//	TMatch * bestMatch = bestMatchEnergy->GetMatch();
//	
//	// go over unmatched nodes - create new potential nodes
//	assert(bestID < (int)unmatched.size());
//	for (int i=0; i<(int)unmatched[bestID].size(); i++)
//	{
//		int nodeID = unmatched[bestID][i];
//		newst->GetNode
//	}
//	
//	// deform new nodes to the point cloud...
	
	if (relabel=='e')	// e-step only
	{
		int changes;
		
		EStep_LabelPoints(bestMatchEnergy, params, changes);
	}
	else if (relabel=='m')	// e and m steps
	{
		assert(false);
	}
}

void TMatchAlg::ExplainNullNodes(TMatchEnergy * match, STemplate * updateTemplate)
{
	assert(false);
}

void TMatchAlg::OptimizeConfigured(const STemplate * sTemplate, TMatchEnergy * energy, TMatchAlgParams params)
{
	s_sProfiler.startedProcess("2_OptimizeConfigured");
	TMatch * match = energy->GetMatch();
	// diminishing null penalties
	double minPenalty=0;
	double maxPenalty=0;
	double currPenalty=0;
	
	if ((params.m_iOptimFlags & TMatchAlgParams::OPT_RescaleNull)!=0)
	{
		minPenalty= energy->GetNullPenalty();
		maxPenalty = minPenalty * 16.;	// vk: constant
		currPenalty = maxPenalty;
		energy->SetNullPenalty(currPenalty);
	}
	
	bool terminate = false;
	int iter;
	std::vector<double> ehist;
	if (params.m_bVerb)
		std::cout<<"Energy Optimization. Model: "<<match->ModelName().c_str()<<". Seed: "<<match->SeedID()<<std::endl;
	
	// label everything greedily
	for (int n=0; n<match->NShapePrimitives(); n++)
		match->SetLabel(n, GreedyBestPntLabel(energy, n));
	match->UpdateLabels(false);
	
	TimeProfiler timer;
	for (iter=0; iter < params.m_iIterations; iter++)
	{
		int mStepIterations = 0;
		int eStepNChanged = 0;
		
		timer.startedProcess("IterTimer");
		
		if (params.m_bVerb)
			std::cout<<"   I="<<(iter<10?"0":"")<<(iter<100?"0":"")<<iter<<std::flush;
		if (iter%2==0)	// E-step (labeling)
		{
			double e0 = energy->Energy();
			terminate = EStep_LabelPoints(energy, params, eStepNChanged);
			//			if (energy->Energy() > e0)	// enforce convergence...
			//				terminate = true;
		}
		else	// M-step (template deformation)
		{
			int mflags = TMatchAlgParams::OPT_GlobalScale | TMatchAlgParams::OPT_GlobalRotate | TMatchAlgParams::OPT_GlobalTrans;
			mflags |= TMatchAlgParams::OPT_NodeXForms | TMatchAlgParams::OPT_NodeShapes;
			if ((params.m_iOptimFlags & (mflags))!=0)
				terminate = MStep_DeformTemplate(energy, params, mStepIterations);
			else
				terminate = true;
			// optimize node existence
			if (currPenalty == minPenalty && (params.m_iOptimFlags & TMatchAlgParams::OPT_NodeExistence)!=0)
				GreedyTryDeletingNodes(energy, params);
		}
        
        terminate = terminate && iter != 0;
        
		ehist.push_back(energy->Energy());
		if (ehist.size() > 10)
		{
			int lid = (int)ehist.size();
			if (ehist[lid] < ehist[lid-8])
				terminate = true;
		}
		
		timer.finishedProcess("IterTimer");
		
		if (params.m_bVerb)
		{
			//const char * sep = "\t";
			ehist[iter] = ehist[iter]<0.000000001 ? 0 : ehist[iter];
			VKString EStr = VKString("  E=")+VKString::number(ehist[iter]);
			while (EStr.length() < 13)
				EStr += " ";
			VKString TStr=VKString("T=")+timer.ProcessLengthStr("IterTimer");
			while (TStr.length() < 10)
				TStr += " ";
			VKString SStr = "S=";
			if (iter%2==0)
				SStr += VKString("E")+VKString::number(eStepNChanged);
			else
				SStr += VKString("M")+VKString::number(mStepIterations);
			while (SStr.length() < 9)
				SStr += " ";
			VKString NStr = VKString("N=")+VKString::number(energy->GetNullPenalty());
			std::cout<<(EStr+TStr+SStr+NStr).c_str()<<std::endl;
		}
		
		if (terminate)
		{
			if (currPenalty<=minPenalty)
			{
				match->UpdateLabels(true);
				break;
			}
			
			if ((params.m_iOptimFlags & TMatchAlgParams::OPT_RescaleNull)!=0)
			{
				currPenalty = vkMax(minPenalty, .5 * currPenalty);	// vk: const
				energy->SetNullPenalty(currPenalty);
			}
		}
	}
	
	if (params.m_bVerb)
	{
		std::cout<<"Converged at "<<iter<<" / "<<params.m_iIterations<<" iterations"<<std::endl;
		std::cout<<"     E=[ ";
		for (int i=0; i<(int)ehist.size(); i++)
			std::cout<<ehist[i]<<" ";
		std::cout<<"]"<<std::endl;
		std::cout<<energy->EnergyStr().c_str()<<std::endl;
	}
	
	delete params.m_pMRFProblem;
	delete params.m_pMRFOptimizer;
	s_sProfiler.finishedProcess("2_OptimizeConfigured");
}

TMatchEnergy * TMatchAlg::OptimizeConfigured(const STemplate * sTemplate, TMatch * match, TMatchAlgParams params)
{
	TMatchEnergy * energy = new TMatchEnergy(match);
	OptimizeConfigured(sTemplate, energy, params);
	
	return energy;
}

// returns true if needs to terminate
bool TMatchAlg::EStep_LabelPoints(TMatchEnergy * energy, TMatchAlgParams params, int & nChanged)
{
	s_sProfiler.startedProcess("3.1_EStep");
	
	TMatch * match = energy->GetMatch();
	nChanged = 0;
	
	if ((params.m_iOptimFlags & TMatchAlgParams::OPT_PntsEdges) !=0 )	// edge terms
	{
		// create mrf problem and optimizer
		if (params.m_EStepAlg==TMatchAlgParams::ESTEP_GraphCutsMRF && params.m_pMRFOptimizer==NULL)
		{
			int nLabels = match->NTemplateNodes();
			int nPoints = match->NShapePrimitives();
			
			const VKGraph * nnEdgesEnergy = energy->GetEdgeEnergy();
			
			assert(nnEdgesEnergy!=NULL);
			const double * pntsEnergy = energy->GetPntsEnergy();
			const double * labelsEnergy = energy->GetLabelEnergy();
			
			double d2i = params.m_iMRFOptimDbl2Int * energy->GetMatch()->NShapePrimitives();
			params.m_pMRFProblem = new MRFProblemArray(nPoints, nLabels, nnEdgesEnergy, pntsEnergy, labelsEnergy, d2i);
			params.m_pMRFOptimizer = new MRFOptimizer(params.m_pMRFProblem);
			params.m_bMRFOptimizerInitLabels = false;
		}

		// optimize using graph cuts		
		assert(params.m_pMRFOptimizer!=NULL && params.m_pMRFProblem!=NULL);
		if (!params.m_bMRFOptimizerInitLabels)
		{
			params.m_bMRFOptimizerInitLabels = true;
			for (int n=0; n<match->NShapePrimitives(); n++)
				params.m_pMRFOptimizer->SetLabel(n, GreedyBestPntLabel(energy, n));
		}
		//s_sProfiler.startedProcess("3.1.3_ComputePntsEnergy");
		params.m_pMRFProblem->SetNewDataCosts(energy->GetPntsEnergy());
		//s_sProfiler.finishedProcess("3.1.3_ComputePntsEnergy");
		
		//s_sProfiler.startedProcess("3.1.5_Optimize");
		params.m_pMRFOptimizer->Optimize(params.m_iMRFIterations);
		//s_sProfiler.finishedProcess("3.1.5_Optimize");
		
		//s_sProfiler.startedProcess("3.1.6_Update");
		// update labeling of points
		for (int n=0; n<match->NShapePrimitives(); n++)
			nChanged += (int)match->SetLabel(n, params.m_pMRFOptimizer->Label(n));
		//s_sProfiler.finishedProcess("3.1.6_Update");
	}
	else if ((params.m_iOptimFlags & TMatchAlgParams::OPT_PntsEdges) == 0)	// no edge terms
	{
		for (int n=0; n<match->NShapePrimitives(); n++)
			nChanged += (int)match->SetLabel(n, GreedyBestPntLabel(energy, n));
	}
	//s_sProfiler.startedProcess("3.1.4_Update");
	match->UpdateLabels(false);
	//s_sProfiler.finishedProcess("3.1.4_Update");
	s_sProfiler.finishedProcess("3.1_EStep");
	return nChanged==0;
}

int TMatchAlg::GreedyBestPntLabel(TMatchEnergy * energy, int n)
{
	TMatch * match = energy->GetMatch();
	int bestLabel = -1;
	double bestMatch = -1;
	for (int l=0; l<match->NTemplateNodes(); l++)
	{
		if (match->Deformation(l).IsIncluded())
		{
			double e = energy->ComputeEGlobalPnt(match->Loc2GlobPid(n), l);			
			if (e < bestMatch || bestLabel<0)
			{
				bestMatch = e;
				bestLabel = l;
			}
		}
	}
	assert(bestLabel!=-1);
	return bestLabel;
}

/**
 * This uses a simple gradient descend algorithm.
 * returns true if needs to terminate
 */
bool TMatchAlg::MStep_DeformTemplate(TMatchEnergy * energy, TMatchAlgParams params, int & iter)
{
	s_sProfiler.startedProcess("3.2_MStep");
	bool term = false;
	if (params.m_MStepAlg==TMatchAlgParams::MSTEP_GradientDescend)
		term = MStep_GradientDescend(energy, params, iter);
	else if (params.m_MStepAlg==TMatchAlgParams::MSTEP_ICP)
		term = MStep_ICP(energy, params, iter);
		
	s_sProfiler.finishedProcess("3.2_MStep");
	
	return term;
}

bool TMatchAlg::MStep_ICP(TMatchEnergy * energy, TMatchAlgParams params, int & iter)
{
//	std::cout<<"Starting M-Step"<<std::endl;
	assert((energy->GetTerms() & TMatchEnergy::ET_PntsOrientation) == 0);	// cannot compare local features w/ ICP
	assert((energy->GetTerms() & TMatchEnergy::ET_DistsModelSpace) != 0);	// measure energy in model space
	assert((energy->GetTerms() & TMatchEnergy::ET_DistsTmpltSpace) == 0);	// not measuring energy in template space
	
	
	// define point sets for icp (all points on template + all points on other shape)
	TMatch * match = energy->GetMatch();
	int nTemplatePnts = match->NTemplatePoints();
	int nShapePoints = match->NShapePrimitives();
	
	int ncorrs = 0;
	
	if (energy->GetTerms() & TMatchEnergy::ET_Pnts)
	{
		for (int n=0; n < match->NTemplateNodes(); n++)
			if (!match->NodeEllipsoid(n).IsType(PartBox::BOXTYPE_NullBox))
				ncorrs += match->NPointsLabeledAsNode(n);
	}
	
	if (energy->GetTerms() & TMatchEnergy::ET_NodeFit2Pnts)
	{
		for (int n=0; n < match->NTemplateNodes(); n++)
			if (match->NPointsLabeledAsNode(n)>0)
				ncorrs += match->NPointsOnTemplateNode(n);
	}
	assert(ncorrs > 0);
	R3Point * pointsT = new R3Point[ncorrs];
	R3Point * pointsM = new R3Point[ncorrs];
	
	int * pointsTIds = new int[ncorrs];
	int * pointsMIds = new int[ncorrs];
	for (int i=0; i<ncorrs; i++)
	{
		pointsTIds[i] = -1;
		pointsMIds[i] = -1;
	}
	
	// fill corresponding points (single ICP step)	
	int offset = 0;
	std::vector<int>  perNodeOffset;
	int nNewVals = 0;
	for (int n=0; n < match->NTemplateNodes(); n++)
	{
		perNodeOffset.push_back(offset);
		offset += FindNearestPointPairs(n, energy, params, &(pointsTIds[offset]), &(pointsMIds[offset]),
										ncorrs-offset, nNewVals);
		assert(offset <= ncorrs);
	}
	assert(offset == ncorrs);
	perNodeOffset.push_back(ncorrs);
	
	// Algorithm: global ICP align, per node ICP align
	bool term = false;
	double eprev = energy->Energy();
	double enext = eprev - 1;

	while (!term && iter < params.m_iICPIterations)
	{
		// global transformation
		bool updated = false;
		double eprevLocal = 100000000;
		double enextLocal = eprevLocal - 1;
		while (!term  && iter < params.m_iICPIterations)
		{
			enextLocal = OptimizeGlobalXForm(energy, params, pointsTIds, pointsMIds, ncorrs);
			if (enextLocal >= eprevLocal) break; // vk: change
			eprevLocal = enextLocal;
			FindNearestPointPairs(energy, params, pointsTIds, pointsMIds, ncorrs, nNewVals, perNodeOffset);
			iter ++ ;
			if (nNewVals==0)	// no changes
				break;
			else
				updated = true;
		}
		
		// per node optimization
		double finalE = 0;
		for (int i=0; i<match->NTemplateNodes(); i++)
		{
			if (match->NPointsLabeledAsNode(i)==0 || match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox))
				continue; // ignore nodes with no points
//			std::cout<<"Optimizing: "<<match->GetSTemplate()->GetNode(match->Deformation(i).NodeID())->NodeName().c_str()<<std::endl;
			int offset = perNodeOffset[i];
			int remaining = perNodeOffset[i+1] - offset;
			int * tids = &(pointsTIds[offset]);
			int * mids = &(pointsMIds[offset]);
			eprevLocal = 100000000;
			enextLocal = eprevLocal - 1;
			while (!term  && (iter < params.m_iICPIterations || params.m_iICPIterations==1))
			{
				enextLocal = OptimizeSingleNodeDeform(i, energy, params, tids, mids, remaining);
				if (enextLocal >= 0)
				{
					if (enextLocal >= eprevLocal) break; // vk: change
					eprevLocal = enextLocal;
				}
				else
					break;
				FindNearestPointPairs(i, energy, params, tids, mids, remaining, nNewVals);
				iter ++ ;
				if (nNewVals==0 || params.m_iICPIterations == 1)	// no changes
					break;
				else
					updated = true;
			}
			finalE += enextLocal;
		}
//		std::cout<<"Updates = "<<updated<<std::endl;
		if (!updated)	// no changes
			break;
		
		enext = energy->Energy();
//		std::cout<<"End of GLOBAL iter: "<<eprev<<" -> "<<enext<<" vs "<<finalE<<std::endl;
//		std::cout<<"    EnergyStr: "<<energy->EnergyStr().c_str()<<std::endl;
		if (enext >= eprev)
			break;
		eprev = enext;
		
		iter++;
	}
	delete [] pointsT;
	delete [] pointsM;
	delete [] pointsTIds;
	delete [] pointsMIds;

	return false;
}

void TMatchAlg::FindNearestPointPairs(TMatchEnergy * energy, const TMatchAlgParams & params,
									  int * pointsTIds, int * pointsMIds, int ncorrs, int & nNewVals,
									  const std::vector<int> & perNodeOffset)
{
	TMatch * match = energy->GetMatch();
	nNewVals = 0;
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		int offset = perNodeOffset[i];
		int nnv=0;
		FindNearestPointPairs(i, energy, params,
							  &(pointsTIds[offset]), &(pointsMIds[offset]), ncorrs-offset, nnv);
		nNewVals += nnv;
	}
}

int TMatchAlg::FindNearestPointPairs(int nodeID, TMatchEnergy * energy, const TMatchAlgParams & params,
									 int * pointsTIds, int * pointsMIds, int remainingSpace, int & nNewVals)
{
	nNewVals = 0;
	TMatch * match = energy->GetMatch();
	if (match->NPointsLabeledAsNode(nodeID)==0 || match->NodeEllipsoid(nodeID).IsType(PartBox::BOXTYPE_NullBox))
		return 0;
		
	int corrid = 0;
	// for each point in template (T) - find closest point in point cloud (P)
	
	if (energy->GetTerms() & TMatchEnergy::ET_NodeFit2Pnts)
	{
		for (int p=0; p<match->NPointsOnTemplateNode(nodeID); p++)
		{
			assert(corrid < remainingSpace);
			// pointsT[corrid] - find point in deformed template space
			int tid = match->GlobalTmpltPnt(nodeID, p);
			R3Point tpnt = match->DeformedTemplateSample(tid);
			match->GlobalXForm().ApplyInverse(tpnt);	// to (normalized) point space
			pointsTIds[corrid] = tid;
			
			// find a point in point cloud closest to tpnt, same kind as tpnt
			
			int gid = match->NearestPointWithLabel(nodeID, tpnt);
			assert(gid>=0);
			
			// check if nearest points changed
			if (pointsMIds[corrid] != gid)
				nNewVals++;
			pointsMIds[corrid] = gid;
			corrid++;
			assert(corrid < remainingSpace);
		}
		//std::cout<<"    node2pnts: "<<nNewVals<<std::endl;
	}
	
	// for each point in P - find closest T
	if (energy->GetTerms() & TMatchEnergy::ET_Pnts)
	{
		for (int p=0; p<match->NPointsLabeledAsNode(nodeID); p++)
		{
			assert(corrid < remainingSpace);
			// tpnt - point cloud deformed to template
			int gid = match->GetPointLabeledAsNode(nodeID, p);
			//R3Point tpnt = match->AlignedToNodePointPosition(nodeID, gid, match->SeedID());
			R3Point tpnt = match->AlignedPointPosition(gid);
			pointsMIds[corrid] = gid;
			
			// find nearest point on nodeID in template
			int tid = match->NearestTemplatePointGlobalID(nodeID, tpnt);
	 
			// check if nearest points changed
			if (pointsTIds[corrid] != tid)
				nNewVals++;
			pointsTIds[corrid] = tid;

			corrid++;
			assert(corrid <= remainingSpace);
		}
		//std::cout<<"    pnts2node: "<<nNewVals<<std::endl;
	}
	//nNewVals = 1;
	return corrid;
} 

double TMatchAlg::OptimizeGlobalXForm(TMatchEnergy * energy, const TMatchAlgParams & params,
									int * pointsTIds, int * pointMIds, int ncorrs)
{
	double distPnts = 0;
	TMatch * match = energy->GetMatch();
	
	// allocate memory
	assert(ncorrs>0);
	R3Point * pntsT, * pntsM;
	TMatchAlg::GetPointArrays(&pntsT, &pntsM, ncorrs);
	
	// find point positions (XYZ)
	for (int i=0; i<ncorrs; i++)
	{
		// template point in normalized points space
		pntsT[i] = match->DeformedTemplateSample(pointsTIds[i]);
		
		// point in normalized space
		R3Point pm = match->NormalizedPosition(pointMIds[i]);
		pntsM[i] = pm;
		match->GlobalXForm().Apply(pm);
		
		distPnts += R3SquaredDistance(pntsT[i], pm);
	}
	
	// find optimal transformation
	//R3Affine icpaffine = R3identity_affine;
	R3Affine icpaffine = R3Affine(CorrsAffineAlign::R3AlignInXY(ncorrs, pntsT, pntsM, NULL, true, true, false));

	energy->GetMatch()->SetGlobalXForm(icpaffine);
	return distPnts;
}

double TMatchAlg::OptimizeSingleNodeDeform(int nodeID, TMatchEnergy * energy, const TMatchAlgParams & params,
										 int * pointsTIds, int * pointsMIds, int ncorrs)
{
	double distPnts = 0;
	
	TMatch * match = energy->GetMatch();
	const TemplateNodeDeform & deform = match->Deformation(nodeID);
	const STemplate * st = match->GetSTemplate();
	const STemplateNode * stn = st->GetNode(deform.NodeID());
	const PartBox * ep = stn->GetPart(match->SeedID());
	const PartOfEdge * pe = stn->PartOfParent();
	assert(pe!=NULL);
	//std::cout<<"Optimizing Node: "<<deform.NodeID()<<" = "<<stn->NodeName().c_str()<<" included="<<ep->IsIncluded()<<std::endl;
	assert(!ep->IsType(PartBox::BOXTYPE_NullBox) && !ep->IsType(PartBox::BOXTYPE_EmptyBox) && ep->IsIncluded());
	int flags, ndims, ancestor;
	
	double Np = (double) match->NShapePrimitives();
	double Nc = (double) ncorrs;

	// allocate memory
	assert(ncorrs>0);
	R3Point * pntsT, * pntsM;
	TMatchAlg::GetPointArrays(&pntsT, &pntsM, ncorrs);

	// define variables based on gaussians for position and scale
	// read from gaussians and re-orient
	R3Triad o = ep->UndeformedOrientation();
	R3Vector d0 = stn->GetMeanScale(match->SeedID());
	R3Vector var_d = 2. * stn->ScaleVariance(match->SeedID(), ancestor);
	bool zero_var_d[3];
	bool allZero = true;
	for (int i=0; i<3; i++)
	{
		zero_var_d[i] = (var_d[i] == 0);
		allZero = allZero && zero_var_d[i];
		var_d[i] = vkMax(var_d[i], 10E-10);
	}
	
	const double * posMean = pe->GetMeanTransform(match->SeedID(), deform.NParts(), deform.PartID(), flags, ndims);
	R3Vector u0(0,0,0);
	//double pntsScale = 1. / (1. * Np);	// TODO: set sigma
	for (int i=0; i<3; i++)
		u0[i] = o[i][0]*posMean[0] + o[i][1]*posMean[1] + o[i][2]*posMean[2];
	R3Vector var_c_orig = 2. * pe->PosVariance(match->SeedID(), deform.NParts(), deform.PartID(), ancestor);
	double var_c[3];
	bool zero_var_c[3];
	for (int i=0; i<3; i++)
	{
		zero_var_c[i] = (var_c_orig[i] == 0);
		allZero = allZero && zero_var_c[i];
		var_c[i] = vkMax(var_c_orig.Dot(o[i]), 10E-10);
	}
	
	if (allZero && (energy->GetTerms() & TMatchEnergy::ET_NodeXForm) && (energy->GetTerms() & TMatchEnergy::ET_NodeShape))
		return -1;

	// find point positions (oriented to node's coordinate frame)
	double m11[3], m12[3], m21[3], m22[3], b1[3], b2[3];

	// terms independent of point pairs
	for (int i=0; i<3; i++)
	{
		m11[i] = 0;		m12[i] = 0;		b1[i] = 0;
		m21[i] = 0;		m22[i] = 0;		b2[i] = 0;
	}
	
	// terms independent of point pairs
	double sw = energy->GetNodeShapeWeight();
	double xw = energy->GetNodeXFormWeight();
	double m2tw = energy->GetModelPntsDistWeight(nodeID);
	double t2mw = energy->GetTmpltPntsDistWeight(nodeID);
	assert(m2tw==t2mw);

	for (int i=0; i<3; i++)
	{
		if (energy->GetTerms() & TMatchEnergy::ET_NodeXForm)
		{
			// dE / dui: independent of point pairs
			m11[i] += xw / var_c[i];			// [  ui  ]
			m12[i] += 0;					// [  di  ]
			b1[i] += xw * u0[i] / var_c[i];		// [  const  ]
		}
		
		if (energy->GetTerms() & TMatchEnergy::ET_NodeShape)
		{
			// dE / ddi: independent of point pairs
			m21[i] += 0;					// [  ui  ]
			m22[i] += sw / var_d[i];			// [  di  ]
			b2[i] += sw * d0[i] / var_d[i];		// [  const  ]
		}
	}
	
	const R3Point & boxCentroid = match->NodeEllipsoid(nodeID).UndeformedCentroid();
	const R3Vector & boxScale = match->NodeEllipsoid(nodeID).UndeformedScale();
	
	// terms due to point pairs
	for (int c=0; c<ncorrs; c++)
	{
		// template point in normalized points space
		R3Point tpnt = match->UndeformedTemplateSample(pointsTIds[c]);

		int pnid, lpid;
		match->LocalPntID(pointsTIds[c], pnid, lpid);
		assert(pnid == nodeID);
		
		// point in normalized space
		R3Point mpnt = match->AlignedPointPosition(pointsMIds[c]);
		{// just for energy computation
			R3Point tmpTemplatePos = match->DeformedTemplateSample(pointsTIds[c]);
			distPnts += R3SquaredDistance(tmpTemplatePos, mpnt) * m2tw;
		}
		//std::cout<<"Dist = "
		
		// project points to local frames
		R3Point ut(0,0,0), um(0,0,0);
		for (int a=0; a<3; a++)
			tpnt[a] = (tpnt[a] - boxCentroid[a]);
		for (int i=0; i<3; i++)
		for (int a=0; a<3; a++)
			ut[i] += (tpnt[a] * o[i][a])/boxScale[i];
		
			for (int i=0; i<3; i++)
		for (int a=0; a<3; a++)
			um[i] += mpnt[a] * o[i][a];

		// go over dimensions independently
		for (int i=0; i<3; i++)
		{
			// dE / dui: point pairs
			m11[i] += 1. * m2tw;			// [  ui  ] - no terms
			m12[i] += ut[i] * m2tw;			// [  di  ]
			b1[i] += um[i] * m2tw;			// [  const  ]
			
			// dE / ddi: point pairs
			m21[i] += ut[i] * m2tw;			// [  ui  ]
			m22[i] += ut[i]*ut[i] * m2tw;	// [  di  ]
			b2[i] += ut[i] * um[i] * m2tw;	// [  const  ]
		}
	}
	
	// find optimal transformation
	double u[3], d[3];
	for (int i=0; i<3; i++)
	{
		double det = m11[i] * m22[i] - m21[i] * m12[i];
		assert(det!=0);
		u[i] = ( m22[i] * b1[i] - m12[i] * b2[i]) / det;
		d[i] = (-m21[i] * b1[i] + m11[i] * b2[i]) / det;
	}
	
	// given u, d - rescale node
	R3Point newpos(0,0,0);
	for (int a=0; a<3; a++)
	for (int i=0; i<3; i++)
		newpos[a] += u[i] * o[i][a];
	for (int a=0; a<3; a++)
	{
		u[a] = newpos[a];
		if (zero_var_c[a] && (energy->GetTerms() & TMatchEnergy::ET_NodeShape))
			u[a] = posMean[a];
	}
	
	bool hadToAbs = false;
	for (int i =0; i<3; i++)
	{
		if (d[i] < 0)
			hadToAbs = true;
	}
	if (hadToAbs)
		std::cout<<"[WARNING] Invalid distance from ICP: ["<<d[0]<<" "<<d[1]<<" "<<d[2]<<"]"<<std::endl;
	for (int i =0; i<3; i++)
	{
		d[i] = vkMax(vkAbs(d[i]), 10E-3);
		if (zero_var_d[i] && (energy->GetTerms() & TMatchEnergy::ET_NodeXForm))
			d[i] = d0[i];
	}
		
	double scalePenalty = 0;
	if (energy->GetTerms() & TMatchEnergy::ET_NodeShape)
	{
		for (int i=0; i<3; i++)
			scalePenalty += sw * pow(d[i]-d0[i], 2.) / var_d[i];
	}
	
	double xformPenalty = 0;
	if (energy->GetTerms() & TMatchEnergy::ET_NodeXForm)
	{
		for (int i=0; i<3; i++)
			xformPenalty += xw * pow(u[i]-posMean[i], 2.) / var_c_orig[i];
	}
	assert(d[0]>0 && d[1]>0 && d[2]>0);
		
	match->SetPCAScale(nodeID, R3Vector(d[0], d[1], d[2]));
	match->SetTransform(nodeID, u);
	match->UpdateEllipsoid(nodeID);
	
	//std::cout<<"ICP Energy: Xf="<<xformPenalty<<" S="<<scalePenalty<<" pnts="<<distPnts<<std::endl;
	
	//return xformPenalty + scalePenalty + distPnts;
	return scalePenalty;
	
}



void TMatchAlg::GreedyTryDeletingNodes(TMatchEnergy * energy, TMatchAlgParams params)
{
	int NPntsToConsider = 4;	// vk: const	- what defines a box
	int NPntsToRemove = 4;		// vk: const
	
	// find null node
	int nullNodeID = -1;
	TMatch * match = energy->GetMatch();
	for (int i=0; i<match->NTemplateNodes(); i++)
		if (match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox))
			nullNodeID = i;
	assert(nullNodeID>=0);

	// try removing the worst-matching node
	std::vector<Sortable> tNodes;
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		const TemplateNodeDeform & deform = match->Deformation(i);
		const PartOfEdge * parent = match->GetSTemplate()->GetNode(deform.NodeID())->PartOfParent();
		assert(deform.PartID()==0);	// TODO: remove all parts if more than 1
		assert(parent!=NULL);
		int npnts = match->NPointsLabeledAsNode(i);
		
		if (parent->GetProbNParts(match->SeedID())->Prob(deform.NParts()) < 1. || npnts < NPntsToConsider)
			tNodes.push_back(Sortable(-energy->NodeEnergy(i), NULL, i));
	}
	std::sort(tNodes.begin(), tNodes.end());
		
	// remember current node ids - assign everything to null
	for (int i=0; i<(int)tNodes.size(); i++)
	{
		int nodeID = tNodes[i].id;
		VKString nodeName = match->GetSTemplate()->GetNode(match->Deformation(nodeID).NodeID())->NodeName();
//		std::cout<<"Trying to delete: "<<nodeName.c_str()<<std::endl;
		assert(nullNodeID>=0);
		std::vector<int> revert;
		double e0 = energy->Energy();
		int npnts = 0;
		for (int p=0; p<match->NShapePrimitives(); p++)
		{
			if (match->LocalNodeID(match->Loc2GlobPid(p)) == nodeID)
			{
				npnts++;
				// assign point to best node
				int bestLabel = -1;
				double bestMatch = 0;
				for (int l=0; l<match->NTemplateNodes(); l++)
				{
					if (match->Deformation(l).IsIncluded())
					{
						double e = energy->ComputeEGlobalPnt(match->Loc2GlobPid(p), l);
						if ((e < bestMatch || bestLabel<0) && l != nodeID)
						{
							bestMatch = e;
							bestLabel = l;
						}
					}
				}

				match->SetLabel(p, bestLabel);
				revert.push_back(p);
			}
		}
		match->UpdateLabels(true);
		double e1 = energy->Energy();
		// enough points to define a box
//		std::cout<<"Decided to Keep: "<<(e1 >= e0 && npnts>3)<<std::endl;
		if (e1 >= e0 && npnts>=NPntsToRemove)	// energy not improved - do not remove node
		{
			for (int r=0; r<(int)revert.size(); r++)
				match->SetLabel(revert[r], nodeID);
			match->UpdateLabels(true);
			break;
		}
	}
}

////////////////// STATIC HELPER FUNCTION //////////////////
void TMatchAlg::SeedAlignments(const STemplate * sTemplate, GeomShape * geom, 
							   TMatchAlgParams params, std::vector<R3Affine> & alignments)
{
	if (params.m_InitAlignments == TMatchAlgParams::ALIGN_IdentityOnly)
		alignments.push_back(R3identity_affine);
	else if (params.m_InitAlignments == TMatchAlgParams::ALIGN_AllXYZ
			 || params.m_InitAlignments == TMatchAlgParams::ALIGN_AllXY)
	{
		int up = -1;
		if (params.m_InitAlignments == TMatchAlgParams::ALIGN_AllXY)
			up = 2;
		
		// for several root-to-root transformations (above) try matching...
		R3Triad axesTmplt(R3Vector(1,0,0), R3Vector(0,1,0), R3Vector(0,0,1));
		R3Triad axesMatch(R3Vector(1,0,0), R3Vector(0,1,0), R3Vector(0,0,1));

		CorrsAffineAlign::GetPCAAlignments(R3Point(0,0,0), 1., axesMatch, R3Point(0,0,0), 1., axesTmplt, 
										   false, true, false, up, alignments);		
	}
	else
		assert(false);	
}

void TMatchAlg::PairwiseMatching(int N1, int N2, const double * vals, std::vector<int> & n1Matches)
{
	std::vector<int> nonZeroRows;
	std::vector<int> nonZeroCols;
	n1Matches.clear();
	for (int i=0; i<N1; i++)
	{
		nonZeroRows.push_back(i);
		n1Matches.push_back(-1);
	}
	for (int i=0; i<N2; i++)
		nonZeroCols.push_back(i);
	
	while (nonZeroRows.size()>0 && nonZeroCols.size()>0)
	{
		int br = 0;
		int bc = 0;
		double bs = vals[br*N2+bc];
		for (int rID=0; rID<(int)nonZeroRows.size(); rID++)
		{
			int r = nonZeroRows[rID];
			for (int cID=0; cID<(int)nonZeroCols.size(); cID++)
			{
				int c = nonZeroCols[cID];
				if (vals[r*N2+c] > bs)
				{
					bs = vals[r*N2+c];
					br = rID;
					bc = cID;
				}
			}
		}
		
		n1Matches[nonZeroRows[br]] = nonZeroCols[bc];
		nonZeroRows.erase(nonZeroRows.begin()+br);
		nonZeroCols.erase(nonZeroCols.begin()+bc);
	}
}



