/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __TMatchAlg_H
#define __TMatchAlg_H

#include "VKGraph.h"
#include "MRFOptimization.h"

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"
#include "ShapeLocalPCA.h"

#include "TimeProfiler.h"

class TMatch;
class STemplate;
class TMatchEnergy;

/**
 * Matching algorithm parameters
 */
class TMatchAlgParams
{
	public:
		TMatchAlgParams();
	
		// labeling algorithm
		enum EStepAlg
		{
			ESTEP_GraphCutsMRF,
		};
	
		// template deformation
		enum MStepAlg
		{
			MSTEP_GradientDescend,
			MSTEP_ICP
		};
	
		enum OptimizationFlags
		{
			OPT_Pnts = 1,
			OPT_PntsEdges = 2,
			
			OPT_GlobalScale = 4,
			OPT_GlobalRotate = 8,
			OPT_GlobalTrans = 16,
			
			OPT_NodeExistence = 32,
			OPT_NodeXForms = 64,
			OPT_NodeShapes = 128,
			
			OPT_RescaleNull = 256
		};
	
		enum InitAlignments
		{
			ALIGN_IdentityOnly,
			ALIGN_AllXYZ,
			ALIGN_AllXY,
			ALIGN_AllPCA,
		};
	
	// be verbose
		bool m_bVerb;
	
	// initial alignments
		bool m_bInitAlignAndConfigQuickly;
		InitAlignments m_InitAlignments;
		int m_iExploreConfigs;
	
	// global iterations params
		int m_iIterations;
		EStepAlg m_EStepAlg;
		MStepAlg m_MStepAlg;
		int m_iOptimFlags;
		
	// mrf parameters
		int m_iMRFIterations;
		double m_iMRFOptimDbl2Int;
	
	// icp parameters
		int m_iICPIterations; 
	
	// gradient descend params
		int m_iNGreedyIterations;
		bool m_bRevert;
		//bool m_bDecoupleNodeEnergy;
	
	// some temporary parameters 
		// if m_EStepAlg==ESTEP_GraphCutsMRF
		MRFProblemArray * m_pMRFProblem;
		MRFOptimizer * m_pMRFOptimizer;
		bool m_bMRFOptimizerInitLabels;
	
	//
};




/**
 * Some template matching algorithms
 */
class TMatchAlg
{
	public:	
		// match model to template: returns all local minima
		static void Optimize(const STemplate * sTemplate, GeomShape * geom, TMatchAlgParams params,
							 std::vector<TMatchEnergy *> & outSortedMatches);
		static TMatchEnergy * OptimizeConfigured(const STemplate * sTemplate, TMatch * seedConfig, TMatchAlgParams params);
		static void OptimizeConfigured(const STemplate * sTemplate, TMatchEnergy * energy, TMatchAlgParams params);
	
		// find best deformation (assuming template is labeled)
		static void OptimizeLabeled(const STemplate * sTemplate, ShapeLabels * labels, TMatchAlgParams params,
									std::vector<TMatchEnergy *> & outSortedMatches, char relabel);
	
		static void ExplainNullNodes(TMatchEnergy * match, STemplate * updateTemplate);
		
		static TimeProfiler s_sProfiler;

		// more detailzed optimization
		static bool EStep_LabelPoints(TMatchEnergy * energy, TMatchAlgParams params, int & nChangedPnts);
		static bool MStep_DeformTemplate(TMatchEnergy * energy, TMatchAlgParams params, int & nIterations);
		static bool MStep_ICP(TMatchEnergy * energy, TMatchAlgParams params, int & nIterations);
		static bool MStep_GradientDescend(TMatchEnergy * energy, TMatchAlgParams params, int & nIterations);
	
	//protected:
		static void OptimizeAligned(const STemplate * sTemplate, GeomShape * geom, TMatchAlgParams params,
									const R3Affine & alignment, std::vector<TMatchEnergy *> & outMatches);
		static int GreedyBestPntLabel(TMatchEnergy * energy, int pntID);

		static void FindNearestPointPairs(TMatchEnergy * energy, const TMatchAlgParams & params,
										  int * pointsTIds, int * pointsMIds, int remainingSpace, int & nNewVals,
										  const std::vector<int> & perNodeOffset);
		static int FindNearestPointPairs(int nodeID, TMatchEnergy * energy, const TMatchAlgParams & params,
										 int * pointsTIds, int * pointsMIds, int remainingSpace, int & nNewVals);
		static double OptimizeGlobalXForm(TMatchEnergy * energy, const TMatchAlgParams & params,
										  int * pointsTIds, int * pointsMIds, int ncorrs);
		static double OptimizeSingleNodeDeform(int nodeID, TMatchEnergy * energy, const TMatchAlgParams & params,
											   int * pointsTIds, int * pointsMIds, int ncorrs);
	
		static void GreedyTryDeletingNodes(TMatchEnergy * energy, TMatchAlgParams params);
		static int GreedyOptimizeNodeShapes(TMatchEnergy * energy, TMatchAlgParams params,
											std::vector<R3Vector> & scales, R3Vector & testScale, 
											double currStep);
		static int GreedyOptimizeNodeXForms(TMatchEnergy * energy, TMatchAlgParams params,
										    std::vector<double *> & transforms, double * testTransform, 
										    double currStep);
		static int GreedyOptimizeGlobalXForm(int & iter, TMatchEnergy * energy, TMatchAlgParams params);
	
		static void SeedAlignments(const STemplate * sTemplate, GeomShape * geom, TMatchAlgParams params,
								   std::vector<R3Affine> & alignments);
		
		static void PairwiseMatching(int N1, int N2, const double * vals, std::vector<int> & n1Matches);
		static int s_iMRFOptimDbl2Int;
	
	protected:
		static void GetPointArrays(R3Point ** pnts1, R3Point ** pnts2, int arraySize);
		static R3Point * s_PointsT;
		static R3Point * s_PointsM;
		static int s_PointsArraySize;
	
//		static double EvalStep(TMatchEnergy * energy, TMatchAlgParams params,
//							   OptimizationFlags optimStep, int nodeID, int dim, double step);
//		static double ChooseStep(TMatchEnergy * energy, TMatchAlgParams params,
//								 OptimizationFlags optimStep, int nodeID, int dim, double step);		
//		static bool GreedyOptimizeFixedLabel(TMatchEnergy * energy, TMatchAlgParams params);
	

};


#endif

