/***************************************************************************
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Sortable.h"
#include "TimeProfiler.h"
#include "MRFOptimization.h"

#include "CorrsAffineAlign.h"

#include "PartOfEdge.h"
#include "TMatchAlg.h"
#include "TMatch.h"
#include "TMatchEnergy.h"
#include "STemplate.h"
#include "STemplateNode.h"

bool TMatchAlg::MStep_GradientDescend(TMatchEnergy * energy, TMatchAlgParams params, int & iter)
{
	TMatch * match = energy->GetMatch();
	
	std::vector<double *> transforms;
	std::vector<R3Vector> scales;
	int xformdims, xformflags;
	
	int ndims = -1;
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		const TemplateNodeDeform & deform = match->Deformation(i);
		const double * tmpxform = deform.GetTransform(xformdims, xformflags);
		if (ndims==-1) ndims = xformdims;
		assert(ndims==xformdims);
		
		transforms.push_back(new double[ndims]);
		for (int j=0; j<ndims; j++)
			transforms[i][j] = tmpxform[j];
		scales.push_back(deform.GetScale());
	}
	
	iter = 0;
	
	double * testTransform = new double[ndims];
	R3Vector testScale;
	int totalChanges = 1;
	int globXFormFlags = (TMatchAlgParams::OPT_GlobalScale
						  | TMatchAlgParams::OPT_GlobalRotate | TMatchAlgParams::OPT_GlobalTrans);
	while (iter < params.m_iNGreedyIterations && totalChanges>0)
	{
		double e0 = energy->Energy();
		//std::cout<<"MDeform: "<<energy->Energy()<<" iter="<<iter<<" / "<<params.m_iNGreedyIterations<<std::endl;
		
		// compute derivatives for each cluster (independent of other ones)
		totalChanges = 0;
		double currStep;
		double minStep;
		double changeRatio = .5; // vk: const
		
		// OPTIMIZE GLOBAL AFFINE MAP
		s_sProfiler.startedProcess("3.2.1_MStepGlobalXForm");
		if (params.m_iOptimFlags & globXFormFlags)
			totalChanges += GreedyOptimizeGlobalXForm(iter, energy, params);
		//std::cout<<"    Global xform: "<<totalChanges<<" I"<<iter<<"\tE="<<energy->Energy()<<std::endl;
		
		s_sProfiler.finishedProcess("3.2.1_MStepGlobalXForm");
		
		// OPTIMIZE NODE TRANSFORMS
		s_sProfiler.startedProcess("3.2.2_MStepNodeXForms");
		if (params.m_iOptimFlags & TMatchAlgParams::OPT_NodeXForms)
		{
			currStep = .05;
			minStep = .01;	// vk: const
			while (iter < params.m_iNGreedyIterations)
			{
				int changes = GreedyOptimizeNodeXForms(energy, params, transforms, testTransform, currStep);
				totalChanges+=changes;
				if (changes==0) currStep *= changeRatio;
				if (currStep<=minStep) break;
				if (changes>0) iter++;
			}
		}
		//std::cout<<"    Node xform: "<<totalChanges<<" I"<<iter<<"\tE="<<energy->Energy()<<std::endl;
		s_sProfiler.finishedProcess("3.2.2_MStepNodeXForms");
		
		// OPTIMIZIE NODE SHAPES
		s_sProfiler.startedProcess("3.2.3_MStepNodeShapes");
		if (params.m_iOptimFlags & TMatchAlgParams::OPT_NodeShapes)
		{
			currStep = .05;
			minStep = .01;	// vk: const
			while (iter < params.m_iNGreedyIterations)
			{
				// i think should also optimize transformations
				int changes = GreedyOptimizeNodeShapes(energy, params, scales, testScale, currStep);
				//std::cout<<"    1. Node scale: "<<totalChanges<<" I"<<iter<<"\tE="<<energy->Energy()<<std::endl;
				//if (params.m_iOptimFlags & TMatchAlgParams::OPT_NodeXForms)
				//	changes += GreedyOptimizeNodeXForms(energy, params, transforms, testTransform, currStep);
				//std::cout<<"    2. Node xform: "<<totalChanges<<" I"<<iter<<"\tE="<<energy->Energy()<<std::endl;
				
				totalChanges+=changes;
				if (changes==0) currStep *= changeRatio;
				if (currStep<=minStep) break;
				if (changes>0) iter++;
			}
		}
		s_sProfiler.finishedProcess("3.2.3_MStepNodeShapes");
		if (energy->Energy() >= e0)
			break;
	}
	
	delete [] testTransform;
	for (int i=0; i<match->NTemplateNodes(); i++)
		delete [] transforms[i];
	return false;
}


int TMatchAlg::GreedyOptimizeNodeShapes(TMatchEnergy * energy, TMatchAlgParams params,
										std::vector<R3Vector> & scales, R3Vector & testScale, double currStep)
{
	std::vector<R3Vector> revertScales;
	if (params.m_bRevert)
		revertScales = scales;
	int changes = 0;
	TMatch * match = energy->GetMatch();
	
	double e0 = energy->Energy();
	std::vector<bool> ignoreNode;
	double epsilon = 10e-4;
	
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		if (match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox) || !match->Deformation(i).IsIncluded())
			continue;
		
		const TemplateNodeDeform & deform = match->Deformation(i);
		R3Vector testScale = scales[i];
		for (int d=0; d<3; d++)
		{
			testScale[d] = scales[i][d]+currStep;
			double eNext = e0;
			if (testScale[0]>=epsilon && testScale[1]>=epsilon && testScale[2]>=epsilon)
			{
				match->SetPCAScale(i, testScale);
				match->UpdateEllipsoid(i);
				eNext = energy->Energy();
			}
			
			double ePrev = e0;
			testScale[d] = scales[i][d]-currStep;
			if (testScale[0]>=epsilon && testScale[1]>=epsilon && testScale[2]>=epsilon)
			{
				match->SetPCAScale(i, testScale);
				match->UpdateEllipsoid(i);
				ePrev = energy->Energy();
			}
			
			// pick optimal
			double change = 0;
			if (ePrev<(e0-epsilon) && ePrev<eNext)
				change = -currStep;
			else if (eNext<(e0-epsilon) && eNext<ePrev)
				change = currStep;
			changes += (int)(change!=0);
			
			// revert or set new optimal
			if (params.m_bRevert)	// change state to optimal
				testScale[d] = scales[i][d];
			else		// change state to previous
				testScale[d] = scales[i][d] + change;
			match->SetPCAScale(i, testScale);
			match->UpdateEllipsoid(i);
			if (!params.m_bRevert)
			{
				double enew = energy->Energy();
				
				//				if (!(enew <= e0 && vkAbs(enew - vkMin(vkMin(ePrev, eNext), e0))<epsilon))
				//				{
				//					std::cout<<"enew="<<enew<<"==min("<<ePrev<<", "<<eNext<<") "<<" vs "<<e0<<std::endl;
				//					assert(false);
				//				}
				e0 = enew;
			}
			
			scales[i][d] += change;
		}
	}
	if (changes>0)
	{
		if (params.m_bRevert)
		{
			const STemplate * st = match->GetSTemplate();
			for (int i=0; i<match->NTemplateNodes(); i++)
				match->SetPCAScale(i, scales[i]);
			match->UpdateEllipsoids();
		}
		
		if (params.m_bRevert && (energy->Energy() >= e0))
		{
			changes = 0;
			assert(scales.size()==revertScales.size());
			scales = revertScales;
			for (int i=0; i<match->NTemplateNodes(); i++)
				match->SetPCAScale(i, scales[i]);
			match->UpdateEllipsoids();
		}
		else if (energy->Energy() > e0)
		{
			std::cout<<"[WARNING] Energy got worse"<<energy->Energy()<<" vs "<<e0<<std::endl;
			//			std::cout<<"Energy got worse: "<<energy->Energy()<<" vs "<<e0<<std::endl;
			//			std::cout<<"Breakdown: "<<energy->EnergyStr().c_str()<<std::endl;
			//			assert(false);
		}
	}
	
	return changes;
}

int TMatchAlg::GreedyOptimizeNodeXForms(TMatchEnergy * energy, TMatchAlgParams params,
										std::vector<double *> & transforms, double * testTransform, double currStep)
{
	int xformdims, xformflags;
	std::vector<double*> revertTranslations;
	TMatch * match = energy->GetMatch();
	if (params.m_bRevert)
	{
		for (int i=0; i<match->NTemplateNodes(); i++)
		{
			const double * xform = match->Deformation(i).GetTransform(xformdims, xformflags);
			assert(xformflags == PartTransforms::PARTXFORM_Translation);
			revertTranslations.push_back(new double[3]);
			for (int d=0; d<3; d++)	// initialize to previous values
				revertTranslations[i][d] = transforms[i][d];
		}
	}
	
	int changes = 0;
	double e0 = energy->Energy();
	VKString e0Str = energy->EnergyStr();
	double epsilon = 10E-4;
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		if (match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox) || !match->Deformation(i).IsIncluded())
			continue;
		
		// transformation. NOTE: currently only translations.
		const double * xform = match->Deformation(i).GetTransform(xformdims, xformflags);
		assert(xformflags == PartTransforms::PARTXFORM_Translation);
		for (int d=0; d<3; d++)	// initialize to previous values
			testTransform[d] = transforms[i][d];
		
		for (int d=0; d<3; d++)
		{
			// test next and previous value
			testTransform[d] = transforms[i][d]+currStep;
			match->SetTransform(i, testTransform);
			match->UpdateEllipsoid(i);
			double eNext = energy->Energy();
			
			testTransform[d] = transforms[i][d]-currStep;
			match->SetTransform(i, testTransform);
			match->UpdateEllipsoid(i);
			double ePrev = energy->Energy();
			
			//std::cout<<"Energy: "<<ePrev<<" -> "<<eNext<<std::endl;
			
			double change = 0;
			if (ePrev<(e0-epsilon) && ePrev<=eNext)
				change = -currStep;
			else if (eNext<(e0-epsilon) && eNext<ePrev)
				change = currStep;
			changes += (int)(change!=0);
			
			// revert or set new optimal
			if (params.m_bRevert)	// change state to optimal
				testTransform[d] = transforms[i][d];
			else		// change state to previous
				testTransform[d] = transforms[i][d] + change;
			
			match->SetTransform(i, testTransform);
			match->UpdateEllipsoid(i);
			
			transforms[i][d] += change;
			if (!params.m_bRevert)
			{
				double enew = energy->Energy();
				VKString newStr = energy->EnergyStr();
				
				//				if (!(enew <= e0 && vkAbs(enew - vkMin(vkMin(ePrev, eNext), e0))<epsilon))
				//				{
				//					energy->UpdateNodeShape(i);
				//					energy->UpdateGlobalXForm();
				//
				//					VKString newNewStr = energy->EnergyStr();
				//					VKString newNewStr2 = energy->EnergyStr();
				//					std::cout<<"enew="<<enew<<"==min("<<ePrev<<", "<<eNext<<") "<<" vs "<<e0<<std::endl;
				//					std::cout<<"Change="<<change<<std::endl;
				//					std::cout<<"Bool1: "<<(ePrev<(e0-epsilon) && ePrev<=eNext)<<std::endl;
				//					std::cout<<"Bool2: "<<(eNext<(e0-epsilon) && eNext<ePrev)<<std::endl;
				//					std::cout<<"e0Str: "<<e0Str.c_str()<<std::endl;
				//					std::cout<<"newStr: "<<newStr.c_str()<<std::endl;
				//					std::cout<<"newNewStr: "<<newNewStr.c_str()<<std::endl;
				//					std::cout<<"newNewStr2: "<<newNewStr2.c_str()<<std::endl;
				//					assert(false);
				//				}
				e0 = enew;
				e0Str = energy->EnergyStr();
			}
		}
	}
	//std::cout<<"Transforms: "<<changes<<std::endl;
	
	if (changes>0)
	{
		if (!params.m_bRevert)
		{
			for (int i=0; i<match->NTemplateNodes(); i++)
				match->SetTransform(i, transforms[i]);
			match->UpdateEllipsoids();
		}
		
		if (params.m_bRevert && energy->Energy() >= e0) // energy got worse: probably oscillating.
		{
			changes = 0;
			for (int i=0; i<(int)transforms.size(); i++)
				for (int d=0; d<3; d++)
					transforms[i][d] = revertTranslations[i][d];
			for (int i=0; i<match->NTemplateNodes(); i++)
				match->SetTransform(i, transforms[i]);
			match->UpdateEllipsoids();
		}
		else if (energy->Energy() > e0)
		{
			std::cout<<"\n[WARNING] Energy got worse"<<energy->Energy()<<" vs "<<e0<<std::endl;
			//assert(false);//??
		}
	}
	
	for (int i=0; i<revertTranslations.size(); i++)
		delete [] revertTranslations[i];
	
	return changes;
}

int TMatchAlg::GreedyOptimizeGlobalXForm(int & iter, TMatchEnergy * energy, TMatchAlgParams params)
{
	TMatch * match = energy->GetMatch();
	
	double currStep;
	double minStep;
	double changeRatio = .5; // vk: const
	
	int totalChanges = 0;
	while (iter < params.m_iNGreedyIterations)
	{
		currStep = 0.05;
		minStep = 0.01; // vk: const
		int totalXFormChanges = 0;
		
		// OPTIMIZE TRANSLATION
		if (params.m_iOptimFlags & TMatchAlgParams::OPT_GlobalTrans)
			while (iter < params.m_iNGreedyIterations)
			{
				int changes = 0;
				double e0 = energy->Energy();
				
				R3Vector translation = energy->GetMatch()->GetTranslation();
				R3Vector testTranslation = translation;
				for (int d=0; d<3; d++)
				{
					//test previous and next value
					testTranslation[d] = translation[d]+currStep;
					match->SetGlobalTranslate(testTranslation);
					double eNext = energy->Energy();
					
					testTranslation[d] = translation[d]-currStep;
					match->SetGlobalTranslate(testTranslation);
					double ePrev = energy->Energy();
					
					// revert back the state
					testTranslation[d] = translation[d];
					match->SetGlobalTranslate(testTranslation);
					match->UpdateEllipsoids();
					
					double change = 0;
					if (ePrev<e0 && ePrev<eNext)
						change = -currStep;
					else if (eNext<e0 && eNext<ePrev)
						change = currStep;
					changes += (int)(change!=0);
					
					translation[d] += change;
				}
				
				match->SetGlobalTranslate(translation);
				match->UpdateEllipsoids();
				if (energy->Energy() > e0)	// energy got worse: probably oscillating
				{
					match->SetGlobalTranslate(testTranslation);
					match->UpdateEllipsoids();
					changes = 0;
				}
				totalXFormChanges+=changes;
				if (changes==0) currStep *= changeRatio;
				if (currStep<=minStep) break;
				if (changes>0) iter++;
			}
		
		
		// OPTIMIZE SCALE
		currStep = 0.05;
		minStep = 0.01; // vk: const
		
		if (params.m_iOptimFlags & TMatchAlgParams::OPT_GlobalScale)
			while (iter < params.m_iNGreedyIterations)
			{
				int changes = 0;
				double e0 = energy->Energy();
				
				double scale = energy->GetMatch()->GetScale()[0];
				double testScale = scale;
				
				testScale = scale + currStep;
				match->SetGlobalScale(R3Vector(testScale, testScale, testScale));
				double eNext = energy->Energy();
				
				testScale = scale - currStep;
				match->SetGlobalScale(R3Vector(testScale, testScale, testScale));
				double ePrev = energy->Energy();
				
				double change = 0;
				if (ePrev<e0 && ePrev<eNext)
					change = -currStep;
				else if (eNext<e0 && eNext<ePrev)
					change = currStep;
				changes += (int)(change!=0);
				scale += change;
				
				match->SetGlobalScale(R3Vector(scale, scale, scale));
				match->UpdateEllipsoids();
				
				if (energy->Energy() > e0)	// energy got worse: probably oscillating
				{
					match->SetGlobalScale(R3Vector(testScale, testScale, testScale));
					match->UpdateEllipsoids();
					changes = 0;
				}
				
				totalXFormChanges+=changes;
				if (changes==0) currStep *= changeRatio;
				if (currStep<=minStep) break;
				if (changes>0) iter++;
			}
		
		// OPTIMIZE ROTATION
		currStep = 3.14159265359 / 10;
		minStep = currStep / 5; // vk: const
		
		if (params.m_iOptimFlags & TMatchAlgParams::OPT_GlobalRotate)
			while (iter < params.m_iNGreedyIterations)
			{
				int changes = 0;
				double e0 = energy->Energy();
				
				R3Vector rotation = energy->GetMatch()->GetRotation();
				R3Vector testRotation = rotation;
				for (int d=0; d<3; d++)
				{
					if (d!=2)
						continue;
					//test previous and next value
					testRotation[d] = rotation[d]+currStep;
					match->SetGlobalRotate(testRotation);
					double eNext = energy->Energy();
					
					testRotation[d] = rotation[d]-currStep;
					match->SetGlobalRotate(testRotation);
					double ePrev = energy->Energy();
					
					// revert back the state
					testRotation[d] = rotation[d];
					match->SetGlobalRotate(testRotation);
					match->UpdateEllipsoids();
					
					double change = 0;
					if (ePrev<e0 && ePrev<eNext)
						change = -currStep;
					else if (eNext<e0 && eNext<ePrev)
						change = currStep;
					changes += (int)(change!=0);
					
					rotation[d] += change;
				}
				
				match->SetGlobalRotate(rotation);
				match->UpdateEllipsoids();
				if (energy->Energy() > e0)	// energy got worse: probably oscillating
				{
					match->SetGlobalRotate(testRotation);
					match->UpdateEllipsoids();
					changes = 0;
				}
				totalXFormChanges+=changes;
				if (changes==0) currStep *= changeRatio;
				if (currStep<=minStep) break;				
				if (changes>0) iter++;
			}
		
		if (totalXFormChanges==0)
			break;
		
		totalChanges += totalXFormChanges;
	}
	
	return totalChanges;
}


