/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __TMatchEnergy_H
#define __TMatchEnergy_H

#include "Sortable.h"

#include "VKGraph.h"

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapeDescriptor.h"
#include "ShapeLocalPCA.h"

#include "ProbDistribution.h"
#include "PartTransforms.h"


#include "TMatch.h"

struct GlobalConstants
{
	static int s_iMinNSamplesForVariance;
	static int s_iMinNSamplesForMean;
};

/** 
 * Helper class for ShapeTemplate
 *
 * Edge that describes scene graph hierarchy. 
 */
class TMatchEnergy
{
	public:
		enum EnergyTerms
		{
			// global flags
			ET_Pnts				= 1<<1,		// penalize pnts -> template 
			ET_NodeFit2Pnts		= 1<<2,		// penalize  template -> pnts
			ET_PntEdges			= 1<<3,		// penalize point -> point smoothness (different labels)
			ET_NodeExists		= 1<<4,		// penalize absence of (high probability) some nodes
			ET_NodeXForm		= 1<<5,		// compute node transformation energy
			ET_NodeShape		= 1<<6,		// compute node scaling energy
			ET_SameWeightParts	= 1<<7,		// all parts have same weight
			
			// per point flags
			ET_PntsDistance		= 1<<10,	// penalize distances
			ET_PntsOrientation	= 1<<11,	// penalize similarity in local orientations of nearest points
			ET_PntsLocalClass	= 1<<12,	// penalize point -> node local orientation similarity (joint boost)
			ET_DistsTmpltSpace  = 1<<13,	// compute all point energy in template space
			ET_DistsModelSpace  = 1<<14,	// compute all point energy in model space
		};

	//	static int GetNhdSizeConstant();
		static void Compatible(TMatch * m1, TMatch * m2, bool & sameParts, bool & sameIncluded);
		static double DeformationDistance(TMatch * m1, TMatch * m2, bool &sameParts, bool &sameIncluded);
		static double LocalDescriptorDistance(TMatch * m1, TMatch * m2);
		static double PntsDistances(TMatch * m1, TMatch * m2);
		static double DeformationDistanceThreshold();

	
		TMatchEnergy(TMatch * match);
		virtual ~TMatchEnergy();
	
		static int DefaultMatchQualityEnergy() ;

		// Total energy of the system, todo	
		virtual double Energy() const;
		virtual double Energy(int terms) const;
		virtual double NodeEnergy(int nodeID) const;
		virtual void SetTerms(int terms);
		virtual int GetTerms() const;
		
		virtual double MissingPartPenalty() const;
	
		virtual VKString EnergyStr() const;
	
		// elements for outside energy computation (e.g. MRF)
		virtual const VKGraph * GetEdgeEnergy() const;
		virtual const double * GetPntsEnergy() const;
		virtual const double * GetLabelEnergy() const;
		virtual void SetEdgeEnergy(const VKGraph * graph);
	
		// changes to match that affect energy
		TMatch * GetMatch();
		virtual void ClearCache();
		virtual void UpdateLabeling();
		virtual void UpdateNodeShape(int nodeID);
		virtual void UpdateNodeXForm(int nodeID);
		virtual void UpdateGlobalXForm();
	
		// adjust energy constants (temporarily?)
		virtual double GetNullPenalty() const;
		virtual void SetNullPenalty(double penalty);

		// Segmentation energy
		//		pnt energy: depends on segmentation and node positions
		virtual double ComputeEGlobalPnt(int globalPntID, int localNodeID) const;
		virtual double LogProbLocalShape(int globalPntID, int localNodeID) const;
	
		// write energy + matching (binary)
		virtual void WriteEnergyAndMatch(FILE * file) const;
		static TMatchEnergy * ReadEnergyAndMatch(FILE * file, const STemplate * st, ModelDatabase * db,
												 double *energy = NULL);

//		// compute pnt energy
//		virtual double ComputeEPnt(const R3Point & p1pos, const R3Triad & p1orient, const double * p1eigs,
//								   const R3Point & p2pos, const R3Triad & p2orient, const double * p2eigs) const;

		virtual double GetNodeShapeWeight() const;
		virtual double GetNodeXFormWeight() const;
	
		virtual double GetModelPntsDistWeight(int nodeID) const;
		virtual double GetTmpltPntsDistWeight(int nodeID) const;
	
	//protected:
		
		//		pnts to current label
		virtual double EPnts(int nodeID) const;
	
		//		pnt similarity energy: depends on node geometry only
		virtual double EEdges() const;	
		//		energy for node existence (penalty). Might penalize if prominent nodes do not exist.
		virtual double ENodeExists(int nodeID) const;
		
		// Deformation energy (except node existence)
		//		transformation penalty (node position / orientation)
		virtual double ENodeXForm(int nodeID) const;
		//		shape penalty (anistropic scale, etc...)
		virtual double ENodeShape(int nodeID) const;
		//		how well points explain node (different from prob. of shape)
		virtual double ENodeFit2Pnts(int nodeID) const;
		
	// parameters: labels, template configuration
		TMatch * m_pMatch;
				
	// some cached data for the energy
		VKGraph * m_pNNGraph;
		double * m_aPerPointEnergy;
		double * m_aPerLabelEnergy;
		bool * m_abRecomputeLocalShapeProb;
		double * m_afPerPointLocShapeLogProb;

	// cached energy terms
		bool GetCachedGlobalValue(EnergyTerms term, double & value) const;
		void SetCacheGlobalValue(EnergyTerms term, double value) const;
		bool GetCachedValue(EnergyTerms term, int nodeID, double & value) const;
		void SetCacheValue(EnergyTerms term, int nodeID, double value) const;
	
		int m_CachedGlobalEnergyFlags;
		std::map<EnergyTerms, double> m_CachedGlobalEnergyTerms;
		std::vector<int> m_CachePerNodeFlags;
		std::vector<std::map<EnergyTerms, double> > m_CachedPerNodeTerms;
	
	// classify points using local shape descriptors
//		virtual double LogProbLocalShape(int gpid, int nodeID) const;
	
	// some constants
		int m_EnergyFlags;	// non-zery energy terms
		//double s_fEllScaleMinVarScale;
		//double s_fEllScaleMinVarOffset;
		//double s_fSubpartXFormMinVar;
		double s_fPnt2NullPenalty;
		double s_fPntInEllipsoidVar;
		double s_fEdgeEnergyWeight;
		//double s_fEnoughSamples;
	
		double s_fPointOrientWeight;	// point orientation weight

//		double s_fPntLocNhdDist;
//		int s_iPntLocNhdSize;	// might want to have one for 'edges' one for PCA estimation
	
		double m_fGlobalEnergyScale;
	
		double s_fXFormWeight;
		double s_fScaleWeight;
	
		double s_fPointEnergyWeight;
	
		double s_ZeroNodeProbPenalty;
	
	// debug: 
		static bool m_bPrintDebug;
	
	// some helper functions
//		void GetEVectorWeights(const double e[3], double w[3], std::vector<int> * outSort) const;
//		double OrientationSimilarity(const R3Triad & a1, const double e1[3],
//									 const R3Triad & a2, const double e2[3]) const;
		
		// returns global point ID with a given label
		virtual int FindNearestPoint(int nodeID, int pntID, const R3Point & tSpacePos) const;
		
		static std::map<const ShapeMesh *, VKGraph * > m_Mesh2Graph;	// to avoid re-computing
};

#endif

