/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __VisModel3D_H
#define __VisModel3D_H

#include "OpenGLHeaders.h"
#include "Model3D.h"
#include "GeomShape.h"
#include "ShapePointCloud.h"
#include "ShapeManifoldMesh.h"
#include "ShapeLabels.h"

#include "VisWindow.h"

class DatabaseVisWindow;
class ShapeMaterials;

class GeometryTextures
{
	public:
		GLuint * m_mMaterialImageToTextureID;
		int m_iNTextures;
		int * w;
		int * h;
};

class VisModel3D
{
	public:
		// ad-hoc way of interpreting optional values
		static const int LABELS_OPTVAL_IS_NODEDEPTH = 0;
		static const int LABELS_OPTVAL_IS_NODEID	= 1000000;	
		
		enum DrawFlags
		{
			DRAW_NONE		= 0,
			DRAW_FACES		= 1,
			DRAW_EDGES		= 2,
			DRAW_VERTICES	= 4,
			DRAW_IDs		= 8,
			DRAW_PDOTS		= 16,
			DRAW_PSPHERES	= 32,
			DRAW_NORMALS	= 64,
			DRAW_MATERIALS  = 128,

			DRAW_MESHES		= 1024,
			DRAW_POINTS		= 2048,
			DRAW_FLAT		= 4096,
		};
	
		enum IntersectBufferType
		{
			IBUF_Pnts = 1,
			IBUF_Meshes = 2
		};

	// static draw functions
		static void DrawPrepareCamera(VisWindow * window, const R3Affine & camera);	
		static void DrawPrepareBasicFaces(VisWindow * window, const R3Affine & camera);
		static void DrawMeshWithSampledValues(const ShapeMesh * mesh, 
											  VisWindow * window, const R3Affine & camera, 
											  ShapePointCloud * pnts, double * values, 
											  double minValue = 0., double maxValue = 1.);
		static void DrawGeomWithLabels(const GeomShape * mesh, const ShapeLabels * labels, 
									   VisWindow * window, const R3Affine & camera, int depth = 1);
		static void DrawMeshWithLabeledGeometry(const ShapeMesh * mesh, ShapeLabels * labels, 
												VisWindow * window, const R3Affine & camera);
		static void DrawMeshFaces(const ShapeMesh * mesh, bool drawMaterials,
								  VisWindow * window, const R3Affine & camera,
								  GLuint * textures = NULL, int nTextures = 0, bool colorMesh = false, 
								  double r = 0, double g = 0, double b = 0, double a = 1.);
	
		static void DrawMeshEdges(const ShapeMesh * mesh, 
								  VisWindow * window, const R3Affine & camera, bool colorMesh = false, 
								  double r = 0, double g = 0, double b = 0, double a = 1.);
	
		static void DrawThumbnail(GLuint textureID, double cx, double cy, double dx, double dy);
	

		static void InitMaterial(ShapeMaterials * materials, int matID,
								 VisWindow * window, int * textureID,
								 GLuint * textures,
								 bool colorMesh = false, 
								 double r = 0, double g = 0, double b = 0, double a = 1.);
	
		static void DrawPntsWithScalar(const ShapePointCloud * pnts, 
									   bool drawIDs, bool drawAsDots, bool drawAsSphers,
									   bool drawNormals,
									   VisWindow * window, 
									   const R3Affine & camera,
									   double r, double rScale,
									   double * perPointVals = NULL,
									   double minValue = 0, double maxValue = 1.);

		static void DrawGeomsToIntersectionBuffer(const std::vector<GeomShape *> & geoms, 
												  VisWindow * window, double r, double rScale,
												  const R3Affine & camera, 
												  const BufferID & bufferID);
	
	// constructor / destructor
		VisModel3D(Model3D * model, int drawFlags);
		virtual ~VisModel3D();
	
	// free as much memory as possible (object will not be revisited for a while).
		virtual void FreeMemory();
		
	// draw
		virtual void Draw(DatabaseVisWindow * window, int cameraID);
		virtual void DrawGeomVals(GeomShape * geom, const VKString & geomValsName, int optVal,
								  DatabaseVisWindow * window, const R3Affine & camera);
		virtual void DrawGeomImg(ShapeImg * imgGeom, GeometryTextures * textures, 
								 DatabaseVisWindow * window, int cameraID);
		virtual void DrawIsectBuffers(DatabaseVisWindow * window, int cameraID,
									  const std::vector<GeomShape*> & geoms, 
									  const std::vector<int> & geomIDs = std::vector<int>(),
									  const R3Affine * camera = NULL);
		virtual GeometryTextures * GetTextures(const VKString & geomName);
		static GeometryTextures * CreateTextures(GeomShape * geom);

	// what to render: flags, geometries list
		virtual bool IsFlagOn(DrawFlags flag);
		virtual void ToggleFlag(DrawFlags drawFlag);
		virtual void SetFlag(DrawFlags drawFlag, bool enable);
		virtual void SetIntersectionBuffer(IntersectBufferType bufferTp, bool enable = true);
		
		virtual void AddRenderableGeomValue(const VKString & geomValueName, int optVal = -1);
		virtual void RemoveRenderableGeomValue(const VKString & geometryName);
	
		virtual void AddRenderableGeometry(const VKString & geometryName);
		virtual void ClearRenderablgeGeometries();
		virtual void RemoveRenderableGeometry(const VKString & geometryName);
		virtual int NumRenderableGeometries();
		virtual GeomShape * GetRenderableGeometry(int geomID);
		virtual Model3D * GetModel();
	
	// cameras
		virtual int SubwindowCameraID(int subwindowID, DatabaseVisWindow * window);
		virtual int CreateNewCamera();
		
		virtual double GetMeshRadius() const;
		virtual double GetRelativePntRadius() const;
		virtual R3Affine GetCameraTransform(const R3Affine & insertObjectSpaceXForm, 
											int cameraID, double * scale = NULL) const;
		virtual R3Affine GetCameraTransform(int cameraID, double * scale = NULL) const;
		
		virtual void WindowResize(int newW, int newH);
		virtual void Rotate(int cameraID, int x, int y, int dx, int dy, VisWindow * window);
		virtual void Translate(int cameraID, int x, int y, int dx, int dy, VisWindow * window);
		virtual void Scale(int cameraID, int x, int y, int dx, int dy, VisWindow * window);
		virtual void Scale(int cameraID, double s, VisWindow * window);
		virtual void UpdateCamera(const R3Vector & upVec, const R3Vector & towardsVec,
								  double relativeScale = 1.);
	
	// interactive selection
		virtual void SelectPoint(const ShapePoint & pnt);
		virtual void SelectPoint(int cameraID, int x, int y, VisWindow * window);
//		virtual const ShapePoint & SelectPointFromPnts(int cameraID, int x, int y, VisWindow * window);
//		virtual const ShapePoint & SelectPointOnMesh(int cameraID, int x, int y, VisWindow * window);
		virtual const ShapePoint & SelectedPoint() const; 
		virtual void UnselectPoint();
		
	// write camera
		virtual void WriteCamera(FILE * file) const;
		virtual void ReadCamera(FILE * file);
	
	protected:
	
		std::map<VKString, GeometryTextures*> m_GeometryTextures;
	
		int m_iEnabledIntersectionBuffers;
		//bool m_bRenderIntersectionBuffers;
		
		virtual ShapePoint QuickIntersection(int cameraID, int x, int y, VisWindow * window);
		std::map<int, int> m_CameraIDToIntBufferID;
		std::vector<BufferID> m_IntersectionBuffers;
		std::vector<std::vector<int> > m_IntersectionGeomIDs;
	
		VKStringList m_RenderableGeometries;
		VKStringList m_RenderableGeomValues;
		std::vector<int> m_RenderableGeomValsOptional;
		std::vector<R3CoordSystem> m_CamerasCoord;
		std::vector<double> m_CamerasScale;
		std::vector<R3Affine> m_CameraPostMap;
	
		R3Point m_Origin;
		double m_fRadius;
		R3Vector m_Towards; 
		R3Vector m_Up;
		double m_fRelativeScale;
		std::map<int, int> m_mSubwindowCamera;
		int m_iDrawFlags;
		Model3D * m_pModel;
	
		double m_fPointRelativeRadius;
	
		ShapePoint m_SelectedPoint;
		
	// delete me data
//		int m_RayX, m_RayY;
};

#endif

