// Source file for GAPS scalar grid class



////////////////////////////////////////////////////////////////////////
// This class stores a regular grid on a sphere. 
// The X dimension represents azimuth (phi)
// measured from the X axis in the counter-clockwise direction.
// The Y dimension of the grid represents downward altitude (theta)
// measured from the positive y axis (the north pole). 
// So, (i,j) on the spherical grid represents the
// point on the sphere with spherical coordinates:
// (2Pi*j/r, Pi*(2i+1)/(2r))
// where r is the resolution of the sampling.
////////////////////////////////////////////////////////////////////////



// Include files

#include "R3Harmonics/R3Harmonics.h"



R2SphericalGrid::
R2SphericalGrid(int phi_resolution, int theta_resolution)
  : grid(phi_resolution, theta_resolution)
{
}



R2SphericalGrid::
R2SphericalGrid(const R2SphericalGrid& g)
  : grid(g.grid)
{
}



R3Point R2SphericalGrid::
WorldPosition(RNCoord x, RNCoord y) const
{
  RNAngle phi = Phi(x);
  RNAngle theta = Theta(y);
  RNScalar sin_theta = sin(theta);
  RNCoord wx = sin_theta * cos(phi);
  RNCoord wy = cos(theta);
  RNCoord wz = sin_theta * sin(phi);
  return R3Point(wx, wy, wz);
}



R2Point R2SphericalGrid::
GridPosition(RNCoord x, RNCoord y, RNCoord z) const
{
  if (RNIsZero(x) && RNIsZero(z)) {
    if (y < 0) return R2Point(ThetaResolution()-1, 0);
    else return R2Point(0,0);
  }
  else {
    RNLength length = sqrt(x*x + y*y + z*z);
    x /= length; y /= length; z /= length;
    RNAngle phi = atan2(z,x);
    if (phi < 0) phi += RN_TWO_PI;
    RNAngle theta = acos(y);
    RNCoord gx = X(phi);
    RNCoord gy = Y(theta);
    return R2Point(gx, gy);
  }
}



RNScalar R2SphericalGrid::
Integral(RNScalar (*f)(RNScalar)) const
{
  // Return integral of f(grid.value), weighting terms by surface area
  // theta_scale has fraction of surface area in slice
  // phi_scale has fraction of surface area in sector
  // mean is weighted average of grid values
  RNScalar mean = 0.0;
  RNScalar py = 1.0;
  RNScalar phi_scale = 1.0 / grid.XResolution();
  for (int j = 0; j < grid.YResolution(); j++) {
    RNScalar y = cos(Theta(j));
    RNScalar dy = py - y; py = y;
    RNScalar theta_scale = dy / (2 * grid.YResolution());
    RNScalar scale = phi_scale * theta_scale;
    for (int i = 0; i < grid.XResolution(); i++) {
      mean += scale * f(GridValue(i, j));
    }
  }

  // Return sum, area of unit sphere is 4*PI 
  return 4 * RN_PI * mean;
}



RNScalar R2SphericalGrid::
Integral(const R2SphericalGrid& g, RNScalar (*f)(RNScalar, RNScalar)) const
{
  // Return integral of f(grid.value, g.grid.value), weighting terms by surface area
  // theta_scale has fraction of surface area in slice
  // phi_scale has fraction of surface area in sector
  // mean is weighted average of grid values
  RNScalar mean = 0.0;
  RNScalar py = 1.0;
  RNScalar phi_scale = 1.0 / grid.XResolution();
  for (int j = 0; j < grid.YResolution(); j++) {
    RNScalar y = cos(Theta(j));
    RNScalar dy = py - y; py = y;
    RNScalar theta_scale = dy / (2 * grid.YResolution());
    RNScalar scale = phi_scale * theta_scale;
    for (int i = 0; i < grid.XResolution(); i++) {
      mean += scale * f(GridValue(i, j), g.GridValue(i, j));
    }
  }

  // Return sum, area of unit sphere is 4*PI 
  return 4 * RN_PI * mean;
}



RNScalar R2SphericalGrid::
GridValue(RNScalar x, RNScalar y) const
{
  // Adjust coordinates (this is different from R2Grid)
  while (x < 0) x += PhiResolution();
  while (x >= PhiResolution()) x -= PhiResolution();
  if (y < 0) y = 0;
  if (y > ThetaResolution()-1) y = ThetaResolution()-1; 

  // Trilinear interpolation
  int ix1 = (int) x;
  int iy1 = (int) y;
  int ix2 = ix1 + 1;
  int iy2 = iy1 + 1;
  if (ix2 >= PhiResolution()) ix2 = 0;  // This is different from R3Grid
  if (iy2 >= ThetaResolution()) iy2 = iy1;
  float dx = x - ix1;
  float dy = y - iy1;
  float value = 0.0;
  value += GridValue(ix1, iy1) * (1.0-dx) * (1.0-dy);
  value += GridValue(ix1, iy2) * (1.0-dx) * dy;
  value += GridValue(ix2, iy1) * dx * (1.0-dy);
  value += GridValue(ix2, iy2) * dx * dy;
  return value;
}



RNScalar R2SphericalGrid::
L1NormFunction(RNScalar value)
{
  // Return absolute value
  return fabs(value);
}



RNScalar R2SphericalGrid::
L2NormFunction(RNScalar value) 
{
  // Return square of value
  return value * value;
}



RNScalar R2SphericalGrid::
AreaFunction(RNScalar value) 
{
  // Return area of grid cell
  return (value != 0.0) ? 1 : 0;
}



RNScalar R2SphericalGrid::
DotFunction(RNScalar value1, RNScalar value2)
{
  // Return product of values
  return value1 * value2;
}



RNScalar R2SphericalGrid::
L1DistanceFunction(RNScalar value1, RNScalar value2) 
{
  // Return absolute value of difference between values
  return fabs(value1 - value2);
}



RNScalar R2SphericalGrid::
L2DistanceFunction(RNScalar value1, RNScalar value2) 
{
  // Return square of difference between values
  RNScalar value = value1 - value2;
  return value * value;
}



int R2SphericalGrid::
ForwardSphericalHarmonicTransform(R2SphericalHarmonics& harmonics) const
{
  // Check resolutions
  int size = ThetaResolution();
  assert(size == PhiResolution());
  assert(size == 2*harmonics.Bandwidth());

  // Perform transform
  return R2ForwardSphericalHarmonicTransform((RNScalar *) GridValues(), (RNScalar *) harmonics.Coefficients(), size); 
}



