// Header file for spherical harmonics class



// Class definition

class R2SphericalHarmonics {
public:
  // Constructors
  R2SphericalHarmonics(int bandwidth = 0);
  R2SphericalHarmonics(const R2SphericalHarmonics& harmonics);
  ~R2SphericalHarmonics(void);

  // Harmonics property functions
  int Bandwidth(void) const;
  int NCoefficients(void) const;
  RNScalar L1Norm(void) const;
  RNScalar L1Norm(int L) const;
  RNScalar L1Norm(int L, int M) const;
  RNScalar L2Norm(void) const;
  RNScalar L2Norm(int L) const;
  RNScalar L2Norm(int L, int M) const;
  RNScalar L2NormSquared(void) const;
  RNScalar L2NormSquared(int L) const;
  RNScalar L2NormSquared(int L, int M) const;

  // Coefficient access functions
  const RNScalar *Coefficients(void) const;
  const RNScalar *Coefficients(int L, int M) const;
  RNScalar Coefficient(int L, int M, int k) const;
  RNScalar RealCoefficient(int L, int M) const;
  RNScalar ImaginaryCoefficient(int L, int M) const;

  // Harmonics manipulation functions
  void Clear(RNScalar value = 0);
  void SetRealCoefficient(int L, int M, RNScalar value);
  void SetImaginaryCoefficient(int L, int M, RNScalar value);

  // Relationship functions
  RNScalar Dot(const R2SphericalHarmonics& harmonics) const;
  RNScalar L1Distance(const R2SphericalHarmonics& harmonics) const;
  RNScalar L2Distance(const R2SphericalHarmonics& harmonics) const;
  RNScalar L2DistanceSquared(const R2SphericalHarmonics& harmonics) const;

  // I/O functions
  int ReadFile(const char *filename);
  int WriteFile(const char *filename) const;
  int Read(FILE *fp = NULL);
  int Write(FILE *fp = NULL) const;
  int Print(FILE *fp = NULL) const;

  // Spherical harmonics transform
  int InverseSphericalHarmonicTransform(R2SphericalGrid& grid) const;

  // Utility functions
  int Index(int L, int M) const;

private:
  RNScalar *coefficients;
  int ncoefficients;
  int bandwidth;
};



// Transform function declarations

extern int R2ForwardSphericalHarmonicTransform(const RNScalar *grid, RNScalar *harmonics, int size);
extern int R2InverseSphericalHarmonicTransform(const RNScalar *harmonics, RNScalar *grid, int size);



// Inline functions

inline int R2SphericalHarmonics::
Bandwidth(void) const
{
  // Return bandwidth
  return bandwidth;
}



inline int R2SphericalHarmonics::
NCoefficients(void) const
{
  // Return ncoefficients
  return ncoefficients;
}



inline RNScalar R2SphericalHarmonics::
L2Norm(void) const
{
  // Return L2 norm
  return sqrt(L2NormSquared());
}



inline RNScalar R2SphericalHarmonics::
L2Norm(int L) const
{
  // Return L2 norm
  return sqrt(L2NormSquared(L));
}



inline RNScalar R2SphericalHarmonics::
L2Norm(int L, int M) const
{
  // Return L2 norm
  return sqrt(L2NormSquared(L, M));
}



inline const RNScalar *R2SphericalHarmonics::
Coefficients(void) const
{
  // Return coefficients
  return coefficients;
}



inline const RNScalar *R2SphericalHarmonics::
Coefficients(int L, int M) const
{
  // Return pointer to pair of coefficients
  return &coefficients[Index(L, M)];
}




inline RNScalar R2SphericalHarmonics::
Coefficient(int L, int M, int k) const
{
  // Return coefficient (0=real, 1=imaginary)
  assert((k >= 0) && (k < 2));
  return coefficients[Index(L, M)+k];
}




inline RNScalar R2SphericalHarmonics::
RealCoefficient(int L, int M) const
{
  // Return real coefficient
  return coefficients[Index(L, M)];
}




inline RNScalar R2SphericalHarmonics::
ImaginaryCoefficient(int L, int M) const
{
  // Return imaginary coefficient
  return coefficients[Index(L, M)+1];
}



inline RNScalar R2SphericalHarmonics::
L2Distance(const R2SphericalHarmonics& harmonics) const
{
  // Return the sqrt of the sum of squared distances between coefficients
  return sqrt(L2DistanceSquared(harmonics));
}



