// Header file for spherical harmonics class



// Class definition

class R3SphericalHarmonics {
public:
  // Constructors
  R3SphericalHarmonics(int bandwidth = 0, int nshells = 0);
  R3SphericalHarmonics(const R3SphericalHarmonics& harmonics);
  ~R3SphericalHarmonics(void);

  // Harmonics property functions
  int Bandwidth(void) const;
  int NCoefficients(void) const;
  int RadiusResolution(void) const;
  RNScalar Weight(int S) const;
  RNScalar L2Norm(void) const;
  RNScalar L2Norm(int S) const;
  RNScalar L2Norm(int S, int L) const;
  RNScalar L2Norm(int S, int L, int M) const;
  RNScalar L2NormSquared(void) const;
  RNScalar L2NormSquared(int S) const;
  RNScalar L2NormSquared(int S, int L) const;
  RNScalar L2NormSquared(int S, int L, int M) const;
  const R2SphericalHarmonics& Shell(int shell) const;

  // Harmonics manipulation functions
  void Clear(RNScalar value = 0);

  // Relationship functions
  RNScalar Dot(const R3SphericalHarmonics& harmonics) const;
  RNScalar L1Distance(const R3SphericalHarmonics& harmonics) const;
  RNScalar L2Distance(const R3SphericalHarmonics& harmonics) const;
  RNScalar L2DistanceSquared(const R3SphericalHarmonics& harmonics) const;

  // I/O functions
  int ReadFile(const char *filename);
  int WriteFile(const char *filename) const;
  int Read(FILE *fp = NULL);
  int Write(FILE *fp = NULL) const;
  int Print(FILE *fp = NULL) const;

  // Spherical harmonics transform
  friend int ForwardSphericalHarmonicTransform(const R3SphericalGrid& grid, R3SphericalHarmonics& harmonics);
  friend int InverseSphericalHarmonicTransform(const R3SphericalHarmonics& harmonics, R3SphericalGrid& grid);

private:
  RNScalar *weights;
  R2SphericalHarmonics **shells;
  int nshells;
};



// Transform function declarations

extern int ForwardSphericalHarmonicTransform(const R3SphericalGrid& grid, R3SphericalHarmonics& harmonics);
extern int InverseSphericalHarmonicTransform(const R3SphericalHarmonics& harmonics, R3SphericalGrid& grid);



// Inline functions

inline int R3SphericalHarmonics::
Bandwidth(void) const
{
  // Return bandwidth
  if (nshells == 0) return 0;
  else if (!shells[0]) return 0;
  else return shells[0]->Bandwidth();
}



inline int R3SphericalHarmonics::
NCoefficients(void) const
{
  // Return number of spherical shells
  if (nshells == 0) return 0;
  else if (!shells[0]) return 0;
  else return nshells * shells[0]->NCoefficients();
}



inline int R3SphericalHarmonics::
RadiusResolution(void) const
{
  // Return number of shells
  return nshells;
}



inline RNScalar R3SphericalHarmonics::
Weight(int S) const
{
  // Return weight for Sth shell
  assert((0 <= S) && (S < nshells));
  return (weights) ? weights[S] : 0;
}


#if 0
inline RNScalar R3SphericalHarmonics::
L2Norm(void) const
{
  // Return L2 norm
  return sqrt(L2NormSquared());
}
#endif


inline RNScalar R3SphericalHarmonics::
L2Norm(int S) const
{
  // Return L2 norm
  return sqrt(L2NormSquared(S));
}



inline RNScalar R3SphericalHarmonics::
L2Norm(int S, int L) const
{
  // Return L2 norm
  return sqrt(L2NormSquared(S, L));
}



inline RNScalar R3SphericalHarmonics::
L2Norm(int S, int L, int M) const
{
  // Return L2 norm
  return sqrt(L2NormSquared(S, L, M));
}



inline const R2SphericalHarmonics& R3SphericalHarmonics::
Shell(int S) const
{
  // Return spherical harmonics for Sth shell
  assert((0 <= S) && (S < nshells));
  R2SphericalHarmonics *shell = shells[S];
  assert(shell);
  return *shell;
}



inline RNScalar R3SphericalHarmonics::
L2Distance(const R3SphericalHarmonics& harmonics) const
{
  // Return the sqrt of the sum of squared distances between coefficients
  return sqrt(L2DistanceSquared(harmonics));
}



