/* Source file for the R3 group class */



/* Include files */

#include "R3Nodes/R3Nodes.h"



/* Class type definitions */

R3_NODE_CLASS_DEFINITIONS(R3GroupNode);



/* Public functions */

int R3InitGroupNode()
{
    /* Return success */
    return TRUE;
}



void R3StopGroupNode()
{
}



R3GroupNode::
R3GroupNode(const char *name, const R3Box& bbox, unsigned long flags)
    : R3InteriorNode(name),
      children(NULL),
      bbox(bbox),
      nfacets(0.0, 0.0),
      flags(flags),
      base_flags(flags)
{
}



R3GroupNode::
R3GroupNode(const R3GroupNode& node)
    : R3InteriorNode(NULL),
      children(NULL),
      bbox(R3null_box),
      nfacets(0.0, 0.0),
      flags(node.base_flags),
      base_flags(node.base_flags)
{
}



R3GroupNode::
~R3GroupNode(void)
{
    // Remove children
    while (NChildren() > 0) RemoveNode(ChildNode(0));
}



const R3Box& R3GroupNode::
BBox(void) const
{
    // Return bounding box
    return bbox;
}



const R3Shape& R3GroupNode::
BShape(void) const
{
    // Return bounding box
    return bbox;
}



const RNInterval R3GroupNode::
NFacets(void) const
{
    // Return number of facets
    return nfacets;
}



const RNFlags R3GroupNode::
Flags(void) const
{
    // Return flags
    return flags;
}



R3Node *R3GroupNode::
ChildNode(int k) const
{
    // Return child node
    return (children) ? children->Kth(k) : NULL;
}



const int R3GroupNode::
NChildren(void) const
{
    // Return number of children
    return (children) ? children->NEntries() : 0;
}



void R3GroupNode::
SetBBox(const R3Box& bbox)
{
    // Save bbox
    R3Box oldbox(this->bbox);

    // Update box 
    this->bbox = bbox;
    
    // Make sure box encloses all children
    R3Node *child;
    RNIterator iterator1;
    R3_FOR_EACH_CHILD_NODE(this, child, iterator1) {
        const R3Box& childbox = child->BBox();
        if (!childbox.IsEmpty()) {
	    const R3Transformation *transformation = Transformation();
	    if (transformation) {
	        R3Box xformed_childbox(childbox);
	        xformed_childbox.Transform(*transformation);
		this->bbox.Union(xformed_childbox);
	    }
	    else {
		this->bbox.Union(childbox);
	    }
	}
    }
    
    // Update parents' boxes
    R3InteriorNode *parent;
    RNIterator iterator2;
    R3_FOR_EACH_PARENT_NODE(this, parent, iterator2) 
        parent->UpdateBBox(this, oldbox, this->bbox);
}



void R3GroupNode::
SetFlags(RNFlags flags)
{
    // Set flags
    RNFlags saveflags(this->base_flags);
    this->base_flags = flags;
    UpdateFlags(this, saveflags, base_flags);
}



void R3GroupNode::
AddFlags(RNFlags flags)
{
    // Add flags
    this->base_flags.Add(flags);
    this->flags.Add(flags);
}



void R3GroupNode::
RemoveFlags(RNFlags flags)
{
    // Remove flags
    RNFlags saveflags(this->base_flags);
    this->base_flags.Remove(flags);
    UpdateFlags(this, saveflags, this->base_flags);
}



void R3GroupNode::
InsertNode(R3Node *node) 
{
    // Insert child
    if (!children) children = new RNVector<R3Node *>();
    children->Insert(node);

    // Add parent to child
    node->parents.Insert(this);

    // Update bounding box
    UpdateBBox(node, R3null_box, node->BBox());

    // Update number of facets
    UpdateNFacets(node, RNzero_interval, node->NFacets());

    // Update flags
    UpdateFlags(node, R3_NULL_NODE_FLAGS, node->Flags());

    // Sort children so that transparent nodes are at end, and non-materialed nodes are in middle
    if (NChildren() >= 2) {
	if (!node->Flags()[R3_MATERIAL_TRANSPARENCY_FLAG]) {
	    RNFlags nodeflags = node->Flags();
	    for (int i = NChildren()-1; i >= 0; i--) {
	        RNFlags prevflags = (i == 0) ? (RNFlags) R3_HAS_MATERIAL_NODE_FLAG : ChildNode(i-1)->Flags();
		if (prevflags[R3_MATERIAL_TRANSPARENCY_FLAG]) continue;
		if ((nodeflags[R3_HAS_MATERIAL_NODE_FLAG]) && (!prevflags[R3_HAS_MATERIAL_NODE_FLAG])) continue;
		if (i != NChildren()-1) children->Swap(i, NChildren()-1);
		break;
	    }
	}
    }
}



void R3GroupNode::
RemoveNode(R3Node *node) 
{
    // Update bounding box
    UpdateBBox(node, node->BBox(), R3null_box);

    // Update number of facets
    UpdateNFacets(node, node->NFacets(), RNzero_interval);

    // Update flags
    UpdateFlags(node, node->Flags(), R3_NULL_NODE_FLAGS);

    // Remove parent from child
    node->parents.Remove(this);

    // Remove child from parent
    assert(children);
    children->Remove(node);
    if (children->IsEmpty()) {
	delete children;
	children = NULL;
    }
}



void R3GroupNode::
UpdateBBox(R3Node *node, const R3Box& oldbox, const R3Box& newbox)
{
    // Just checking 
    assert(NChildren() > 0);

    // Get transformed newbox
    R3Box xformed_newbox(newbox);
    const R3Transformation *transformation = Transformation();
    if (transformation && (!newbox.IsEmpty())) 
        xformed_newbox.Transform(*transformation);

    // Check if have only one child
    if (NChildren() == 1) {
	// Update box
	R3Box savebox(bbox);
	bbox = xformed_newbox;

	// Update parents' boxes
	R3InteriorNode *parent;
	RNIterator iterator2;
	R3_FOR_EACH_PARENT_NODE(this, parent, iterator2) 
	    parent->UpdateBBox(this, savebox, bbox);
    }
    else {
        // Check if child box grew
        if (R3Contains(newbox, oldbox)) {
	    // Child box grew or stayed same - union into bbox 
	    if (!R3Contains(bbox, xformed_newbox)) {
	        R3Box savebox(bbox);
		bbox.Union(xformed_newbox);
		
		// Update parents' boxes
		R3InteriorNode *parent;
		RNIterator iterator2;
		R3_FOR_EACH_PARENT_NODE(this, parent, iterator2) 
		    parent->UpdateBBox(this, savebox, bbox);
	    }
	}
	else {
	    // Child box shrunk - recompute box from scratch
	    R3Node *child;
	    RNIterator iterator;
	    R3Box savebox(bbox);
	    bbox = R3null_box;
	    R3_FOR_EACH_CHILD_NODE(this, child, iterator) {
		if (child == node) {
		    bbox.Union(xformed_newbox);
		}
		else {
		    R3Box childbox = child->BBox();
		    if (transformation && (!childbox.IsEmpty())) 
		        childbox.Transform(*transformation);
		    bbox.Union(childbox);
		}
	    }

	    // Update parents' boxes
	    if (!R3Contains(bbox, savebox)) {
	        R3InteriorNode *parent;
		RNIterator iterator2;
		R3_FOR_EACH_PARENT_NODE(this, parent, iterator2) 
		    parent->UpdateBBox(this, savebox, bbox);
	    }
	}
    }
}



void R3GroupNode::
UpdateNFacets(R3Node *node, const RNInterval& oldnfacets, const RNInterval& newnfacets)
{
    // Just checking 
    assert(NChildren() > 0);

    // Remember original nfacets
    RNInterval savenfacets(nfacets);
	    
    // Update nfacets
    nfacets += RNInterval(newnfacets.Min() - oldnfacets.Min(), 
			  newnfacets.Max() - oldnfacets.Max());

    // Update cull flags
    if ((NChildren() <= 1) || (nfacets.Max() <= NChildren())) {
        flags.Remove(R3_CULL_CHILDREN_NODE_FLAG);
        base_flags.Remove(R3_CULL_CHILDREN_NODE_FLAG);
    }
    else {
        flags.Add(R3_CULL_CHILDREN_NODE_FLAG);
	base_flags.Add(R3_CULL_CHILDREN_NODE_FLAG);
    }

    // Update parents' nfacets
    if (nfacets != savenfacets) {
        R3InteriorNode *parent;
	RNIterator iterator2;
	R3_FOR_EACH_PARENT_NODE(this, parent, iterator2) 
	    parent->UpdateNFacets(this, savenfacets, nfacets);
    }
}



void R3GroupNode::
UpdateFlags(R3Node *node, RNFlags oldflags, RNFlags newflags)
{
    // Just checking 
    assert(NChildren() > 0);

    // Remember original flags
    RNFlags saveflags(flags);
	    
    // Check if have only one child
    if (NChildren() == 1) {
	// Update flags
	flags = base_flags | (newflags & R3_CHILD_NODE_FLAGS_MASK);

	// Not more than one child -- definitely do not cull
	base_flags.Remove(R3_CULL_CHILDREN_NODE_FLAG);
	flags.Remove(R3_CULL_CHILDREN_NODE_FLAG);
    }
    else {
        // Update base flags
        if (nfacets.Max() <= NChildren()) base_flags.Remove(R3_CULL_CHILDREN_NODE_FLAG);
        else base_flags.Add(R3_CULL_CHILDREN_NODE_FLAG);

	// Compute flags
	if (newflags.Contains(oldflags)) {
	    // Update flags
	    if (node != this) flags.Add(newflags & R3_CHILD_NODE_FLAGS_MASK);
	    else flags = base_flags | ((newflags | flags) & R3_CHILD_NODE_FLAGS_MASK);
	}
	else {
	    // Compute child flags from scratch
	    R3Node *child;
	    RNIterator iterator;
	    RNFlags childflags = R3_NULL_NODE_FLAGS;
	    R3_FOR_EACH_CHILD_NODE(this, child, iterator) {
		if (child == node) childflags.Add(newflags);
		else childflags.Add(child->Flags());
	    }
	    
	    // Update flags
	    flags = base_flags | (childflags & R3_CHILD_NODE_FLAGS_MASK);
	}
    }

    // Update parent flags
    if (flags != saveflags) {
        R3InteriorNode *parent;
	RNIterator iterator2;
	R3_FOR_EACH_PARENT_NODE(this, parent, iterator2) 
	  parent->UpdateFlags(this, saveflags, flags);
    }
}


//#ifdef SHAPE_IS_NOT_GROUP
//
//void R3GroupNode::
//Draw(R3Session& session, R3TraversalState& state)
//{
//    // ###################################################################
//    // This is here for speed only -- direct access to children nodes ???
//    // ###################################################################
//
//    // Check if have any children
//    if (!children) return;
//
//    // Update statistics
//    R3node_total_nodes_drawn++;
//
//    // Push state
//    state.PushNode(this);
//    
//    // Check if should cull node
//    if ((state.Flags()[R3_TRAVERSAL_STATE_BEAM_FLAG]) && 
//	(Flags()[R3_CULL_CHILDREN_NODE_FLAG])) {
//        // Draw each child node 
//        for (int i = 0; i < children->NEntries(); i++) {
//  	    // Get kth child
//	    R3Node *child = children->Kth(i);
//
//	    // Check if should check child visibility
//	    if (child->NFacets().Max() >= R3min_nfacets_for_frustum_cull) {
//	        // Check if bounding shape is at least partially visible
//	        R3node_total_nodes_checked_for_draw++;
//	        if (!R3Intersects(state.VisibilityBeam(), child->BBox())) continue;
//
//		// Check if can unset state BEAM flag for child traversal
//		if ((child->NChildren() > 1) && 
//		    (R3Contains(state.VisibilityBeam(), child->BBox()))) {
//		    // Draw completely visible child 
//		    state.Flags().Remove(R3_TRAVERSAL_STATE_BEAM_FLAG);
//		    child->Draw(session, state);
//		    state.Flags().Add(R3_TRAVERSAL_STATE_BEAM_FLAG);
//		}
//		else {
//		    // Draw partially visible child 
//		    child->Draw(session, state);
//
//		    // Outline child bounding shape
//		    if (R3node_show_bounds) child->BBox().Outline();
//		}	
//	    }
//	    else {
//	        // Draw simple child, irrespective of whether it is visible
//	        // state.Flags().Remove(R3_TRAVERSAL_STATE_BEAM_FLAG);
//	        child->Draw(session, state);
//		// state.Flags().Add(R3_TRAVERSAL_STATE_BEAML_FLAG);
//	    }
//        }
//    }
//    else {
//        // Draw each child node 
//        for (int i = 0; i < children->NEntries(); i++) 
//            children->Kth(i)->Draw(session, state);
//    }
//
//    // Pop state
//    state.PopNode(this);
//}
//
//#else
//
//void R3GroupNode::
//Draw(R3Session& session, R3TraversalState& state)
//{
//    // ###################################################################
//    // This is here for speed only -- direct access to children nodes ???
//    // ###################################################################
//
//    // Check if have any children
//    if (!children) return;
//
//    // Update statistics
//    R3node_total_nodes_drawn++;
//
//    // Push state
//    state.PushNode(this);
//    
//    // Check if should cull node
//    if ((state.Flags()[R3_TRAVERSAL_STATE_BEAM_FLAG]) && 
//	(Flags()[R3_CULL_CHILDREN_NODE_FLAG])) {
//        // Draw each child node 
//        for (int i = 0; i < children->NEntries(); i++) {
//  	    // Get kth child
//	    R3GroupNode *child = (R3GroupNode *) children->Kth(i);
//
//	    // Check if should check child visibility
//	    if (child->nfacets.Max() >= R3min_nfacets_for_frustum_cull) {
//	        // Check if bounding shape is at least partially visible
//	        R3node_total_nodes_checked_for_draw++;
//	        if (!R3Intersects(state.VisibilityBeam(), child->bbox)) continue;
//
//		// Check if can unset state BEAM flag for child traversal
//		if ((child->children) && 
//		    (child->children->NEntries() > 1) && 
//		    (R3Contains(state.VisibilityBeam(), child->bbox))) {
//		    // Draw completely visible child 
//		    state.Flags().Remove(R3_TRAVERSAL_STATE_BEAM_FLAG);
//		    child->Draw(session, state);
//		    state.Flags().Add(R3_TRAVERSAL_STATE_BEAM_FLAG);
//		}
//		else {
//		    // Draw partially visible child 
//		    child->Draw(session, state);
//
//		    // Outline child bounding shape
//		    if (R3node_show_bounds) child->bbox.Outline();
//		}	
//	    }
//	    else {
//	        // Draw simple child, irrespective of whether it is visible
//	        // state.Flags().Remove(R3_TRAVERSAL_STATE_BEAM_FLAG);
//	        child->Draw(session, state);
//		// state.Flags().Add(R3_TRAVERSAL_STATE_BEAML_FLAG);
//	    }
//        }
//    }
//    else {
//        // Draw each child node 
//        for (int i = 0; i < children->NEntries(); i++) 
//            children->Kth(i)->Draw(session, state);
//    }
//
//    // Pop state
//    state.PopNode(this);
//}
//
//#endif
//
//
//
//int R3GroupNode::
//Simulate(R3Session& session, R3TraversalState& state)
//{
//    // ###################################################################
//    // This is here for speed only -- direct access to children nodes ???
//    // ###################################################################
//
//    // Check if have any children
//    if (!children) return FALSE;
//
//    // Simulate each child node
//    int status = FALSE;
//    for (int i = 0; i < children->NEntries(); i++) {
//        R3Node *child = children->Kth(i);
//	if (child->Flags()[R3_SIMULATE_NODE_FLAGS])
//	    status |= child->Simulate(session, state);
//    }
//
//    // Return status
//    return status;
//}
//
//
//
