/* Source file for the R3 session class */



/* Include files */

#include "R3Nodes/R3Nodes.h"



/* Public flags */

const RNFlags R3_SESSION_INSIDE_SEQUENCE_FLAG = 0x1;
const RNFlags R3_SESSION_BEGINNING_SEQUENCE_FLAG = 0x2;



/* Public variables */

R3Session R3null_session;
RNTime R3start_time;



/* Private varaibles */

RNScalar R3session_min_frame_rate = 10.0;
RNScalar R3session_target_frame_rate = 15.0;
RNScalar R3session_max_frame_rate = 20.0;
RNScalar R3session_min_lod_threshold = 1.0E-4; // tanangle/facet
RNScalar R3session_max_lod_threshold = 0.01; // tanangle/facet (was 0.0008)
RNBoolean R3session_update_lod_threshold = TRUE;



/* Public functions */

int R3InitSession()
{
    /* Read start time */
    R3start_time.Read();

    /* Return success */
    return TRUE;
}



void R3StopSession()
{
}



R3Session::
R3Session(void)
    : R3View(),
      id(0),
      frame(NULL),
      frames(),
      simulation_time(0.0),
      frame_rate(15.0),
      lod_threshold(0.0),
      delta_lod_threshold(0.0),
      flags(0)
{
}



R3Session::
R3Session(const R3Session& session)
    : R3View(session),
      id(R3next_mark++),
      frame(NULL),
      frames(),
      simulation_time(0.0),
      frame_rate(15.0),
      lod_threshold(0.0),
      delta_lod_threshold(0.0),
      flags(0)
{
}



R3Session::
R3Session(const R3View& view)
    : R3View(view),
      id(R3next_mark++),
      frame(NULL),
      frames(),
      simulation_time(0.0),
      frame_rate(15.0),
      lod_threshold(0.0),
      delta_lod_threshold(0.0),
      flags(0)
{
}



R3Session::
R3Session(const R3Camera& camera, const R2Viewport& viewport, R3Model *model)
    : R3View(camera, viewport, model),
      id(R3next_mark++),
      frame(NULL),
      frames(),
      simulation_time(0.0),
      frame_rate(15.0),
      lod_threshold(0.0),
      delta_lod_threshold(0.0),
      flags(0)
{
}



R3Session::
~R3Session(void)
{
}



void R3Session::
BeginSequence(void) 
{
    // Just checking
    assert(!frame);

    // Set sequence flags
    flags.Add(R3_SESSION_BEGINNING_SEQUENCE_FLAG);

    // Reset sequence variables
    delta_lod_threshold = 0.0;
}



void R3Session::
EndSequence(void) 
{
    // Just checking
    assert(!frame);

    // Reet sequence flags
    flags.Remove(R3_SESSION_BEGINNING_SEQUENCE_FLAG |
		 R3_SESSION_INSIDE_SEQUENCE_FLAG);

    // Reset sequence variables
    delta_lod_threshold = 0.0;
}



void R3Session::
RefreshFrame(void) 
{
    // Just checking
    assert(!frame);

    // Refresh frame
    BeginFrame();
    SimulateFrame();
    DrawFrame();
    EndFrame();
}



void R3Session::
BeginFrame(void) 
{
    // Just checking
    assert(!frame);

    // Re-evaluate simulation time
    simulation_time = R3start_time.Elapsed();

    // Re-evaulate frame rate
    if (flags[R3_SESSION_INSIDE_SEQUENCE_FLAG]) {
        // Update frame rate 
        assert(frames.Tail());
        RNScalar previous_frame_time = frames.Tail()->ElapsedTime();
	RNScalar previous_frame_rate = (previous_frame_time > 0.0167) ? 1.0 / previous_frame_time : 60.0;
	frame_rate = 0.5 * frame_rate + 0.5 * previous_frame_rate;
	assert(frame_rate > 0.0);

        // Update lod threshold
	if (R3session_update_lod_threshold) {
	    if ((frame_rate < R3session_min_frame_rate) || (frame_rate > R3session_max_frame_rate)) {
	        RNScalar target_frame_rate = 0.5 * R3session_target_frame_rate + 0.5 * frame_rate;
		RNScalar denom = (frame_rate > target_frame_rate) ? frame_rate : target_frame_rate;
		RNScalar delta = (target_frame_rate - frame_rate) / denom;
		if ((delta * delta_lod_threshold) < 0.0) delta *= (delta < 0.0) ? -delta : delta;
		delta_lod_threshold = delta * delta * delta;
		assert(RNIsGreaterOrEqual(delta_lod_threshold, -1.0) && RNIsLessOrEqual(delta_lod_threshold, 1.0));
		if (lod_threshold >= R3session_min_lod_threshold) lod_threshold *= 1.0 + delta_lod_threshold;
		else if (delta > 0.0) lod_threshold = R3session_min_lod_threshold;
		if (lod_threshold < R3session_min_lod_threshold) lod_threshold = 0.0;
		else if (lod_threshold > R3session_max_lod_threshold) lod_threshold = R3session_max_lod_threshold;
		// printf("%6.4f %12.10f %12.10f %12.10f\n", frame_rate, lod_threshold, delta_lod_threshold, delta);
	    }
	}
    }

    // Create a new frame
    frame = new R3Frame(this);
    assert(frame);

    // Reset pick/draw statistics
    R3ResetNodeStatistics();
}



void R3Session::
SimulateFrame(void)
{
    // Just checking
    assert(frame);

    // Simulate frame
    frame->Simulate();
}


void R3Session::
DrawFrame(void)
{
    // Just checking
    assert(frame);

    // Draw frame
    frame->Draw();
}



void R3Session::
EndFrame(void) 
{
    // Just checking
    assert(frame);

    // Add frame to the queue
    frames.InsertHead(frame);

    // Delete all but current frame from queue
    while (frames.NEntries() > 1) {
	R3Frame *old = frames.Tail();
	frames.RemoveTail();
	delete old;
    }

    // Update sequence flags
    if (flags[R3_SESSION_BEGINNING_SEQUENCE_FLAG]) 
        flags.Add(R3_SESSION_INSIDE_SEQUENCE_FLAG);

    // Update frame
    frame = NULL;
}



int R3Session::
SendPickMessages(int x, int y, R3MessageType type, const void *buffer, int length)
{
    // Initialize status 
    int status = FALSE;

    // Pick node
    R3TraversalPath hit_path;
    if (Pick(x, y, &hit_path)) {
        for (int i = hit_path.NNodes() - 1; i >= 0; i--) {
	    R3Node *node = hit_path.KthNode(i);
	    status |= SendMessage(R3null_state, node, type, buffer, length);
	}
    }

    // Return whether anybody processed message
    return status;
}



