//
// $Id: VsExprNodes.h,v 1.1 2000/01/17 18:47:59 min Exp min $
//

#ifndef __VSEXPRNODES_H
#define __VSEXPRNODES_H

#include "VsExprNode.h"

#include <string.h>
#include <stdlib.h>
#include "R3Graphics/R3Graphics.h"
#include "Fields.h"
#include "Files.h"
#include "VsStatements.h"
#include "Files.h"
#include "VsExpression.h"





static const int MAX_OPERATOR_LENGTH = 4;

class VsBinaryExprNode : public VsExprNode
{

public:

  VsBinaryExprNode() {}
  VsBinaryExprNode(VsExprNode *new_left, char *new_op, VsExprNode *new_right) {
    left = new_left;
    strcpy(my_op, new_op);
    right = new_right;
  }
  ~VsBinaryExprNode() {
    if (left) delete left;
    if (right) delete right;
  }

  Field *get_field_ptr() { return NULL; }
  Field *evaluate();

  void set_left(VsExprNode *new_left) { left = new_left; }
  void set_right(VsExprNode *new_right) { right = new_right; }
  void set_op(char *new_op) { strcpy(my_op, new_op); }


private:

  VsExprNode *left;
  VsExprNode *right;

  char my_op[MAX_OPERATOR_LENGTH];


};  // VsBinaryExprNode class





class VsConstantNode : public VsExprNode
{

public:

  VsConstantNode() {}
  VsConstantNode(char *new_value_string);
  ~VsConstantNode() {}


  virtual VsNodeType get_node_type() { return VsConstantNodeType; }

  Field *evaluate() { return &my_value; }

  Field *get_field_ptr() { return &my_value; }

  // again, make this so it supports all number types
  Field *operator-(Field *other) {
    return new SFFloat(my_value.get_float() - 
		       ((SFFloat *) other)->get_float());
  }


private:

  SFFloat my_value;

  char value_string[MAX_TOKEN_LENGTH];


};  // VsConstantNode class





class VsStringConstantNode : public VsExprNode
{

public:

  VsStringConstantNode() {}
  VsStringConstantNode(SFString *new_string) { my_string = new_string; }
  VsStringConstantNode(char *new_value_string) {
    my_string = new SFString(new_value_string);
  }  // constructor
  ~VsStringConstantNode() {
    if (my_string) delete my_string;
  }


  virtual VsNodeType get_node_type() { return VsStringConstantNodeType; }

  Field *evaluate() { return my_string; }

  Field *get_field_ptr() { return my_string; }


private:

  SFString *my_string;


};  // VsStringConstantNode class





#include "VsSymbolNode.h"





class VsVariableNode : public VsSymbolNode
{

public:

  VsVariableNode() : VsSymbolNode() {}
  VsVariableNode(char *new_id);
  VsVariableNode(char *new_id, Field *new_field_ptr, FieldType new_type);
  VsVariableNode(VsVariableNode& other);

  ~VsVariableNode();


  virtual VsNodeType get_node_type() { return VsVariableNodeType; }

  void parse(char *field_type, int get_default);

  VsSymbolNode *create_copy() {
    return new VsVariableNode(*this);
  }
  Field *evaluate() { return field_ptr; }
  Field *get_field_ptr() { return field_ptr; }
  void set_field_ptr(Field *new_field_ptr) { field_ptr = new_field_ptr; }

  FieldType get_field_type() { return field_type; }

  char *get_field_var_name() { return get_name(); }

  void set_changed(int new_value) { changed = new_value; }
  int get_changed() { return changed; }


private:

  Field *parse_field_type();

  Field *field_ptr;
  FieldType field_type;
  int changed;

};  // VsVariableNode class





#include "VsSymbolTable.h"

class VsFunctionNode : public VsSymbolNode
{

public:

  VsFunctionNode() {}
  ~VsFunctionNode() {}


  virtual VsNodeType get_node_type() { return VsFunctionNodeType; }

  void parse(Tokenfile *scriptfile);
  void parse_arguments(Tokenfile *tokenfile_ptr);

  Field *evaluate() { return evaluate(0, NULL); }
  Field *evaluate(int argc, Field **argv);


private:

  VsSymbolTable local_vars;
  VsSymbolTable arguments;
  VsStatementBlock statements;


};  // VsFunctionNode class





class VsArrayNode : public VsVariableNode
{

public:

  VsArrayNode() { node_ptr = NULL; }
  VsArrayNode(VsVariableNode& var_ptr) : VsVariableNode(var_ptr) {
    node_ptr = &var_ptr;
  }
  ~VsArrayNode() {}


  VsNodeType get_node_type() { return VsArrayNodeType; }

  void parse(Tokenfile *tokenfile_ptr);

  Field *evaluate();


private:

  VsExpression index;
  VsExprNode *node_ptr;


};  // VsArrayNode class





class VsFunctionCallNode : public VsExprNode
{

public:

  VsFunctionCallNode() {}
  ~VsFunctionCallNode() {
    if (function_ptr) delete function_ptr;
    int nr_arguments = arguments.NEntries();
    for(int i=0; i<nr_arguments; i++) delete arguments.Kth(i);
  }  // destructor


  VsNodeType get_node_type() { return VsFunctionCallNodeType; }

  void parse(Tokenfile *tokenfile_ptr);

  Field *evaluate();

  void set_function_ptr(VsFunctionNode *new_function_ptr) {
    function_ptr = new_function_ptr;
  }
  VsFunctionNode *get_function_ptr() { return function_ptr; }


protected:

  VsFunctionNode *function_ptr;
  RNArray<VsExpression *> arguments;
  Field **actual_arguments;


};  // VsFunctionCallNode





static const char browser_methods[][32] = {
  "getName", "getVersion", "getCurrentSpeed", "getCurrentFrameRate",
  "getWorldURL", "replaceWorld", "createVrmlFromString",
  "createVrmlFromURL", "addRoute", "deleteRoute", "loadURL",
  "setDescription", "\0" };

class VsBrowserCallNode : public VsFunctionCallNode
{

public:

  VsBrowserCallNode() {
    browser_name.set_string((char*)"RING VRML2 Browser");
    browser_version.set_string((char*)"version 0.9");
  }
  ~VsBrowserCallNode() {}


  VsNodeType get_node_type() { return VsBrowserCallNodeType; }

  void parse(Tokenfile *tokenfile_ptr);

  Field *evaluate();


private:

  int function_index;

  SFString browser_name;
  SFString browser_version;


};  // class VsBrowserCallNode





class VsMathCallNode : public VsFunctionCallNode
{

public:

  VsMathCallNode() {
    function_ptr = NULL;
  }
  ~VsMathCallNode() {}


  VsNodeType get_node_type() { return VsMathCallNodeType; }

  void parse(Tokenfile *tokenfile_ptr);

  Field *evaluate();


private:

  int is_constant;
  int parse_constant_name(Tokenfile *tokenfile_ptr);
  SFFloat my_value;

  int parse_function_name(Tokenfile *tokenfile_ptr);

  // pointer to function
  double (*function_ptr_zero_args)();
  double (*function_ptr_one_arg)(double);
  double (*function_ptr_two_args)(double, double);

  int nr_parameters;


};  // VsMathCallNode class

//
// need these function for _every_ function that has both a double
// and a float version because of stupid compiler bug...
//

double my_abs(double x);
double my_acos(double x);
double my_asin(double x);
double my_atan(double x);
double my_ceil(double x);
double my_cos(double x);
double my_exp(double x);
double my_floor(double x);
double my_log(double x);
double my_max(double x, double y);
double my_min(double x, double y);
double my_pow(double x, double y);
double my_random();
double my_round(double x);
double my_sin(double x);
double my_sqrt(double x);
double my_tan(double x);





class VsConstructorNode : public VsFunctionCallNode
{

public:

  VsConstructorNode() {}
  ~VsConstructorNode() {}


  VsNodeType get_node_type() { return VsConstructorNodeType; }

  void parse(Tokenfile *tokenfile_ptr);

  Field *evaluate();

  Field *get_field_ptr() { return field_ptr; }
  FieldVarType get_field_var_type() { return field_var_type; }


private:

  Field *field_ptr;
  FieldVarType field_var_type;


};  // VsConstructorNode class


#endif






